﻿using MeshVR;
using MVR.FileManagementSecure;
using SimpleJSON;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text.RegularExpressions;
using UnityEngine;

namespace Naturalis
{
    sealed class AutoUpgrade : MVRScript
    {
        const string NATURALIS_VERSION = "1.0.2";
        readonly Regex _naturalisBMRegex = new Regex(@"^plugin#\d+_Naturalis.BootyMagic", RegexOptions.Compiled);
        readonly Regex _naturalisTMRegex = new Regex(@"^plugin#\d+_Naturalis.TittyMagic", RegexOptions.Compiled);
        readonly Regex _tittyMagicRegex = new Regex(@"^plugin#\d+_TittyMagic.Script", RegexOptions.Compiled);

        PresetManager _presetManager;
        PresetManagerControl _presetManagerControl;

        public override void Init()
        {
            StartCoroutine(InitCo());
        }

        IEnumerator InitCo()
        {
            yield return new WaitForEndOfFrame();
            while(SuperController.singleton.isLoading)
            {
                yield return null;
            }

            if(_restoringFromJSON)
            {
                manager.RemovePluginWithUID(GetPluginId(name));
                yield break;
            }

            _presetManager = containingAtom.GetStorableByID("PluginPresets").GetComponentInChildren<PresetManager>();
            _presetManagerControl = containingAtom.presetManagerControls.Find(pmc => pmc.name == "PluginPresets");

            var naturalisBM = FindStorablesByRegexMatch(_naturalisBMRegex).FirstOrDefault();
            var naturalisTM = FindStorablesByRegexMatch(_naturalisTMRegex).FirstOrDefault();
            var naturalis = naturalisBM != null ? naturalisBM : naturalisTM;
            if(naturalis != null)
            {
                string versionStr = naturalis.IsStringJSONParam("Naturalis Version") ? naturalis.GetStringParamValue("Naturalis Version") : "";
                if(string.IsNullOrEmpty(versionStr))
                {
                    pluginLabelJSON.val = "Unidentified Naturalis version. ???";
                }
                else if(versionStr == NATURALIS_VERSION)
                {
                    LogMessage($"Naturalis v{NATURALIS_VERSION} is already added.");
                    manager.RemovePluginWithUID(GetPluginId(name));
                    yield break;
                }
                else if(versionStr != "1.0.0")
                {
                    pluginLabelJSON.val = "Only upgrade from v1.0.0 (13.var) is supported";
                }
                else
                {
                    if(naturalisTM == null)
                    {
                        StartCoroutine(MigrateFromNaturalis1_x_BM(naturalisTM));
                    }
                    else if(naturalisBM == null)
                    {
                        StartCoroutine(MigrateFromNaturalis1_x_TM(naturalisTM));
                    }
                    else
                    {
                        StartCoroutine(MigrateFromNaturalis1_x(naturalisBM, naturalisTM));
                    }
                }

                yield break;
            }

            var tittyMagic = FindStorablesByRegexMatch(_tittyMagicRegex).FirstOrDefault();
            if(tittyMagic != null)
            {
                string versionStr = tittyMagic.IsStringJSONParam("version") ? tittyMagic.GetStringParamValue("version") : "";
                if(string.IsNullOrEmpty(versionStr))
                {
                    pluginLabelJSON.val = "TittyMagic version not identified / too old";
                }
                else
                {
                    if(versionStr.StartsWith("v"))
                    {
                        versionStr = versionStr.Substring(1);
                    }

                    if(!versionStr.StartsWith("5."))
                    {
                        pluginLabelJSON.val = "Only upgrade from TittyMagic 5 is supported";
                    }
                    else if(versionStr.Split('-').Length > 1)
                    {
                        pluginLabelJSON.val = $"Upgrading prerelease v{versionStr} is unsupported";
                    }
                    else
                    {
                        StartCoroutine(MigrateFromTittyMagic5_x(tittyMagic, versionStr));
                    }
                }

                yield break;
            }

            StartCoroutine(MigrateFromNone());
        }

        IEnumerator MigrateFromNaturalis1_x_BM(JSONStorable naturalisBM)
        {
            var bootyMagicJSON = naturalisBM.GetJSON(true, true, true);
            manager.RemovePluginWithUID(GetPluginId(naturalisBM.name));
            AddNaturalis("BootyMagic");

            yield return new WaitForSeconds(1);
            var naturalisBMNew = FindStorablesByRegexMatch(_naturalisBMRegex).First();
            while(!naturalisBMNew || naturalisBMNew.GetBoolParamValue("internal_isCalibrating"))
            {
                yield return null;
            }

            yield return new WaitForSeconds(0.1f);
            pluginLabelJSON.val = "Restoring plugin params...";
            naturalisBMNew.RestoreFromJSON(bootyMagicJSON);

            yield return new WaitForSeconds(1.5f);

            /* Remove this script from atom. GetPluginId called again because of plugin uid being reassignment due to deleting Naturalis */
            manager.RemovePluginWithUID(GetPluginId(name));
        }

        IEnumerator MigrateFromNaturalis1_x_TM(JSONStorable naturalisTM)
        {
            var tittyMagicJSON = naturalisTM.GetJSON(true, true, true);
            manager.RemovePluginWithUID(GetPluginId(naturalisTM.name));
            AddNaturalis("TittyMagic");

            yield return new WaitForSeconds(1);
            var naturalisTMNew = FindStorablesByRegexMatch(_naturalisTMRegex).First();
            while(!naturalisTMNew || naturalisTMNew.GetBoolParamValue("internal_isCalibrating"))
            {
                yield return null;
            }

            yield return new WaitForSeconds(0.1f);
            pluginLabelJSON.val = "Restoring plugin params...";
            naturalisTMNew.RestoreFromJSON(tittyMagicJSON);

            yield return new WaitForSeconds(1.5f);

            /* Remove this script from atom. GetPluginId called again because of plugin uid being reassignment due to deleting Naturalis */
            manager.RemovePluginWithUID(GetPluginId(name));
        }

        IEnumerator MigrateFromNaturalis1_x(JSONStorable naturalisBM, JSONStorable naturalisTM)
        {
            var bootyMagicJSON = naturalisBM.GetJSON(true, true, true);
            var tittyMagicJSON = naturalisTM.GetJSON(true, true, true);
            manager.RemovePluginWithUID(GetPluginId(naturalisBM.name));
            AddNaturalis("Naturalis");

            yield return new WaitForSeconds(1);
            var naturalisBMNew = FindStorablesByRegexMatch(_naturalisBMRegex).First();
            var naturalisTMNew = FindStorablesByRegexMatch(_naturalisTMRegex).First();
            while(
                !naturalisBMNew ||
                !naturalisTMNew ||
                naturalisBMNew.GetBoolParamValue("internal_isCalibrating") ||
                naturalisTMNew.GetBoolParamValue("internal_isCalibrating")
            )
            {
                yield return null;
            }

            yield return new WaitForSeconds(0.1f);
            pluginLabelJSON.val = "Restoring plugin params...";
            naturalisBMNew.RestoreFromJSON(bootyMagicJSON);
            naturalisTMNew.RestoreFromJSON(tittyMagicJSON);

            yield return new WaitForSeconds(1.5f);

            /* Remove this script from atom. GetPluginId called again because of plugin uid being reassignment due to deleting Naturalis */
            manager.RemovePluginWithUID(GetPluginId(name));
        }

        IEnumerator MigrateFromTittyMagic5_x(JSONStorable tittyMagic5, string version)
        {
            var tittyMagicJSON = ConvertToNaturalisTittyMagicJSON(tittyMagic5.GetJSON(true, true, true), version);
            manager.RemovePluginWithUID(GetPluginId(tittyMagic5.name));
            AddNaturalis("Naturalis");

            yield return new WaitForSeconds(1);
            var naturalisTMNew = FindStorablesByRegexMatch(_naturalisTMRegex).FirstOrDefault();
            while(!naturalisTMNew || naturalisTMNew.GetBoolParamValue("internal_isCalibrating"))
            {
                yield return null;
            }

            yield return new WaitForSeconds(0.1f);
            pluginLabelJSON.val = "Restoring plugin params...";
            naturalisTMNew.RestoreFromJSON(tittyMagicJSON);

            yield return new WaitForSeconds(1.5f);

            /* Remove this script from atom. GetPluginId called again because of plugin uid being reassignment due to deleting TittyMagic 5. */
            manager.RemovePluginWithUID(GetPluginId(name));
        }

        IEnumerator MigrateFromNone()
        {
            AddNaturalis("Naturalis");
            yield return new WaitForSeconds(1);
            manager.RemovePluginWithUID(GetPluginId(name));
        }

        /* Converts TittyMagic 5 JSON to Naturalis TittyMagic JSON.
         * - LDP is disabled (even if you have it enabled by default in Naturalis User Preferences).
         * - Auto-Adjust Gravity Counterforce is enabled (default in Naturalis).
         * - Mass Offset is not applied, set to 0 instead since the mass calculation is different in Naturalis.
         * - All physics parameter offsets are also not applied, except for Angle Targets.
         *   The default physics values differ too much with Naturalis that the result could be quite different with the TittyMagic 5 offset values.
         *   With soft physics parameters, the gradient also complicates things.
         * - Quickness is converted from the 0 to 100 range to the -1 to 1 range, with 70 in TittyMagic corresponding to 0 in Naturalis.
         * - The overall gravity physics multipliers are converted to Naturalis' parameter specific multipliers.
         * - Everything else should be migrated as is - softness, soft physics on/off, self-collision, adaptive friction, nipple erection,
         *   gravity offset morphing, and hard collider properties.
         */
        static JSONClass ConvertToNaturalisTittyMagicJSON(JSONClass jc, string matchedVersion)
        {
            float baseGravity = GetFloatValue(jc, "gravityPhysicsBase", 1f);
            float upGravity = GetFloatValue(jc, "gravityPhysicsUp", 1f) * baseGravity;
            float downGravity = GetFloatValue(jc, "gravityPhysicsDown", 1f) * baseGravity;
            float forwardGravity = GetFloatValue(jc, "gravityPhysicsForward", 1f) * baseGravity;
            float backGravity = GetFloatValue(jc, "gravityPhysicsBack", 1f) * baseGravity;
            float leftRightGravity = GetFloatValue(jc, "gravityPhysicsLeftRight", 1f) * baseGravity;

            float legacyQuickness = GetFloatValue(jc, "breastQuickness", 70f);
            float quickness = legacyQuickness <= 70f
                ? Mathf.Lerp(-1f, 0f, legacyQuickness / 70f)
                : Mathf.Lerp(0f, 1f, (legacyQuickness - 70f) / 30f);

            var result = new JSONClass
            {
                ["id"] = "plugin#0_Naturalis.TittyMagic",
                ["enabled"] = { AsBool = true },
                ["Breast Soft Physics Enabled"] = { AsBool = GetBoolValue(jc, "on", false) },
                ["Local Distortion Physics Enabled"] = { AsBool = false },
                ["Breast Soft Physics Self Collide"] = { AsBool = GetBoolValue(jc, "allowSelfCollision", false) },
                ["Auto-Adjust Gravity Counterforce"] = { AsBool = true },
                ["Use Adaptive Friction"] = { AsBool = GetBoolValue(jc, "enableAdaptiveFriction", false) },
                // ["Breast Joint Mass Offset"] = { AsFloat = GetFloatValue(jc, "massOffset", 0f) },
                // ["Center Of Gravity Offset"] = { AsFloat = GetFloatValue(jc, "centerOfGravityPercentOffset", 0f) },
                // ["Spring Offset"] = { AsFloat = GetFloatValue(jc, "springOffset", 0f) },
                // ["Damper Offset"] = { AsFloat = GetFloatValue(jc, "damperOffset", 0f) },
                // ["In/Out Spring Offset"] = { AsFloat = GetFloatValue(jc, "positionSpringZOffset", 0f) },
                // ["In/Out Damper Offset"] = { AsFloat = GetFloatValue(jc, "positionDamperZOffset", 0f) },
                ["Up/Down Angle Target Offset"] = { AsFloat = GetFloatValue(jc, "targetRotationXOffset", 0f) },
                ["Right/Left Angle Target Offset"] = { AsFloat = GetFloatValue(jc, "targetRotationYOffset", 0f) },
                // ["Fat Spring Offset"] = { AsFloat = GetFloatValue(jc, "softVerticesSpringOffset", 0f) },
                // ["Fat Spring Gradient Middle Shift"] = { AsFloat = -0.4f },
                // ["Fat Spring Gradient Outer Multiplier"] = { AsFloat = GetFloatValue(jc, "softVerticesSpringOffset", 0f) },
                // ["Fat Spring Gradient Center Multiplier"] = { AsFloat = 1.5f },
                // ["Fat Spring Gradient Nipple Multiplier"] = { AsFloat = 2.25f },
                // ["Fat Damper Offset"] = { AsFloat = GetFloatValue(jc, "softVerticesDamperOffset", 0f) },
                // ["Fat Damper Gradient Middle Shift"] = { AsFloat = 0f },
                // ["Fat Damper Gradient Outer Multiplier"] = { AsFloat = 1.8f },
                // ["Fat Damper Gradient Center Multiplier"] = { AsFloat = 1.5f },
                // ["Fat Damper Gradient Nipple Multiplier"] = { AsFloat = 1.8f },
                // ["Fat Mass Offset"] = { AsFloat = GetFloatValue(jc, "softVerticesMassOffset", 0f) },
                // ["Fat Mass Gradient Middle Shift"] = { AsFloat = 0f },
                // ["Fat Mass Gradient Outer Multiplier"] = { AsFloat = 1f },
                // ["Fat Mass Gradient Center Multiplier"] = { AsFloat = 1f },
                // ["Fat Mass Gradient Nipple Multiplier"] = { AsFloat = 1f },
                // ["Fat Collider Radius Offset"] = { AsFloat = GetFloatValue(jc, "softVerticesColliderRadiusOffset", 0f) },
                // ["Fat Collider Radius Gradient Middle Shift"] = { AsFloat = 0f },
                // ["Fat Collider Radius Gradient Outer Multiplier"] = { AsFloat = 1.1f },
                // ["Fat Collider Radius Gradient Center Multiplier"] = { AsFloat = 1.25f },
                // ["Fat Collider Radius Gradient Nipple Multiplier"] = { AsFloat = 0.33f },
                // ["Fat Collider Depth Offset"] = { AsFloat = GetFloatValue(jc, "softVerticesColliderAdditionalNormalOffsetOffset", 0f) },
                // ["Fat Collider Depth Gradient Middle Shift"] = { AsFloat = 0f },
                // ["Fat Collider Depth Gradient Outer Multiplier"] = { AsFloat = 1f },
                // ["Fat Collider Depth Gradient Center Multiplier"] = { AsFloat = 1f },
                // ["Fat Collider Depth Gradient Nipple Multiplier"] = { AsFloat = 1f },
                // ["Fat Distance Limit Offset"] = { AsFloat = GetFloatValue(jc, "softVerticesDistanceLimitOffset", 0f) },
                // ["Fat Distance Limit Gradient Middle Shift"] = { AsFloat = -0.8f },
                // ["Fat Distance Limit Gradient Outer Multiplier"] = { AsFloat = 1f },
                // ["Fat Distance Limit Gradient Center Multiplier"] = { AsFloat = 0.8f },
                // ["Fat Distance Limit Gradient Nipple Multiplier"] = { AsFloat = 0.86f },
                // ["Fat Back Force Offset"] = { AsFloat = GetFloatValue(jc, "softVerticesBackForceOffset", 0f) },
                // ["Fat Back Force Max Force Offset"] = { AsFloat = GetFloatValue(jc, "softVerticesBackForceMaxForceOffset", 0f) },
                // ["Fat Back Force Threshold Offset"] = { AsFloat = GetFloatValue(jc, "softVerticesBackForceThresholdDistanceOffset", 0f) },
                ["Twist Angle Target"] = { AsFloat = GetFloatValue(jc, "targetRotationZ", 1) },
                ["Gravity Counterforce (%)"] = { AsFloat = 10f },
                ["Spring Up Gravity"] = { AsFloat = upGravity },
                ["Spring Back Gravity"] = { AsFloat = backGravity },
                ["Spring Forward Gravity"] = { AsFloat = forwardGravity },
                ["Spring Sideways In Gravity"] = { AsFloat = leftRightGravity },
                ["Spring Sideways Out Gravity"] = { AsFloat = leftRightGravity },
                ["Damper Back Gravity"] = { AsFloat = backGravity },
                ["Damper Forward Gravity"] = { AsFloat = forwardGravity },
                ["Damper Sideways In Gravity"] = { AsFloat = leftRightGravity },
                ["In/Out Spring Back Gravity"] = { AsFloat = backGravity },
                ["In/Out Spring Forward Gravity"] = { AsFloat = forwardGravity },
                ["Up/Down Angle Target Up Gravity"] = { AsFloat = upGravity },
                ["Up/Down Angle Target Down Gravity"] = { AsFloat = downGravity },
                ["Right/Left Angle Target Sideways In Gravity"] = { AsFloat = leftRightGravity },
                ["Right/Left Angle Target Sideways Out Gravity"] = { AsFloat = leftRightGravity },
                ["Force Physics Base Multiplier"] = { AsFloat = GetFloatValue(jc, "forcePhysicsBase", 1) },
                ["Force Physics Depth In"] = { AsFloat = GetFloatValue(jc, "forcePhysicsBack", 1) },
                ["Force Physics Depth Out"] = { AsFloat = GetFloatValue(jc, "forcePhysicsForward", 1) },
                ["Friction Offset"] = { AsFloat = GetFloatValue(jc, "frictionOffset", 0f) },
                ["Dry Skin Friction"] = { AsFloat = 1f },
                ["Pectoral1 Radius Offset"] = { AsFloat = GetFloatValue(jc, "pectoral1ColliderRadius", 0f) },
                ["Pectoral1 X Offset"] = { AsFloat = GetFloatValue(jc, "pectoral1ColliderCenterX", 0f) },
                ["Pectoral1 Y Offset"] = { AsFloat = GetFloatValue(jc, "pectoral1ColliderCenterY", 0f) },
                ["Pectoral1 Z Offset"] = { AsFloat = GetFloatValue(jc, "pectoral1ColliderCenterZ", 0f) },
                ["Pectoral2 Radius Offset"] = { AsFloat = GetFloatValue(jc, "pectoral2ColliderRadius", 0f) },
                ["Pectoral2 X Offset"] = { AsFloat = GetFloatValue(jc, "pectoral2ColliderCenterX", 0f) },
                ["Pectoral2 Y Offset"] = { AsFloat = GetFloatValue(jc, "pectoral2ColliderCenterY", 0f) },
                ["Pectoral2 Z Offset"] = { AsFloat = GetFloatValue(jc, "pectoral2ColliderCenterZ", 0f) },
                ["Pectoral3 Radius Offset"] = { AsFloat = GetFloatValue(jc, "pectoral3ColliderRadius", 0f) },
                ["Pectoral3 X Offset"] = { AsFloat = GetFloatValue(jc, "pectoral3ColliderCenterX", 0f) },
                ["Pectoral3 Y Offset"] = { AsFloat = GetFloatValue(jc, "pectoral3ColliderCenterY", 0f) },
                ["Pectoral3 Z Offset"] = { AsFloat = GetFloatValue(jc, "pectoral3ColliderCenterZ", 0f) },
                ["Pectoral4 Radius Offset"] = { AsFloat = GetFloatValue(jc, "pectoral4ColliderRadius", 0f) },
                ["Pectoral4 X Offset"] = { AsFloat = GetFloatValue(jc, "pectoral4ColliderCenterX", 0f) },
                ["Pectoral4 Y Offset"] = { AsFloat = GetFloatValue(jc, "pectoral4ColliderCenterY", 0f) },
                ["Pectoral4 Z Offset"] = { AsFloat = GetFloatValue(jc, "pectoral4ColliderCenterZ", 0f) },
                ["Pectoral5 Radius Offset"] = { AsFloat = GetFloatValue(jc, "pectoral5ColliderRadius", 0f) },
                ["Pectoral5 X Offset"] = { AsFloat = GetFloatValue(jc, "pectoral5ColliderCenterX", 0f) },
                ["Pectoral5 Y Offset"] = { AsFloat = GetFloatValue(jc, "pectoral5ColliderCenterY", 0f) },
                ["Pectoral5 Z Offset"] = { AsFloat = GetFloatValue(jc, "pectoral5ColliderCenterZ", 0f) },
                ["Gravity Offset Morphing"] = { AsFloat = GetFloatValue(jc, "gravityOffsetMorphing", 1f) },
                ["Force Morphing Base Multiplier"] = { AsFloat = GetFloatValue(jc, "forceMorphingBase", 1f) },
                ["Force Morphing Up"] = { AsFloat = GetFloatValue(jc, "forceMorphingUp", 1f) },
                ["Force Morphing Depth In"] = { AsFloat = GetFloatValue(jc, "forceMorphingBack", 1f) },
                ["Force Morphing Depth Out"] = { AsFloat = GetFloatValue(jc, "forceMorphingForward", 1f) },
                ["Nipple Erection"] = { AsFloat = GetFloatValue(jc, "nippleErection", 0f) },
                ["Breast Softness"] = { AsFloat = GetFloatValue(jc, "breastSoftness", 70f) },
                ["Breast Quickness Offset"] = { AsFloat = quickness },
            };

            if(matchedVersion == "5.0.0")
            {
                result["Pectoral1 Mass Multiplier"].AsFloat = GetFloatValue(jc, "pectoral1ColliderForce", 0.50f);
                result["Pectoral2 Mass Multiplier"].AsFloat = GetFloatValue(jc, "pectoral2ColliderForce", 0.50f);
                result["Pectoral3 Mass Multiplier"].AsFloat = GetFloatValue(jc, "pectoral3ColliderForce", 0.50f);
                result["Pectoral4 Mass Multiplier"].AsFloat = GetFloatValue(jc, "pectoral4ColliderForce", 0.50f);
                result["Pectoral5 Mass Multiplier"].AsFloat = GetFloatValue(jc, "pectoral5ColliderForce", 0.50f);
                result["Hard Colliders Base Mass Multiplier"].AsFloat = GetFloatValue(jc, "combinedColliderForce", 1.00f);
            }
            else
            {
                result["Pectoral1 Mass Multiplier"].AsFloat = GetFloatValue(jc, "pectoral1CollisionForce", 0.50f);
                result["Pectoral2 Mass Multiplier"].AsFloat = GetFloatValue(jc, "pectoral2CollisionForce", 0.50f);
                result["Pectoral3 Mass Multiplier"].AsFloat = GetFloatValue(jc, "pectoral3CollisionForce", 0.50f);
                result["Pectoral4 Mass Multiplier"].AsFloat = GetFloatValue(jc, "pectoral4CollisionForce", 0.50f);
                result["Pectoral5 Mass Multiplier"].AsFloat = GetFloatValue(jc, "pectoral5CollisionForce", 0.50f);
                result["Hard Colliders Base Mass Multiplier"].AsFloat = GetFloatValue(jc, "baseCollisionForce", 1.00f);
            }

            if(matchedVersion.StartsWith("5.0.") || matchedVersion.StartsWith("5.1."))
            {
                result["Force Morphing Sideways In"].AsFloat = GetFloatValue(jc, "forceMorphingLeftRight", 1.00f);
                result["Force Morphing Sideways Out"].AsFloat = GetFloatValue(jc, "forceMorphingLeftRight", 1.00f);
            }
            else
            {
                result["Force Morphing Sideways In"].AsFloat = GetFloatValue(jc, "forceMorphingSidewaysIn", 1f);
                result["Force Morphing Sideways Out"].AsFloat = GetFloatValue(jc, "forceMorphingSidewaysOut", 1f);
            }

            result["pluginLabel"] = jc["pluginLabel"].Value;
            return result;
        }

        static float GetFloatValue(JSONClass jc, string key, float defaultValue)
        {
            return jc.HasKey(key) ? jc[key].AsFloat : defaultValue;
        }

        static bool GetBoolValue(JSONClass jc, string key, bool defaultValue)
        {
            return jc.HasKey(key) ? jc[key].AsBool : defaultValue;
        }

        void AddNaturalis(string cslist)
        {
            pluginLabelJSON.val = "Adding Naturalis...";
            string tmpPresetName = $"tmp_{Guid.NewGuid().ToString().Substring(0, 7)}";
            string savePath = $"Custom/Atom/Person/Plugins/Preset_{tmpPresetName}.vap";
            string packagePath = GetPackagePath(GetPluginId(name));
            var storables = new JSONArray();
            storables.Add(new JSONClass
            {
                ["id"] = "PluginManager",
                ["plugins"] = new JSONClass
                {
                    ["plugin#0"] = $"{packagePath}Custom/Scripts/everlaster/Naturalis/{cslist}.cslist",
                },
            });
            if(cslist == "Naturalis" || cslist == "BootyMagic")
            {
                storables.Add(new JSONClass
                {
                    ["id"] = "plugin#0_Naturalis.BootyMagic",
                    ["Local Distortion Physics Enabled"] = { AsBool = false },
                });
            }
            if(cslist == "Naturalis" || cslist == "TittyMagic")
            {
                storables.Add(new JSONClass
                {
                    ["id"] = "plugin#0_Naturalis.TittyMagic1",
                    ["Local Distortion Physics Enabled"] = { AsBool = false },
                });
            }
            var presetJSON = new JSONClass
            {
                ["setUnlistedParamsToDefault"] = { AsBool = true },
                ["storables"] = storables,
            };
            FileManagerSecure.WriteAllText(savePath, presetJSON.ToString(string.Empty));

            string currentPresetName = _presetManagerControl.GetStringParamValue("presetName");
            _presetManagerControl.SetStringParamValue("presetName", tmpPresetName);
            _presetManager.MergeLoadPreset();
            _presetManager.DeletePreset();
            _presetManagerControl.SetStringParamValue("presetName", currentPresetName);
        }

        bool _restoringFromJSON;

        public override void RestoreFromJSON(
            JSONClass jc,
            bool restorePhysical = true,
            bool restoreAppearance = true,
            JSONArray presetAtoms = null,
            bool setMissingToDefault = true
        )
        {
            _restoringFromJSON = true;
        }

        IEnumerable<JSONStorable> FindStorablesByRegexMatch(Regex regex)
        {
            var storables = FindStorablesByRegexMatchInternal(regex);
            var list = storables.ToList();
            list.RemoveAll(storable => !storable || !storable.containingAtom);
            return list;
        }

        IEnumerable<JSONStorable> FindStorablesByRegexMatchInternal(Regex regex)
        {
            var storableIds = containingAtom.GetStorableIDs();
            foreach(string id in storableIds)
            {
                if(regex.IsMatch(id))
                {
                    yield return containingAtom.GetStorableByID(id);
                }
            }
        }

        static string GetPluginId(string pluginStorableName)
        {
            return pluginStorableName.Substring(0, pluginStorableName.IndexOf('_'));
        }

        string GetPackagePath(string pluginUid)
        {
            string packageId = GetPackageId(pluginUid);
            return packageId == null ? "" : $"{packageId}:/";
        }

        string GetPackageId(string pluginUid)
        {
            string filename = manager.GetJSON()["plugins"][pluginUid].Value;
            int idx = filename.IndexOf(":/", StringComparison.Ordinal);
            return idx >= 0 ? filename.Substring(0, idx) : null;
        }

        static void LogError(string str)
        {
            SuperController.LogError($"{nameof(Naturalis)}.{nameof(AutoUpgrade)}: {str}");
        }

        static void LogMessage(string str)
        {
            SuperController.LogMessage($"{nameof(Naturalis)}.{nameof(AutoUpgrade)}: {str}");
        }
    }
}
