﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using Naturalis.BM;
using Naturalis.BM.Components;
using Naturalis.BM.Handlers;
using Naturalis.BM.UI;
using Naturalis.Common;
using Naturalis.Common.Components;
using Naturalis.Common.Handlers;
using Naturalis.Common.UI;
using SimpleJSON;
using UnityEngine;

namespace Naturalis
{
    sealed class BootyMagic : ScriptBase
    {
        public static BootyMagic script { get; private set; }
        public static Loggr logger { get; private set; }

        public override bool ShouldIgnore()
        {
            return false;
        }

        const string CALIBRATION_DATA_VERSION = "1.0.0";

#region *** UI Listeners ***

        protected override Action OnUIEnabled()
        {
            return UIEnabled;
        }

        void UIEnabled()
        {
            if(!Person.IsFemale())
            {
                return;
            }

            if(tabs.MainTabIsActive())
            {
                if(!tabs.NestedWindowIsActive())
                {
                    tabs.mainWindow.Refresh();
                }
            }
            else if(tabs.HardCollidersTabIsActive())
            {
                if(!enabled)
                {
                    tabs.GoToMainWindow();
                }

                if(HardColliderHandler.colliderVisualizer)
                {
                    HardColliderHandler.colliderVisualizer.ShowPreviewsJSON.val = true;
                }
            }
            else if(tabs.PhysicsTabIsActive())
            {
                if(!tabs.NestedWindowIsActive())
                {
                    if(SoftPhysicsHandler.softPhysicsSynced)
                    {
                        tabs.physicsWindow.Refresh(SoftPhysicsHandler.softPhysicsOn);
                    }
                }
                else if(tabs.GetActiveNestedWindow() is SoftParameterWindow)
                {
                    SoftPhysicsHandler.softJointDiagnostics.DrawServiceSetEnabled(true);
                    var softParameterWindow = (SoftParameterWindow) tabs.GetActiveNestedWindow();
                    softParameterWindow.Refresh();
                    softParameterWindow.SetShowColliderPreviews();
                }
            }
        }

        protected override Action OnUIDisabled()
        {
            return UIDisabled;
        }

        void UIDisabled()
        {
            if(!Person.IsFemale())
            {
                return;
            }

            if(Settings.instance.calibrateOnUIClosedJsb.val && calibrationHelper.pending)
            {
                StartCalibration();
            }

            if(HardColliderHandler.colliderVisualizer)
            {
                HardColliderHandler.colliderVisualizer.ShowPreviewsJSON.val = false;
            }

            if(SoftPhysicsHandler.colliderVisualizer)
            {
                SoftPhysicsHandler.colliderVisualizer.ShowPreviewsJSON.val = false;
            }

            tabs.activeTabWindow.ClosePopups();

            if(tabs.MainTabIsActive())
            {
                if(tabs.GetActiveNestedWindow() is DevWindow)
                {
                    SoftPhysicsHandler.softJointDiagnostics?.DrawServiceSetEnabled(false);
                    ((DevWindow) tabs.GetActiveNestedWindow()).ClearDiagnostics();
                }
                else if(tabs.GetActiveNestedWindow() is SettingsWindow)
                {
                    tabs.mainWindow.OnReturn();
                }
            }
            else if(tabs.PhysicsTabIsActive())
            {
                if(tabs.GetActiveNestedWindow() is SoftParameterWindow)
                {
                    SoftPhysicsHandler.softJointDiagnostics?.DrawServiceSetEnabled(false);
                }
            }
        }

#endregion

        Person _person;
        public Glute leftGlute { get; private set; }
        public Glute rightGlute { get; private set; }
        public PresetHandler presetHandler { get; private set; }
        MorphListener _gluteMorphListener;

        public StorableFloat softnessJsf { get; private set; }
        public StorableFloat quicknessJsf { get; private set; }

        public static float softness { get; private set; }
        public static float quickness { get; private set; }

        public static Bindings bindings { get; private set; } // Todo shared
        public CalibrationHelper calibrationHelper { get; private set; }
        public UIModManager uiModManager { get; private set; }
        public static Tabs tabs { get; private set; }

#region *** Init ***

        public static bool IsInitialized()
        {
            return script.initialized == true;
        }

        public override void Init()
        {
            script = this;
            logger = new Loggr(Constant.BOOTY_MAGIC);
            StartCoroutine(InitCo());
        }

        IEnumerator InitCo()
        {
            Settings.instance = gameObject.AddComponent<Settings>();
            Settings.instance.Setup(this);
            yield return WaitForSettingsInstance();

            if(Settings.instance.initialized == false)
            {
                yield break;
            }

            _person = Person.instance;

            if(!_person.isFemale)
            {
                var errorJss = new JSONStorableString("Info", $"<b>{Constant.BOOTY_MAGIC} is only supported on a female character.</b>");
                var textField = CreateTextField(errorJss);
                textField.backgroundColor = Color.clear;
                initialized = false;
                yield break;
            }

            /* Used to store version in save JSON and communicate version to other plugin instances */
            this.NewJSONStorableString(Constant.NATURALIS_VERSION, Settings.NATURALIS_VERSION);

            presetHandler = new PresetHandler(this, Constant.BOOTY_MAGIC);
            SetupHandlers();
            SetupGluteMorphListener();
            SetupStorables();
            SetupCallbacks();
            Integration.instance.Init();
            SetupCalibrationHelper();

            InitBindings(); // Might already be setup in OnBindingsListRequested.
            SuperController.singleton.BroadcastMessage("OnActionsProviderAvailable", this, SendMessageOptions.DontRequireReceiver);

            SetupUI();

            if(!restoringFromJson && enabled)
            {
                StartCalibration(true);
            }

            initialized = true;
        }

        void SetupHandlers()
        {
            MainPhysicsHandler.Init();
            SoftPhysicsHandler.Init();
            leftGlute = new Glute(Side.LEFT);
            rightGlute = new Glute(Side.RIGHT);
            GravityPhysicsHandler.Init();
            FrictionHandler.instance.Init();

            HardColliderHandler.Init();

            ForceMorphHandler.Init();
        }

        void SetupGluteMorphListener()
        {
            var indices = VertexIndices.VolumeCalculationVerts(Side.LEFT).Concat(VertexIndices.VolumeCalculationVerts(Side.RIGHT));
            _gluteMorphListener = new MorphListener(
                new HashSet<int>(indices),
                new HashSet<string>
                {
                    "FBMFitnessDetails",
                    "FBMHeight",
                    "NyssaH",
                    "Norma",
                    "Alessa",
                    "Eisa",
                    "Izarra",
                    "bellybulge",
                    "Athletic",
                    "Body Tone",
                    "Bodybuilder Details",
                    "Ribs Width",
                    "Distort 1",
                    "Distort 2",
                    "FBMBodyTone",
                    "FBMBodybuilderDetails",
                    "Center Gap Smooth",
                    "PBMSacralDimples",
                    "Glute Line Sharpness",
                    "G3F_ThighsInnerSlim",
                    "Inner Thigh Gap2",
                    "Inner Thigh Gap",
                    "!Bend Helper",
                    "Pubic Area Width",
                    "Centre Crease",
                    "WaistBellyIn",
                    "BackLessCurve",
                    "TeethFixer",
                    "Thigh inner contour",
                    "G3F_Hip Leg Space",
                    "Doggy Style Adjust",
                    "Glute crease",
                    "deepbulge_bot",
                },
                new FrequencyRunner(this.NewStorableFloat("Glute Size Changed Check Frequency", 3, 1, 10, false))
            );
            _gluteMorphListener.ProcessMorphs(_person.geometry.morphBank1);
        }

        void SetupStorables()
        {
            softnessJsf = this.NewStorableFloat("Glute Softness", 70f, 0f, 100f);
            softnessJsf.setCallbackFunction = _ => SetCalibrationPending(softnessJsf);

            quicknessJsf = this.NewStorableFloat("Glute Quickness Offset", 0.00f, -1.00f, 1.00f);
            quicknessJsf.setCallbackFunction = _ => SetCalibrationPending(quicknessJsf);

            this.NewJSONStorableAction(Constant.CALIBRATE, () => StartCalibration(true));
            this.NewJSONStorableAction(Constant.CALIBRATE_ON_NEXT_PENDING, () => calibrationHelper.calibrateOnNextPending = true);

            var syncSoftPhysicsJsb = this.NewStorableBool(
                Constant.INTERNAL_SYNC_SOFT_PHYSICS,
                UserPreferences.singleton.softPhysics &&
                containingAtom.GetStorableByID("SoftBodyPhysicsEnabler").GetBoolParamValue("enabled") &&
                SoftPhysicsHandler.gluteSoftPhysicsOnJsb.val
            );
            syncSoftPhysicsJsb.setCallbackFunction = OnSoftPhysicsToggled;
            this.NewJSONStorableAction(
                Constant.INTERNAL_ON_PHYSICS_RATE_CHANGED,
                () =>
                {
                    if(!calibrationHelper.pending && !calibrationHelper.calibrateOnNextPending)
                    {
                        logger.LogMessage("Physics rate changed - calibration set to pending.");
                    }

                    calibrationHelper.updateMassWhenCalibrating = true;
                    calibrationHelper.SetCalibrationPending(Constant.PHYSICS_RATE_CHANGED);
                }
            );
            this.NewJSONStorableAction(
                Constant.INTERNAL_ON_CHARACTER_CHANGED,
                () =>
                {
                    FrictionHandler.instance.Refresh();
                    uiModManager.SetFrictionUIVisibility();
                }
            );
            var unsetPendingJss = this.NewJSONStorableString(Constant.INTERNAL_UNSET_CALIBRATION_PENDING, "");
            unsetPendingJss.setJSONCallbackFunction = jss =>
            {
                calibrationHelper.SetCalibrationPending(jss.val, false);
                jss.valNoCallback = "";
            };
        }

        void SetupCallbacks()
        {
            _person.scaleJsf.setCallbackFunction += OnAtomScaleChanged;
            SuperController.singleton.onBeforeSceneSaveHandlers += OnBeforeSceneSave;
            SuperController.singleton.onSceneSavedHandlers += OnSceneSaved;
        }

        void SetupCalibrationHelper()
        {
            calibrationHelper = gameObject.AddComponent<CalibrationHelper>();
            calibrationHelper.Init(
                Constant.BOOTY_MAGIC,
                this.NewStorableBool(Constant.INTERNAL_IS_CALIBRATING, false),
                Settings.tittyMagic,
                Integration.instance.OtherCalibrationInProgress,
                () => StartCalibration()
            );
        }

        // https://github.com/vam-community/vam-plugins-interop-specs/blob/main/keybindings.md
        public void OnBindingsListRequested(List<object> bindingsList)
        {
            InitBindings(); // Might already be setup in Init.
            bindingsList.Add(bindings.@namespace);
            bindingsList.AddRange(bindings.GetActionsList());
        }

        void InitBindings()
        {
            if(bindings)
            {
                return;
            }

            bindings = gameObject.AddComponent<Bindings>();
            bindings.Init(this, nameof(BootyMagic));
        }

        void SetupUI()
        {
            uiModManager = gameObject.AddComponent<UIModManager>();
            uiModManager.ModifyAtomUI();

            tabs = new Tabs
            {
                mainWindow = new MainWindow(),
                physicsWindow = new PhysicsWindow(),
                morphsWindow = new MorphsWindow(),
                hardCollidersWindow = new HardCollidersWindow(),
            };
            tabs.CreateNavigationButtons();
            tabs.GoToMainWindow();
        }

#endregion

#region *** Update ***

        void Update()
        {
            if(initialized != true)
            {
                return;
            }

            try
            {
                if(!updatePaused && _gluteMorphListener.ChangeDetected())
                {
                    calibrationHelper.updateMassWhenCalibrating = true;
                    calibrationHelper.SetCalibrationPending(Constant.SIZE_CHANGED);
                }

                var nestedWindow = tabs.GetActiveNestedWindow();
                if(nestedWindow is ParameterWindow && pluginUIEventsListener.isEnabled)
                {
                    ((ParameterWindow) nestedWindow).UpdateDiagnostics();
                }
                else if(nestedWindow is SoftParameterWindow && pluginUIEventsListener.isEnabled)
                {
                    ((SoftParameterWindow) nestedWindow).UpdateDiagnostics();
                }
                else if(nestedWindow is DevWindow)
                {
                    (nestedWindow as DevWindow).Update();
                }
            }
            catch(Exception e)
            {
                logger.LogError($"{nameof(Update)} error: {e}");
            }
        }

        public static bool updatePaused { get; private set; }

        void FixedUpdate()
        {
            if(initialized != true ||
                restoringFromJson ||
                !_person.skin ||
                containingAtom.IsFreezeGrabbing()
            )
            {
                return;
            }

            if(updatePaused)
            {
                if(!string.IsNullOrEmpty(_gravitySimulationDirection))
                {
                    SimulateGravity(_gravitySimulationDirection);
                }

                return;
            }

            try
            {
                leftGlute.Measure();
                rightGlute.Measure();

                HardColliderHandler.UpdateDistanceDiffs();
                GravityEffectCalc.CalculateAngles();

                HardColliderHandler.UpdateFriction();
                GravityPhysicsHandler.Update();

                if(!Settings.instance.MorphUpdatesPaused())
                {
                    ForceMorphHandler.Update();
                }

                SoftPhysicsHandler.softJointsUpdater.Run();
            }
            catch(Exception e)
            {
                logger.LogError($"{nameof(FixedUpdate)} error: {e}");
                updatePaused = true;
            }
        }

        bool _updatingMass;
        string _gravitySimulationDirection;

        void SimulateGravity(string direction)
        {
            GravityEffectCalc.OverrideAngles(direction);
            GravityPhysicsHandler.ApplyGravitySimulationForce(direction);

            if(!_updatingMass)
            {
                if(direction != Direction.DOWN)
                {
                    leftGlute.Measure();
                    rightGlute.Measure();
                }

                GravityPhysicsHandler.Update(true);
                ForceMorphHandler.Update();
            }
        }

#endregion

#region *** Calibration ***

        public void SetCalibrationPending(IStorable storable)
        {
            if(restoringFromJson)
            {
                return;
            }

            calibrationHelper.SetCalibrationPending(storable.Name(), !storable.IsCalibrated());
        }

        public void StartCalibration(bool forceUpdateMass = false)
        {
            if(restoringFromJson || _inCalibrationCo)
            {
                return;
            }

            if(!enabled)
            {
                logger.LogMessage($"Enable {Constant.BOOTY_MAGIC} to calibrate physics and morphs.");
                return;
            }

            if(forceUpdateMass)
            {
                calibrationHelper.updateMassWhenCalibrating = true;
            }

            _calibrationCo = StartCoroutine(CalibrationCo());
        }

        bool _inCalibrationCo;
        bool _firstCalibrationDone;
        Coroutine _calibrationCo;

        IEnumerator CalibrationCo(JSONClass calibrationData = null)
        {
            _inCalibrationCo = true;
            updatePaused = true;

            /* Ensure everything is ready for calibration to proceed in case of e.g. appearance preset load */
            {
                /* Ensures glute soft physics enabled value is correct on first calibration when soft physics enabled is synced to SoftPhysicsHandler */
                while(!Settings.personMonitor.initialized)
                {
                    yield return null;
                }

                /* Can occur when loading a look and VaM pauses to load assets */
                while(!_person.SelectedCharacterIsReady())
                {
                    yield return null;
                }

                /* Must be unfrozen before checking for hard collider rigidbodies */
                yield return CalibrationHelper.WaitForAtomPhysicsUnfrozen(containingAtom);

                Settings.personMonitor.Check();
                Settings.personMonitor.enabled = false;

                HardColliderHandler.ConfigureAndReturnAutoColliders();

                /* Sanity check. Might occur (?) even after hard colliders are enabled by settings monitor */
                float timeout = Time.unscaledTime + 5;
                while(!HardColliderHandler.RigidbodiesFound() && Time.unscaledTime < timeout)
                {
                    yield return null;
                }

                if(!HardColliderHandler.RigidbodiesFound())
                {
                    logger.LogError("Calibration error: collider rigidbodies not found after 5 seconds of waiting.");
                    yield break;
                }
            }

            bool doFullCalibration = false;
            if(calibrationData != null)
            {
                try
                {
                    CalibrateFromJson(calibrationData);
                }
                catch(Exception e)
                {
                    doFullCalibration = true;
                    logger.LogError($"Error loading calibration data from JSON, continuing with full calibration. Error: {e}");
                }
            }
            else
            {
                doFullCalibration = true;
            }

            if(doFullCalibration)
            {
                yield return FullCalibrationCo();
            }

            SaveCalibratedValues();
            calibrationHelper.updateMassWhenCalibrating = false;
            updatePaused = false;
            _inCalibrationCo = false;
            _firstCalibrationDone = true;

            /* Update UI */
            if(tabs.MainTabIsActive())
            {
                if(!tabs.NestedWindowIsActive())
                {
                    tabs.mainWindow.Refresh();
                }
            }
            else if(tabs.PhysicsTabIsActive())
            {
                if(!tabs.NestedWindowIsActive())
                {
                    tabs.physicsWindow.Refresh(SoftPhysicsHandler.softPhysicsOn);
                }
                else if(tabs.GetActiveNestedWindow() is SoftParameterWindow)
                {
                    ((SoftParameterWindow) tabs.GetActiveNestedWindow()).Refresh();
                }
            }
        }

        void CalibrateFromJson(JSONClass jc)
        {
            SoftPhysicsHandler.SyncSoftPhysicsOn();
            MainPhysicsHandler.UpdateMassValueAndAmounts(jc["mass"].AsFloat);
            softness = CalculateSoftness();
            quickness = quicknessJsf.val;

            GravityPhysicsHandler.AutoCalculateGravityCounterforcePct(false);
            ForceMorphHandler.SetMultipliers();
            HardColliderHandler.UpdateFrictionSizeMultipliers();

            MainPhysicsHandler.UpdatePhysics();
            SoftPhysicsHandler.UpdatePhysics();
            FrictionHandler.instance.CalculateFriction();

            leftGlute.CalibrateFromJson(jc["left_glute"].AsObject);
            rightGlute.CalibrateFromJson(jc["right_glute"].AsObject);
            SoftPhysicsHandler.CalibrateFromJson(jc["soft_joints"].AsObject);
            HardColliderHandler.SyncCollidersMass();
            HardColliderHandler.CalibrateFromJson(jc["hard_colliders"].AsObject);
            HardColliderHandler.SyncAllOffsets();

            calibrationHelper.EnableSettingsMonitor();
        }

        IEnumerator FullCalibrationCo()
        {
            /* Must be called before calibration pose is applied because the current pose must be stored without duplicate morphs having >0 values. */
            ZeroDynamicAdjustments();
            _person.ZeroDuplicateBootyMagicMorphs();

            yield return calibrationHelper.Begin();
            if(calibrationHelper.cancelling)
            {
                calibrationHelper.cancelling = false;
                yield break;
            }

            while(_gluteMorphListener.MorphValueChanged())
            {
                yield return new WaitForSeconds(0.1f);
            }

            SoftPhysicsHandler.SyncSoftPhysicsOn();
            _person.SetGlutesUseGravity(false);

            {
                _gravitySimulationDirection = Direction.DOWN;
                yield return UpdateMassAndPhysics();
                MainPhysicsHandler.SetGravitySimulationDamper();
                yield return new WaitForSeconds(0.25f);

                float minTime = Time.time + 0.40f;
                float maxTime = Time.time + 1.00f;
                const float tolerance = 0.003f;
                while(
                    Time.time < minTime ||
                    (leftGlute.Moved(tolerance) || rightGlute.Moved(tolerance)) &&
                    Time.time < maxTime
                )
                {
                    yield return null;
                }

                leftGlute.Calibrate();
                rightGlute.Calibrate();
                if(SoftPhysicsHandler.LdpUpdateIsEnabled())
                {
                    SoftPhysicsHandler.CalibrateSoftJointsDistortion(Direction.DOWN);
                }

                HardColliderHandler.SyncCollidersMass();
                HardColliderHandler.CalibrateColliders();
                HardColliderHandler.SyncAllOffsets();
                _gravitySimulationDirection = null;
            }

            if(SoftPhysicsHandler.LdpUpdateIsEnabled())
            {
                foreach(string direction in new[] { Direction.FORWARD, Direction.UP, Direction.LEFT, Direction.RIGHT, Direction.BACK })
                {
                    _gravitySimulationDirection = direction;
                    yield return new WaitForSeconds(0.25f);

                    float timeout = Time.time + 1.50f;
                    while(SoftPhysicsHandler.SoftJointsMoved(0.008f) && Time.time < timeout)
                    {
                        yield return null;
                    }

                    SoftPhysicsHandler.CalibrateSoftJointsDistortion(direction);
                    _gravitySimulationDirection = null;
                }
            }

            _person.SetGlutesUseGravity(true);
            MainPhysicsHandler.UnsetGravitySimulationDamper();

            yield return calibrationHelper.FinishCo();
        }

        void ZeroDynamicAdjustments()
        {
            leftGlute.ResetMeasurements();
            rightGlute.ResetMeasurements();
            SoftPhysicsHandler.ResetSoftJointsDistortion();
            HardColliderHandler.ResetDistanceDiffs();
            GravityPhysicsHandler.ResetAll();
            ForceMorphHandler.ResetAll();
        }

        float CalculateSoftness()
        {
            return Calc.SoftnessBaseCurve(softnessJsf.val / 100f);
        }

        IEnumerator UpdateMassAndPhysics()
        {
            _updatingMass = true;

            if(calibrationHelper.updateMassWhenCalibrating)
            {
                softness = 0;
                quickness = 0;
                MainPhysicsHandler.UpdatePhysics(false);
                SoftPhysicsHandler.UpdatePhysics(false);

                yield return new WaitForSeconds(0.25f);
                /* Mass is calculated multiple times because each new mass value changes the exact glute
                 * shape and therefore the estimated volume.
                 */
                MainPhysicsHandler.CalculateMass();
                yield return Utils.WaitAndRepeat(
                    MainPhysicsHandler.CalculateMass,
                    (int) Mathf.Lerp(4, 12, MainPhysicsHandler.normalizedRealMass),
                    0.13f
                );
            }

            softness = CalculateSoftness();
            quickness = quicknessJsf.val;

            GravityPhysicsHandler.AutoCalculateGravityCounterforcePct(false);
            ForceMorphHandler.SetMultipliers();
            HardColliderHandler.UpdateFrictionSizeMultipliers();
            MainPhysicsHandler.UpdatePhysics();
            SoftPhysicsHandler.UpdatePhysics();
            FrictionHandler.instance.CalculateFriction();
            _updatingMass = false;
        }

        void SaveCalibratedValues()
        {
            softnessJsf.SaveCalibrated();
            quicknessJsf.SaveCalibrated();
            ForceMorphHandler.SaveCalibrated();
            MainPhysicsHandler.SaveCalibrated();
            GravityPhysicsHandler.SaveCalibrated();
            SoftPhysicsHandler.SaveCalibrated();
        }

#endregion

        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            var jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
            EnsureStoreDynamicDefaults(jc);
            Utils.RemoveInternalKeys(jc);

            if(_savingScene && _firstCalibrationDone && !calibrationHelper.pending)
            {
                var calibrationData = new JSONClass
                {
                    ["version"] = CALIBRATION_DATA_VERSION,
                    ["mass"] = { AsFloat = MainPhysicsHandler.calibratedMass },
                    ["left_glute"] = leftGlute.SerializeCalibrationData(),
                    ["right_glute"] = rightGlute.SerializeCalibrationData(),
                    ["soft_joints"] = SoftPhysicsHandler.SerializeCalibrationData(),
                    ["hard_colliders"] = HardColliderHandler.SerializeCalibrationData(),
                };
                jc["calibration_data"] = calibrationData;
            }

            needsStore = true;
            return jc;
        }

        /* When default value is changed at runtime, a value matching the changed default value will not save to JSON unless the default value is reset. */
        static void EnsureStoreDynamicDefaults(JSONClass jc)
        {
            jc[GravityPhysicsHandler.gravityCounterforceJsf.name].AsFloat = GravityPhysicsHandler.gravityCounterforceJsf.val;
        }

        public static bool restoringFromJson { get; private set; }

        public override void RestoreFromJSON(
            JSONClass jc,
            bool restorePhysical = true,
            bool restoreAppearance = true,
            JSONArray presetAtoms = null,
            bool setMissingToDefault = true
        )
        {
            restoringFromJson = true;

            /* Disable early to allow correct enabled value to be used during Init */
            if(jc.HasKey("enabled") && !jc["enabled"].AsBool)
            {
                enabled = false;
            }

            /* Prevent overriding versionJss.val from JSON. Version stored in JSON just for information,
             * but could be intercepted here and used to save a "loadedFromVersion" value.
             */
            if(jc.HasKey(Constant.NATURALIS_VERSION))
            {
                jc[Constant.NATURALIS_VERSION] = Settings.NATURALIS_VERSION;
            }

            /* Naturalis v1.0.0 -> v1.0.1 fixes */
            JSONUtils.RenameKey(jc, "In/Out Damper Back Gravity", "In/Out Spring Back Gravity");
            JSONUtils.RenameKey(jc, "In/Out Damper Forward Gravity", "In/Out Spring Forward Gravity");

            StartCoroutine(
                RestoreFromJSONCo(
                    jc,
                    restorePhysical,
                    restoreAppearance,
                    presetAtoms,
                    setMissingToDefault
                )
            );
        }

        IEnumerator RestoreFromJSONCo(
            JSONClass jc,
            bool restorePhysical,
            bool restoreAppearance,
            JSONArray presetAtoms,
            bool setMissingToDefault
        )
        {
            while(initialized == null)
            {
                yield return null;
            }

            if(initialized == false)
            {
                yield break;
            }

            ConvertColliderParamNames(jc);
            base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);

            restoringFromJson = false;
            if(enabled)
            {
                JSONClass calibrationData = null;
                if(jc.HasKey("calibration_data"))
                {
                    var tmpData = jc["calibration_data"].AsObject;
                    if(
                        tmpData.HasKey("version") &&
                        new SemVer(tmpData["version"].Value).IsMajorMinorEqual(CALIBRATION_DATA_VERSION)
                    )
                    {
                        calibrationData = tmpData;
                    }
                }

                if(calibrationData != null)
                {
                    StartCoroutine(CalibrationCo(calibrationData));
                }
                else
                {
                    StartCalibration(true);
                }
            }
        }

        public override void RestoreFromPresetJSON(JSONClass jc)
        {
            if(initialized != true || restoringFromJson)
            {
                logger.LogMessage("Preset file must be loaded after initialization.");
                return;
            }

            ConvertColliderParamNames(jc);
            base.RestoreFromJSON(jc);

            if(Settings.instance.calibrateOnPresetLoadedJsb.val)
            {
                StartCalibration(true);
            }
        }

        /* Naturalis v1.0 beta 4 and earlier -> Naturalis v1.0 */
        static void ConvertColliderParamNames(JSONClass jc)
        {
            string[] colliderIds =
            {
                "Glute1",
                "Glute1 (1)",
                "Glute1 (2)",
                "Glute1 (3)",
                "Glute1 (4)",
                "Glute1 (5)",
                "Glute1 (6)",
                "Glute1 (7)",
            };

            string[] paramNames =
            {
                Constant.MASS_MULTIPLIER,
                Constant.RADIUS_OFFSET,
                Constant.X_OFFSET,
                Constant.Y_OFFSET,
                Constant.Z_OFFSET,
            };

            foreach(string colliderId in colliderIds)
            {
                foreach(string paramName in paramNames)
                {
                    string key = colliderId.ToLower() + paramName;
                    if(jc.HasKey(key))
                    {
                        float value = jc[key].AsFloat;
                        jc.Remove(key);
                        jc[$"{colliderId} {paramName}"].AsFloat = value;
                    }
                }
            }
        }

#region Callbacks

        void OnAtomScaleChanged(float _)
        {
            if(!enabled)
            {
                return;
            }

            if(!calibrationHelper.pending && !calibrationHelper.calibrateOnNextPending)
            {
                logger.LogMessage("Person atom scale changed - calibration set to pending.");
            }

            calibrationHelper.updateMassWhenCalibrating = true;
            calibrationHelper.SetCalibrationPending(Constant.SIZE_CHANGED);
        }

        static bool _savingScene;

        static void OnBeforeSceneSave()
        {
            _savingScene = true;
            updatePaused = true;
            ForceMorphHandler.ResetAll();
        }

        void OnSceneSaved()
        {
            _savingScene = false;
            if(!calibrationHelper.calibratingJsb.val)
            {
                updatePaused = false;
            }
        }

        void OnSoftPhysicsToggled(bool value)
        {
            SoftPhysicsHandler.ReverseSyncSoftPhysicsOn();
            if(!calibrationHelper.calibrateOnNextPending)
            {
                logger.LogMessage($"Soft physics {(value ? "enabled" : "disabled")} - updates paused, calibration set to pending.");
            }

            updatePaused = true;
            calibrationHelper.updateMassWhenCalibrating = true;
            calibrationHelper.SetCalibrationPending(Constant.SOFT_PHYSICS_CHANGED);
            SoftPhysicsHandler.softPhysicsSynced = false;

            if(tabs.PhysicsTabIsActive())
            {
                if(tabs.GetActiveNestedWindow() is SoftParameterWindow)
                {
                    if(!value)
                    {
                        tabs.physicsWindow.OnReturn();
                    }
                }
                else if(!value)
                {
                    tabs.physicsWindow.Refresh(false);
                }
            }

            if(!value)
            {
                SoftPhysicsHandler.softJointDiagnostics.visualizeLdpJsb.val = false;
            }
        }

#endregion

        public RectTransform GetLeftUIContent()
        {
            return leftUIContent;
        }

        public RectTransform GetRightUIContent()
        {
            return rightUIContent;
        }

        void OnEnable()
        {
            if(initialized != true)
            {
                return;
            }

            try
            {
                Settings.personMonitor.Check();
                _person.EnableUseColliders();
                HardColliderHandler.EnableMultiplyFriction();
                if(HardColliderHandler.colliderVisualizer)
                {
                    HardColliderHandler.colliderVisualizer.enabled = true;
                }

                SoftPhysicsHandler.EnableMultiplyFriction();
                if(SoftPhysicsHandler.colliderVisualizer)
                {
                    SoftPhysicsHandler.colliderVisualizer.enabled = true;
                }

                SoftPhysicsHandler.softJointDiagnostics?.DrawServiceSetEnabled(true);
                StartCalibration(true);
                uiModManager.enabled = true;
            }
            catch(Exception e)
            {
                logger.LogError($"{nameof(OnEnable)} error: {e}");
            }
        }

        void OnDisable()
        {
            if(initialized != true)
            {
                return;
            }

            try
            {
                if(_calibrationCo != null)
                {
                    StopCoroutine(_calibrationCo);
                }

                HardColliderHandler.RestoreOriginalPhysics();
                if(HardColliderHandler.colliderVisualizer)
                {
                    HardColliderHandler.colliderVisualizer.enabled = false;
                }

                MainPhysicsHandler.RestoreOriginalPhysics();
                SoftPhysicsHandler.RestoreOriginalPhysics();
                if(SoftPhysicsHandler.colliderVisualizer)
                {
                    SoftPhysicsHandler.colliderVisualizer.enabled = false;
                }

                SoftPhysicsHandler.softJointDiagnostics?.DrawServiceSetEnabled(false);
                uiModManager.enabled = false;

                calibrationHelper.Finish(Settings.tittyMagic.IsEnabledNullSafe());
                ForceMorphHandler.ResetAll();
            }
            catch(Exception e)
            {
                logger.LogError($"{nameof(OnDisable)} error: {e}");
            }
        }

        new void OnDestroy()
        {
            try
            {
                base.OnDestroy();
                Destroy(calibrationHelper);
                Destroy(bindings);

                /* Nullify static reference fields to let GC collect unreachable instances */
                ForceMorphHandler.Destroy();
                FrictionHandler.Destroy();
                GravityPhysicsHandler.Destroy();
                HardColliderHandler.Destroy();
                MainPhysicsHandler.Destroy();
                SoftPhysicsHandler.Destroy();

                script = null;
                tabs = null;
                logger = null;
                bindings = null;

                if(Person.instance?.scaleJsf != null)
                {
                    Person.instance.scaleJsf.setCallbackFunction -= OnAtomScaleChanged;
                }

                Destroy(uiModManager);
                Integration.Destroy();
                CalibrationHelper.Destroy();

                SuperController.singleton.onSceneSavedHandlers -= OnSceneSaved;
                SuperController.singleton.onBeforeSceneSaveHandlers -= OnBeforeSceneSave;
                SuperController.singleton.BroadcastMessage("OnActionsProviderDestroyed", this, SendMessageOptions.DontRequireReceiver);
            }
            catch(Exception e)
            {
                if(initialized == true)
                {
                    SuperController.LogError($"{nameof(OnDestroy)} error: {e}");
                }
                else
                {
                    Debug.LogError($"{nameof(OnDestroy)} error: {e}");
                }
            }
        }
    }
}
