using System;
using Naturalis.Common;
using UnityEngine;

namespace Naturalis.BM.Handlers
{
    static class GravityEffectCalc
    {
        public static float gluteRollL { get; private set; }
        public static float glutePitchL { get; private set; }
        public static float gluteRollR { get; private set; }
        public static float glutePitchR { get; private set; }
        public static float hipRoll { get; private set; }
        public static float hipPitch { get; private set; }
        public static float hipPitch2X { get; private set; }

        public static void CalculateAngles()
        {
            var lGluteRotation = Person.GetGluteRotation(Side.LEFT);
            var rGluteRotation = Person.GetGluteRotation(Side.RIGHT);
            var hipRotation = Person.GetHipRotation();

            gluteRollL = Roll(lGluteRotation);
            glutePitchL = Pitch(lGluteRotation);
            gluteRollR = Roll(rGluteRotation);
            glutePitchR = Pitch(rGluteRotation);
            hipRoll = Roll(hipRotation);
            hipPitch = Pitch(hipRotation);
            hipPitch2X = 2 * hipPitch;
        }

        static float Roll(Quaternion q)
        {
            return 2 * InverseLerpToPi(Mathf.Asin(2 * q.x * q.y + 2 * q.z * q.w));
        }

        static float Pitch(Quaternion q)
        {
            return InverseLerpToPi(Mathf.Atan2(2 * q.x * q.w - 2 * q.y * q.z, 1 - 2 * q.x * q.x - 2 * q.z * q.z));
        }

        static float InverseLerpToPi(float val)
        {
            return val > 0
                ? Mathf.InverseLerp(0, Mathf.PI, val)
                : -Mathf.InverseLerp(0, Mathf.PI, -val);
        }

        // ReSharper disable once UnusedMember.Global
        public static float UpEffect(
            float pitch,
            float roll,
            float multiplier,
            float additionalRollEffect
        )
        {
            float effect = Mathf.Abs(pitch) * RollMultiplier(roll) / 2;
            return (effect + additionalRollEffect) * multiplier / 2;
        }

        // ReSharper disable once UnusedMember.Global
        public static float DownEffect(float pitch, float roll, float multiplier)
        {
            return (2 - Mathf.Abs(pitch) / 2) * RollMultiplier(roll) * multiplier / 2;
        }

        public static float DepthAdjustByAngle(float pitch)
        {
            // leaning forward
            if(pitch >= 0)
            {
                // upright
                if(pitch < 1)
                {
                    return pitch;
                }

                // upside down
                return 2 - pitch;
            }

            // leaning back
            // upright
            if(pitch >= -1)
            {
                return -pitch;
            }

            // upside down
            return 2 + pitch;
        }

        public static float UpDownAdjustByAngle(float pitch)
        {
            // leaning forward
            if(pitch >= 0)
            {
                // upright
                if(pitch < 1)
                {
                    return 1 - pitch;
                }

                // upside down
                return pitch - 1;
            }

            // leaning back
            // upright
            if(pitch >= -1)
            {
                return 1 + pitch;
            }

            // upside down
            return -pitch - 1;
        }

        public static float RollMultiplier(float roll)
        {
            return 1 - Mathf.Abs(roll);
        }

        public static void OverrideAngles(string direction)
        {
            switch(direction)
            {
                case Direction.UP:
                {
                    hipPitch = 2;
                    glutePitchL = 2;
                    glutePitchR = 2;
                    hipRoll = 0;
                    gluteRollL = 0;
                    gluteRollR = 0;
                    break;
                }
                case Direction.DOWN:
                {
                    hipPitch = 0;
                    glutePitchL = 0;
                    glutePitchR = 0;
                    hipRoll = 0;
                    gluteRollL = 0;
                    gluteRollR = 0;
                    break;
                }
                case Direction.FORWARD:
                {
                    hipPitch = 1.0f;
                    glutePitchL = 1.0f;
                    glutePitchR = 1.0f;
                    hipRoll = 0;
                    gluteRollL = -0.2f;
                    gluteRollR = 0.2f;
                    break;
                }
                case Direction.BACK:
                {
                    hipPitch = -1.0f;
                    glutePitchL = -1.0f;
                    glutePitchR = -1.0f;
                    hipRoll = 0;
                    gluteRollL = 0.25f;
                    gluteRollR = -0.25f;
                    break;
                }
                case Direction.LEFT:
                {
                    hipPitch = 2;
                    glutePitchL = 2;
                    glutePitchR = 2;
                    hipRoll = 1;
                    gluteRollL = 0.65f;
                    gluteRollR = 0.90f;
                    break;
                }
                case Direction.RIGHT:
                {
                    hipPitch = 2;
                    glutePitchL = 2;
                    glutePitchR = 2;
                    hipRoll = -1;
                    gluteRollL = -0.90f;
                    gluteRollR = -0.65f;
                    break;
                }
                default: throw new ArgumentException($"Invalid direction {direction}");
            }
        }
    }
}
