﻿using System.Collections.Generic;
using Naturalis.BM.UI;
using Naturalis.Common;
using Naturalis.Common.Components;
using Naturalis.Common.Handlers.Configs;
using UnityEngine;
using static Direction;

namespace Naturalis.BM.Handlers
{
    static class GravityPhysicsHandler
    {
        public static StorableFloat gravityCounterforceJsf { get; private set; }
        public static JSONStorableBool autoAdjustGravityCounterforceJsb { get; private set; }

        static bool _useGravityCounterforce;
        static Vector3 _gravityCounterForce;

        public static void Init()
        {
            gravityCounterforceJsf = BootyMagic.script.NewStorableFloat("Gravity Counterforce (%)", 0f, 0f, 100f);
            gravityCounterforceJsf.setCallbackFunction = OnGravityCounterforceAdjusted;
            autoAdjustGravityCounterforceJsb = BootyMagic.script.NewStorableBool("Auto-Adjust Gravity Counterforce", true);
            autoAdjustGravityCounterforceJsb.setCallbackFunction = OnAutoAdjustGravityCounterforceToggled;
            LoadSettings();
        }

        static void OnGravityCounterforceAdjusted(float value)
        {
            // will still change value if called via trigger, but will not trigger calculation
            if(!autoAdjustGravityCounterforceJsb.val)
            {
                CalculateGravityCounterForce(value);
                BootyMagic.script.SetCalibrationPending(gravityCounterforceJsf);
            }
        }

        static void CalculateGravityCounterForce(float value)
        {
            int counterForcePercent = Mathf.RoundToInt(value);
            gravityCounterforceJsf.valNoCallback = counterForcePercent;
            if(counterForcePercent > 0)
            {
                _useGravityCounterforce = true;
                _gravityCounterForce = -Physics.gravity * counterForcePercent / 100;
            }
            else
            {
                _useGravityCounterforce = false;
                _gravityCounterForce = Vector3.zero;
            }
        }

        static void OnAutoAdjustGravityCounterforceToggled(bool value)
        {
            if(BootyMagic.restoringFromJson)
            {
                return;
            }

            if(value)
            {
                AutoCalculateGravityCounterforcePct();
            }

            var parameterWindow = BootyMagic.tabs.GetActiveNestedWindow() as ParameterWindow;
            if(parameterWindow != null && parameterWindow.GetId() == GLUTE_JOINT_MASS.NAME)
            {
                parameterWindow.Refresh();
            }
        }

        public static void AutoCalculateGravityCounterforcePct(bool setPending = true)
        {
            float calculatedValue = Mathf.Lerp(50, 90, 1 - MainPhysicsHandler.normalizedMass);
            float rounded = Mathf.Round(calculatedValue);
            gravityCounterforceJsf.defaultVal = rounded;
            if(autoAdjustGravityCounterforceJsb.val)
            {
                CalculateGravityCounterForce(calculatedValue);
                if(setPending)
                {
                    BootyMagic.script.SetCalibrationPending(gravityCounterforceJsf);
                }
            }
        }

        static void LoadSettings()
        {
            var spring = MainPhysicsHandler.spring;
            var inOutSpring = MainPhysicsHandler.inOutSpring;
            var upDownAngleTarget = MainPhysicsHandler.upDownAngleTarget;
            var rightLeftAngleTarget = MainPhysicsHandler.rightLeftAngleTarget;

            spring.SetGravityCalculators(SpringCalculators());
            inOutSpring.SetGravityCalculators(InOutSpringCalculators());
            upDownAngleTarget.SetGravityCalculators(UpDownAngleTargetCalculators());
            rightLeftAngleTarget.SetGravityCalculators(RightLeftAngleTargetCalculators());

            spring.SetGravityMultiplierCallbacks();
            inOutSpring.SetGravityMultiplierCallbacks();
            upDownAngleTarget.SetGravityMultiplierCallbacks();
            rightLeftAngleTarget.SetGravityMultiplierCallbacks();

            spring.gravityInfoText =
                "Spring decreases when leaning in any direction. This makes glutes move" +
                " and morph more easily due to gravity.";
            inOutSpring.gravityInfoText =
                "In/Out Spring decreases when leaning forward or back in order to make depth" +
                " morphing take effect more easily due to gravity.";
            upDownAngleTarget.gravityInfoText =
                "Up/Down Angle Target is negative when upright (pulling glutes down)," +
                " and positive when upside down (pulling glutes up).";
            rightLeftAngleTarget.gravityInfoText =
                "Right/Left Angle Target is positive when leaning to the right and negative" +
                " when leaning to the left (the other way around for the right glute).";
        }

        static PhysicsCalculator SpringCalculator(string side, string direction)
        {
            return new PhysicsCalculator
            {
                multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(SPRING.NAME, side), direction, side == Side.LEFT),
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 0,
                    multiplierA = -2.5f,
                    multiplierB = -7.5f,
                    curveA = x => 1 - x,
                },
                secondaryConfig = new MultiplierConfig
                {
                    baseValue = 0,
                    multiplierA = -2.5f,
                    multiplierB = -7.5f,
                    curveA = x => 1 - x,
                },
            };
        }

        static PhysicsCalculator ForwardSpringCalculator(string side)
        {
            return new PhysicsCalculator
            {
                multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(SPRING.NAME, side), FORWARD, side == Side.LEFT),
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 0,
                    multiplierA = 15f,
                    multiplierB = 5f,
                    curveA = x => 1 - x,
                },
                secondaryConfig = new MultiplierConfig
                {
                    baseValue = 0,
                    multiplierA = 15f,
                    multiplierB = 5f,
                    curveA = x => 1 - x,
                },
            };
        }

        static Dictionary<string, Dictionary<string, PhysicsCalculator>> SpringCalculators()
        {
            return new Dictionary<string, Dictionary<string, PhysicsCalculator>>
            {
                {
                    Side.LEFT, new Dictionary<string, PhysicsCalculator>
                    {
                        { UP, SpringCalculator(Side.LEFT, UP) },
                        { BACK, SpringCalculator(Side.LEFT, BACK) },
                        { FORWARD, ForwardSpringCalculator(Side.LEFT) },
                        { SIDEWAYS_IN, SpringCalculator(Side.LEFT, SIDEWAYS_IN) },
                        { SIDEWAYS_OUT, SpringCalculator(Side.LEFT, SIDEWAYS_OUT) },
                    }
                },
                {
                    Side.RIGHT, new Dictionary<string, PhysicsCalculator>
                    {
                        { UP, SpringCalculator(Side.RIGHT, UP) },
                        { BACK, SpringCalculator(Side.RIGHT, BACK) },
                        { FORWARD, ForwardSpringCalculator(Side.RIGHT) },
                        { SIDEWAYS_IN, SpringCalculator(Side.RIGHT, SIDEWAYS_IN) },
                        { SIDEWAYS_OUT, SpringCalculator(Side.RIGHT, SIDEWAYS_OUT) },
                    }
                },
            };
        }

        static PhysicsCalculator InOutSpringCalculator(string side, string direction)
        {
            return new PhysicsCalculator
            {
                multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(IN_OUT_SPRING.NAME, side), direction, side == Side.LEFT),
                primaryConfig = new MultiplierConfig
                {
                    baseValue = -50f,
                    multiplierA = -100f,
                    multiplierB = -125f,
                    curveA = x => Calc.Exponential1(1 - x, 0.3f, 10f, 0.84f, 1.01f),
                    curveB = Calc.InOutSpringSoftnessCurve,
                },
                secondaryConfig = new MultiplierConfig
                {
                    baseValue = -50f,
                    multiplierA = -100f,
                    multiplierB = -125f,
                    curveA = x => Calc.Exponential1(1 - x, 0.3f, 10f, 0.84f, 1.01f),
                    curveB = Calc.InOutSpringSoftnessCurve,
                },
            };
        }

        static Dictionary<string, Dictionary<string, PhysicsCalculator>> InOutSpringCalculators()
        {
            return new Dictionary<string, Dictionary<string, PhysicsCalculator>>
            {
                {
                    Side.LEFT, new Dictionary<string, PhysicsCalculator>
                    {
                        { BACK, InOutSpringCalculator(Side.LEFT, BACK) },
                        { FORWARD, InOutSpringCalculator(Side.LEFT, FORWARD) },
                    }
                },
                {
                    Side.RIGHT, new Dictionary<string, PhysicsCalculator>
                    {
                        { BACK, InOutSpringCalculator(Side.RIGHT, BACK) },
                        { FORWARD, InOutSpringCalculator(Side.RIGHT, FORWARD) },
                    }
                },
            };
        }

        static PhysicsCalculator UpAngleTargetCalculator(string side)
        {
            return new PhysicsCalculator
            {
                multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(UP_DOWN_ANGLE_TARGET.NAME, side), UP, side == Side.LEFT),
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 2.00f,
                    multiplierA = 8.00f,
                    multiplierB = 4.00f,
                    curveB = Calc.AngleTargetSoftnessCurve,
                },
                secondaryConfig = new MultiplierConfig
                {
                    baseValue = 2.00f,
                    multiplierA = 8.00f,
                    multiplierB = 4.00f,
                    curveB = Calc.AngleTargetSoftnessCurve,
                },
            };
        }

        static PhysicsCalculator DownAngleTargetCalculator(string side)
        {
            return new PhysicsCalculator
            {
                multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(UP_DOWN_ANGLE_TARGET.NAME, side), DOWN, side == Side.LEFT),
                primaryConfig = new MultiplierConfig
                {
                    baseValue = -2.00f,
                    multiplierA = -8.00f,
                    multiplierB = -4.00f,
                    curveB = Calc.AngleTargetSoftnessCurve,
                },
                secondaryConfig = new MultiplierConfig
                {
                    baseValue = -2.00f,
                    multiplierA = -8.00f,
                    multiplierB = -4.00f,
                    curveB = Calc.AngleTargetSoftnessCurve,
                },
            };
        }

        static Dictionary<string, Dictionary<string, PhysicsCalculator>> UpDownAngleTargetCalculators()
        {
            return new Dictionary<string, Dictionary<string, PhysicsCalculator>>
            {
                {
                    Side.LEFT, new Dictionary<string, PhysicsCalculator>
                    {
                        { UP, UpAngleTargetCalculator(Side.LEFT) },
                        { DOWN, DownAngleTargetCalculator(Side.LEFT) },
                    }
                },
                {
                    Side.RIGHT, new Dictionary<string, PhysicsCalculator>
                    {
                        { UP, UpAngleTargetCalculator(Side.RIGHT) },
                        { DOWN, DownAngleTargetCalculator(Side.RIGHT) },
                    }
                },
            };
        }

        static Dictionary<string, Dictionary<string, PhysicsCalculator>> RightLeftAngleTargetCalculators()
        {
            return new Dictionary<string, Dictionary<string, PhysicsCalculator>>
            {
                {
                    Side.LEFT, new Dictionary<string, PhysicsCalculator>
                    {
                        {
                            SIDEWAYS_IN, new PhysicsCalculator
                            {
                                multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(RIGHT_LEFT_ANGLE_TARGET.NAME, Side.LEFT), SIDEWAYS_IN, true),
                                primaryConfig = new MultiplierConfig
                                {
                                    baseValue = 2.00f,
                                    multiplierA = 8.00f,
                                    multiplierB = 4.00f,
                                    curveB = Calc.AngleTargetSoftnessCurve,
                                },
                                secondaryConfig = new MultiplierConfig
                                {
                                    baseValue = 2.00f,
                                    multiplierA = 8.00f,
                                    multiplierB = 4.00f,
                                    curveB = Calc.AngleTargetSoftnessCurve,
                                },
                            }
                        },
                        {
                            SIDEWAYS_OUT, new PhysicsCalculator
                            {
                                multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(RIGHT_LEFT_ANGLE_TARGET.NAME, Side.LEFT), SIDEWAYS_OUT, true),
                                primaryConfig = new MultiplierConfig
                                {
                                    baseValue = -3.00f,
                                    multiplierA = -12.00f,
                                    multiplierB = -6.00f,
                                    curveB = Calc.AngleTargetSoftnessCurve,
                                },
                                secondaryConfig = new MultiplierConfig
                                {
                                    baseValue = -3.00f,
                                    multiplierA = -12.00f,
                                    multiplierB = -6.00f,
                                    curveB = Calc.AngleTargetSoftnessCurve,
                                },
                            }
                        },
                    }
                },
                {
                    Side.RIGHT, new Dictionary<string, PhysicsCalculator>
                    {
                        {
                            SIDEWAYS_IN, new PhysicsCalculator
                            {
                                multiplierJsf =
                                    NewGravityMultiplierJsf(Utils.ParamName(RIGHT_LEFT_ANGLE_TARGET.NAME, Side.RIGHT), SIDEWAYS_IN, false),
                                primaryConfig = new MultiplierConfig
                                {
                                    baseValue = -2.00f,
                                    multiplierA = -8.00f,
                                    multiplierB = -4.00f,
                                    curveB = Calc.AngleTargetSoftnessCurve,
                                },
                                secondaryConfig = new MultiplierConfig
                                {
                                    baseValue = -2.00f,
                                    multiplierA = -8.00f,
                                    multiplierB = -4.00f,
                                    curveB = Calc.AngleTargetSoftnessCurve,
                                },
                            }
                        },
                        {
                            SIDEWAYS_OUT, new PhysicsCalculator
                            {
                                multiplierJsf = NewGravityMultiplierJsf(
                                    Utils.ParamName(RIGHT_LEFT_ANGLE_TARGET.NAME, Side.RIGHT),
                                    SIDEWAYS_OUT,
                                    false
                                ),
                                primaryConfig = new MultiplierConfig
                                {
                                    baseValue = 3.00f,
                                    multiplierA = 12.00f,
                                    multiplierB = 6.00f,
                                    curveB = Calc.AngleTargetSoftnessCurve,
                                },
                                secondaryConfig = new MultiplierConfig
                                {
                                    baseValue = 3.00f,
                                    multiplierA = 12.00f,
                                    multiplierB = 6.00f,
                                    curveB = Calc.AngleTargetSoftnessCurve,
                                },
                            }
                        },
                    }
                },
            };
        }

        static StorableFloat NewGravityMultiplierJsf(string paramName, string direction, bool register)
        {
            return BootyMagic.script.NewStorableFloat($"{paramName} {direction} Gravity", 1, 0, 2, register);
        }

        public static void SaveCalibrated()
        {
            gravityCounterforceJsf.SaveCalibrated();
        }

        static bool _simulatingGravity;
        static bool _softPhysicsOn;
        static float _mass;
        static float _softness;
        static float _rollL;
        static float _rollR;
        static float _pitchL;
        static float _pitchR;

        public static void Update(bool simulatingGravity = false)
        {
            _simulatingGravity = simulatingGravity;
            _softPhysicsOn = SoftPhysicsHandler.softPhysicsOn;
            _mass = MainPhysicsHandler.normalizedMass;
            _softness = BootyMagic.softness;
            _rollL = Calc.SmoothStep((GravityEffectCalc.gluteRollL + GravityEffectCalc.hipRoll) / 2);
            _rollR = Calc.SmoothStep((GravityEffectCalc.gluteRollR + GravityEffectCalc.hipRoll) / 2);
            _pitchL = 2 * Calc.SmoothStep(GravityEffectCalc.glutePitchL);
            _pitchR = 2 * Calc.SmoothStep(GravityEffectCalc.glutePitchR);
            if(!_simulatingGravity)
            {
                ApplyGravityCounterforce();
            }

            AdjustSidewaysPhysics();
            AdjustVerticalPhysics();
            AdjustHorizontalPhysics();
        }

        static void AdjustSidewaysPhysics()
        {
            float effectL = Mathf.Abs(_rollL) / 2;
            if(_rollL > 0)
            {
                ResetLeftGlute(SIDEWAYS_IN);
                UpdateLeftGlute(SIDEWAYS_OUT, effectL);
            }
            else
            {
                ResetLeftGlute(SIDEWAYS_OUT);
                UpdateLeftGlute(SIDEWAYS_IN, effectL);
            }

            float effectR = Mathf.Abs(_rollR) / 2;
            if(_rollR > 0)
            {
                ResetRightGlute(SIDEWAYS_OUT);
                UpdateRightGlute(SIDEWAYS_IN, effectR);
            }
            else
            {
                ResetRightGlute(SIDEWAYS_IN);
                UpdateRightGlute(SIDEWAYS_OUT, effectR);
            }
        }

        static void AdjustVerticalPhysics()
        {
            float effectL = GravityEffectCalc.UpDownAdjustByAngle(_pitchL) * GravityEffectCalc.RollMultiplier(_rollL) / 2;
            float effectR = GravityEffectCalc.UpDownAdjustByAngle(_pitchR) * GravityEffectCalc.RollMultiplier(_rollR) / 2;

            if(_pitchL > 1 || _pitchL < -1)
            {
                // upside down, leaning forward or back
                ResetLeftGlute(DOWN);
                UpdateLeftGlute(UP, effectL);
            }
            else
            {
                ResetLeftGlute(UP);
                UpdateLeftGlute(DOWN, effectL);
            }

            if(_pitchR > 1 || _pitchR < -1)
            {
                // upside down, leaning forward or back
                ResetRightGlute(DOWN);
                UpdateRightGlute(UP, effectR);
            }
            else
            {
                ResetRightGlute(UP);
                UpdateRightGlute(DOWN, effectR);
            }
        }

        static void AdjustHorizontalPhysics()
        {
            float effectL = GravityEffectCalc.DepthAdjustByAngle(_pitchL) * GravityEffectCalc.RollMultiplier(_rollL) / 2;
            float effectR = GravityEffectCalc.DepthAdjustByAngle(_pitchR) * GravityEffectCalc.RollMultiplier(_rollR) / 2;

            if(_pitchL > 0)
            {
                // leaning forward, upright or upside down
                ResetLeftGlute(BACK);
                UpdateLeftGlute(FORWARD, effectL);
            }
            else
            {
                // leaning back
                ResetLeftGlute(FORWARD);
                UpdateLeftGlute(BACK, effectL);
            }

            if(_pitchR > 0)
            {
                // leaning forward, upright or upside down
                ResetRightGlute(BACK);
                UpdateRightGlute(FORWARD, effectR);
            }
            else
            {
                // leaning back
                UpdateRightGlute(BACK, effectR);
                ResetRightGlute(FORWARD);
            }
        }

        public static void ApplyGravitySimulationForce(string direction)
        {
            Person.ApplyGravityToGlutes(direction, _gravityCounterForce);
        }

        static void ApplyGravityCounterforce()
        {
            if(!_useGravityCounterforce)
            {
                return;
            }

            Person.ApplyForceToGlutes(_gravityCounterForce);
        }

        static void UpdateLeftGlute(string direction, float effect)
        {
            MainPhysicsHandler.spring.left.UpdateGravityValue(direction, effect, _mass, _softness, _softPhysicsOn);
            if(!_simulatingGravity)
            {
                MainPhysicsHandler.damper.left.UpdateGravityValue(direction, effect, _mass, _softness, _softPhysicsOn);
            }

            MainPhysicsHandler.inOutSpring.left.UpdateGravityValue(direction, effect, _mass, _softness, _softPhysicsOn);
            MainPhysicsHandler.upDownAngleTarget.left.UpdateGravityValue(direction, effect, _mass, _softness, _softPhysicsOn);
            MainPhysicsHandler.rightLeftAngleTarget.left.UpdateGravityValue(direction, effect, _mass, _softness, _softPhysicsOn);
        }

        static void UpdateRightGlute(string direction, float effect)
        {
            MainPhysicsHandler.spring.right.UpdateGravityValue(direction, effect, _mass, _softness, _softPhysicsOn);
            if(!_simulatingGravity)
            {
                MainPhysicsHandler.damper.right.UpdateGravityValue(direction, effect, _mass, _softness, _softPhysicsOn);
            }

            MainPhysicsHandler.inOutSpring.right.UpdateGravityValue(direction, effect, _mass, _softness, _softPhysicsOn);
            MainPhysicsHandler.upDownAngleTarget.right.UpdateGravityValue(direction, effect, _mass, _softness, _softPhysicsOn);
            MainPhysicsHandler.rightLeftAngleTarget.right.UpdateGravityValue(direction, effect, _mass, _softness, _softPhysicsOn);
        }

        public static void ResetAll()
        {
            foreach(string direction in new[]
            {
                SIDEWAYS_IN,
                SIDEWAYS_OUT,
                UP,
                DOWN,
                FORWARD,
                BACK,
            })
            {
                ResetLeftGlute(direction);
                ResetRightGlute(direction);
            }
        }

        static void ResetLeftGlute(string direction)
        {
            MainPhysicsHandler.spring.left.ResetGravityValue(direction);
            if(!_simulatingGravity)
            {
                MainPhysicsHandler.damper.left.ResetGravityValue(direction);
            }

            MainPhysicsHandler.inOutSpring.left.ResetGravityValue(direction);
            MainPhysicsHandler.upDownAngleTarget.left.ResetGravityValue(direction);
            MainPhysicsHandler.rightLeftAngleTarget.left.ResetGravityValue(direction);
        }

        static void ResetRightGlute(string direction)
        {
            MainPhysicsHandler.spring.right.ResetGravityValue(direction);
            if(!_simulatingGravity)
            {
                MainPhysicsHandler.damper.right.ResetGravityValue(direction);
            }

            MainPhysicsHandler.inOutSpring.right.ResetGravityValue(direction);
            MainPhysicsHandler.upDownAngleTarget.right.ResetGravityValue(direction);
            MainPhysicsHandler.rightLeftAngleTarget.right.ResetGravityValue(direction);
        }

        public static void Destroy()
        {
            gravityCounterforceJsf = null;
            autoAdjustGravityCounterforceJsb = null;
        }
    }
}
