using System;
using System.Collections.Generic;
using System.Linq;
using Naturalis.BM.Models;
using Naturalis.Common;
using Naturalis.Common.Components;
using Naturalis.Common.Models;
using SimpleJSON;
using UnityEngine;

namespace Naturalis.BM.Handlers
{
    static class HardColliderHandler
    {
        public static ColliderVisualizer colliderVisualizer { get; private set; }
        public static JSONStorableStringChooser colliderGroupsJssc { get; private set; }
        public static List<HardColliderGroup> hardColliderGroups { get; private set; }
        public static JSONStorableFloat baseMassMultiplierJsf { get; private set; }
        public static JSONStorableString allCollidersMassDisplayValueJss { get; private set; }
        public static JSONStorableBool highlightAllJsb { get; private set; }

        static float _frictionSizeMultiplierLeft;
        static float _frictionSizeMultiplierRight;

        static bool _initialized;

        public static void Init()
        {
            var autoColliderPairs = ConfigureAndReturnAutoColliders();
            InitColliderVisualizer(); // depends on alt colliders being enabled first
            hardColliderGroups = autoColliderPairs.Select(pair => NewHardColliderGroup(pair.left, pair.right)).ToList();

            EnableMultiplyFriction();

            var options = hardColliderGroups.Select(c => c.visualizerId).ToList();
            var displayOptions = hardColliderGroups.Select(c => c.id).ToList();
            colliderGroupsJssc = new JSONStorableStringChooser(
                "colliderGroup",
                options,
                displayOptions,
                "",
                "Collider"
            )
            {
                setCallbackFunction = value =>
                {
                    colliderVisualizer.SelectablesJSON.val = value;
                    SyncSizeAuto();
                },
                val = options[0],
            };

            colliderVisualizer.GroupsJSON.setJSONCallbackFunction = jssc =>
            {
                colliderVisualizer.SelectedPreviewOpacityJSON.val = jssc.val == "Off" ? 0 : 1;
                if(jssc.val != "Off")
                {
                    colliderVisualizer.SelectablesJSON.val = colliderGroupsJssc.val;
                }

                SyncSizeAuto();
            };
            colliderVisualizer.XRayPreviewsJSON.setJSONCallbackFunction = _ => SyncSizeAuto();

            baseMassMultiplierJsf = BootyMagic.script.NewStorableFloat($"Hard Colliders Base {Constant.MASS_MULTIPLIER}", 1.00f, 0.01f, 2.00f);
            baseMassMultiplierJsf.setCallbackFunction = _ => SyncCollidersMass();
            allCollidersMassDisplayValueJss = new JSONStorableString("allCollidersMassDisplayValue", "");
            highlightAllJsb = new JSONStorableBool(
                "Highlight All Colliders",
                false,
                value => colliderVisualizer.PreviewOpacityJSON.val = value ? 1.00f : 0.67f
            );
            _initialized = true;
        }

        static void InitColliderVisualizer()
        {
            colliderVisualizer = BootyMagic.script.gameObject.AddComponent<ColliderVisualizer>();
            var groups = new List<Group>
            {
                new Group("Off", @"$off"), //match nothing
                new Group("Both glutes", @"^((AutoCollider)?FemaleAutoColliders)?[LR]Glute1"),
                new Group("Left glute", @"^((AutoCollider)?FemaleAutoColliders)?LGlute1"),
            };
            colliderVisualizer.Init(BootyMagic.script, groups);
            colliderVisualizer.PreviewOpacityJSON.val = 0.67f;
            colliderVisualizer.PreviewOpacityJSON.defaultVal = 0.67f;
            colliderVisualizer.SelectedPreviewOpacityJSON.val = 1;
            colliderVisualizer.SelectedPreviewOpacityJSON.defaultVal = 1;
            colliderVisualizer.GroupsJSON.val = "Left glute";
            colliderVisualizer.GroupsJSON.defaultVal = "Left glute";
            colliderVisualizer.HighlightMirrorJSON.val = true;

            foreach(string option in new[] { "Select...", "Other", "All" })
            {
                colliderVisualizer.GroupsJSON.choices.Remove(option);
            }
        }

        public static List<AutoColliderPair> ConfigureAndReturnAutoColliders()
        {
            var autoColliderGroups = BootyMagic.script.containingAtom.GetComponentsInChildren<AutoColliderGroup>();

            const string nameStart = "FemaleAutoColliders";
            var lGlute = autoColliderGroups.First(ag => ag.name == $"{nameStart}LGlute");
            var rGlute = autoColliderGroups.First(ag => ag.name == $"{nameStart}RGlute");
            var lGluteAlt = autoColliderGroups.First(ag => ag.name == $"{nameStart}LGluteAlt");
            var rGluteAlt = autoColliderGroups.First(ag => ag.name == $"{nameStart}RGluteAlt");

            lGlute.enabled = false;
            rGlute.enabled = false;
            lGluteAlt.enabled = true;
            rGluteAlt.enabled = true;

            var leftAutoColliders = lGluteAlt.GetAutoColliders();
            var rightAutoColliders = rGluteAlt.GetAutoColliders();

            IgnoreSoftColliders(leftAutoColliders);
            IgnoreSoftColliders(rightAutoColliders);

            if(!_initialized)
            {
                var pairs = new List<AutoColliderPair>();
                for(int i = 0; i < leftAutoColliders.Length; i++)
                {
                    pairs.Add(new AutoColliderPair { left = leftAutoColliders[i], right = rightAutoColliders[i] });
                }

                return pairs;
            }

            return null;
        }

        static void IgnoreSoftColliders(AutoCollider[] autoColliders)
        {
            foreach(var softVerticesGroup in Person.instance.glutePhysicsMesh.softVerticesGroups)
            {
                foreach(var softVerticesSet in softVerticesGroup.softVerticesSets)
                {
                    foreach(var ac in autoColliders)
                    {
                        Physics.IgnoreCollision(softVerticesSet.jointCollider, ac.jointCollider);
                        if(softVerticesSet.jointCollider2)
                        {
                            Physics.IgnoreCollision(softVerticesSet.jointCollider2, ac.jointCollider);
                        }
                    }
                }
            }
        }

        static Dictionary<string, Scaler> GetScalingConfig(string id)
        {
            switch(id)
            {
                case "Glute1":
                    return new Dictionary<string, Scaler>
                    {
                        {
                            Constant.RADIUS_OFFSET, new Scaler
                            {
                                offset = 0.00f,
                                max = 40,
                            }
                        },
                        {
                            Constant.X_OFFSET, new Scaler
                            {
                                offset = 0.04f,
                                max = 40,
                            }
                        },
                        {
                            Constant.Y_OFFSET, new Scaler
                            {
                                offset = -0.36f,
                                max = 40,
                            }
                        },
                        {
                            Constant.Z_OFFSET, new Scaler
                            {
                                offset = -0.05f,
                                max = 40,
                            }
                        },
                    };
                case "Glute1 (1)":
                    return new Dictionary<string, Scaler>
                    {
                        {
                            Constant.RADIUS_OFFSET, new Scaler
                            {
                                offset = 0.00f,
                                max = 40,
                            }
                        },
                        {
                            Constant.X_OFFSET, new Scaler
                            {
                                offset = 0.00f,
                                max = 40,
                            }
                        },
                        {
                            Constant.Y_OFFSET, new Scaler
                            {
                                offset = 0.00f,
                                max = 40,
                            }
                        },
                        {
                            Constant.Z_OFFSET, new Scaler
                            {
                                offset = 0.00f,
                                max = 40,
                            }
                        },
                    };
                case "Glute1 (2)":
                    return new Dictionary<string, Scaler>
                    {
                        {
                            Constant.RADIUS_OFFSET, new Scaler
                            {
                                offset = 0.00f,
                                max = 40,
                            }
                        },
                        {
                            Constant.X_OFFSET, new Scaler
                            {
                                offset = 0.57f,
                                max = 40,
                            }
                        },
                        {
                            Constant.Y_OFFSET, new Scaler
                            {
                                offset = -1.06f,
                                max = 40,
                                massCurve = x => 0.12f * x,
                            }
                        },
                        {
                            Constant.Z_OFFSET, new Scaler
                            {
                                offset = -0.47f,
                                max = 40,
                                massCurve = x => -0.14f * x,
                            }
                        },
                    };
                case "Glute1 (3)":
                    return new Dictionary<string, Scaler>
                    {
                        {
                            Constant.RADIUS_OFFSET, new Scaler
                            {
                                offset = 0.00f,
                                max = 40,
                            }
                        },
                        {
                            Constant.X_OFFSET, new Scaler
                            {
                                offset = 0.00f,
                                max = 40,
                            }
                        },
                        {
                            Constant.Y_OFFSET, new Scaler
                            {
                                offset = -0.67f,
                                max = 40,
                                massCurve = x => 0.43f * x,
                            }
                        },
                        {
                            Constant.Z_OFFSET, new Scaler
                            {
                                offset = -0.57f,
                                max = 40,
                                massCurve = x => -0.60f * x,
                            }
                        },
                    };
                case "Glute1 (4)":
                    return new Dictionary<string, Scaler>
                    {
                        {
                            Constant.RADIUS_OFFSET, new Scaler
                            {
                                offset = 0.00f,
                                max = 40,
                                massCurve = x => -0.21f * x,
                            }
                        },
                        {
                            Constant.X_OFFSET, new Scaler
                            {
                                offset = 0.00f,
                                max = 40,
                            }
                        },
                        {
                            Constant.Y_OFFSET, new Scaler
                            {
                                offset = -0.18f,
                                max = 40,
                                massCurve = x => -0.13f * x,
                            }
                        },
                        {
                            Constant.Z_OFFSET, new Scaler
                            {
                                offset = -0.06f,
                                max = 40,
                                massCurve = x => -0.37f * x,
                            }
                        },
                    };
                case "Glute1 (5)":
                    return new Dictionary<string, Scaler>
                    {
                        {
                            Constant.RADIUS_OFFSET, new Scaler
                            {
                                offset = 0.00f,
                                max = 40,
                            }
                        },
                        {
                            Constant.X_OFFSET, new Scaler
                            {
                                offset = 0.00f,
                                max = 40,
                            }
                        },
                        {
                            Constant.Y_OFFSET, new Scaler
                            {
                                offset = 0.00f,
                                max = 40,
                                massCurve = x => -0.17f * x,
                            }
                        },
                        {
                            Constant.Z_OFFSET, new Scaler
                            {
                                offset = 0.00f,
                                max = 40,
                                massCurve = x => -0.12f * x,
                            }
                        },
                    };
                case "Glute1 (6)":
                    return new Dictionary<string, Scaler>
                    {
                        {
                            Constant.RADIUS_OFFSET, new Scaler
                            {
                                offset = 0.00f,
                                max = 40,
                            }
                        },
                        {
                            Constant.X_OFFSET, new Scaler
                            {
                                offset = 0.00f,
                                max = 40,
                            }
                        },
                        {
                            Constant.Y_OFFSET, new Scaler
                            {
                                offset = 0.00f,
                                max = 40,
                                massCurve = x => -0.16f * x,
                            }
                        },
                        {
                            Constant.Z_OFFSET, new Scaler
                            {
                                offset = 0.00f,
                                max = 40,
                                massCurve = x => 0.04f * x,
                            }
                        },
                    };
                case "Glute1 (7)":
                    return new Dictionary<string, Scaler>
                    {
                        {
                            Constant.RADIUS_OFFSET, new Scaler
                            {
                                offset = 0.00f,
                                max = 40,
                            }
                        },
                        {
                            Constant.X_OFFSET, new Scaler
                            {
                                offset = -0.04f,
                                max = 40,
                            }
                        },
                        {
                            Constant.Y_OFFSET, new Scaler
                            {
                                offset = -0.82f,
                                max = 40,
                            }
                        },
                        {
                            Constant.Z_OFFSET, new Scaler
                            {
                                offset = -0.23f,
                                max = 40,
                                massCurve = x => -0.15f * x,
                            }
                        },
                    };
                default: throw new ArgumentException($"Invalid id {id}");
            }
        }

        static HardColliderGroup NewHardColliderGroup(AutoCollider leftAc, AutoCollider rightAc)
        {
            string id = leftAc.name.Replace("AutoColliderFemaleAutoCollidersL", ""); // -> Glute1, Glute1 (1), Glute1 (2) etc.
            var leftCollider = new HardCollider(leftAc);
            var rightCollider = new HardCollider(rightAc);
            var scalingConfig = GetScalingConfig(id);
            string visualizerId = colliderVisualizer.SelectablesJSON.choices.Find(option => option.EndsWith($"L{id}"));
            var hardColliderGroup = new HardColliderGroup(
                id,
                visualizerId,
                leftCollider,
                rightCollider,
                scalingConfig[Constant.RADIUS_OFFSET],
                scalingConfig[Constant.X_OFFSET],
                scalingConfig[Constant.Y_OFFSET],
                scalingConfig[Constant.Z_OFFSET],
                1.0f // TODO
            )
            {
                massMultiplierJsf = BootyMagic.script.NewStorableFloat($"{id} {Constant.MASS_MULTIPLIER}", 1.00f, 0.01f, 2.00f),
                radiusJsf = BootyMagic.script.NewStorableFloat($"{id} {Constant.RADIUS_OFFSET}", 0, -1f, 1f),
                rightJsf = BootyMagic.script.NewStorableFloat($"{id} {Constant.X_OFFSET}", 0, -1f, 1f),
                upJsf = BootyMagic.script.NewStorableFloat($"{id} {Constant.Y_OFFSET}", 0, -1f, 1f),
                lookJsf = BootyMagic.script.NewStorableFloat($"{id} {Constant.Z_OFFSET}", 0, -1f, 1f),
            };

            hardColliderGroup.massMultiplierJsf.setCallbackFunction = _ => SyncColliderMass(
                hardColliderGroup,
                AllCollidersMass() / hardColliderGroups.Count,
                true
            );
            hardColliderGroup.radiusJsf.setCallbackFunction = _ => SyncRadius(hardColliderGroup);
            hardColliderGroup.rightJsf.setCallbackFunction = _ => SyncPosition(hardColliderGroup);
            hardColliderGroup.upJsf.setCallbackFunction = _ => SyncPosition(hardColliderGroup);
            hardColliderGroup.lookJsf.setCallbackFunction = _ => SyncPosition(hardColliderGroup);

            return hardColliderGroup;
        }

        static void SyncRadius(HardColliderGroup group)
        {
            group.UpdateDimensions(MainPhysicsHandler.normalizedMass);
            SyncSizeAuto();
        }

        static void SyncPosition(HardColliderGroup group)
        {
            group.UpdatePosition(MainPhysicsHandler.normalizedMass);
            SyncSizeAuto();
        }

        public static void SyncAllOffsets()
        {
            float mass = MainPhysicsHandler.normalizedRealMass;
            foreach(var group in hardColliderGroups)
            {
                group.UpdateDimensions(mass);
                group.UpdatePosition(mass);
            }

            SyncSizeAuto();
        }

        static void SyncSizeAuto()
        {
            foreach(var group in hardColliderGroups)
            {
                group.AutoColliderSizeSet();
            }

            float averageRadius = hardColliderGroups.Average(group => group.left.autoCollider.colliderRadius);
            foreach(var group in hardColliderGroups)
            {
                group.UpdateMaxFrictionalDistance(
                    _frictionSizeMultiplierLeft,
                    _frictionSizeMultiplierRight,
                    averageRadius
                );
            }

            colliderVisualizer.SyncPreviews();
        }

        // TODO
        public static void UpdateFrictionSizeMultipliers()
        {
            _frictionSizeMultiplierLeft = 1f;
            _frictionSizeMultiplierRight = 1f;
        }

        public static void UpdateFriction()
        {
            foreach(var group in hardColliderGroups)
            {
                group.UpdateFriction(FrictionHandler.instance.maxHardColliderFriction);
            }
        }

        static float AllCollidersMass()
        {
            return baseMassMultiplierJsf.val * MainPhysicsHandler.collidersMassAmount;
        }

        public static void SyncCollidersMass()
        {
            if(hardColliderGroups == null)
            {
                return;
            }

            float massPerCollider = AllCollidersMass() / hardColliderGroups.Count;
            foreach(var group in hardColliderGroups)
            {
                SyncColliderMass(group, massPerCollider);
            }

            UpdateAllCollidersMassDisplayValue();
        }

        static void SyncColliderMass(HardColliderGroup group, float mass, bool updateDisplayValue = false)
        {
            group.UpdateRigidbodyMass(group.massMultiplierJsf.val * mass);

            if(updateDisplayValue)
            {
                UpdateAllCollidersMassDisplayValue();
            }
        }

        static void UpdateAllCollidersMassDisplayValue()
        {
            float total = 0;
            foreach(var group in hardColliderGroups)
            {
                total += group.left.GetRigidbodyMass();
            }

            string displayValue = $"{Calc.RoundToDecimals(total)} kg".Bold();
            allCollidersMassDisplayValueJss.val = $"All Colliders Mass: {displayValue}";
        }

        public static void CalibrateColliders()
        {
            var gluteCenterLeft = GluteCenter(VertexIndices.VolumeCalculationVerts(Side.LEFT));
            var gluteCenterRight = GluteCenter(VertexIndices.VolumeCalculationVerts(Side.RIGHT));
            foreach(var group in hardColliderGroups)
            {
                group.Calibrate(gluteCenterLeft, gluteCenterRight, Person.instance.hipRb);
            }
        }

        public static void UpdateDistanceDiffs()
        {
            var gluteCenterLeft = GluteCenter(VertexIndices.VolumeCalculationVerts(Side.LEFT));
            var gluteCenterRight = GluteCenter(VertexIndices.VolumeCalculationVerts(Side.RIGHT));
            foreach(var group in hardColliderGroups)
            {
                group.UpdateDistanceDiffs(gluteCenterLeft, gluteCenterRight, Person.instance.hipRb);
            }
        }

        static Vector3 GluteCenter(int[] vertexIndices)
        {
            var vertices = new Vector3[vertexIndices.Length];
            for(int i = 0; i < vertexIndices.Length; i++)
            {
                vertices[i] = Person.instance.skin.rawSkinnedVerts[vertexIndices[i]];
            }

            return Calc.RelativePosition(Person.instance.hipRb, Calc.AveragePosition(vertices));
        }

        public static void ResetDistanceDiffs()
        {
            foreach(var group in hardColliderGroups)
            {
                group.ResetDistanceDiffs();
            }
        }

        public static void EnableMultiplyFriction()
        {
            foreach(var group in hardColliderGroups)
            {
                group.EnableMultiplyFriction();
            }
        }

        public static JSONClass SerializeCalibrationData()
        {
            var jc = new JSONClass();
            foreach(var group in hardColliderGroups)
            {
                var groupJson = new JSONClass
                {
                    ["left_dist"] = { AsFloat = group.left.sqrNeutralDistance },
                    ["right_dist"] = { AsFloat = group.right.sqrNeutralDistance },
                };
                jc[group.GetJsonId()] = groupJson;
            }

            return jc;
        }

        public static void CalibrateFromJson(JSONClass jc)
        {
            foreach(var group in hardColliderGroups)
            {
                var groupJson = jc[group.GetJsonId()].AsObject;
                group.left.sqrNeutralDistance = groupJson["left_dist"].AsFloat;
                group.right.sqrNeutralDistance = groupJson["right_dist"].AsFloat;
            }
        }

        public static void RestoreOriginalPhysics()
        {
            /* Required for restoring default collider mass. Enabled in case disabled
             * programmatically, and not yet restored back on by PersonMonitor.
             */
            Person.instance.EnableUseColliders();
            foreach(var group in hardColliderGroups)
            {
                group.RestoreDefaults();
            }
        }

        public static void Destroy()
        {
            UnityEngine.Object.Destroy(colliderVisualizer);
            colliderVisualizer = null;
            colliderGroupsJssc = null;
            hardColliderGroups = null;
            baseMassMultiplierJsf = null;
            allCollidersMassDisplayValueJss = null;
            highlightAllJsb = null;
        }

        public static bool RigidbodiesFound()
        {
            return hardColliderGroups.All(group => group.HasRigidbodies());
        }
    }
}
