﻿using Naturalis.BM.Models;
using Naturalis.Common;
using Naturalis.Common.Handlers.Configs;
using Naturalis.Common.Models;
using UnityEngine;
using static Side;

namespace Naturalis.BM.Handlers
{
    static class MainPhysicsHandler
    {
        // AdjustJoints.joint1.slerpDrive.maximumForce value
        const float DEFAULT_SLERP_MAX_FORCE = 500; // TODO check

        public static float calibratedMass { get; private set; }

        public static float collidersMassAmount { get; private set; }
        public static float normalizedMass { get; private set; }
        public static float normalizedRealMass { get; private set; }

        public static GlutePhysicsParameterGroup[] parameterGroups { get; private set; }

        public static GlutePhysicsParameterGroup gluteMass { get; private set; }
        public static GlutePhysicsParameterGroup centerOfGravity { get; private set; }
        public static GlutePhysicsParameterGroup spring { get; private set; }
        public static GlutePhysicsParameterGroup damper { get; private set; }
        public static GlutePhysicsParameterGroup inOutSpring { get; private set; }
        public static GlutePhysicsParameterGroup inOutDamper { get; private set; }
        public static GlutePhysicsParameterGroup upDownAngleTarget { get; private set; }

        public static GlutePhysicsParameterGroup rightLeftAngleTarget { get; private set; }
        // ReSharper restore MemberCanBePrivate.Global

        public static void Init()
        {
            LoadSettings();
        }

        public static void CalculateMass()
        {
            float leftVolume = CalculateVolume(VertexIndices.VolumeCalculationVerts(LEFT));
            float rightVolume = CalculateVolume(VertexIndices.VolumeCalculationVerts(RIGHT));
            float volume = (leftVolume + rightVolume) / 2;
            gluteMass.UpdateMinMax(
                SoftPhysicsHandler.softPhysicsOn ? GLUTE_JOINT_MASS.SOFT_MIN : GLUTE_JOINT_MASS.HARD_MIN,
                SoftPhysicsHandler.softPhysicsOn ? GLUTE_JOINT_MASS.SOFT_MAX : GLUTE_JOINT_MASS.HARD_MAX
            );
            /* Interpolates volume range [0.77, 2.70] to mass range [1.00, 4.00] */
            UpdateMassValueAndAmounts(1.55f * volume - 0.19f);
        }

        public static void UpdateMassValueAndAmounts(float mass)
        {
            calibratedMass = mass;
            float min = gluteMass.left.baseValueJsf.min;
            float max = gluteMass.left.baseValueJsf.max;
            float jointToCollidersRatio = Mathf.Lerp(0.92f, 0.78f, Mathf.InverseLerp(min, max, mass));
            gluteMass.UpdateMassValue(jointToCollidersRatio * mass);
            gluteMass.UpdateOffsetMinMax();
            collidersMassAmount = (1 - jointToCollidersRatio) * mass;
            normalizedMass = Mathf.InverseLerp(min, max, gluteMass.left.activeValueJsf.val);
            normalizedRealMass = Mathf.InverseLerp(min, max, gluteMass.left.baseValueJsf.val);
        }

        static float CalculateVolume(int[] vertexIndices)
        {
            var rawSkinnedVerts = Person.instance.skin.rawSkinnedVerts;
            var positions = new Vector3[vertexIndices.Length];
            for(int i = 0; i < vertexIndices.Length; i++)
            {
                positions[i] = Calc.RelativePosition(Person.instance.hipRb, rawSkinnedVerts[vertexIndices[i]]);
            }

            return 1000 * 0.5f * VolumeCalculator.SphereVolume(positions);
        }

        static void LoadSettings()
        {
            gluteMass = new GlutePhysicsParameterGroup
            {
                name = GLUTE_JOINT_MASS.NAME,
                valueFormat = "F3",
                left = GluteMass(LEFT),
                right = GluteMass(RIGHT),
                requiresCalibration = true,
                infoText = GLUTE_JOINT_MASS.NAME.Bold() +
                    " is the mass of the glute joint. " +
                    "\n\nSince mass represents glute size, other physics parameters are adjusted based on its value." +
                    "\n\nAdditional inertia is provided by the mass of glute hard colliders.",
            };

            centerOfGravity = new GlutePhysicsParameterGroup
            {
                name = CENTER_OF_GRAVITY.NAME,
                valueFormat = "F2",
                left = CenterOfGravity(LEFT),
                right = CenterOfGravity(RIGHT),
                requiresCalibration = true,
                infoText = CENTER_OF_GRAVITY.NAME.Bold() +
                    " is the position of the glute joint's center of mass." +
                    "\n\nAt 0, the center of mass is inside the body at the glute joint. At 1, it is at the skin surface.",
            };

            spring = new GlutePhysicsParameterGroup
            {
                name = SPRING.NAME,
                valueFormat = "F0",
                left = Spring(LEFT),
                right = Spring(RIGHT),
                requiresCalibration = true,
                infoText = SPRING.NAME.Bold() +
                    " pushes the glute joint towards its angle target." +
                    "\n\nThe angle target is defined by the Up/Down and Left/Right Angle Target parameters.",
            };

            damper = new GlutePhysicsParameterGroup
            {
                name = DAMPER.NAME,
                valueFormat = "F2",
                left = Damper(LEFT),
                right = Damper(RIGHT),
                dependsOnPhysicsRate = true,
                infoText = DAMPER.NAME.Bold() +
                    " reduces oscillation around the joint's angle target." +
                    "\n\nThe higher the damper, the quicker glutes will stop swinging.",
            };

            inOutSpring = new GlutePhysicsParameterGroup
            {
                name = IN_OUT_SPRING.NAME,
                valueFormat = "F0",
                left = InOutSpring(LEFT),
                right = InOutSpring(RIGHT),
                requiresCalibration = true,
                infoText = IN_OUT_SPRING.NAME.Bold() +
                    " pushes the glute joint towards its position target along the Z axis." +
                    "\n\nDirectional force morphing along the forward-back axis depends on In/Out Spring being suitably low" +
                    " for the given glute mass.",
            };

            inOutDamper = new GlutePhysicsParameterGroup
            {
                name = IN_OUT_DAMPER.NAME,
                valueFormat = "F0",
                left = InOutDamper(LEFT),
                right = InOutDamper(RIGHT),
                dependsOnPhysicsRate = true,
                infoText = IN_OUT_DAMPER.NAME.Bold() + " reduces oscillation around the joint position target along the Z axis.",
            };

            var upDownTargetL = UpDownAngleTarget(LEFT);
            var rightLeftTargetL = RightLeftAngleTarget(LEFT);

            var upDownTargetR = UpDownAngleTarget(RIGHT);
            var rightLeftTargetR = RightLeftAngleTarget(RIGHT);

            rightLeftTargetL.sync = value => SyncAngleTarget(LEFT, value, upDownTargetL.activeValueJsf.val);
            rightLeftTargetR.sync = value => SyncAngleTarget(RIGHT, value, upDownTargetR.activeValueJsf.val);

            upDownAngleTarget = new GlutePhysicsParameterGroup
            {
                name = UP_DOWN_ANGLE_TARGET.NAME,
                valueFormat = "F2",
                left = rightLeftTargetL,
                right = rightLeftTargetR,
                requiresCalibration = true,
                infoText = UP_DOWN_ANGLE_TARGET.NAME.Bold() +
                    " is the vertical target angle of the glute joint." +
                    " Negative values pull glutes down, positive values push them up." +
                    "\n\nThe offset shifts the center around which the final angle is calculated based on the person's pose" +
                    " (see Gravity Physics below).",
            };

            rightLeftAngleTarget = new GlutePhysicsParameterGroup
            {
                name = RIGHT_LEFT_ANGLE_TARGET.NAME,
                valueFormat = "F2",
                left = upDownTargetL,
                right = upDownTargetR,
                requiresCalibration = true,
                rightInverted = true,
                infoText = RIGHT_LEFT_ANGLE_TARGET.NAME.Bold() +
                    " is the horizontal target angle of the glute joint." +
                    " Negative values pull glutes apart, positive values push them together.",
            };

            parameterGroups = new[]
            {
                centerOfGravity,
                spring,
                damper,
                inOutSpring,
                inOutDamper,
                upDownAngleTarget,
                rightLeftAngleTarget,
            };

            gluteMass.SetCallbackFunctions();
            foreach(var group in parameterGroups)
            {
                group.SetCallbackFunctions();
            }
        }

#region *** Parameter Setup ***

        static PhysicsParameter GluteMass(string side)
        {
            var parameter = new PhysicsParameter(
                BootyMagic.script,
                side,
                GLUTE_JOINT_MASS.NAME,
                GLUTE_JOINT_MASS.SOFT_MIN,
                GLUTE_JOINT_MASS.SOFT_MAX
            )
            {
                displayValueJss = new JSONStorableString($"{GLUTE_JOINT_MASS.NAME} {side} Value Alt", ""),
                buildDisplayValueString = value => $"{GLUTE_JOINT_MASS.NAME}: {$"{Calc.RoundToDecimals(value, gluteMass.roundFactor)} kg".Bold()}",
                sync = value => SyncGluteMass(Person.instance.gluteRbs[side], value),
            };
            return parameter;
        }

        static PhysicsParameter CenterOfGravity(string side)
        {
            var parameter = new PhysicsParameter(BootyMagic.script, side, CENTER_OF_GRAVITY.NAME, CENTER_OF_GRAVITY.MIN, CENTER_OF_GRAVITY.MAX)
            {
                sync = value => SyncCenterOfGravity(Person.instance.gluteRbs[side], value),
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 0.33f,
                    multiplierA = 0.75f,
                    multiplierB = 1.33f,
                },
                secondaryConfig = new MultiplierConfig
                {
                    baseValue = 0.27f,
                    multiplierA = 0.75f,
                    multiplierB = 1.33f,
                },
                quicknessConfig = new MultiplierConfig
                {
                    baseValue = -0.06f,
                    multiplierA = 0.60f,
                    multiplierB = 0.00f,
                },
                slownessConfig = new MultiplierConfig
                {
                    baseValue = 0.06f,
                    multiplierA = 0.60f,
                    multiplierB = 0.00f,
                },
            };
            return parameter;
        }

        static PhysicsParameter Spring(string side)
        {
            var parameter = new PhysicsParameter(BootyMagic.script, side, SPRING.NAME, SPRING.MIN, SPRING.MAX)
            {
                sync = value => SyncSpring(Person.instance.gluteJoints[side], Person.instance.gluteRbs[side], value),
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 66f,
                    multiplierA = 0.80f,
                    multiplierB = -0.46f,
                    curveA = x => Calc.InverseSmoothStep(x, 1.00f, 0.23f, 0.30f),
                    curveB = x => Calc.InverseSmoothStep(x, 1.00f, 0.10f, 0.13f),
                },
                secondaryConfig = new MultiplierConfig
                {
                    baseValue = 55f,
                    multiplierA = 0.80f,
                    multiplierB = -0.52f,
                    curveA = x => Calc.InverseSmoothStep(x, 1.00f, 0.23f, 0.30f),
                    curveB = x => Calc.InverseSmoothStep(x, 1.00f, 0.10f, 0.13f),
                },
                quicknessConfig = new MultiplierConfig
                {
                    baseValue = 6f,
                    multiplierA = 0.67f,
                    multiplierB = 0.00f,
                    curveA = x => Calc.InverseSmoothStep(x, 1.00f, 0.23f, 0.30f),
                },
                slownessConfig = new MultiplierConfig
                {
                    baseValue = -9f,
                    multiplierA = 0.67f,
                    multiplierB = 0.00f,
                    curveA = x => Calc.InverseSmoothStep(x, 1.00f, 0.23f, 0.30f),
                },
            };

            return parameter;
        }

        static PhysicsParameter Damper(string side)
        {
            var parameter = new PhysicsParameter(BootyMagic.script, side, DAMPER.NAME, DAMPER.MIN, DAMPER.MAX)
            {
                sync = value => SyncDamper(Person.instance.gluteJoints[side], Person.instance.gluteRbs[side], value),
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 0.58f,
                    multiplierA = 1.00f,
                    multiplierB = -0.55f,
                    curveA = x => Calc.InverseSmoothStep(x, 1.00f, 0.31f, 0.80f),
                    curveB = x => Calc.InverseSmoothStep(x, 1.00f, 0.24f, 0.61f),
                },
                secondaryConfig = new MultiplierConfig
                {
                    baseValue = 0.58f,
                    multiplierA = 1.00f,
                    multiplierB = -0.55f,
                    curveA = x => Calc.InverseSmoothStep(x, 1.00f, 0.31f, 0.80f),
                    curveB = x => Calc.InverseSmoothStep(x, 1.00f, 0.24f, 0.61f),
                },
                quicknessConfig = new MultiplierConfig
                {
                    baseValue = -0.03f,
                    multiplierA = 8.375f,
                    multiplierB = 0.00f,
                    curveA = x => Calc.InverseSmoothStep(x, 1.00f, 0.31f, 0.80f),
                },
                slownessConfig = new MultiplierConfig
                {
                    baseValue = 0.015f,
                    multiplierA = 8.375f,
                    multiplierB = 0.00f,
                    curveA = x => Calc.InverseSmoothStep(x, 1.00f, 0.31f, 0.80f),
                },
            };

            return parameter;
        }

        static PhysicsParameter InOutSpring(string side)
        {
            var parameter = new PhysicsParameter(BootyMagic.script, side, IN_OUT_SPRING.NAME, IN_OUT_SPRING.MIN, IN_OUT_SPRING.MAX)
            {
                sync = value => SyncInOutSpring(Person.instance.gluteJoints[side], Person.instance.gluteRbs[side], value),
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 900f,
                    multiplierA = -0.15f,
                    multiplierB = -0.30f,
                    curveA = x => 1 - x, // TODO smoothstep
                },
                secondaryConfig = new MultiplierConfig
                {
                    baseValue = 750f,
                    multiplierA = -0.12f,
                    multiplierB = -0.24f,
                },
                quicknessConfig = new MultiplierConfig
                {
                    baseValue = 35f,
                    multiplierA = -0.28f,
                    multiplierB = 0.00f,
                    curveA = x => 1 - x,
                },
                slownessConfig = new MultiplierConfig
                {
                    baseValue = -52f,
                    multiplierA = -0.28f,
                    multiplierB = 0.00f,
                    curveA = x => 1 - x,
                },
            };
            return parameter;
        }

        static PhysicsParameter InOutDamper(string side)
        {
            var parameter = new PhysicsParameter(BootyMagic.script, side, IN_OUT_DAMPER.NAME, IN_OUT_DAMPER.MIN, IN_OUT_DAMPER.MAX)
            {
                sync = value => SyncInOutDamper(Person.instance.gluteJoints[side], Person.instance.gluteRbs[side], value),
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 7.50f,
                    multiplierA = 0.67f,
                    multiplierB = -0.67f,
                },
                secondaryConfig = new MultiplierConfig
                {
                    baseValue = 7.50f,
                    multiplierA = 0.67f,
                    multiplierB = -0.67f,
                },
                quicknessConfig = new MultiplierConfig
                {
                    baseValue = -2.00f,
                    multiplierA = 1.33f,
                    multiplierB = 0.00f,
                },
                slownessConfig = new MultiplierConfig
                {
                    baseValue = 3.00f,
                    multiplierA = 1.33f,
                    multiplierB = 0.00f,
                },
            };
            return parameter;
        }

        static PhysicsParameter UpDownAngleTarget(string side)
        {
            var parameter = new PhysicsParameter(
                BootyMagic.script,
                side,
                UP_DOWN_ANGLE_TARGET.NAME,
                UP_DOWN_ANGLE_TARGET.MIN,
                UP_DOWN_ANGLE_TARGET.MAX
            )
            {
                primaryConfig = new MultiplierConfig(),
                secondaryConfig = new MultiplierConfig(),
            };
            return parameter;
        }

        static PhysicsParameter RightLeftAngleTarget(string side)
        {
            var parameter = new PhysicsParameter(
                BootyMagic.script,
                side,
                RIGHT_LEFT_ANGLE_TARGET.NAME,
                RIGHT_LEFT_ANGLE_TARGET.MIN,
                RIGHT_LEFT_ANGLE_TARGET.MAX
            )
            {
                primaryConfig = new MultiplierConfig(),
                secondaryConfig = new MultiplierConfig(),
            };
            return parameter;
        }

#endregion

#region *** Sync Functions ***

        // Reimplements AdjustJoints.cs method SyncMass
        static void SyncGluteMass(Rigidbody rb, float value)
        {
            if(!BootyMagic.script.enabled)
            {
                return;
            }

            if(Mathf.Abs(rb.mass - value) <= 0.001f)
            {
                return;
            }

            rb.mass = value;
            rb.WakeUp();
        }

        // Reimplements AdjustJoints.cs method SyncCenterOfGravity
        static void SyncCenterOfGravity(Rigidbody rb, float value)
        {
            if(!BootyMagic.script.enabled)
            {
                return;
            }

            var newCenterOfMass = Vector3.Lerp(Person.instance.gluteControl.lowCenterOfGravity, Person.instance.gluteControl.highCenterOfGravity,
                value);
            if(Calc.VectorIsEqualWithin(1 / 100f, rb.centerOfMass, newCenterOfMass))
            {
                return;
            }

            rb.centerOfMass = newCenterOfMass;
            rb.WakeUp();
        }

        static void SyncSpring(ConfigurableJoint joint, Rigidbody rb, float value)
        {
            if(!BootyMagic.script.enabled)
            {
                return;
            }

            // see AdjustJoints.cs method ScaleChanged
            float scalePow = Mathf.Pow(1.7f, Person.instance.gluteControl.scale - 1f);

            float scaledSpring = value * scalePow;
            if(Mathf.Abs(joint.slerpDrive.positionSpring - scaledSpring) <= 1)
            {
                return;
            }

            var slerpDrive = joint.slerpDrive;
            slerpDrive.positionSpring = scaledSpring;
            slerpDrive.maximumForce = DEFAULT_SLERP_MAX_FORCE * scalePow;
            joint.slerpDrive = slerpDrive;

            var angularXDrive = joint.angularXDrive;
            angularXDrive.positionSpring = scaledSpring;
            joint.angularXDrive = angularXDrive;

            var angularYZDrive = joint.angularYZDrive;
            angularYZDrive.positionSpring = scaledSpring;
            joint.angularYZDrive = angularYZDrive;

            var angularXLimitSpring = joint.angularXLimitSpring;
            angularXLimitSpring.spring = scaledSpring * Person.instance.gluteControl.limitSpringMultiplier;
            joint.angularXLimitSpring = angularXLimitSpring;

            var angularYZLimitSpring = joint.angularYZLimitSpring;
            angularYZLimitSpring.spring = scaledSpring * Person.instance.gluteControl.limitSpringMultiplier;
            joint.angularYZLimitSpring = angularYZLimitSpring;

            rb.WakeUp();
        }

        static void SyncDamper(ConfigurableJoint joint, Rigidbody rb, float value)
        {
            if(!BootyMagic.script.enabled)
            {
                return;
            }

            // see AdjustJoints.cs method ScaleChanged
            float scalePow = Mathf.Pow(1.7f, Person.instance.gluteControl.scale - 1f);

            float scaledDamper = value * scalePow;
            if(Mathf.Abs(joint.slerpDrive.positionDamper - scaledDamper) <= 0.01f)
            {
                return;
            }

            var slerpDrive = joint.slerpDrive;
            slerpDrive.positionDamper = scaledDamper;
            slerpDrive.maximumForce = DEFAULT_SLERP_MAX_FORCE * scalePow;
            joint.slerpDrive = slerpDrive;

            var angularXDrive = joint.angularXDrive;
            angularXDrive.positionDamper = scaledDamper;
            joint.angularXDrive = angularXDrive;

            var angularYZDrive = joint.angularYZDrive;
            angularYZDrive.positionDamper = scaledDamper;
            joint.angularYZDrive = angularYZDrive;

            var angularXLimitSpring = joint.angularXLimitSpring;
            angularXLimitSpring.damper = scaledDamper * Person.instance.gluteControl.limitDamperMultiplier;
            joint.angularXLimitSpring = angularXLimitSpring;

            var angularYZLimitSpring = joint.angularYZLimitSpring;
            angularYZLimitSpring.damper = scaledDamper * Person.instance.gluteControl.limitDamperMultiplier;
            joint.angularYZLimitSpring = angularYZLimitSpring;

            rb.WakeUp();
        }

        // Reimplements AdjustJoints.cs method SyncJointPositionZDrive
        static void SyncInOutSpring(ConfigurableJoint joint, Rigidbody rb, float value)
        {
            if(!BootyMagic.script.enabled)
            {
                return;
            }

            var zDrive = joint.zDrive;
            if(Mathf.Abs(zDrive.positionSpring - value) <= 1)
            {
                return;
            }

            zDrive.positionSpring = value;
            joint.zDrive = zDrive;
            rb.WakeUp();
        }

        // Reimplements AdjustJoints.cs method SyncJointPositionZDrive
        static void SyncInOutDamper(ConfigurableJoint joint, Rigidbody rb, float value)
        {
            if(!BootyMagic.script.enabled)
            {
                return;
            }

            var zDrive = joint.zDrive;
            if(Mathf.Abs(zDrive.positionDamper - value) <= 1)
            {
                return;
            }

            zDrive.positionDamper = value;
            joint.zDrive = zDrive;
            rb.WakeUp();
        }

        static void SyncAngleTarget(string side, float rotationX, float rotationY)
        {
            if(!BootyMagic.script.enabled)
            {
                return;
            }

            var rotation = Person.GetGluteJointRotation(side);
            rotation.x = rotationX;
            rotation.y = rotationY;
            Person.SetGluteRotation(side, rotation);
        }

#endregion

        public static void SaveCalibrated()
        {
            gluteMass.SaveCalibrated();
            foreach(var group in parameterGroups)
            {
                group.SaveCalibrated();
            }
        }

        public static void UpdatePhysics(bool updateOffsetMinMax = true)
        {
            foreach(var group in parameterGroups)
            {
                group.UpdateBaseValue(
                    normalizedMass,
                    BootyMagic.softness,
                    BootyMagic.quickness,
                    SoftPhysicsHandler.softPhysicsOn
                );
                if(updateOffsetMinMax)
                {
                    group.UpdateOffsetMinMax();
                }
            }
        }

        public static void SetGravitySimulationDamper()
        {
            float interpolant = Calc.InverseSmoothStep(normalizedMass, 1, 0.50f, 0.10f);
            damper.OverrideValue(Mathf.Lerp(2.5f, 5f, interpolant));
        }

        public static void UnsetGravitySimulationDamper()
        {
            damper.UpdateBaseValue(
                normalizedMass,
                BootyMagic.softness,
                BootyMagic.quickness,
                SoftPhysicsHandler.softPhysicsOn
            );
        }

        public static void RestoreOriginalPhysics()
        {
            if(BootyMagic.IsInitialized())
            {
                return;
            }

            foreach(string name in new[]
            {
                "mass",
                "centerOfGravityPercent",
                "spring",
                "damper",
                "positionSpringZ",
                "positionDamperZ",
                "targetRotationX",
                "targetRotationY",
            })
            {
                /* Set a value that is different from the original, then restore the original
                 * in order to trigger VaM's internal sync
                 */
                var paramJsf = Person.instance.gluteControl.GetFloatJSONParam(name);
                float original = paramJsf.val;
                paramJsf.valNoCallback = Mathf.Abs(paramJsf.val - paramJsf.min) > 0.01f
                    ? paramJsf.min
                    : paramJsf.max;
                paramJsf.val = original;
            }
        }

        public static void Destroy()
        {
            parameterGroups = null;

            gluteMass = null;
            centerOfGravity = null;
            spring = null;
            damper = null;
            inOutSpring = null;
            inOutDamper = null;
            rightLeftAngleTarget = null;
            upDownAngleTarget = null;
        }
    }
}
