using System.Collections.Generic;
using System.Linq;
using Naturalis.BM.Models;
using Naturalis.BM.UI;
using Naturalis.Common;
using Naturalis.Common.Components;
using Naturalis.Common.Handlers.Configs;
using Naturalis.Common.Models;
using SimpleJSON;
using UnityEngine;
using static Side;
using static SoftVerticesGroupName;

namespace Naturalis.BM.Handlers
{
    static class SoftPhysicsHandler
    {
        public static ColliderVisualizer colliderVisualizer { get; private set; }
        public static SoftJointDiagnostics softJointDiagnostics { get; private set; }

        static Dictionary<string, SoftJoint[]> _allSoftJoints = new Dictionary<string, SoftJoint[]>();
        public static Dictionary<string, SoftJoint[]> outerSoftJoints { get; private set; }
        public static Dictionary<string, SoftJoint[]> mainSoftJoints { get; private set; }
        public static Dictionary<string, SoftJoint[]> centerSoftJoints { get; private set; }
        static Dictionary<string, SoftJoint[]> _mainNonGradientJoints = new Dictionary<string, SoftJoint[]>();
        static Dictionary<string, SoftJoint[]> _lowerReferenceGradientJoints = new Dictionary<string, SoftJoint[]>();
        static Dictionary<string, SoftJoint[]> _lowerNonGradientJoints = new Dictionary<string, SoftJoint[]>();

        static Dictionary<string, SoftJointGradientGroup[]> _gradientGroups = new Dictionary<string, SoftJointGradientGroup[]>();

        static Dictionary<string, DAZPhysicsMeshSoftVerticesGroup> _softVerticesGroups;
        static Dictionary<string, Dictionary<string, DAZPhysicsMeshSoftVerticesSet[]>> _softVerticesSets =
            new Dictionary<string, Dictionary<string, DAZPhysicsMeshSoftVerticesSet[]>>();

        public static SoftPhysicsParameterGroup[] softParameterGroups { get; private set; }
        public static PhysicsParameterGroup[] parameterGroups { get; private set; }

        public static GluteSoftPhysicsParameterGroup fatSpring { get; private set; }
        public static GluteSoftPhysicsParameterGroup fatDamper { get; private set; }
        public static GluteSoftPhysicsParameterGroup fatMass { get; private set; }
        public static GluteSoftPhysicsParameterGroup fatColliderRadius { get; private set; }
        public static GluteSoftPhysicsParameterGroup fatColliderDepth { get; private set; }
        public static GluteSoftPhysicsParameterGroup fatDistanceLimit { get; private set; }

        public static GlutePhysicsParameterGroup fatBackForce { get; private set; }
        public static GlutePhysicsParameterGroup fatBackForceMaxForce { get; private set; }
        public static GlutePhysicsParameterGroup fatBackForceThreshold { get; private set; }

        public static FrequencyRunner softJointsUpdater { get; private set; }

        public static JSONStorableBool gluteSoftPhysicsOnJsb { get; private set; }
        public static StorableBool ldpEnabledJsb { get; private set; }

        public static JSONStorableBool showColliderPreviewsJsb { get; private set; }

        static bool _pluginSoftPhysicsToggled;
        public static bool softPhysicsSynced { get; set; }
        public static bool softPhysicsOn { get; private set; }
        static bool _ldpOn;

        public static bool LdpUpdateIsEnabled()
        {
            return _ldpOn && Settings.personMonitor.GluteSoftPhysicsIsEnabled();
        }

        public static void Init()
        {
            InitColliderVisualizer();
            SetupSoftGroupsAndJoints();
            EnableMultiplyFriction();
            softJointDiagnostics = new SoftJointDiagnostics(_allSoftJoints[LEFT]);
            softJointsUpdater = new FrequencyRunner(Settings.instance.ldpUpdateRateJssc, () => UpdateSoftJointsDistortion());
            SetupStorables();
            LoadSettings();
            softPhysicsOn = gluteSoftPhysicsOnJsb.val;
        }

        static void InitColliderVisualizer()
        {
            colliderVisualizer = BootyMagic.script.gameObject.AddComponent<ColliderVisualizer>();
            var groups = new List<Group>
            {
                new Group("Off", @"$off"), //match nothing
                new Group("Both glutes", @"PhysicsMeshJoint(left|right) glute?\d+$"),
                new Group("Left glute", @"PhysicsMeshJointleft glute?\d+$"),
                new Group("Right glute", @"PhysicsMeshJointright glute?\d+$"),
            };
            colliderVisualizer.Init(BootyMagic.script, groups);
            colliderVisualizer.PreviewOpacityJSON.val = 0.67f;
            colliderVisualizer.PreviewOpacityJSON.defaultVal = 0.67f;
            colliderVisualizer.SelectedPreviewOpacityJSON.val = 1;
            colliderVisualizer.SelectedPreviewOpacityJSON.defaultVal = 1;
            colliderVisualizer.GroupsJSON.val = "Left glute";
            colliderVisualizer.GroupsJSON.defaultVal = "Left glute";
            colliderVisualizer.HighlightMirrorJSON.val = true;
            colliderVisualizer.ShowPreviewsJSON.val = false;

            foreach(string option in new[] { "Select...", "Other", "All" })
            {
                colliderVisualizer.GroupsJSON.choices.Remove(option);
            }

            showColliderPreviewsJsb = new JSONStorableBool("Show Collider Previews", true, (bool _) =>
            {
                if(BootyMagic.tabs.PhysicsTabIsActive() && BootyMagic.tabs.GetActiveNestedWindow() is SoftParameterWindow)
                {
                    ((SoftParameterWindow) BootyMagic.tabs.GetActiveNestedWindow()).SetShowColliderPreviews();
                }
            });
        }

        static void SetupSoftGroupsAndJoints()
        {
            SetupSoftVerticesGroups();
            SetupSoftVerticesSets(LEFT);
            SetupSoftVerticesSets(RIGHT);
            outerSoftJoints = new Dictionary<string, SoftJoint[]>();
            mainSoftJoints = new Dictionary<string, SoftJoint[]>();
            centerSoftJoints = new Dictionary<string, SoftJoint[]>();
            SetupSoftJoints(LEFT);
            SetupSoftJoints(RIGHT);
        }

        static void SetupSoftVerticesGroups()
        {
            var groups = Person.instance.glutePhysicsMesh.softVerticesGroups;
            _softVerticesGroups = new Dictionary<string, DAZPhysicsMeshSoftVerticesGroup>
            {
                { LEFT, groups.Find(group => group.name == "left glute") },
                { RIGHT, groups.Find(group => group.name == "right glute") },
            };
        }

        static void SetupSoftVerticesSets(string side)
        {
            _softVerticesSets[side] = new Dictionary<string, DAZPhysicsMeshSoftVerticesSet[]>();
            var sets = _softVerticesGroups[side].softVerticesSets;
            foreach(string group in new[] { OUTER, MAIN, CENTER, LOWER })
            {
                var indices = VertexIndices.AllSoftVerts(side, group);
                _softVerticesSets[side][group] = sets.Where(set => indices.Contains(set.targetVertex)).ToArray();
            }
        }

        static void SetupSoftJoints(string side)
        {
            var allSJs = new List<SoftJoint>();
            var outerSJs = new List<SoftJoint>();
            var mainSJs = new List<SoftJoint>();
            var centerSJs = new List<SoftJoint>();
            var lowerSJs = new List<SoftJoint>();

            foreach(var kvp in _softVerticesSets[side])
            {
                string category = kvp.Key;
                var softVerticesSets = kvp.Value;
                foreach(var set in softVerticesSets)
                {
                    var softJoint = new SoftJoint(set)
                    {
                        id = $"{side}.{category}.{set.targetVertex.ToString()}",
                    };
                    switch(category)
                    {
                        case OUTER:
                            outerSJs.Add(softJoint);
                            break;
                        case MAIN:
                            mainSJs.Add(softJoint);
                            break;
                        case CENTER:
                            centerSJs.Add(softJoint);
                            break;
                        case LOWER:
                            lowerSJs.Add(softJoint);
                            break;
                    }

                    allSJs.Add(softJoint);
                }
            }

            _allSoftJoints[side] = allSJs.ToArray();
            outerSoftJoints[side] = outerSJs.ToArray();
            mainSoftJoints[side] = mainSJs.ToArray();
            centerSoftJoints[side] = centerSJs.ToArray();
            _lowerNonGradientJoints[side] = lowerSJs.ToArray();

            var additionalSpringMultiplierVerts = VertexIndices.AdditionalSpringMultiplierVerts(side);
            var nearbyVerts = VertexIndices.NearbyVerts(side);
            var adjacentVerts = VertexIndices.AdjacentVerts(side);
            foreach(var softJoint in allSJs)
            {
                int target = softJoint.softVerticesSet.targetVertex;
                if(nearbyVerts.ContainsKey(target))
                {
                    int[] nearbyTarget = nearbyVerts[target];
                    softJoint.nearbyJoints = allSJs.Where(x => nearbyTarget.Contains(x.softVerticesSet.targetVertex)).ToArray();
                }
                else
                {
                    // lower verts do not have nearby verts
                    softJoint.nearbyJoints = new SoftJoint[] { };
                }

                if(adjacentVerts.ContainsKey(target))
                {
                    int[] adjacentTarget = adjacentVerts[target];
                    softJoint.adjacentJoints = allSJs.Where(x => adjacentTarget.Contains(x.softVerticesSet.targetVertex)).ToArray();
                }
                else
                {
                    // outer, center and lower verts do not have adjacent verts
                    softJoint.adjacentJoints = new SoftJoint[] { };
                }

                if(additionalSpringMultiplierVerts.ContainsKey(target))
                {
                    softJoint.additionalSpringMultiplier = additionalSpringMultiplierVerts[target];
                }
            }

            var gradientGroupsList = new List<SoftJointGradientGroup>();
            foreach(int[] indexGroup in VertexIndices.CenterToOuterVerts(side))
            {
                var gradientGroup = new SoftJointGradientGroup
                {
                    center = centerSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[0])),
                    main1 = mainSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[1])),
                };

                switch(indexGroup.Length)
                {
                    case 3:
                        gradientGroup.outer = outerSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[2]));
                        break;
                    case 4:
                        gradientGroup.main2 = mainSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[2]));
                        gradientGroup.outer = outerSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[3]));
                        break;
                    case 5:
                        gradientGroup.main2 = mainSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[2]));
                        gradientGroup.main3 = mainSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[3]));
                        gradientGroup.outer = outerSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[4]));
                        break;
                    case 6:
                        gradientGroup.main2 = mainSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[2]));
                        gradientGroup.main3 = mainSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[3]));
                        gradientGroup.main4 = mainSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[4]));
                        gradientGroup.outer = outerSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[5]));
                        break;
                }

                gradientGroupsList.Add(gradientGroup);
            }

            _gradientGroups[side] = gradientGroupsList.ToArray();

            var mainNonGradientIndices = VertexIndices.MainNonGradientVerts(side);
            _mainNonGradientJoints[side] = _allSoftJoints[side]
                .Where(x => mainNonGradientIndices.Contains(x.softVerticesSet.targetVertex))
                .ToArray();

            var lowerReferenceGradientIndices = VertexIndices.LowerReferenceGradientVerts(side);
            _lowerReferenceGradientJoints[side] = _allSoftJoints[side]
                .Where(x => lowerReferenceGradientIndices.Contains(x.softVerticesSet.targetVertex))
                .ToArray();
        }

        static bool MatchByVertex(SoftJoint softJoint, int vertex)
        {
            return softJoint.softVerticesSet.targetVertex == vertex;
        }

        static void SetupStorables()
        {
            gluteSoftPhysicsOnJsb = BootyMagic.script.NewStorableBool(
                Constant.GLUTE_SOFT_PHYSICS_ENABLED,
                Person.instance.glutePhysicsMesh.on
            );
            gluteSoftPhysicsOnJsb.setCallbackFunction = value =>
            {
                ldpEnabledJsb.val = Settings.instance.glutesLdpDefaultEnabledJsb.val && value;
                if(!_reverseSyncingSoftPhysics)
                {
                    _pluginSoftPhysicsToggled = true;
                    if(BootyMagic.script.enabled)
                    {
                        Person.instance.glutePhysicsMesh.on = value;
                    }
                    else
                    {
                        BootyMagic.logger.LogMessage($"Enable the plugin to set {gluteSoftPhysicsOnJsb.name}.");
                    }

                    if(BootyMagic.tabs.MainTabIsActive())
                    {
                        BootyMagic.tabs.mainWindow.Refresh();
                    }
                }

                _reverseSyncingSoftPhysics = false;
            };

            ldpEnabledJsb = BootyMagic.script.NewStorableBool(
                "Local Distortion Physics Enabled",
                Settings.instance.glutesLdpDefaultEnabledJsb.val && Person.instance.glutePhysicsMesh.on
            );
            ldpEnabledJsb.setJSONCallbackFunction = jsb =>
            {
                var storable = (StorableBool) jsb;
                if(storable.val)
                {
                    if(!gluteSoftPhysicsOnJsb.val)
                    {
                        BootyMagic.logger.LogMessage("Enable glute soft physics to use local distortion physics.");
                    }
                    else
                    {
                        BootyMagic.script.SetCalibrationPending(storable);
                    }
                }
                else
                {
                    if(softPhysicsOn)
                    {
                        BootyMagic.script.SetCalibrationPending(storable);
                    }
                }

                softJointDiagnostics.visualizeLdpJsb.val = jsb.val;

                if(BootyMagic.tabs.MainTabIsActive())
                {
                    BootyMagic.tabs.mainWindow.Refresh();
                }
            };

            if(ldpEnabledJsb.val)
            {
                softJointDiagnostics.visualizeLdpJsb.val = true;
            }
        }

        static void LoadSettings()
        {
            fatSpring = new GluteSoftPhysicsParameterGroup
            {
                name = FAT_SPRING.NAME,
                left = FatSpring(LEFT),
                right = FatSpring(RIGHT),
                valueFormat = "F1",
                visualizeRelativeDistortion = true,
                visualizationLimit = 250f,
                distortionLimit = 500f,
                infoText = FAT_SPRING.NAME.Bold() +
                    " holds each soft joint in its target position." +
                    "\n\nLow fat spring makes glute fat soft and slow." +
                    " High fat spring makes it rigid and quick to return to its normal shape.",
            };

            fatDamper = new GluteSoftPhysicsParameterGroup
            {
                name = FAT_DAMPER.NAME,
                left = FatDamper(LEFT),
                right = FatDamper(RIGHT),
                valueFormat = "F2",
                visualizeRelativeDistortion = true,
                visualizationLimit = 2.50f,
                distortionLimit = 5.00f,
                infoText = FAT_DAMPER.NAME.Bold() +
                    " reduces oscillation of each soft joint around its target position." +
                    "\n\nLow fat damper makes glute fat jiggle more easily.",
            };

            fatMass = new GluteSoftPhysicsParameterGroup
            {
                name = FAT_MASS.NAME,
                left = FatMass(LEFT),
                right = FatMass(RIGHT),
                valueFormat = "F3",
                visualizeRelativeDistortion = true,
                visualizationLimit = 0.150f,
                distortionLimit = 0.300f,
                infoText = FAT_MASS.NAME.Bold() +
                    " determines the mass of each soft joint" +
                    " and the density of glute fat overall." +
                    " Increasing glute size without increasing fat mass reduces density.",
            };

            fatColliderRadius = new GluteSoftPhysicsParameterGroup
            {
                name = FAT_COLLIDER_RADIUS.NAME,
                left = FatColliderRadius(LEFT),
                right = FatColliderRadius(RIGHT),
                valueFormat = "F4",
                hasSoftColliderVisualization = true,
                visualizeRelativeDistortion = false,
                visualizationLimit = 0.0100f,
                distortionLimit = 0.0200f,
                infoText = FAT_COLLIDER_RADIUS.NAME.Bold() +
                    " is a scaling factor for the size of the" +
                    " colliders connected to each soft joint." +
                    "\n\nThe actual size varies per collider and automatically scales with glute size.",
            };

            fatColliderDepth = new GluteSoftPhysicsParameterGroup
            {
                name = FAT_COLLIDER_DEPTH.NAME,
                valueFormat = "F4",
                left = FatColliderDepth(LEFT),
                right = FatColliderDepth(RIGHT),
                hasSoftColliderVisualization = true,
                visualizeRelativeDistortion = false,
                visualizationLimit = 0.0020f,
                distortionLimit = 0.0020f,
                infoText = FAT_COLLIDER_DEPTH.NAME.Bold() +
                    " offsets collider positions relative to skin surface." +
                    "\n\nNegative values pull colliders out from the glute, positive values push them into the glute.",
            };

            fatDistanceLimit = new GluteSoftPhysicsParameterGroup
            {
                name = FAT_DISTANCE_LIMIT.NAME,
                valueFormat = "F3",
                left = FatDistanceLimit(LEFT),
                right = FatDistanceLimit(RIGHT),
                visualizeRelativeDistortion = true,
                visualizationLimit = 0.020f,
                distortionLimit = 0.020f,
                infoText = FAT_DISTANCE_LIMIT.NAME.Bold() + " is the maximum distance each soft joint can move away from its target position.",
            };

            softParameterGroups = new SoftPhysicsParameterGroup[]
            {
                fatSpring,
                fatDamper,
                fatMass,
                fatColliderRadius,
                fatColliderDepth,
                fatDistanceLimit,
            };

            fatBackForce = new GlutePhysicsParameterGroup
            {
                name = FAT_BACK_FORCE.NAME,
                valueFormat = "F2",
                left = FatBackForce(LEFT),
                right = FatBackForce(RIGHT),
                infoText = FAT_BACK_FORCE.NAME.Bold() +
                    " applies a force to the glute joint based on the movement of each soft joint." +
                    "\n\nLow back force (not 0) helps move the glute with collision, and adds a dampening effect." +
                    " High force can create a feedback loop that spirals out of control.",
            };

            fatBackForceMaxForce = new GlutePhysicsParameterGroup
            {
                name = FAT_BACK_FORCE_MAX_FORCE.NAME,
                valueFormat = "F2",
                left = FatBackForceMaxForce(LEFT),
                right = FatBackForceMaxForce(RIGHT),
                infoText = FAT_BACK_FORCE_MAX_FORCE.NAME.Bold() +
                    " sets an upper limit on the magnitude of back force." +
                    $"\n\nAlong with {FAT_BACK_FORCE_THRESHOLD.NAME}, this can be used to prevent an out of control feedback loop.",
            };

            fatBackForceThreshold = new GlutePhysicsParameterGroup
            {
                name = FAT_BACK_FORCE_THRESHOLD.NAME,
                valueFormat = "F3",
                left = FatBackForceThreshold(LEFT),
                right = FatBackForceThreshold(RIGHT),
                infoText = FAT_BACK_FORCE_THRESHOLD.NAME.Bold() +
                    " sets a minimum distance each soft joint needs to move" +
                    " for back force to be applied, ensuring that small movements don't cause the whole glute to move." +
                    $"\n\nAlong with {FAT_BACK_FORCE_MAX_FORCE.NAME}, this can be used to prevent an out of control feedback loop.",
            };

            parameterGroups = new PhysicsParameterGroup[]
            {
                fatBackForce,
                fatBackForceMaxForce,
                fatBackForceThreshold,
            };

            foreach(var group in softParameterGroups)
            {
                group.SetCallbackFunctions();
            }

            foreach(var group in parameterGroups)
            {
                group.SetCallbackFunctions();
            }

            var springJsf = fatSpring.left.baseValueJsf;
            var damperJsf = fatDamper.left.baseValueJsf;
            var massJsf = fatMass.left.baseValueJsf;
            var colliderRadiusJsf = fatColliderRadius.left.baseValueJsf;
            var colliderDepthJsf = fatColliderDepth.left.baseValueJsf;
            var distanceLimitJsf = fatDistanceLimit.left.baseValueJsf;

            SoftJoint.minSpring = springJsf.min;
            SoftJoint.maxSpring = springJsf.max * Constant.GRADIENT_MULTI_MAX;
            SoftJoint.minDamper = damperJsf.min;
            SoftJoint.maxDamper = damperJsf.max * Constant.GRADIENT_MULTI_MAX;
            SoftJoint.minMass = massJsf.min;
            SoftJoint.maxMass = massJsf.max * Constant.GRADIENT_MULTI_MAX;
            SoftJoint.minColliderRadius = colliderRadiusJsf.min;
            SoftJoint.maxColliderRadius = colliderRadiusJsf.max * Constant.GRADIENT_MULTI_MAX;
            SoftJoint.minColliderDepth = colliderDepthJsf.min * Constant.GRADIENT_MULTI_MAX;
            SoftJoint.maxColliderDepth = colliderDepthJsf.max * Constant.GRADIENT_MULTI_MAX;
            SoftJoint.minDistanceLimit = distanceLimitJsf.min;
            SoftJoint.maxDistanceLimit = distanceLimitJsf.max * Constant.GRADIENT_MULTI_MAX;
        }

        public static Rigidbody[] GetPositionalRigidbodies(string side, string group)
        {
            var rigidbodies = new List<Rigidbody>();
            foreach(var set in _softVerticesSets[side][group])
            {
                var indices = VertexIndices.PositionalSoftVerts(side, group);
                if(indices.Contains(set.targetVertex))
                {
                    rigidbodies.Add(set.jointRB);
                }
            }

            return rigidbodies.ToArray();
        }

#region *** Parameter Setup ***

        static float FatSpringSoftnessCurve(float x)
        {
            return Calc.InverseSmoothStep(x, 1.00f, 0.00f, 0.11f);
        }

        static SoftPhysicsParameter NewSoftPhysicsParameter(string side, string name, float min, float max)
        {
            var parameter = new SoftPhysicsParameter(BootyMagic.script, Constant.BOOTY_MAGIC, side, name, min, max);
            parameter.sync = (value, additionalOffset) => SyncGradientParameter(side, name, parameter.gradientConfig, value, additionalOffset);
            return parameter;
        }

        static SoftPhysicsParameter FatSpring(string side)
        {
            var parameter = NewSoftPhysicsParameter(side, FAT_SPRING.NAME, FAT_SPRING.MIN, FAT_SPRING.MAX);
            parameter.primaryConfig = new MultiplierConfig
            {
                baseValue = 1000f,
                multiplierA = 0.00f,
                multiplierB = -0.915f,
                curveB = FatSpringSoftnessCurve,
            };
            parameter.secondaryConfig = new MultiplierConfig
            {
                baseValue = 1000f,
                multiplierA = 0.00f,
                multiplierB = -0.878f,
                curveB = FatSpringSoftnessCurve,
            };
            parameter.quicknessConfig = new MultiplierConfig
            {
                baseValue = 50f,
                multiplierA = 0.00f,
                multiplierB = -0.75f,
            };
            parameter.slownessConfig = new MultiplierConfig
            {
                baseValue = -125f,
                multiplierA = 0.00f,
                multiplierB = -0.960f,
            };

            var gradientConfig = parameter.gradientConfig;
            gradientConfig.middleShiftJsf.SetValueAndDefault(-0.40f);
            gradientConfig.outerMultiplierJsf.SetValueAndDefault(1.60f);
            gradientConfig.centerMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.outerToMiddleCurvature = 0.60f;
            gradientConfig.outerToMiddleMidPoint = 0.60f;

            var ldpConfig = parameter.ldpConfig;
            ldpConfig.compressionEffectJsf.SetValueAndDefault(0.62f);
            ldpConfig.compressionSensitivityJsf.SetValueAndDefault(0.70f);
            ldpConfig.stretchingEffectJsf.SetValueAndDefault(0.62f);
            ldpConfig.stretchingSensitivityJsf.SetValueAndDefault(0.70f);

            return parameter;
        }

        static SoftPhysicsParameter FatDamper(string side)
        {
            var parameter = NewSoftPhysicsParameter(side, FAT_DAMPER.NAME, FAT_DAMPER.MIN, FAT_DAMPER.MAX);
            parameter.primaryConfig = new MultiplierConfig
            {
                baseValue = 2.53f,
                multiplierA = 0.18f,
                multiplierB = -0.59f,
                curveB = x => Calc.Exponential1(x, 1.90f, 1.74f, 1.17f),
            };
            parameter.secondaryConfig = new MultiplierConfig
            {
                baseValue = 2.16f,
                multiplierA = 0.18f,
                multiplierB = -0.59f,
                curveB = x => Calc.Exponential1(x, 1.90f, 1.74f, 1.17f),
            };
            parameter.quicknessConfig = new MultiplierConfig
            {
                baseValue = -1.92f,
                multiplierA = 0.18f,
                multiplierB = -0.59f,
                curveB = x => Calc.Exponential1(x, 1.90f, 1.74f, 1.17f),
            };
            parameter.slownessConfig = new MultiplierConfig
            {
                baseValue = 0.96f,
                multiplierA = 0.18f,
                multiplierB = -0.59f,
                curveB = x => Calc.Exponential1(x, 1.90f, 1.74f, 1.17f),
            };

            var gradientConfig = parameter.gradientConfig;
            gradientConfig.middleShiftJsf.SetValueAndDefault(-0.20f);
            gradientConfig.outerMultiplierJsf.SetValueAndDefault(1.60f);
            gradientConfig.centerMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.outerToMiddleCurvature = 0.60f;
            gradientConfig.outerToMiddleMidPoint = 0.60f;

            var ldpConfig = parameter.ldpConfig;
            ldpConfig.compressionEffectJsf.SetValueAndDefault(-0.50f);
            ldpConfig.compressionSensitivityJsf.SetValueAndDefault(0.80f);
            ldpConfig.stretchingEffectJsf.SetValueAndDefault(-0.50f);
            ldpConfig.stretchingSensitivityJsf.SetValueAndDefault(0.80f);

            return parameter;
        }

        static SoftPhysicsParameter FatMass(string side)
        {
            var parameter = NewSoftPhysicsParameter(side, FAT_MASS.NAME, FAT_MASS.MIN, FAT_MASS.MAX);
            parameter.primaryConfig = new MultiplierConfig
            {
                baseValue = 0.096f,
                multiplierA = 0.33f,
                multiplierB = 0.60f,
                curveA = x => Calc.Exponential1(x, 1.91f, 1.7f, 0.82f),
                curveB = x => Calc.Exponential1(x, 2.30f, 1.74f, 1.17f),
            };
            parameter.secondaryConfig = new MultiplierConfig
            {
                baseValue = 0.096f,
                multiplierA = 0.33f,
                multiplierB = 0.60f,
                curveA = x => Calc.Exponential1(x, 1.91f, 1.7f, 0.82f),
                curveB = x => Calc.Exponential1(x, 2.30f, 1.74f, 1.17f),
            };
            parameter.quicknessConfig = new MultiplierConfig
            {
                baseValue = -0.081f,
                multiplierA = 0.33f,
                multiplierB = 0.60f,
                curveB = x => Calc.Exponential1(x, 2.30f, 1.74f, 1.17f),
            };
            parameter.slownessConfig = new MultiplierConfig
            {
                baseValue = 0.081f,
                multiplierA = 0.33f,
                multiplierB = 0.60f,
                curveB = x => Calc.Exponential1(x, 2.30f, 1.74f, 1.17f),
            };

            var gradientConfig = parameter.gradientConfig;
            gradientConfig.middleShiftJsf.SetValueAndDefault(0.00f);
            gradientConfig.outerMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.centerMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.outerToMiddleCurvature = 0.60f;
            gradientConfig.outerToMiddleMidPoint = 0.60f;

            var ldpConfig = parameter.ldpConfig;
            ldpConfig.compressionEffectJsf.SetValueAndDefault(0.70f);
            ldpConfig.compressionSensitivityJsf.SetValueAndDefault(0.80f);
            ldpConfig.stretchingEffectJsf.SetValueAndDefault(-0.35f);
            ldpConfig.stretchingSensitivityJsf.SetValueAndDefault(0.80f);

            return parameter;
        }

        static SoftPhysicsParameter FatColliderRadius(string side)
        {
            var parameter = NewSoftPhysicsParameter(side, FAT_COLLIDER_RADIUS.NAME, FAT_COLLIDER_RADIUS.MIN, FAT_COLLIDER_RADIUS.MAX);
            parameter.primaryConfig = new MultiplierConfig
            {
                baseValue = 0.0236f,
                multiplierA = 1.03f,
                multiplierB = 0.20f,
            };
            parameter.secondaryConfig = new MultiplierConfig
            {
                baseValue = 0.0272f,
                multiplierA = 1.03f,
                multiplierB = 0.20f,
            };

            var gradientConfig = parameter.gradientConfig;
            gradientConfig.middleShiftJsf.SetValueAndDefault(0.00f);
            gradientConfig.outerMultiplierJsf.SetValueAndDefault(1.10f);
            gradientConfig.centerMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.outerToMiddleCurvature = 0.60f;
            gradientConfig.outerToMiddleMidPoint = 0.60f;

            var ldpConfig = parameter.ldpConfig;
            ldpConfig.compressionEffectJsf.SetValueAndDefault(0.10f);
            ldpConfig.compressionSensitivityJsf.SetValueAndDefault(0.70f);
            ldpConfig.stretchingEffectJsf.SetValueAndDefault(0.40f);
            ldpConfig.stretchingSensitivityJsf.SetValueAndDefault(0.70f);

            parameter.usesRealMass = true;
            return parameter;
        }

        static SoftPhysicsParameter FatColliderDepth(string side)
        {
            var parameter = NewSoftPhysicsParameter(side, FAT_COLLIDER_DEPTH.NAME, FAT_COLLIDER_DEPTH.MIN, FAT_COLLIDER_DEPTH.MAX);
            parameter.primaryConfig = new MultiplierConfig();
            parameter.secondaryConfig = new MultiplierConfig();

            var gradientConfig = parameter.gradientConfig;
            gradientConfig.middleShiftJsf.SetValueAndDefault(0.00f);
            gradientConfig.outerMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.centerMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.outerToMiddleCurvature = 0.60f;
            gradientConfig.outerToMiddleMidPoint = 0.60f;

            var ldpConfig = parameter.ldpConfig;
            ldpConfig.compressionEffectJsf.SetValueAndDefault(0.00f);
            ldpConfig.compressionSensitivityJsf.SetValueAndDefault(0.80f);
            ldpConfig.stretchingEffectJsf.SetValueAndDefault(1.00f);
            ldpConfig.stretchingSensitivityJsf.SetValueAndDefault(0.90f);

            return parameter;
        }

        static SoftPhysicsParameter FatDistanceLimit(string side)
        {
            var parameter = NewSoftPhysicsParameter(side, FAT_DISTANCE_LIMIT.NAME, FAT_DISTANCE_LIMIT.MIN, FAT_DISTANCE_LIMIT.MAX);
            parameter.primaryConfig = new MultiplierConfig
            {
                baseValue = 0.0204f,
                multiplierA = 1.17f,
                multiplierB = 0.26f,
                curveA = x => Calc.InverseSmoothStep(x, 1.00f, 0.22f, 0.10f),
            };
            parameter.secondaryConfig = new MultiplierConfig
            {
                baseValue = 0.0255f,
                multiplierA = 0.82f,
                multiplierB = 0.26f,
                curveA = x => Calc.InverseSmoothStep(x, 1.00f, 0.22f, 0.10f),
            };

            var gradientConfig = parameter.gradientConfig;
            gradientConfig.middleShiftJsf.SetValueAndDefault(0.00f);
            gradientConfig.outerMultiplierJsf.SetValueAndDefault(0.80f);
            gradientConfig.centerMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.outerToMiddleCurvature = 0.60f;
            gradientConfig.outerToMiddleMidPoint = 0.60f;

            var ldpConfig = parameter.ldpConfig;
            ldpConfig.compressionEffectJsf.SetValueAndDefault(0.00f);
            ldpConfig.compressionSensitivityJsf.SetValueAndDefault(0.80f);
            ldpConfig.stretchingEffectJsf.SetValueAndDefault(0.70f);
            ldpConfig.stretchingSensitivityJsf.SetValueAndDefault(0.80f);

            parameter.usesRealMass = true;
            return parameter;
        }

        static PhysicsParameter FatBackForce(string side)
        {
            var parameter = new PhysicsParameter(
                BootyMagic.script,
                side,
                FAT_BACK_FORCE.NAME,
                FAT_BACK_FORCE.MIN,
                FAT_BACK_FORCE.MAX
            )
            {
                sync = value => SyncBackForce(side, value),
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 2.88f,
                    multiplierA = 2.19f,
                    multiplierB = 8.75f,
                    curveA = x => Calc.InverseSmoothStep(x, 1.00f, 0.50f, 0.88f),
                    curveB = x => 1 - FatSpringSoftnessCurve(x),
                },
                secondaryConfig = new MultiplierConfig
                {
                    baseValue = 1.44f,
                    multiplierA = 1.10f,
                    multiplierB = 17.50f,
                    curveA = x => Calc.InverseSmoothStep(x, 1.00f, 0.50f, 0.88f),
                    curveB = x => 1 - FatSpringSoftnessCurve(x),
                },
                quicknessConfig = new MultiplierConfig
                {
                    baseValue = -0.36f,
                    multiplierA = 8.25f,
                    multiplierB = 0.00f,
                    curveA = x => Calc.InverseSmoothStep(x, 1.00f, 0.50f, 0.88f),
                },
                slownessConfig = new MultiplierConfig
                {
                    baseValue = 0.72f,
                    multiplierA = 8.25f,
                    multiplierB = 0.00f,
                },
            };

            return parameter;
        }

        static PhysicsParameter FatBackForceMaxForce(string side)
        {
            var parameter = new PhysicsParameter(
                BootyMagic.script,
                side,
                FAT_BACK_FORCE_MAX_FORCE.NAME,
                FAT_BACK_FORCE_MAX_FORCE.MIN,
                FAT_BACK_FORCE_MAX_FORCE.MAX
            )
            {
                sync = value => SyncBackForceMaxForce(side, value),
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 50.00f,
                    multiplierA = 0.00f,
                    multiplierB = 0.00f,
                },
                secondaryConfig = new MultiplierConfig
                {
                    baseValue = 50.00f,
                    multiplierA = 0.00f,
                    multiplierB = 0.00f,
                },
            };
            return parameter;
        }

        static PhysicsParameter FatBackForceThreshold(string side)
        {
            var parameter = new PhysicsParameter(
                BootyMagic.script,
                side,
                FAT_BACK_FORCE_THRESHOLD.NAME,
                FAT_BACK_FORCE_THRESHOLD.MIN,
                FAT_BACK_FORCE_THRESHOLD.MAX
            )
            {
                sync = value => SyncBackForceThresholdDistance(side, value),
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 0.001f,
                    multiplierA = 0.00f,
                    multiplierB = 0.00f,
                },
                secondaryConfig = new MultiplierConfig
                {
                    baseValue = 0.001f,
                    multiplierA = 0.00f,
                    multiplierB = 0.00f,
                },
            };
            return parameter;
        }

#endregion

#region *** Sync Functions ***

        static void SyncGradientParameter(string side, string paramName, GradientConfig gradientConfig, float value, float additionalOffset)
        {
            if(!BootyMagic.IsInitialized() || !BootyMagic.script.enabled || BootyMagic.restoringFromJson)
            {
                // Syncing parameter before first time calibration (e.g. due to offset storable callback) results in NaN errors
                return;
            }

            var gradientGroups = _gradientGroups[side];
            foreach(var group in gradientGroups)
            {
                group.SetGradientBaseValues(paramName, gradientConfig, value);
                group.SetAdditionalOffsets(paramName, additionalOffset);
            }

            /* Calculate remaining physics values for joints not included in gradient.
             * Values are based on selected neighboring joints.
             */
            foreach(var joint in _mainNonGradientJoints[side])
            {
                joint.SetBaseValue(
                    paramName,
                    joint.adjacentJoints.Average(x => x.GetBaseValue(paramName))
                );
                joint.SetAdditionalOffset(paramName, additionalOffset);
            }

            /* Calculate remaining physics value for lower non-gradient joints close to the vagina. */
            float lowerValue = _lowerReferenceGradientJoints[side].Average(x => x.GetBaseValue(paramName));
            foreach(var joint in _lowerNonGradientJoints[side])
            {
                joint.SetBaseValue(paramName, lowerValue);
                joint.SetAdditionalOffset(paramName, additionalOffset);
            }
        }

        static void SyncBackForce(string side, float value)
        {
            if(!BootyMagic.script.enabled)
            {
                return;
            }

            _softVerticesGroups[side].jointBackForce = value;
        }

        static void SyncBackForceMaxForce(string side, float value)
        {
            if(!BootyMagic.script.enabled)
            {
                return;
            }

            _softVerticesGroups[side].jointBackForceMaxForce = value;
        }

        static void SyncBackForceThresholdDistance(string side, float value)
        {
            if(!BootyMagic.script.enabled)
            {
                return;
            }

            _softVerticesGroups[side].jointBackForceThresholdDistance = value;
        }

        static bool _reverseSyncingSoftPhysics;

        /* Update value if external value changed */
        public static void ReverseSyncSoftPhysicsOn()
        {
            if(!_pluginSoftPhysicsToggled)
            {
                _reverseSyncingSoftPhysics = true;
                gluteSoftPhysicsOnJsb.val = Person.instance.glutePhysicsMesh.on;
            }

            _pluginSoftPhysicsToggled = false;
        }

        public static void SyncSoftPhysicsOn()
        {
            softPhysicsOn = Settings.personMonitor.GluteSoftPhysicsIsEnabled();
            _ldpOn = softPhysicsOn && ldpEnabledJsb.val;
            if(!softPhysicsSynced)
            {
                BootyMagic.script.leftGlute.SetSoftPhysicsBased(softPhysicsOn);
                BootyMagic.script.rightGlute.SetSoftPhysicsBased(softPhysicsOn);
            }

            softPhysicsSynced = true;
        }

        public static void SyncFriction(string side, float friction)
        {
            // TODO convert to gradient
            float centerFriction = friction + 0.33f * (1 - friction);

            foreach(var softJoint in mainSoftJoints[side])
            {
                SetSoftJointFriction(softJoint, friction);
            }

            foreach(var softJoint in outerSoftJoints[side])
            {
                SetSoftJointFriction(softJoint, friction);
            }

            foreach(var softJoint in centerSoftJoints[side])
            {
                SetSoftJointFriction(softJoint, centerFriction);
            }
        }

        static void SetSoftJointFriction(SoftJoint softJoint, float value)
        {
            var material = softJoint.softVerticesSet.jointCollider.material;
            material.dynamicFriction = value;
            material.staticFriction = value;
        }

#endregion

        public static void SetShowColliderPreviews(bool hasSoftColliderVisualization)
        {
            colliderVisualizer.ShowPreviewsJSON.val = showColliderPreviewsJsb.val && hasSoftColliderVisualization;
        }

        public static void SaveCalibrated()
        {
            ldpEnabledJsb.SaveCalibrated();
        }

        public static void UpdatePhysics(bool updateOffsetMinMax = true)
        {
            if(!softPhysicsOn)
            {
                return;
            }

            foreach(var kvp in _gradientGroups)
            {
                foreach(var group in kvp.Value)
                {
                    group.CalculateDistances();
                }
            }

            foreach(var group in softParameterGroups)
            {
                group.UpdateBaseValue(
                    group.left.usesRealMass
                        ? MainPhysicsHandler.normalizedRealMass
                        : MainPhysicsHandler.normalizedMass,
                    BootyMagic.softness,
                    BootyMagic.quickness,
                    LdpUpdateIsEnabled()
                );
                if(updateOffsetMinMax)
                {
                    group.UpdateOffsetMinMax();
                }
            }

            foreach(var group in parameterGroups)
            {
                group.UpdateBaseValue(
                    MainPhysicsHandler.normalizedMass,
                    BootyMagic.softness,
                    BootyMagic.quickness,
                    softPhysicsOn
                );
                if(updateOffsetMinMax)
                {
                    group.UpdateOffsetMinMax();
                }
            }
        }

        public static bool SoftJointsMoved(float tolerance)
        {
            return _allSoftJoints[LEFT].Any(softJoint => softJoint.Moved(tolerance)) ||
                _allSoftJoints[RIGHT].Any(softJoint => softJoint.Moved(tolerance));
        }

        public static void CalibrateSoftJointsDistortion(string direction)
        {
            foreach(var kvp in _allSoftJoints)
            {
                foreach(var softJoint in kvp.Value)
                {
                    softJoint.CalibrateDistortion(direction);
                }
            }
        }

        public static void UpdateSoftJointsDistortion(bool force = false)
        {
            UpdateSoftJointsDistortion(LEFT, force);
            UpdateSoftJointsDistortion(RIGHT, force);
        }

        public static void UpdateSoftJointsDistortion(string side, bool force = false)
        {
            if(!LdpUpdateIsEnabled())
            {
                return;
            }

            foreach(var softJoint in _allSoftJoints[side])
            {
                if(softJoint.UpdateDistortion(force))
                {
                    softJoint.SyncSpringAndDamperDistortion(side);
                    softJoint.SyncMassDistortion(side);
                    softJoint.SyncColliderRadiusAndDepthDistortion(side);
                    softJoint.SyncDistanceLimitDistortion(side);
                }
            }
        }

        public static void ResetSoftJointsDistortion()
        {
            foreach(var kvp in _allSoftJoints)
            {
                foreach(var softJoint in kvp.Value)
                {
                    softJoint.ResetDistortion();
                }
            }
        }

        public static void EnableMultiplyFriction()
        {
            foreach(var kvp in _allSoftJoints)
            {
                foreach(var softJoint in kvp.Value)
                {
                    softJoint.EnableMultiplyFriction();
                }
            }
        }

        public static JSONClass SerializeCalibrationData()
        {
            var jc = new JSONClass();
            if(!LdpUpdateIsEnabled())
            {
                return jc;
            }

            foreach(var kvp in _allSoftJoints)
            {
                foreach(var softJoint in kvp.Value)
                {
                    jc[softJoint.id] = softJoint.SerializeCalibrationData();
                }
            }

            return jc;
        }

        public static void CalibrateFromJson(JSONClass jc)
        {
            if(!LdpUpdateIsEnabled())
            {
                return;
            }

            foreach(var kvp in _allSoftJoints)
            {
                foreach(var softJoint in kvp.Value)
                {
                    softJoint.CalibrateFromJson(jc[softJoint.id].AsObject);
                }
            }
        }

        public static void RestoreOriginalPhysics()
        {
            if(!BootyMagic.IsInitialized())
            {
                return;
            }

            foreach(string name in Person.instance.glutePhysicsMesh.GetFloatParamNames())
            {
                /* Set a value that is different from the original, then restore the original
                 * in order to trigger VaM's internal sync
                 */
                var paramJsf = Person.instance.glutePhysicsMesh.GetFloatJSONParam(name);
                float original = paramJsf.val;
                paramJsf.valNoCallback = Mathf.Abs(paramJsf.val - paramJsf.min) > 0.01f
                    ? paramJsf.min
                    : paramJsf.max;
                paramJsf.val = original;
            }

            foreach(var kvp in _allSoftJoints)
            {
                foreach(var softJoint in kvp.Value)
                {
                    softJoint.RestoreOriginalFriction();
                }
            }
        }

        public static void Destroy()
        {
            Object.Destroy(colliderVisualizer);
            colliderVisualizer = null;
            softJointDiagnostics?.Destroy();
            softJointDiagnostics = null;
            showColliderPreviewsJsb = null;

            _allSoftJoints = null;
            outerSoftJoints = null;
            mainSoftJoints = null;
            centerSoftJoints = null;

            _gradientGroups = null;
            _mainNonGradientJoints = null;
            _lowerReferenceGradientJoints = null;
            _lowerNonGradientJoints = null;

            _softVerticesGroups = null;
            _softVerticesSets = null;

            parameterGroups = null;
            softParameterGroups = null;

            fatSpring = null;
            fatDamper = null;
            fatMass = null;
            fatColliderRadius = null;
            fatColliderDepth = null;
            fatDistanceLimit = null;

            fatBackForce = null;
            fatBackForceMaxForce = null;
            fatBackForceThreshold = null;

            softJointsUpdater = null;
            gluteSoftPhysicsOnJsb = null;
            ldpEnabledJsb = null;
            showColliderPreviewsJsb = null;
        }
    }
}
