using System;
using System.Linq;
using Naturalis.BM.Handlers;
using Naturalis.Common;
using SimpleJSON;
using UnityEngine;

namespace Naturalis.BM.Components
{
    sealed class Glute
    {
        readonly string _side;
        readonly Rigidbody _gluteRb;
        readonly Transform _gluteRbT;
        readonly Rigidbody _hipRb;
        readonly Transform _hipRbT;

        public Glute(string side)
        {
            _side = side;
            _gluteRb = Person.instance.gluteRbs[_side];
            _gluteRbT = _gluteRb.transform;
            _hipRb = Person.instance.hipRb;
            _hipRbT = _hipRb.transform;
        }

        Func<Vector3> _calculateSimulatedJointPosition;
        Func<Vector3> _calculateRelativePosition;
        Func<float> _calculateRelativeDepth;

        public void SetSoftPhysicsBased(bool softPhysicsOn)
        {
            _calculateSimulatedJointPosition = () =>
            {
                var originRotationDiff = _hipRb.rotation * Quaternion.Inverse(_calibratedOriginRotation);
                return _hipRb.position + originRotationDiff * _calibratedOffsetFromHipRb;
            };

            if(softPhysicsOn)
            {
                var softJointOuterRbs = SoftPhysicsHandler.GetPositionalRigidbodies(_side, SoftVerticesGroupName.OUTER);
                var softJointMainRbs = SoftPhysicsHandler.GetPositionalRigidbodies(_side, SoftVerticesGroupName.MAIN);
                var softJointCenterRbs = SoftPhysicsHandler.centerSoftJoints[_side].Select(softJoint => softJoint.jointRb).ToArray();
                var combinedRbs = softJointMainRbs
                    .Concat(softJointCenterRbs)
                    .Concat(softJointOuterRbs)
                    .ToArray();

                _calculateRelativePosition = () => Calc.RelativePosition(
                    _simulatedJointPosition,
                    _hipRbT,
                    _gluteRbT,
                    Calc.AverageRigidbodiesPosition(combinedRbs)
                );
                _calculateRelativeDepth = () =>
                {
                    var centerToGlute = Calc.RelativePosition(
                        _simulatedJointPosition,
                        _hipRbT,
                        _gluteRbT,
                        Calc.AverageRigidbodiesPosition(softJointCenterRbs)
                    );
                    var outerToGlute = Calc.RelativePosition(
                        _simulatedJointPosition,
                        _hipRbT,
                        _gluteRbT,
                        Calc.AverageRigidbodiesPosition(softJointOuterRbs)
                    );
                    float softJointsDepth = (centerToGlute - outerToGlute).z;
                    return (JointsRelativeDepth() + softJointsDepth) / 2;
                };
            }
            else
            {
                /* Position of center skin vertices relative to glute */
                _centerVertexIndices = VertexIndices.CenterPositionalVertsSoftPhysicsOff(_side);
                _calculateRelativePosition = SkinVerticesRelativePosition;
                _calculateRelativeDepth = JointsRelativeDepth;
            }
        }

        int[] _centerVertexIndices;

        Vector3 SkinVerticesRelativePosition()
        {
            var rawSkinnedVerts = Person.instance.skin.rawSkinnedVerts;
            var vertices = new Vector3[_centerVertexIndices.Length];
            for(int i = 0; i < _centerVertexIndices.Length; i++)
            {
                vertices[i] = rawSkinnedVerts[_centerVertexIndices[i]];
            }

            return Calc.RelativePosition(_simulatedJointPosition, _hipRbT, Calc.AveragePosition(vertices));
        }

        float JointsRelativeDepth()
        {
            var jointsRelativePosition = Calc.RelativePosition(
                _simulatedJointPosition,
                _hipRbT,
                _gluteRbT,
                _hipRb.position
            );
            return jointsRelativePosition.z;
        }

        float _prevXAngle;
        float _prevYAngle;

        public bool Moved(float tolerance)
        {
            float xAngle = CalculateXAngle();
            float yAngle = CalculateYAngle();
            bool moved = Mathf.Abs(xAngle - _prevXAngle) > tolerance || Mathf.Abs(yAngle - _prevYAngle) > tolerance;
            _prevXAngle = xAngle;
            _prevYAngle = yAngle;
            return moved;
        }

        Vector3 _calibratedOffsetFromHipRb;
        Quaternion _calibratedOriginRotation;
        Vector2 _angleXBaseVector;
        Vector2 _angleYBaseVector;

        Vector3 _simulatedJointPosition;
        Vector3 _positionRelativeToOrigin;
        public float angleY { get; private set; }
        public float angleX { get; private set; }
        public float depthDiff { get; private set; }
        float _depthBase;

        public void Calibrate()
        {
            _calibratedOffsetFromHipRb = _gluteRb.position - _hipRb.position;
            _calibratedOriginRotation = _hipRb.rotation;

            _simulatedJointPosition = _calculateSimulatedJointPosition();
            _positionRelativeToOrigin = _calculateRelativePosition();
            _angleXBaseVector = new Vector2(_positionRelativeToOrigin.z, _positionRelativeToOrigin.x);
            _angleYBaseVector = new Vector2(_positionRelativeToOrigin.z, _positionRelativeToOrigin.y);
            _depthBase = _calculateRelativeDepth();
        }

        public void Measure()
        {
            _simulatedJointPosition = _calculateSimulatedJointPosition();
            _positionRelativeToOrigin = _calculateRelativePosition();
            angleX = CalculateXAngle();
            angleY = CalculateYAngle();
            depthDiff = _depthBase - _calculateRelativeDepth();
        }

        float CalculateXAngle()
        {
            return Vector2.SignedAngle(_angleXBaseVector, new Vector2(_positionRelativeToOrigin.z, _positionRelativeToOrigin.x));
        }

        float CalculateYAngle()
        {
            return Vector2.SignedAngle(_angleYBaseVector, new Vector2(_positionRelativeToOrigin.z, _positionRelativeToOrigin.y));
        }

        public void ResetMeasurements()
        {
            angleY = 0;
            angleX = 0;
            depthDiff = 0;
        }

        public JSONClass SerializeCalibrationData()
        {
            var jc = new JSONClass
            {
                ["offset_from_origin"] = JSONUtils.SerializeVector3(_calibratedOffsetFromHipRb),
                ["origin_rotation"] = JSONUtils.SerializeQuaternion(_calibratedOriginRotation),
                ["angle_x_base_vector"] = JSONUtils.SerializeVector2(_angleXBaseVector),
                ["angle_y_base_vector"] = JSONUtils.SerializeVector2(_angleYBaseVector),
                ["depth_base"] = { AsFloat = _depthBase },
            };
            return jc;
        }

        public void CalibrateFromJson(JSONClass jc)
        {
            _calibratedOffsetFromHipRb = JSONUtils.DeserializeVector3(jc["offset_from_origin"]);
            _calibratedOriginRotation = JSONUtils.DeserializeQuaternion(jc["origin_rotation"]);
            _angleXBaseVector = JSONUtils.DeserializeVector2(jc["angle_x_base_vector"]);
            _angleYBaseVector = JSONUtils.DeserializeVector2(jc["angle_y_base_vector"]);
            _depthBase = jc["depth_base"].AsFloat;
        }
    }
}
