﻿using System.Collections.Generic;
using System.Linq;
using Naturalis.BM.Models;
using Naturalis.Common;
using Naturalis.Common.Diagnostics;
using Naturalis.Common.Models;
using UnityEngine;

namespace Naturalis.BM
{
    sealed class SoftJointDiagnostics
    {
        readonly static Color _gridColor = new Color(1f, 1f, 1f, 0.035f);

        readonly DrawService _drawService;
        readonly SoftJoint[] _softJoints;

        public JSONStorableBool visualizeGradientJsb { get; }
        public JSONStorableBool visualizeLdpJsb { get; }
        public JSONStorableStringChooser softJointRbDiagnosticJssc { get; }

        SoftJoint _selected;

        public SoftJointDiagnostics(SoftJoint[] softJoints)
        {
            _softJoints = softJoints;

            var options = new List<string> { "Select..." };
            options.AddRange(_softJoints.Select(softJoint => softJoint.id));

            softJointRbDiagnosticJssc = new JSONStorableStringChooser(
                "softJointRb",
                options,
                options,
                "Select...",
                "Soft Joint",
                value => _selected = _softJoints.FirstOrDefault(softJoint => softJoint.id == value)
            );

            _drawService = BootyMagic.script.gameObject.AddComponent<DrawService>();
            _drawService.Init(Color.gray, true);

            visualizeGradientJsb = new JSONStorableBool("Visualize Gradient", true);
            visualizeLdpJsb = new JSONStorableBool("Visualize Local Distortion", false);
        }

        public void DrawServiceSetEnabled(bool value)
        {
            if(!_drawService)
            {
                return;
            }

            ClearDiagnostics();
            _drawService.enabled = value;
        }

        public void ClearDiagnostics()
        {
            _drawService.lines.Clear();
        }

        public void DrawDiagnostics(SoftPhysicsParameterGroup group, string paramName)
        {
            if(BootyMagic.updatePaused || !_drawService.enabled)
            {
                return;
            }

            var param = group.left;
            float visualizationMin = -Mathf.Max(group.visualizationLimit, param.offsetJsf.min - param.offsetJsf.val);
            float visualizationMax = Mathf.Min(group.visualizationLimit, param.offsetJsf.max - param.offsetJsf.val);

            if(visualizeGradientJsb.val)
            {
                float baseWithOffset = param.baseValueJsf.val + param.offsetJsf.val;
                foreach(var softJoint in _softJoints)
                {
                    var color = JointRelativeColor(
                        softJoint.GetBaseValue(paramName) - baseWithOffset,
                        baseWithOffset,
                        visualizationMin,
                        visualizationMax
                    );
                    _drawService.lines.AddRange(PositionMarker(softJoint.jointRb.position, color, 0.002f));
                }
            }

            if(visualizeLdpJsb.val)
            {
                if(group.visualizeRelativeDistortion)
                {
                    foreach(var softJoint in _softJoints)
                    {
                        float jointBaseValue = softJoint.GetBaseValue(paramName);
                        var color = JointRelativeColor(
                            softJoint.GetDistortionValue(paramName),
                            jointBaseValue,
                            visualizationMin,
                            visualizationMax
                        );
                        _drawService.lines.AddRange(JointConnections(softJoint, color));
                    }
                }
                else
                {
                    foreach(var softJoint in _softJoints)
                    {
                        var color = JointColor(softJoint.GetDistortionValue(paramName), visualizationMin, visualizationMax);
                        _drawService.lines.AddRange(JointConnections(softJoint, color));
                    }
                }
            }
        }

        public void DrawSelectedJoint()
        {
            if(_selected != null)
            {
                var pos = _selected.jointRb.position;
                _drawService.lines.AddRange(PositionMarker(pos, Color.green));
                _drawService.lines.AddRange(JointConnections(_selected, Color.white));
            }
        }

        public string GetSelectedJointValues(SoftPhysicsParameterGroup group)
        {
            string paramName = group.name;
            int roundFactor = group.roundFactor;

            if(_selected != null)
            {
                string str = "";
                float distance = Calc.RoundToDecimals(10 * _selected.distortion, 10);

                float paramBase = _selected.GetBaseValue(paramName);
                float paramOffset = group.left.offsetJsf.val;
                float paramBaseWithOffset = paramBase + paramOffset;

                float baseWithOffsetValue = Calc.RoundToDecimals(paramBaseWithOffset, roundFactor);
                float baseValue = Calc.RoundToDecimals(paramBase, roundFactor);
                float offsetValue = Calc.RoundToDecimals(paramOffset, roundFactor);
                float distortionValue = Calc.RoundToDecimals(_selected.GetDistortionValue(paramName), roundFactor);
                float realValue = Calc.RoundToDecimals(_selected.GetRealValue(paramName), roundFactor);
                str += $"\nBase+Offset value: {baseWithOffsetValue.ToString().Bold()}";
                str += $"\nBase Value: {baseValue.ToString().Bold()}";
                str += $"\nOffset Value: {offsetValue.ToString().Bold()}";
                str += $"\nDistortion: {distance.ToString()} mm";
                str += $"\nDistortion Value: {distortionValue.ToString().Bold()}";
                str += $"\nReal Value: {realValue.ToString().Bold()}";
                return "\n".Size(6) + str;
            }

            return "";
        }

        static IEnumerable<LineV3> PositionMarker(Vector3 pos, Color color, float size = 0.003f)
        {
            return new List<LineV3>
            {
                new LineV3
                {
                    from = pos + new Vector3(-size, 0, 0),
                    to = pos + new Vector3(size, 0, 0),
                    color = color,
                },
                new LineV3
                {
                    from = pos + new Vector3(0, -size, 0),
                    to = pos + new Vector3(0, size, 0),
                    color = color,
                },
                new LineV3
                {
                    from = pos + new Vector3(0, 0, -size),
                    to = pos + new Vector3(0, 0, size),
                    color = color,
                },
            };
        }

        static IEnumerable<LineV3> JointConnections(SoftJoint joint, Color color)
        {
            var origin = joint.jointRb.position;
            foreach(var target in joint.nearbyJoints)
            {
                yield return new LineV3 { from = origin, to = (origin + target.jointRb.position) / 2, color = color };
            }
        }

        static Color JointColor(float diff, float min, float max)
        {
            return diff < 0
                ? Color.Lerp(Color.cyan, _gridColor, Mathf.InverseLerp(min, 0, diff))
                : Color.Lerp(_gridColor, Color.red, Mathf.InverseLerp(0, max, diff));

        }

        static Color JointRelativeColor(float diff, float offset, float min, float max)
        {
            if(diff < 0)
            {
                min -= offset;
                return min >= 0
                    ? Color.cyan
                    : Color.Lerp(Color.cyan, _gridColor, Mathf.InverseLerp(min, 0, diff));

            }

            max += offset;
            return max <= 0
                ? Color.red
                : Color.Lerp(_gridColor, Color.red, Mathf.InverseLerp(0, max, diff));

        }

        public void Destroy()
        {
            Object.Destroy(_drawService);
        }
    }
}
