using System;
using System.Collections.Generic;
using System.Linq;
using Naturalis.BM.Handlers;
using Naturalis.Common;
using Naturalis.Common.Diagnostics;
using Naturalis.Common.Models;
using Naturalis.Common.UI;
using UnityEngine;
using UnityEngine.Events;

namespace Naturalis.BM.UI
{
    sealed class DevWindow : WindowBase
    {
        readonly DrawService _drawService;
        readonly int[] _centerVertexIndices;

        readonly JSONStorableString _leftDebugArea;
        readonly JSONStorableString _jointDiagnosticsJss;
        readonly JSONStorableStringChooser _softParamJssc;
        SoftPhysicsParameterGroup _selectedParam;

        public DevWindow(UnityAction onReturnToParent = null) : base(BootyMagic.script, nameof(DevWindow), onReturnToParent)
        {
            {
                _drawService = BootyMagic.script.gameObject.AddComponent<DrawService>();
                _drawService.Init(Color.white, true);
                _centerVertexIndices = VertexIndices.CenterPositionalVertsSoftPhysicsOff(Side.LEFT);
            }

            _leftDebugArea = new JSONStorableString("leftDebugArea", "");
            _jointDiagnosticsJss = new JSONStorableString("jointDiagnostics", "");
            var choices = SoftPhysicsHandler.softParameterGroups.Select(group => group.name).ToList();
            _softParamJssc = new JSONStorableStringChooser("softPhysicsParams", choices, choices.First(), "Params")
            {
                setCallbackFunction = OnParamOptionSelected,
            };
            OnParamOptionSelected(_softParamJssc.val);
        }

        public Vector3 Deg2Rad(Vector3 eulerAngles)
        {
            float x = Mathf.Deg2Rad * eulerAngles.x;
            float y = Mathf.Deg2Rad * eulerAngles.y;
            float z = Mathf.Deg2Rad * eulerAngles.z;
            return new Vector3(x, y, z);
        }

        static Vector3 VerticesAveragePosition(int[] indices)
        {
            var vertices = new Vector3[indices.Length];
            for(int i = 0; i < indices.Length; i++)
            {
                vertices[i] = Person.instance.skin.rawSkinnedVerts[indices[i]];
            }

            return Calc.AveragePosition(vertices);
        }

        static IEnumerable<LineV3> PositionMarker(Vector3 pos, Color color, float size = 0.003f)
        {
            return new List<LineV3>
            {
                new LineV3
                {
                    from = pos + new Vector3(-size, 0, 0),
                    to = pos + new Vector3(size, 0, 0),
                    color = color,
                },
                new LineV3
                {
                    from = pos + new Vector3(0, -size, 0),
                    to = pos + new Vector3(0, size, 0),
                    color = color,
                },
                new LineV3
                {
                    from = pos + new Vector3(0, 0, -size),
                    to = pos + new Vector3(0, 0, size),
                    color = color,
                },
            };
        }

        void OnParamOptionSelected(string value)
        {
            _selectedParam = SoftPhysicsHandler.softParameterGroups.First(group => group.name == value);
        }

        protected override void OnBuild()
        {
            AddSpacer(50, true);
            AddElement(() =>
            {
                var textField = script.CreateTextField(_leftDebugArea, false);
                textField.UItext.fontSize = 28;
                textField.height = Constant.UI_MAX_HEIGHT - 130;
                return textField;
            });

            AddElement(() =>
            {
                var popup = script.CreatePopupAuto(SoftPhysicsHandler.softJointDiagnostics.softJointRbDiagnosticJssc, true, 480f);
                popup.popup.labelText.color = Color.black;
                return popup;
            });

            AddElement(() =>
            {
                var popup = script.CreatePopupAuto(_softParamJssc, true);
                popup.popup.labelText.color = Color.black;
                return popup;
            });

            AddElement(() =>
            {
                var textField = script.CreateTextField(_jointDiagnosticsJss, true);
                textField.UItext.fontSize = 28;
                textField.height = 250;
                return textField;
            });
        }

        protected override void OnClose()
        {
            ClearDiagnostics();
            SoftPhysicsHandler.softJointDiagnostics.DrawServiceSetEnabled(false);
        }

        public void Update()
        {
            // DebugAngles(BootyMagic.script.leftGlute);
            // ClearDiagnostics();
            // DrawDiagnostics();
            // DebugSkinVerts();
            // DebugSoftJoints();
        }

        // void DebugAngles(Glute glute)
        // {
        //     if(Time.frameCount % 6 != 0)
        //     {
        //         return;
        //     }
        //
        //     float angleY = Calc.RoundToDecimals(glute.angleY, 100);
        //     float angleX = Calc.RoundToDecimals(glute.angleX, 100);
        //     float depthDiff = Calc.RoundToDecimals(glute.depthDiff, 100);
        //     float angleYPct = Calc.RoundToDecimals(100 * angleY / ForceMorphHandler._upDownLimit, 10);
        //     float angleXPct = Calc.RoundToDecimals(100 * angleX / ForceMorphHandler._sidewaysLimit, 10);
        //     float depthDiffPct = Calc.RoundToDecimals(100 * depthDiff / ForceMorphHandler._depthInLimit, 10);
        //     string angleYPctString = angleYPct < 100 ? angleYPct.ToString() : $"{angleYPct}".Color(Color.red);
        //     string angleXPctString = angleXPct < 100 ? angleXPct.ToString() : $"{angleXPct}".Color(Color.red);
        //     string depthDiffPctString = depthDiffPct < 100 ? depthDiffPct.ToString() : $"{depthDiffPct}".Color(Color.red);
        //     _leftDebugArea.val =
        //         $"\nangleY {angleY}" +
        //         $"\nangleX {angleX}" +
        //         $"\ndepthDiff {depthDiff}" +
        //         $"\nangleYPct {angleYPctString}" +
        //         $"\nangleXPct {angleXPctString}" +
        //         $"\ndepthDiffPct {depthDiffPctString}" +
        //         $"";
        // }

        public void ClearDiagnostics()
        {
            _drawService.lines.Clear();
        }

        void DrawDiagnostics()
        {
        }

        void DebugSkinVerts()
        {
            try
            {
                string str = "";
                int[] indices = VertexIndices.VolumeCalculationVerts(Side.LEFT);
                var rawSkinnedVerts = Person.instance.skin.rawSkinnedVerts;
                foreach(int i in indices)
                {
                    var vertex = rawSkinnedVerts[i];
                    str +=
                        $"\n\n{i}:" +
                        $" x {Calc.RoundToDecimals(vertex.x)}" +
                        $" y {Calc.RoundToDecimals(vertex.y)}" +
                        $" z {Calc.RoundToDecimals(vertex.z)}";
                }

                _leftDebugArea.val = str;
            }
            catch(Exception e)
            {
                Debug.Log($"{nameof(DebugSkinVerts)}: {e}");
            }
        }

        void DebugSoftJoints()
        {
            try
            {
                if(_selectedParam != null)
                {
                    if(Time.frameCount % 6 == 0)
                    {
                        _jointDiagnosticsJss.val = SoftPhysicsHandler.softJointDiagnostics.GetSelectedJointValues(_selectedParam);
                    }
                }

                SoftPhysicsHandler.softJointDiagnostics.ClearDiagnostics();
                SoftPhysicsHandler.softJointDiagnostics.DrawSelectedJoint();
                if(_selectedParam != null)
                {
                    SoftPhysicsHandler.softJointDiagnostics.DrawDiagnostics(_selectedParam, _selectedParam.name);
                }
            }
            catch(Exception e)
            {
                Debug.Log($"{nameof(DebugSoftJoints)}: {e}");
            }
        }
    }
}
