using System;
using System.Collections.Generic;
using Naturalis.BM.Handlers;
using Naturalis.Common.UI;
using UnityEngine;

namespace Naturalis.BM.UI
{
    sealed class HardCollidersWindow : WindowBase, IWindow
    {
        readonly Dictionary<string, UIDynamic> _colliderSectionElements = new Dictionary<string, UIDynamic>();
        readonly ColliderVisualizer _visualizer = HardColliderHandler.colliderVisualizer;

        public HardCollidersWindow() : base(BootyMagic.script, nameof(HardCollidersWindow))
        {
        }

        protected override void OnBuild()
        {
            BuildLeftSide();
            BuildRightSide();
            AddShowPreviewsPopupChangeHandler();
            AddColliderPopupChangeHandler();
            _visualizer.enabled = true;
            _visualizer.ShowPreviewsJSON.val = true;
        }

        void BuildLeftSide(bool rightSide = false)
        {
            AddHeaderTextField("Configure Hard Colliders", rightSide);

            AddInfoTextField(
                "Hard colliders apply a force to the back to the glute joint," +
                " making the glute move in response to collision.",
                rightSide
            );

            AddElement(HardColliderHandler.baseMassMultiplierJsf.name, () =>
            {
                var slider = script.CreateSlider(HardColliderHandler.baseMassMultiplierJsf, rightSide);
                slider.valueFormat = "F2";
                slider.label = "Base Mass Multiplier";
                return slider;
            });

            AddElement(() =>
                CreateValueTextField(HardColliderHandler.allCollidersMassDisplayValueJss, 28, 45, rightSide)
            );

            AddInfoTextField(
                "One third of Glute Mass is applied to hard colliders, the rest is applied to the glute joint." +
                "\n\n<b>Collider mass</b> affects the glute joint's inertia as well as the amount of force" +
                " applied back to the glute joint." +
                "\n\n<b>Radius</b> adjusts the size of the selected collider." +
                "\n\n<b>X, Y and Z offsets</b> can be used to modify the position" +
                " of the selected collider." +
                "\n\nThe closer the collider matches the volume of the glute, the more" +
                " accurately it responds to touch.",
                rightSide,
                525
            );

            AddElement(_visualizer.GroupsJSON.name, () =>
            {
                var popup = script.CreatePopup(_visualizer.GroupsJSON, rightSide);
                popup.label = "Show Previews";
                popup.popup.labelText.color = Color.black;
                popup.popup.popupPanel.offsetMin += new Vector2(5, popup.popup.popupPanelHeight + 55);
                popup.popup.popupPanel.offsetMax += new Vector2(5, popup.popup.popupPanelHeight + 55);
                return popup;
            });

            AddElement(_visualizer.XRayPreviewsJSON.name, () =>
            {
                var toggle = script.CreateToggle(_visualizer.XRayPreviewsJSON, rightSide);
                toggle.label = "Use XRay Previews";
                return toggle;
            });

            AddElement(HardColliderHandler.highlightAllJsb.name, () =>
                script.CreateToggle(HardColliderHandler.highlightAllJsb, rightSide)
            );
        }

        void BuildRightSide(bool rightSide = true)
        {
            AddSpacer(40, rightSide);
            AddElement(HardColliderHandler.colliderGroupsJssc.name, () =>
            {
                var popup = script.CreatePopupAuto(HardColliderHandler.colliderGroupsJssc, rightSide, 360f);
                popup.popup.labelText.color = Color.black;
                return popup;
            });
        }

        protected override void OnClose()
        {
            _visualizer.ShowPreviewsJSON.val = false;
            _visualizer.enabled = false;
        }

        void RebuildColliderSection(string colliderId, bool rightSide = true)
        {
            var hardColliderGroup = HardColliderHandler.hardColliderGroups
                .Find(group => group.visualizerId == colliderId);

            AddColliderSectionElement(hardColliderGroup.massMultiplierJsf.name, () =>
            {
                var slider = script.CreateSlider(hardColliderGroup.massMultiplierJsf, rightSide);
                slider.valueFormat = "F2";
                slider.label = "Mass Multiplier";
                return slider;
            });

            AddColliderSectionElement(hardColliderGroup.left.massDisplayValueJss.name, () =>
                CreateValueTextField(hardColliderGroup.left.massDisplayValueJss, 28, 45, rightSide)
            );

            AddColliderSectionElement(hardColliderGroup.radiusJsf.name, () =>
            {
                var slider = script.CreateSlider(hardColliderGroup.radiusJsf, rightSide);
                slider.valueFormat = "F2";
                slider.label = "Radius Offset";
                return slider;
            });

            AddColliderSectionElement(hardColliderGroup.rightJsf.name, () =>
            {
                var slider = script.CreateSlider(hardColliderGroup.rightJsf, rightSide);
                slider.valueFormat = "F2";
                slider.label = "X Offset";
                return slider;
            });

            AddColliderSectionElement(hardColliderGroup.upJsf.name, () =>
            {
                var slider = script.CreateSlider(hardColliderGroup.upJsf, rightSide);
                slider.valueFormat = "F2";
                slider.label = "Y Offset";
                return slider;
            });

            AddColliderSectionElement(hardColliderGroup.lookJsf.name, () =>
            {
                var slider = script.CreateSlider(hardColliderGroup.lookJsf, rightSide);
                slider.valueFormat = "F2";
                slider.label = "Z Offset";
                return slider;
            });
        }

        void AddColliderSectionElement(string key, Func<UIDynamic> createElement)
        {
            _colliderSectionElements[key] = createElement();
        }

        public new void Clear()
        {
            base.Clear();
            ClearColliderSection();
            RemoveShowPreviewsPopupChangeHandler();
            RemoveColliderPopupChangeHandler();
        }

        void ClearColliderSection()
        {
            foreach(var kvp in _colliderSectionElements)
            {
                script.RemoveElement(kvp.Value);
            }

            _colliderSectionElements.Clear();
        }

        void AddShowPreviewsPopupChangeHandler()
        {
            var uiDynamicPopup = GetElementAs<UIDynamicPopup>(_visualizer.GroupsJSON.name);
            if(uiDynamicPopup)
            {
                uiDynamicPopup.popup.onValueChangeHandlers += OnShowPreviewsPopupValueChanged;
                GetElement(_visualizer.XRayPreviewsJSON.name).SetActiveStyle(uiDynamicPopup.popup.currentValue != "Off");
            }
        }

        void RemoveShowPreviewsPopupChangeHandler()
        {
            var popup = GetElementAs<UIDynamicPopup>(_visualizer.GroupsJSON.name);
            if(popup)
            {
                popup.popup.visible = false;
                popup.popup.onValueChangeHandlers -= OnShowPreviewsPopupValueChanged;
            }
        }

        void OnShowPreviewsPopupValueChanged(string value)
        {
            GetElement(_visualizer.XRayPreviewsJSON.name).SetActiveStyle(value != "Off");
            GetElement(HardColliderHandler.highlightAllJsb.name).SetActiveStyle(value != "Off");
        }

        void AddColliderPopupChangeHandler()
        {
            var popup = GetElementAs<UIDynamicPopup>(HardColliderHandler.colliderGroupsJssc.name);
            if(popup)
            {
                popup.popup.onValueChangeHandlers += OnColliderPopupValueChanged;
                RebuildColliderSection(HardColliderHandler.colliderGroupsJssc.val);
            }
        }

        void RemoveColliderPopupChangeHandler()
        {

            var popup = GetElementAs<UIDynamicPopup>(HardColliderHandler.colliderGroupsJssc.name);
            if(popup)
            {
                popup.popup.visible = false;
                popup.popup.onValueChangeHandlers -= OnColliderPopupValueChanged;
            }
        }

        void OnColliderPopupValueChanged(string value)
        {
            ClearColliderSection();
            RebuildColliderSection(value);
        }
    }
}
