using Naturalis.BM.Handlers;
using Naturalis.Common.UI;
using UnityEngine;

namespace Naturalis.BM.UI
{
    sealed class MainWindow : WindowBase
    {
        JSONStorableAction configureColliderFrictionAction { get; }
        JSONStorableAction openSettingsWindowAction { get; }
        public JSONStorableAction openDevWindowAction { get; }
        public JSONStorableAction openDevMorphWindowAction { get; }

        public MainWindow() : base(BootyMagic.script, nameof(MainWindow))
        {
            openSettingsWindowAction = new JSONStorableAction(
                "openSettings",
                () =>
                {
                    ClearSelf();
                    activeNestedWindow = nestedWindows.Find(window => window.GetId() == nameof(SettingsWindow));
                    activeNestedWindow.Rebuild();
                }
            );
            nestedWindows.Add(new SettingsWindow(BootyMagic.script, OnReturn));

            configureColliderFrictionAction = new JSONStorableAction(
                "configureColliderFriction",
                () =>
                {
                    if(!BootyMagic.script.enabled)
                    {
                        BootyMagic.logger.LogMessage("Enable the plugin to configure collider friction.");
                        return;
                    }

                    BootyMagic.bindings.actions[$"{Constant.OPEN_UI}_{Constant.FRICTION}"].actionCallback();
                }
            );

            if(Settings.envIsDevelopment)
            {
                openDevWindowAction = new JSONStorableAction(
                    "openDevWindow",
                    () =>
                    {
                        ClearSelf();
                        activeNestedWindow = nestedWindows.Find(window => window.GetId() == nameof(DevWindow));
                        activeNestedWindow.Rebuild();
                    }
                );
                nestedWindows.Add(new DevWindow(OnReturn));

                openDevMorphWindowAction = new JSONStorableAction(
                    "openDevMorphWindow",
                    () =>
                    {
                        ClearSelf();
                        activeNestedWindow = nestedWindows.Find(window => window.GetId() == nameof(DevMorphWindow));
                        activeNestedWindow.Rebuild();
                    }
                );
                nestedWindows.Add(new DevMorphWindow(OnReturn));
            }
        }

        protected override void OnBuild()
        {
            BuildLeftSide();
            BuildRightSide();
            Refresh();
        }

        void BuildLeftSide(bool rightSide = false)
        {
            AddSpacer(5, rightSide);
            AddElement(() =>
            {
                var slider = script.CreateSlider(BootyMagic.script.softnessJsf, rightSide);
                slider.valueFormat = "F0";
                return slider;
            });

            AddInfoTextField(
                "Softness, quickness and mass affect physics base values, the effect of gravity," +
                " the amount of morphing, and hard collider properties. Use them to determine" +
                " the overall behavior, then fine tune physics, morphs and colliders as needed.",
                rightSide,
                180
            );

            AddElement(SoftPhysicsHandler.gluteSoftPhysicsOnJsb.name, () =>
                script.CreateToggle(SoftPhysicsHandler.gluteSoftPhysicsOnJsb, rightSide)
            );

            AddInfoTextField(
                $"Simulate soft-body dynamics for glutes. Soft physics is required to get the most out of {Constant.BOOTY_MAGIC}.",
                rightSide
            );

            AddSpacer(470, rightSide);
            AddElement(() =>
            {
                var jsUrl = BootyMagic.script.presetHandler.savePresetJsu;
                var button = script.CreateButton(jsUrl.name, rightSide);
                jsUrl.fileBrowseButton = button.button;
                return button;
            });
            AddElement(() =>
            {
                var jsUrl = BootyMagic.script.presetHandler.loadPresetJsu;
                var button = script.CreateButton(jsUrl.name, rightSide);
                jsUrl.fileBrowseButton = button.button;
                return button;
            });
        }

        void BuildRightSide(bool rightSide = true)
        {
            AddSpacer(5, rightSide);
            AddElement(() =>
            {
                var slider = script.CreateSlider(BootyMagic.script.quicknessJsf, rightSide);
                slider.valueFormat = "F2";
                return slider;
            });

            AddElement(() =>
            {
                var slider = script.CreateSlider(MainPhysicsHandler.gluteMass.left.offsetJsf, rightSide);
                slider.valueFormat = MainPhysicsHandler.gluteMass.valueFormat;
                return slider;
            });

            AddElement(() =>
                CreateValueTextField(MainPhysicsHandler.gluteMass.left.displayValueJss, 28, 45, rightSide)
            );

            AddElement(SoftPhysicsHandler.ldpEnabledJsb.name, () =>
                script.CreateToggle(SoftPhysicsHandler.ldpEnabledJsb, rightSide)
            );

            AddInfoTextField(
                "Adjust soft physics parameters of individual soft joints dynamically with stretching and compression.",
                rightSide
            );

            AddElement(configureColliderFrictionAction.name, () =>
            {
                var button = script.CreateButton(configureColliderFrictionAction.name, rightSide);
                configureColliderFrictionAction.RegisterButton(button);
                button.label = "Configure Collider Friction";
                return button;
            });

            AddInfoTextField(
                "Collider friction controls how easily colliders stick to or slip away from hands or other colliding objects.",
                rightSide
            );

            AddSpacer(290, rightSide);

            AddElement(() =>
            {
                var button = script.CreateButton(openSettingsWindowAction.name, rightSide);
                openSettingsWindowAction.RegisterButton(button);
                button.label = "Naturalis User Preferences";
                return button;
            });

            AddVersionTextField();
        }

        void AddVersionTextField()
        {
            var versionJss = new JSONStorableString("version", "");
            var versionTextField = CreateVersionTextField(versionJss);
            AddElement(versionTextField);
            BootyMagic.script.AddTextFieldToJss(versionTextField, versionJss);
        }

        public void Refresh()
        {
            var gluteSoftPhysicsOnToggle = GetElement(SoftPhysicsHandler.gluteSoftPhysicsOnJsb.name);
            var ldpEnabledToggle = GetElement(SoftPhysicsHandler.ldpEnabledJsb.name);
            gluteSoftPhysicsOnToggle.SetActiveStyle(
                Settings.personMonitor.GluteSoftPhysicsIsEnabled(),
                false,
                SoftPhysicsHandler.gluteSoftPhysicsOnJsb.val && script.enabled
            );
            ldpEnabledToggle.SetActiveStyle(
                SoftPhysicsHandler.softPhysicsOn && Settings.personMonitor.GluteSoftPhysicsIsEnabled(),
                false,
                true
            );
        }

        protected override void OnClose()
        {
            if(activeNestedWindow is SettingsWindow)
            {
                activeNestedWindow.Clear();
                activeNestedWindow = null;
            }
        }
    }
}
