using Naturalis.BM.Handlers;
using Naturalis.Common.Models;
using Naturalis.Common.UI;

namespace Naturalis.BM.UI
{
    sealed class PhysicsWindow : WindowBase
    {
        public PhysicsWindow() : base(BootyMagic.script, nameof(PhysicsWindow))
        {
            nestedWindows.Add(new ParameterWindow(MainPhysicsHandler.gluteMass, OnReturn));

            foreach(var group in MainPhysicsHandler.parameterGroups)
            {
                nestedWindows.Add(new ParameterWindow(group, OnReturn));
            }

            foreach(var group in SoftPhysicsHandler.softParameterGroups)
            {
                nestedWindows.Add(new SoftParameterWindow(group, OnReturn));
            }

            foreach(var group in SoftPhysicsHandler.parameterGroups)
            {
                nestedWindows.Add(new ParameterWindow(group, OnReturn));
            }
        }

        /* Buttons are active if soft physics is currently enabled and in sync, and settingsMonitor agrees */

        static bool IsSoftPhysicsOnAndSynced()
        {
            return SoftPhysicsHandler.softPhysicsOn &&
                SoftPhysicsHandler.softPhysicsSynced &&
                Settings.personMonitor.GluteSoftPhysicsIsEnabled();
        }

        protected override void OnBuild()
        {
            BuildLeftSide();
            BuildRightSide();

            UpdateMainParamButtons();
            UpdateSoftParamButtons(IsSoftPhysicsOnAndSynced());
        }

        void BuildLeftSide(bool rightSide = false)
        {
            AddHeaderTextField("Joint Physics Parameters", rightSide);
            AddSpacer(5, rightSide);
            AddPhysicsParamButton(MainPhysicsHandler.gluteMass, rightSide);
            foreach(var group in MainPhysicsHandler.parameterGroups)
            {
                AddPhysicsParamButton(group, rightSide);
            }
        }

        void BuildRightSide(bool rightSide = true)
        {
            AddHeaderTextField("Soft Physics Parameters", rightSide);
            AddSpacer(5, rightSide);
            foreach(var group in SoftPhysicsHandler.softParameterGroups)
            {
                AddSoftPhysicsParamButton(group, rightSide);
            }

            foreach(var group in SoftPhysicsHandler.parameterGroups)
            {
                AddSoftPhysicsParamButton(group, rightSide);
            }
        }

        void AddPhysicsParamButton(PhysicsParameterGroup param, bool rightSide)
        {
            AddElement($"{param.name}Button", () =>
            {
                var button = script.CreateButton(param.name, rightSide);
                button.height = 102f;
                button.buttonText.fontSize = 30;
                var nestedWindow = nestedWindows.Find(window => window.GetId() == param.name);
                button.AddListener(() =>
                {
                    ClearSelf();
                    activeNestedWindow = nestedWindow;
                    activeNestedWindow.Rebuild();
                });

                return button;
            });
        }

        void AddSoftPhysicsParamButton(SoftPhysicsParameterGroup group, bool rightSide)
        {
            AddSoftPhysicsParamButton(group.name, rightSide);
        }

        void AddSoftPhysicsParamButton(PhysicsParameterGroup group, bool rightSide)
        {
            AddSoftPhysicsParamButton(group.name, rightSide);
        }

        void AddSoftPhysicsParamButton(string paramName, bool rightSide)
        {
            AddElement($"{paramName}Button", () =>
            {
                var button = script.CreateButton(paramName, rightSide);
                button.height = 102f;
                button.buttonText.fontSize = 30;
                var nestedWindow = nestedWindows.Find(window => window.GetId() == paramName);
                button.AddListener(() =>
                {
                    ClearSelf();
                    activeNestedWindow = nestedWindow;
                    activeNestedWindow.Rebuild();
                });

                return button;
            });
        }

        void UpdateMainParamButtons()
        {
            SyncParamButton(MainPhysicsHandler.gluteMass.name, MainPhysicsHandler.gluteMass.IsModified());

            foreach(var group in MainPhysicsHandler.parameterGroups)
            {
                SyncParamButton(group.name, group.IsModified());
            }
        }

        void UpdateSoftParamButtons(bool active)
        {
            foreach(var group in SoftPhysicsHandler.softParameterGroups)
            {
                SyncParamButton(group.name, group.IsModified(), active);
            }

            foreach(var group in SoftPhysicsHandler.parameterGroups)
            {
                SyncParamButton(group.name, group.IsModified(), active);
            }
        }

        public void Refresh(bool softPhysicsEnabled)
        {
            UpdateSoftParamButtons(softPhysicsEnabled);
        }

        public void OnMainParamAdjusted(PhysicsParameterGroup group)
        {
            if(!BootyMagic.tabs.IsActive(this))
            {
                return;
            }

            bool active = true;
            if(
                group == SoftPhysicsHandler.fatBackForce &&
                group == SoftPhysicsHandler.fatBackForceThreshold &&
                group == SoftPhysicsHandler.fatBackForceMaxForce
            )
            {
                active = IsSoftPhysicsOnAndSynced();
            }

            SyncParamButton(group.name, group.IsModified(), active);
        }

        public void OnSoftParamAdjusted(SoftPhysicsParameterGroup group)
        {
            if(!BootyMagic.tabs.IsActive(this))
            {
                return;
            }

            SyncParamButton(group.name, group.IsModified(), IsSoftPhysicsOnAndSynced());
        }

        void SyncParamButton(string paramName, bool modified, bool active = true)
        {
            var button = GetElementAs<UIDynamicButton>($"{paramName}Button");
            string label = $"  {paramName}";
            if(modified && active)
            {
                label += " *".Bold();
            }

            button.label = label;
            button.SetActiveStyle(active, true);
        }
    }
}
