﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;

namespace Naturalis.Common.Components
{
    sealed class CalibrationHelper : MonoBehaviour
    {
        Person _person;
        string _domainName;
        public JSONStorableBool calibratingJsb { get; private set; }
        Func<bool> _localOtherIsCalibrating;
        Func<bool> _globalOtherIsCalibrating;
        MotionAnimationMaster _motionAnimationMaster;
        Action _calibrationAction;
        public bool calibrateOnNextPending { get; set; }

        readonly HashSet<string> _pendingReasons = new HashSet<string>();
        public bool pending { get; private set; }
        public bool cancelling { get; set; }
        public bool updateMassWhenCalibrating { get; set; } // May have incorrect state when calibration pending gets cancelled.

        public void Init(
            string domainName,
            JSONStorableBool calibratingJsb,
            JSONStorable otherDomain,
            Func<bool> globalOtherIsCalibrating,
            Action calibrationAction
        )
        {
            _person = Person.instance;
            _domainName = domainName;
            this.calibratingJsb = calibratingJsb;
            this.calibratingJsb.setCallbackFunction = _ => UpdateButtonText();
            _localOtherIsCalibrating = () => otherDomain.GetBoolParamValueNullSafe(Constant.INTERNAL_IS_CALIBRATING);
            _globalOtherIsCalibrating = globalOtherIsCalibrating;
            _motionAnimationMaster = SuperController.singleton.motionAnimationMaster;
            _calibrationAction = calibrationAction;
        }

        UIDynamicButton _calibrationButton;

        public UIDynamicButton calibrationButton
        {
            set
            {
                _calibrationButton = value;
                UpdateButtonText();
            }
        }

        public void SetCalibrationPending(string reason, bool value = true)
        {
            if(value)
            {
                _pendingReasons.Add(reason);
            }
            else
            {
                _pendingReasons.Remove(reason);
            }

            pending = _pendingReasons.Any();
            if(pending && calibrateOnNextPending)
            {
                _calibrationAction();
                calibrateOnNextPending = false;
            }
            else
            {
                UpdateButtonText();
            }
        }

        void ResetCalibrationPending()
        {
            calibrateOnNextPending = false;
            pending = false;
            _pendingReasons.Clear();
            UpdateButtonText();
        }

        bool _showCalibrationDone;

        void UpdateButtonText()
        {
            if(!_calibrationButton)
            {
                return;
            }

            if(calibratingJsb.val)
            {
                _calibrationButton.label = $"Calibrating {_domainName}...";
                _calibrationButton.SetActiveStyle(false, true);
            }
            else if(pending)
            {
                _calibrationButton.label = $"{_domainName} Calibration Pending".Color(new Color(0.75f, 0.00f, 0.00f));
                _calibrationButton.SetActiveStyle(true);
            }
            else if(_showCalibrationDone)
            {
                _calibrationButton.label = "Calibration Done";
                _calibrationButton.SetActiveStyle(false, true);
            }
            else
            {
                _calibrationButton.label = $"  Calibrate {_domainName}";
                _calibrationButton.SetActiveStyle(gameObject.GetComponent<MVRScript>().IsEnabledNullSafe(), true);
            }
        }

        public static IEnumerator WaitForAtomPhysicsUnfrozen(Atom atom)
        {
            if(atom.isPhysicsFrozen)
            {
                Loggr.Message($"Calibration will start when physics is unfrozen on atom '{atom.uid}'.");
            }

            while(atom.isPhysicsFrozen)
            {
                yield return null;
            }
        }

        bool _queued;
        bool _deferOnGlobalActions;
        bool? _globalFrozen;
        bool? _playing;

        bool _deferOnLocalActions;
        bool? _collisionEnabled;

        public IEnumerator Begin()
        {
            ResetCalibrationPending();

            if(calibratingJsb.val)
            {
                if(!_queued)
                {
                    _queued = true;
                }
                else
                {
                    cancelling = true;
                    yield break;
                }
            }

            while(_queued)
            {
                yield return null;
            }

            calibratingJsb.val = true;

            /* The instance which started calibrating first has control */
            if(_globalOtherIsCalibrating())
            {
                _deferOnGlobalActions = true;
            }
            else
            {
                if(Settings.instance.freezeMotionSoundJsb.val && _globalFrozen == null)
                {
                    _globalFrozen = Utils.GlobalAnimationFrozen();
                    if(!_globalFrozen.Value)
                    {
                        SuperController.singleton.SetFreezeAnimation(true);
                    }
                }
                else if(Settings.instance.pauseSceneAnimationJsb.val && _playing == null)
                {
                    _playing = _motionAnimationMaster.activeWhilePlaying.activeSelf;
                    if(_playing.Value)
                    {
                        _motionAnimationMaster.StopPlayback();
                    }
                }
            }

            if(_localOtherIsCalibrating())
            {
                _deferOnLocalActions = true;
            }
            else
            {
                if(_collisionEnabled == null)
                {
                    _collisionEnabled = _person.collisionEnabledJsb.val;
                    _person.collisionEnabledJsb.val = false;
                }

                SaveOtherAtomControllerStates();
                ApplyCalibrationPose();
            }
        }

        static List<FreeControllerState> _freeControllerStates = new List<FreeControllerState>();

        void SaveOtherAtomControllerStates()
        {
            _freeControllerStates.Clear();
            var rigidbodies = _person.atom.linkableRigidbodies;
            var atoms = SuperController.singleton.GetAtoms().Where(atom => atom.uid != _person.atom.uid);

            foreach(var controller in atoms.SelectMany(atom => atom.freeControllers))
            {
                if(rigidbodies.Contains(controller.linkToRB) && FreeControllerState.ControllerIsLinked(controller))
                {
                    var state = new FreeControllerState(controller);
                    state.SetStateOn();
                    _freeControllerStates.Add(state);
                }
            }
        }

        static string _preCalibrationPoseName;
        static PresetParams _posePresetParams;

        void ApplyCalibrationPose()
        {
            _posePresetParams = _person.GetPosePresetParams();
            _preCalibrationPoseName = $"calibration_{Guid.NewGuid().ToString().Substring(0, 7)}";
            _person.SavePosePreset(_preCalibrationPoseName, storeRootNodeInPose: true);
            string calibrationPosePresetName = _person.atom.mainController.GetBoolParamValue("detachControl")
                ? Constant.NO_ROOT_CALIBRATION_POSE
                : Constant.DEFAULT_CALIBRATION_POSE;
            _person.LoadPosePreset(calibrationPosePresetName, true);
        }

        public IEnumerator FinishCo()
        {
            if(!calibratingJsb.val)
            {
                yield break;
            }

            if(!_deferOnLocalActions && !_queued)
            {
                while(_localOtherIsCalibrating())
                {
                    yield return new WaitForSecondsRealtime(0.1f);
                }

                RestorePersonCollisionAndPose();
                RestoreOtherAtomControllerStates();
            }

            if(!_deferOnGlobalActions && !_queued)
            {
                while(_globalOtherIsCalibrating())
                {
                    yield return new WaitForSecondsRealtime(0.1f);
                }

                if(_globalFrozen.HasValue)
                {
                    if(!_globalFrozen.Value)
                    {
                        SuperController.singleton.SetFreezeAnimation(false);
                    }

                    _globalFrozen = null;
                }
                else if(_playing.HasValue)
                {
                    if(_playing.Value)
                    {
                        _motionAnimationMaster.StartPlayback();
                    }

                    _playing = null;
                }
            }

            if(!_queued)
            {
                calibratingJsb.val = false;
                EnableSettingsMonitor();
                StartCoroutine(FinishButtonUpdateCo());
            }

            _deferOnLocalActions = false;
            _deferOnGlobalActions = false;
            _queued = false;
        }

        IEnumerator FinishButtonUpdateCo()
        {
            _showCalibrationDone = true;
            UpdateButtonText();
            yield return new WaitForSeconds(1);
            _showCalibrationDone = false;
            UpdateButtonText();
        }

        public void Finish(bool otherIsEnabled)
        {
            if(!calibratingJsb.val)
            {
                ResetCalibrationPending();
                return;
            }

            if(!otherIsEnabled)
            {
                RestorePersonCollisionAndPose();
                RestoreOtherAtomControllerStates();

                if(_globalFrozen.HasValue)
                {
                    if(!_globalFrozen.Value)
                    {
                        SuperController.singleton.SetFreezeAnimation(false);
                    }

                    _globalFrozen = null;
                }
                else if(_playing.HasValue)
                {
                    if(_playing.Value)
                    {
                        _motionAnimationMaster.StartPlayback();
                    }

                    _playing = null;
                }
            }

            calibratingJsb.val = false;
            _deferOnLocalActions = false;
            _deferOnGlobalActions = false;
            _queued = false;
        }

        public void EnableSettingsMonitor()
        {
            Settings.personMonitor.enabled = !_localOtherIsCalibrating();
        }

        void RestorePersonCollisionAndPose()
        {
            if(!calibratingJsb.val)
            {
                return;
            }

            _person.LoadPosePreset(_preCalibrationPoseName, false);
            _person.DeletePosePreset(_preCalibrationPoseName);
            _person.SetPosePresetParams(_posePresetParams);

            if(_collisionEnabled == true)
            {
                _person.collisionEnabledJsb.val = true;
            }

            _collisionEnabled = null;
            _preCalibrationPoseName = null;
        }

        static void RestoreOtherAtomControllerStates()
        {
            foreach(var state in _freeControllerStates)
            {
                state.Restore();
            }

            _freeControllerStates.Clear();
        }

        public static void Destroy()
        {
            _freeControllerStates = null;
        }
    }
}
