using System;
using System.Collections;
using System.Collections.Generic;
using UnityEngine;

namespace Naturalis.Common.Components
{
    sealed class UIMod : MonoBehaviour
    {
        readonly List<RectTransformReposition> _repositionedRects = new List<RectTransformReposition>();
        readonly List<GameObject> _inactivatedUIGameObjects = new List<GameObject>();
        readonly List<GameObject> _customUIGameObjects = new List<GameObject>();

        public string id { get; private set; }
        Transform _container;
        string _targetName;
        UnityEventsListener _containerUIEventsListener;
        public UnityEventsListener targetUIEventsListener { get; set; }

        public bool IsVisible()
        {
            return targetUIEventsListener && targetUIEventsListener.isEnabled;
        }

        Func<UIMod, IEnumerator> _applyChanges;
        public Func<bool> checkPluginEnabled { get; set; }
        public bool useDisabledStyle { get; set; }

        public void Init(
            string uiModId,
            Transform container,
            string targetTransformName,
            Func<UIMod, IEnumerator> changesFunc
        )
        {
            id = uiModId;
            _container = container;
            _targetName = targetTransformName;
            _applyChanges = changesFunc;
        }

        public void Apply()
        {
            if(_container.gameObject.activeInHierarchy)
            {
                StartCoroutine(ApplyCo());
            }
            /* Add listener to wait for the transform to be active in hierarchy */
            else
            {
                _containerUIEventsListener = _container.gameObject.AddComponent<UnityEventsListener>();
                _containerUIEventsListener.onEnable.AddListener(() => StartCoroutine(ApplyCo()));
            }
        }

        bool _modified;
        public Transform target { get; private set; }

        IEnumerator ApplyCo()
        {
            if(_modified)
            {
                yield break;
            }

            Destroy(_containerUIEventsListener);
            _containerUIEventsListener = null;

            /* Set modified immediately to prevent double apply in case eventListener
             * gets enabled multiple times in a short period of time
             */
            _modified = true;

            if(_targetName != null)
            {
                float timeout = Time.unscaledTime + 1;
                while(Time.unscaledTime < timeout)
                {
                    yield return new WaitForSeconds(0.1f);
                    target = _container.Find(_targetName);
                }

                if(!target)
                {
                    Loggr.Error($"Failed to modify {id} UI on target {_targetName} - could not find transform.");
                    yield break;
                }
            }

            yield return _applyChanges(this);
            if(!checkPluginEnabled())
            {
                Disable();
            }
        }

        public void InactivateChildren()
        {
            foreach(Transform child in target)
            {
                if(child.gameObject.activeSelf)
                {
                    _inactivatedUIGameObjects.Add(child.gameObject);
                    child.gameObject.SetActive(false);
                }
            }
        }

        public void AddCustomObject(GameObject go)
        {
            _customUIGameObjects.Add(go);
        }

        public void RepositionTransform(Transform t, Vector2 originalPosition, Vector2 newPosition)
        {
            _repositionedRects.Add(new RectTransformReposition(t.GetComponent<RectTransform>(), originalPosition, newPosition));
        }

        public void Enable()
        {
            if(!_modified)
            {
                Apply();
            }
            else
            {
                foreach(var go in _customUIGameObjects)
                {
                    if(useDisabledStyle)
                    {
                        var elements = go.GetComponentsInChildren<UIDynamic>();
                        foreach(var element in elements)
                        {
                            element.SetActiveStyle(true, false, true);
                        }
                    }
                    else
                    {
                        go.SetActive(true);
                    }
                }

                foreach(var go in _inactivatedUIGameObjects)
                {
                    go.SetActive(false);
                }

                foreach(var repositionedRect in _repositionedRects)
                {
                    repositionedRect.Apply();
                }
            }
        }

        public void Disable()
        {
            foreach(var go in _customUIGameObjects)
            {
                if(useDisabledStyle)
                {
                    var elements = go.GetComponentsInChildren<UIDynamic>();
                    foreach(var element in elements)
                    {
                        element.SetActiveStyle(false, true);
                    }
                }
                else
                {
                    go.SetActive(false);
                }
            }

            foreach(var go in _inactivatedUIGameObjects)
            {
                go.SetActive(true);
            }

            foreach(var repositionedRect in _repositionedRects)
            {
                repositionedRect.RestoreOriginal();
            }
        }

        void OnDestroy()
        {
            DestroyImmediate(_containerUIEventsListener);
            DestroyImmediate(targetUIEventsListener);
            foreach(var go in _customUIGameObjects)
            {
                Destroy(go);
            }
        }
    }
}
