﻿using Naturalis.Common.Components;
using UnityEngine;

namespace Naturalis.Common.Handlers.Configs
{
    sealed class LocalDistortionPhysicsConfig
    {
        float _compressionMax;
        float _compressionMultiplier;
        float _compressionSlope;

        public StorableFloat compressionEffectJsf { get; set; }
        public StorableFloat compressionSensitivityJsf { get; set; }

        float _stretchingMax;
        float _stretchingMultiplier;
        float _stretchingSlope;

        public StorableFloat stretchingEffectJsf { get; set; }
        public StorableFloat stretchingSensitivityJsf { get; set; }

        public void SetCompressionMaxAndMultiplier(float value)
        {
            _compressionMax = CalculateMax(value);
            _compressionMultiplier = CalculateMultiplier(value);
        }

        public void SetStretchingMaxAndMultiplier(float value)
        {
            _stretchingMax = CalculateMax(value);
            _stretchingMultiplier = CalculateMultiplier(value);
        }

        public void SetCompressionSlope(float value)
        {
            _compressionSlope = CalculateSlope(value);
        }

        public void SetStretchingSlope(float value)
        {
            _stretchingSlope = CalculateSlope(value);
        }

        static float CalculateMax(float value)
        {
            return value > 0
                ? Calc.RoundToDecimals(value / 2, 100)
                : Calc.RoundToDecimals(-value / 2, 100);
        }

        static float CalculateMultiplier(float value)
        {
            return value > 0
                ? Calc.RoundToDecimals(Calc.Exponential1(value, 0.7f, 4.9f, 1.5f), 100)
                : -Calc.RoundToDecimals(Calc.Exponential1(-value, 0.7f, 4.9f, 1.5f), 100);
        }

        static float CalculateSlope(float value)
        {
            return Mathf.Lerp(-0.33f, 0.75f, 1 - value);
        }

        public float CalculateDistortionOffset(
            float limit,
            float distortionBase,
            float distortion
        )
        {
            if(distortion < 0) // compression
            {
                if(_compressionMultiplier == 0)
                {
                    return 0;
                }

                float value = Calc.InverseSmoothStep(
                    -distortion / (_compressionMax * distortionBase),
                    1,
                    _compressionSlope,
                    0.5f
                );

                return _compressionMultiplier < 0
                    ? -Mathf.Lerp(0, limit, -_compressionMultiplier * value)
                    : Mathf.Lerp(0, limit, _compressionMultiplier * value);
            }
            else // stretching
            {
                if(_stretchingMultiplier == 0)
                {
                    return 0;
                }

                float value = Calc.InverseSmoothStep(
                    distortion / (_stretchingMax * distortionBase),
                    1,
                    _stretchingSlope,
                    0.5f
                );

                return _stretchingMultiplier < 0
                    ? -Mathf.Lerp(0, limit, -_stretchingMultiplier * value)
                    : Mathf.Lerp(0, limit, _stretchingMultiplier * value);
            }
        }

        public bool IsModified()
        {
            return compressionEffectJsf.IsModified() ||
                compressionSensitivityJsf.IsModified() ||
                stretchingEffectJsf.IsModified() ||
                stretchingSensitivityJsf.IsModified();
        }
    }
}
