using MVR.FileManagementSecure;
using SimpleJSON;
using System.Collections.Generic;

namespace Naturalis.Common.Handlers
{
    sealed class PresetHandler
    {
        readonly ScriptBase _script;
        readonly string _domainName;
        readonly string _presetDir;
        string _lastDir;
        public JSONStorableUrl loadPresetJsu { get; }
        public JSONStorableUrl savePresetJsu { get; }

        const string PRESET_EXT = "json";

        public PresetHandler(ScriptBase script, string domainName)
        {
            _script = script;
            _domainName = domainName;
            _presetDir = $"{FileUtils.SAVES_DIR}/{domainName}Presets";
            LastDirFallback();

            loadPresetJsu = new JSONStorableUrl($"Load {domainName} Preset", string.Empty, "json", _lastDir)
            {
                allowFullComputerBrowse = true,
                allowBrowseAboveSuggestedPath = true,
                hideExtension = true,
                showDirs = true,
                beginBrowseWithObjectCallback = BeginLoadBrowse,
                endBrowseWithObjectCallback = jsu => OnLoadPathSelected(jsu.val),
            };

            savePresetJsu = new JSONStorableUrl($"Save {domainName} Preset", string.Empty, "json", _lastDir)
            {
                allowFullComputerBrowse = true,
                allowBrowseAboveSuggestedPath = true,
                hideExtension = true,
                showDirs = true,
                beginBrowseWithObjectCallback = BeginSaveBrowse,
                endBrowseWithObjectCallback = jsu => OnSavePathSelected(jsu.val),
                shortCuts = new List<ShortCut>(), // clear shortcuts
            };

            var loadPresetWithPathJsu = new JSONStorableUrl("_loadPresetWithPath", string.Empty, "json", _lastDir)
            {
                allowFullComputerBrowse = true,
                allowBrowseAboveSuggestedPath = true,
                hideExtension = true,
                showDirs = true,
                beginBrowseWithObjectCallback = BeginLoadBrowse,
            };

            var loadPresetWithPathJSON = new JSONStorableActionPresetFilePath(
                "LoadPresetWithPath",
                OnLoadPathSelected,
                loadPresetWithPathJsu
            );

            script.RegisterPresetFilePathAction(loadPresetWithPathJSON);
        }

        // see e.g. DAZCharacterTextureControl.BeginBrowse
        void BeginLoadBrowse(JSONStorableUrl jsu)
        {
            FileUtils.EnsureDirExists(_presetDir);
            LastDirFallback();
            jsu.shortCuts = FileUtils.GetShortCutsForDirectory(_presetDir);
        }

        void BeginSaveBrowse(JSONStorableUrl jsu)
        {
            FileUtils.EnsureDirExists(_presetDir);
            LastDirFallback();
            _script.DeferSetSaveFileBrowser(PRESET_EXT); // JSONStorableUrl calls BeginSaveBrowse before creating the file browser
        }

        void OnLoadPathSelected(string path)
        {
            if(string.IsNullOrEmpty(path))
            {
                return;
            }

            OnPathSelected(ref path);
            var jc = _script.LoadJSON(path).AsObject;
            if(jc != null && ValidatePreset(jc, path))
            {
                jc["enabled"].AsBool = _script.enabled;
                _script.RestoreFromPresetJSON(jc);
            }

            loadPresetJsu.val = ""; // ensure empty value when next browse cancelled
        }

        bool ValidatePreset(JSONClass jc, string path)
        {
            string id = jc["id"].Value;
            if(!id.EndsWith($"Naturalis.{_domainName}"))
            {
                Loggr.Error($"Loaded preset is not a valid {_domainName} preset. Path: {path}", false);
                return false;
            }

            return true;
        }

        void OnSavePathSelected(string path)
        {
            if(string.IsNullOrEmpty(path))
            {
                return;
            }

            OnPathSelected(ref path);
            if(!path.ToLower().EndsWith(PRESET_EXT.ToLower()))
            {
                path += "." + PRESET_EXT;
            }

            var jc = _script.GetJSON();
            jc.Remove("enabled");
            _script.SaveJSON(jc, path);
            SuperController.singleton.DoSaveScreenshot(path);
            savePresetJsu.val = ""; // ensure empty value when next browse cancelled
        }

        void OnPathSelected(ref string path)
        {
            path = SuperController.singleton.NormalizePath(path);
            string containingDir = path.Substring(0, path.LastIndexOfAny(new[] { '/', '\\' }));
            FileUtils.EnsureDirExists(containingDir);
            _lastDir = containingDir + @"\";
            loadPresetJsu.suggestedPath = _lastDir;
        }

        void LastDirFallback()
        {
            if(!FileUtils.DirectoryExists(_lastDir))
            {
                _lastDir = _presetDir;
            }
        }
    }
}
