﻿using System;
using System.Collections.Generic;
using System.Linq;
using Naturalis.Common.Components;
using Naturalis.Common.Handlers.Configs;
using UnityEngine;

namespace Naturalis.Common.Models
{
    sealed class PhysicsParameter : ParameterBase
    {
        public JSONStorableFloat activeValueJsf { get; }
        public JSONStorableString displayValueJss { get; set; }

        public Dictionary<string, PhysicsCalculator> gravityCalculators = new Dictionary<string, PhysicsCalculator>();
        readonly Dictionary<string, float> _additiveGravityAdjustments = new Dictionary<string, float>();

        public Dictionary<string, PhysicsCalculator> forceCalculators = new Dictionary<string, PhysicsCalculator>();
        readonly Dictionary<string, float> _additiveForceAdjustments = new Dictionary<string, float>();

        public Func<float, string> buildDisplayValueString { get; set; }
        public Action<float> sync { private get; set; }

        public PhysicsParameter(MVRScript script, string side, string name, float min, float max)
        {
            string paramName = Utils.ParamName(name, side);
            baseValueJsf = new StorableFloat($"{paramName} Base Value", 0, min, max);
            activeValueJsf = new StorableFloat($"{paramName} Value", 0, min, max);
            offsetJsf = script.NewStorableFloat($"{paramName} Offset", 0, -max, max, side == Side.LEFT);
            calculator = new PhysicsCalculator();
            quicknessOffsetCalculator = new PhysicsCalculator();
            slownessOffsetCalculator = new PhysicsCalculator();
        }

        public bool HasAdjustableGravityPhysics()
        {
            return gravityCalculators.Count > 0 &&
                gravityCalculators.All(x => x.Value.multiplierJsf != null);
        }

        void Sync()
        {
            float value = offsetJsf.val +
                baseValueJsf.val +
                _additiveGravityAdjustments.Values.Sum() +
                _additiveForceAdjustments.Values.Sum();
            activeValueJsf.val = value;
            displayValueJss?.SetVal(buildDisplayValueString(value));
            sync?.Invoke(value); // sync is null for target rotation parameters
        }

        public void UpdateMinMax(float min, float max)
        {
            baseValueJsf.min = min;
            baseValueJsf.max = max;
            activeValueJsf.min = min;
            activeValueJsf.max = max;
        }

        public void UpdateMassValue(float mass)
        {
            baseValueJsf.val = mass;
            Sync();
        }

        public void UpdateBaseValue(float mass, float softness, float quickness, float physicsRateMultiplier, bool usePrimary)
        {
            float value = calculator.Calculate(mass, softness, usePrimary);

            if(quickness > 0)
            {
                float maxQuicknessOffset = quicknessOffsetCalculator.Calculate(mass, softness, true);
                value += Mathf.Lerp(0, maxQuicknessOffset, quickness);
            }

            if(quickness < 0)
            {
                float maxSlownessOffset = slownessOffsetCalculator.Calculate(mass, softness, true);
                value += Mathf.Lerp(0, maxSlownessOffset, -quickness);
            }

            baseValueJsf.val = physicsRateMultiplier * value;
            Sync();
        }

        // For temp use, call UpdateValue after
        public void OverrideValue(float value)
        {
            activeValueJsf.val = value;
            sync?.Invoke(value);
        }

        public void UpdateOffsetValue(float value)
        {
            offsetJsf.valNoCallback = value;
            Sync();
        }

        public void UpdateGravityValue(string direction, float effect, float mass, float softness, bool softPhysicsOn)
        {
            if(!gravityCalculators.ContainsKey(direction))
            {
                return;
            }

            float gravityValue = gravityCalculators[direction].CalculateDynamic(effect, mass, softness, softPhysicsOn);
            _additiveGravityAdjustments[direction] = gravityValue;
            Sync();
        }

        public void ResetGravityValue(string direction)
        {
            if(!gravityCalculators.ContainsKey(direction))
            {
                return;
            }

            _additiveGravityAdjustments[direction] = 0;
            Sync();
        }

        public void UpdateForceValue(string direction, float effect, float mass, float softness, bool softPhysicsOn)
        {
            if(!forceCalculators.ContainsKey(direction))
            {
                return;
            }

            float forceValue = forceCalculators[direction].CalculateDynamic(effect, mass, softness, softPhysicsOn);
            _additiveForceAdjustments[direction] = forceValue;
            Sync();
        }

        public void ResetForceValue(string direction)
        {
            if(!forceCalculators.ContainsKey(direction))
            {
                return;
            }

            _additiveForceAdjustments[direction] = 0;
            Sync();
        }

        public void SaveCalibrated()
        {
            offsetJsf.SaveCalibrated();
            if(HasAdjustableGravityPhysics())
            {
                foreach(var kvp in gravityCalculators)
                {
                    kvp.Value.multiplierJsf.SaveCalibrated();
                }
            }
        }
    }
}
