﻿using System;
using Naturalis.Common.Components;
using Naturalis.Common.Handlers.Configs;
using UnityEngine;

namespace Naturalis.Common.Models
{
    sealed class SoftPhysicsParameter : ParameterBase
    {
        public PhysicsCalculator inverseFrictionCalculator { get; set; }
        float _additiveFrictionAdjustment;

        public GradientConfig gradientConfig { get; }
        public LocalDistortionPhysicsConfig ldpConfig { get; }

        public bool usesRealMass { get; set; }
        public Action<float, float> sync { private get; set; }

        public SoftPhysicsParameter(MVRScript script, string domain, string side, string name, float min, float max)
        {
            string paramName = Utils.ParamName(name, side);
            bool register = side == Side.LEFT;

            baseValueJsf = new StorableFloat($"{paramName} Base Value", 0, min, max);
            offsetJsf = script.NewStorableFloat($"{paramName} Offset", 0, -max, max, register);
            calculator = new PhysicsCalculator();
            quicknessOffsetCalculator = new PhysicsCalculator();
            slownessOffsetCalculator = new PhysicsCalculator();

            gradientConfig = new GradientConfig
            {
                middleShiftJsf = script.NewStorableFloat($"{paramName} Gradient Middle Shift", 0.00f, -1.00f, 1.00f, register),
                outerMultiplierJsf = NewGroupMultiplierJsf(script, paramName, "Outer", register),
                centerMultiplierJsf = NewGroupMultiplierJsf(script, paramName, "Center", register),
            };

            if(domain == Constant.TITTY_MAGIC)
            {
                gradientConfig.nippleMultiplierJsf = NewGroupMultiplierJsf(script, paramName, "Nipple", register);
            }

            ldpConfig = new LocalDistortionPhysicsConfig
            {
                compressionEffectJsf = script.NewStorableFloat($"{paramName} Local Compression Effect", 0, -3.00f, 3.00f, register),
                compressionSensitivityJsf = script.NewStorableFloat($"{paramName} Local Compression Sensitivity", 0, 0.00f, 1.00f, register),
                stretchingEffectJsf = script.NewStorableFloat($"{paramName} Local Stretching Effect", 0, -3.00f, 3.00f, register),
                stretchingSensitivityJsf = script.NewStorableFloat($"{paramName} Local Stretching Sensitivity", 0, 0.00f, 1.00f, register),
            };
        }

        static StorableFloat NewGroupMultiplierJsf(MVRScript script, string paramName, string groupName, bool register)
        {
            return script.NewStorableFloat(
                $"{paramName} Gradient {groupName} Multiplier",
                1.00f,
                0.01f,
                Constant.GRADIENT_MULTI_MAX,
                register
            );
        }

        public void Sync()
        {
            sync(baseValueJsf.val + offsetJsf.val, _additiveFrictionAdjustment);
        }

        public void UpdateBaseValue(float mass, float softness, float quickness, float physicsRateMultiplier, bool usePrimary)
        {
            baseValueJsf.val = NewBaseValue(mass, softness, quickness, physicsRateMultiplier, usePrimary);
            Sync();
        }

        public void UpdateOffsetValue(float value)
        {
            offsetJsf.valNoCallback = value;
            Sync();
        }

        float NewBaseValue(float mass, float softness, float quickness, float physicsRateMultiplier, bool usePrimary)
        {
            float value = calculator.Calculate(mass, softness, usePrimary);

            if(quickness > 0)
            {
                float maxQuicknessOffset = quicknessOffsetCalculator.Calculate(mass, softness, true);
                value += Mathf.Lerp(0, maxQuicknessOffset, quickness);
            }

            if(quickness < 0)
            {
                float maxSlownessOffset = slownessOffsetCalculator.Calculate(mass, softness, true);
                value += Mathf.Lerp(0, maxSlownessOffset, -quickness);
            }

            return physicsRateMultiplier * value;
        }

        public void UpdateInverseFrictionValue(float friction, float mass, float softness)
        {
            float value = inverseFrictionCalculator.CalculateDynamic(friction, mass, softness, true);
            _additiveFrictionAdjustment = Calc.RoundToDecimals(value, 10000);
            Sync();
        }

        public void ResetInverseFrictionValue()
        {
            _additiveFrictionAdjustment = 0;
            Sync();
        }
    }
}
