﻿using System;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;

namespace Naturalis.Common
{
    sealed class MorphListener
    {
        const float FILTER_STRENGTH = 0.005f;

        readonly HashSet<int> _vertexIndices;
        readonly HashSet<string> _excludeMorphsNames;
        readonly FrequencyRunner _runner;

        public MorphListener(HashSet<int> vertexIndices, HashSet<string> excludeMorphsNames, FrequencyRunner runner)
        {
            _vertexIndices = vertexIndices;
            _excludeMorphsNames = excludeMorphsNames;
            _runner = runner;
        }

        public void ProcessMorphs(DAZMorphBank morphBank)
        {
            if(morphBank.morphs == null)
            {
                return;
            }

            var listenedMorphs = new Dictionary<DAZMorph, float>();
            foreach(var morph in morphBank.morphs)
            {
                try
                {
                    if(
                        morph.visible &&
                        !morph.isPoseControl &&
                        morph.group != null &&
                        !morph.group.Contains("Pose/") &&
                        !listenedMorphs.ContainsKey(morph)
                    )
                    {
                        if(!_excludeMorphsNames.Contains(morph.morphName) && InSet(morph, FILTER_STRENGTH))
                        {
                            listenedMorphs.Add(morph, morph.morphValue);
                        }
                    }
                }
                catch(Exception e)
                {
                    Debug.LogError($"Unable to add morph '{morph.morphName}'. Error: {e}");
                }
            }

            if(_morphs == null)
            {
                _morphs = listenedMorphs.Keys.ToArray();
                _values = listenedMorphs.Values.ToArray();
            }
            else
            {
                var tmpMorphs = _morphs.ToList();
                tmpMorphs.AddRange(listenedMorphs.Keys);
                _morphs = tmpMorphs.ToArray();
                var tmpValues = _values.ToList();
                tmpValues.AddRange(listenedMorphs.Values);
                _values = tmpValues.ToArray();
            }
        }

        DAZMorph[] _morphs;
        float[] _values;

        public bool ChangeDetected()
        {
            return Settings.instance.checkForSizeChangesJsb.val && _runner.Run(MorphValueChanged);
        }

        public bool MorphValueChanged()
        {
            for(int i = 0; i < _morphs.Length; i++)
            {
                var dazMorph = _morphs[i];
                float oldValue = _values[i];
                float newValue = dazMorph.morphValue;
                if(Mathf.Abs(newValue - oldValue) > 0.001f)
                {
                    _values[i] = newValue;
                    return true;
                }
            }

            return false;
        }

        bool InSet(DAZMorph morph, float filterStrength)
        {
            if(morph.deltas == null)
            {
                morph.LoadDeltas();
            }

            float hitDelta = 0.0f;
            // ReSharper disable once AssignNullToNotNullAttribute
            float hitDeltaMax = morph.deltas.Sum(x => x.delta.magnitude) * filterStrength;
            foreach(var delta in morph.deltas)
            {
                if(_vertexIndices.Contains(delta.vertex))
                {
                    hitDelta += delta.delta.magnitude;
                    if(hitDelta >= hitDeltaMax)
                    {
                        return true;
                    }
                }
            }

            return false;
        }
    }
}
