﻿using System.Collections;
using System.Collections.Generic;
using System.Linq;
using MeshVR;
using Naturalis.Common.Components;
using UnityEngine;
using static Side;

namespace Naturalis.Common
{
    class Person
    {
        public static Person instance { get; private set; } = new Person();

        public static string GetId()
        {
            return instance.atom.uid;
        }

        public Atom atom { get; set; }
        public DAZCharacterSelector geometry { get; private set; }
        public DAZSkinV2 skin { get; private set; }

        GenerateDAZMorphsControlUI _morphsControlUI;
        public JSONStorable softBodyPhysicsEnabler { get; private set; }
        public JSONStorableBool collisionEnabledJsb { get; private set; }
        public JSONStorableFloat scaleJsf { get; private set; }

        public AdjustJoints gluteControl { get; private set; }
        public DAZPhysicsMesh glutePhysicsMesh { get; private set; }
        public Rigidbody hipRb { get; private set; }
        public Dictionary<string, ConfigurableJoint> gluteJoints { get; private set; }
        Dictionary<string, DAZBone> _gluteBones;
        public Dictionary<string, Rigidbody> gluteRbs { get; private set; }

        public AdjustJoints breastControl { get; private set; }
        public DAZPhysicsMesh breastPhysicsMesh { get; private set; }
        public Rigidbody chestRb { get; private set; }
        public Dictionary<string, ConfigurableJoint> pectoralJoints { get; private set; }
        Dictionary<string, DAZBone> _pectoralBones;
        public Dictionary<string, Rigidbody> pectoralRbs { get; private set; }
        public Dictionary<string, Rigidbody> nippleRbs { get; private set; }

        static string _morphsPath;
        static string _packagePrefix;
        PresetManagerControl _posePmc;
        PresetManager _posePm;
        PresetManager _morphPm;

        public Transform atomUIContent { get; private set; }

#region Public status

        public bool characterReady { get; private set; }
        public bool skinReady { get; private set; }

        public bool isFemale { get; private set; }

        public static bool IsFemale()
        {
            return instance.isFemale;
        }

        public static bool PectoralCollisionsEnabled(bool enabled)
        {
            return enabled
                ? instance.pectoralRbs[LEFT].detectCollisions || instance.pectoralRbs[RIGHT].detectCollisions
                : !instance.pectoralRbs[LEFT].detectCollisions || !instance.pectoralRbs[RIGHT].detectCollisions;
        }

        public static bool HasDummySkin()
        {
            return instance.geometry.selectedCharacter.name == "femaledummy";
        }

#endregion

        public IEnumerator WaitForGeometryAndSkinReady(int limit)
        {
            float timeout = Time.unscaledTime + limit;
            while(!(characterReady && skinReady) && Time.unscaledTime < timeout)
            {
                geometry = (DAZCharacterSelector) atom.GetStorableByID("geometry");
                characterReady = geometry.selectedCharacter.ready;
                skinReady = HasDummySkin() || atom.GetStorableByID("skin");
                yield return new WaitForSecondsRealtime(0.5f);
            }
        }

        public IEnumerator RefreshSkin()
        {
            characterReady = false;
            skinReady = false;
            skin = null;

            const int limit = 15;
            yield return WaitForGeometryAndSkinReady(limit);

            if(!characterReady)
            {
                Loggr.Error(
                    $"Selected character {geometry.selectedCharacter.name} was not ready after {limit.ToString()} seconds of waiting. " +
                    "Try reloading"
                );
                yield break;
            }

            if(!skinReady)
            {
                Loggr.Error(
                    $"Person skin materials not found after {limit.ToString()} seconds of waiting. " +
                    "Try reloading"
                );
                yield break;
            }

            skin = atom.GetComponentInChildren<DAZCharacter>().skin;
        }

        public bool SelectedCharacterIsReady()
        {
            return geometry.selectedCharacter.ready;
        }

        public void Setup()
        {
            isFemale = !geometry.selectedCharacter.isMale;
            skin = atom.GetComponentInChildren<DAZCharacter>().skin;
            _morphsControlUI = isFemale ? geometry.morphsControlUI : geometry.morphsControlUIOtherGender;
            softBodyPhysicsEnabler = atom.GetStorableByID("SoftBodyPhysicsEnabler");
            collisionEnabledJsb = atom.GetStorableByID("AtomControl").GetBoolJSONParam("collisionEnabled");
            scaleJsf = atom.GetStorableByID("rescaleObject").GetFloatJSONParam("scale");

            if(Utils.BootyMagicExists() && isFemale)
            {
                gluteControl = (AdjustJoints) atom.GetStorableByID("GluteControl");
                glutePhysicsMesh = (DAZPhysicsMesh) atom.GetStorableByID("LowerPhysicsMesh");
                hipRb = Utils.FindRigidbody(atom, "hip");
                gluteJoints = new Dictionary<string, ConfigurableJoint>
                {
                    { LEFT, gluteControl.joint1 },
                    { RIGHT, gluteControl.joint2 },
                };
                _gluteBones = new Dictionary<string, DAZBone>
                {
                    { LEFT, gluteJoints[LEFT].GetComponent<DAZBone>() },
                    { RIGHT, gluteJoints[RIGHT].GetComponent<DAZBone>() },
                };
                gluteRbs = new Dictionary<string, Rigidbody>
                {
                    { LEFT, gluteJoints[LEFT].GetComponent<Rigidbody>() },
                    { RIGHT, gluteJoints[RIGHT].GetComponent<Rigidbody>() },
                };
            }

            if(Utils.TittyMagicExists())
            {
                breastControl = (AdjustJoints) atom.GetStorableByID(isFemale ? "BreastControl" : "PectoralControl");
                breastPhysicsMesh = (DAZPhysicsMesh) atom.GetStorableByID("BreastPhysicsMesh");
                chestRb = Utils.FindRigidbody(atom, "chest");
                pectoralJoints = new Dictionary<string, ConfigurableJoint>
                {
                    { LEFT, breastControl.joint2 },
                    { RIGHT, breastControl.joint1 },
                };
                _pectoralBones = new Dictionary<string, DAZBone>
                {
                    { LEFT, pectoralJoints[LEFT].GetComponent<DAZBone>() },
                    { RIGHT, pectoralJoints[RIGHT].GetComponent<DAZBone>() },
                };
                pectoralRbs = new Dictionary<string, Rigidbody>
                {
                    { LEFT, pectoralJoints[LEFT].GetComponent<Rigidbody>() },
                    { RIGHT, pectoralJoints[RIGHT].GetComponent<Rigidbody>() },
                };
                if(isFemale)
                {
                    nippleRbs = new Dictionary<string, Rigidbody>
                    {
                        { LEFT, Utils.FindRigidbody(atom, "lNipple") },
                        { RIGHT, Utils.FindRigidbody(atom, "rNipple") },
                    };
                }
            }

            string packagePath = string.IsNullOrEmpty(Settings.packageId) ? "" : $"{Settings.packageId}:/";
            _morphsPath = $"{packagePath}Custom/Atom/Person/Morphs/female/everlaster";
            _packagePrefix = string.IsNullOrEmpty(Settings.packageId) ? "" : $"{Settings.packageId}:";

            _posePmc = atom.presetManagerControls.Find(pmc => pmc.name == "PosePresets");
            _posePm = atom.GetStorableByID("PosePresets").GetComponentInChildren<PresetManager>();
            _posePm.postLoadEvent.AddListener(OnPosePresetLoaded);

            _morphPm = atom.GetStorableByID("MorphPresets").GetComponentInChildren<PresetManager>();
            _morphPm.postLoadEvent.AddListener(ZeroAllDuplicateMorphs);

            atomUIContent = atom.transform.Find("UI/UIPlaceHolderModel/UIModel/Canvas/Panel/Content");

            SaveOriginalUseColliders();
            EnableUseColliders();
        }

        void OnPosePresetLoaded()
        {
            if(_posePmc.GetStringParamValue("presetName").StartsWith($"everlaster/{nameof(Naturalis)}/"))
            {
                return;
            }

            ZeroAllDuplicateMorphs();
        }

        static bool _originalUseAdvancedColliders;
        static bool _originalUseAuxBreastColliders;

        static void SaveOriginalUseColliders()
        {
            if(!instance.isFemale)
            {
                return;
            }

            _originalUseAdvancedColliders = instance.geometry.useAdvancedColliders;

            if(Utils.TittyMagicExists())
            {
                _originalUseAuxBreastColliders = instance.geometry.useAuxBreastColliders;
            }
        }

        public void EnableUseColliders()
        {
            if(!isFemale)
            {
                return;
            }

            /* Advanced colliders must be enabled for collider visualizer, force morphing and hard collider handler */
            geometry.useAdvancedColliders = true;

            if(Utils.TittyMagicExists())
            {
                geometry.useAuxBreastColliders = true;
            }
        }

        public static void RestoreOriginalUseColliders()
        {
            if(!instance.isFemale)
            {
                return;
            }

            instance.geometry.useAdvancedColliders = _originalUseAdvancedColliders;

            if(Utils.TittyMagicExists())
            {
                instance.geometry.useAuxBreastColliders = _originalUseAuxBreastColliders;
            }
        }

        public void SetGlutesUseGravity(bool value)
        {
            gluteRbs[LEFT].useGravity = value;
            gluteRbs[RIGHT].useGravity = value;
        }

        public void SetBreastsUseGravity(bool value)
        {
            pectoralRbs[LEFT].useGravity = value;
            pectoralRbs[RIGHT].useGravity = value;
        }

        public static Vector3 GetGluteJointRotation(string side)
        {
            return side == LEFT
                ? instance.gluteControl.smoothedJoint2TargetRotation
                : instance.gluteControl.smoothedJoint1TargetRotation;
        }

        public static void SetGluteRotation(string side, Vector3 rotation)
        {
            instance._gluteBones[side].baseJointRotation = rotation;
        }

        public static Vector3 GetPectoralJointRotation(string side)
        {
            return side == LEFT
                ? instance.breastControl.smoothedJoint2TargetRotation
                : instance.breastControl.smoothedJoint1TargetRotation;
        }

        public static void SetPectoralRotation(string side, Vector3 rotation)
        {
            instance._pectoralBones[side].baseJointRotation = rotation;
        }

        public static void SetPectoralCollisions(bool value)
        {
            instance.pectoralRbs[LEFT].detectCollisions = value;
            instance.pectoralRbs[RIGHT].detectCollisions = value;
        }

        public static DAZMorph GetBreastMorph(string fileName)
        {
            return instance.GetMorph($"TittyMagic/{fileName}.vmi");
        }

        public static DAZMorph GetGluteMorph(string fileName)
        {
            return instance.GetMorph($"BootyMagic/{fileName}.vmi");
        }

        DAZMorph GetMorph(string subPath)
        {
            string morphUid = $"{_morphsPath}/{subPath}";
            var dazMorph = _morphsControlUI.GetMorphByUid(morphUid);
            if(dazMorph == null)
            {
                Loggr.Error($"Morph with uid '{morphUid}' not found!");
            }

            return dazMorph;
        }

        public static Collider GetBreastHardCollider(string id)
        {
            return instance.geometry.auxBreastColliders.ToList().Find(c => c.name.Contains(id));
        }

        public static Quaternion GetGluteRotation(string side)
        {
            return instance.gluteRbs[side].rotation;
        }

        public static Quaternion GetHipRotation()
        {
            return instance.hipRb.rotation;
        }

        public static Quaternion GetPectoralRotation(string side)
        {
            return instance.pectoralRbs[side].rotation;
        }

        public static Quaternion GetChestRotation()
        {
            return instance.chestRb.rotation;
        }

        public static void ApplyGravityToGlutes(string direction, Vector3 additionalForce)
        {
            var gravityForce = Utils.DirectionVector(instance.hipRb, direction) * Physics.gravity.magnitude;
            instance.gluteRbs[LEFT].AddForce(gravityForce + additionalForce, ForceMode.Acceleration);
            instance.gluteRbs[RIGHT].AddForce(gravityForce + additionalForce, ForceMode.Acceleration);
        }

        public static void ApplyGravityToBreasts(string direction, Vector3 additionalForce)
        {
            var gravityForce = Utils.DirectionVector(instance.chestRb, direction) * Physics.gravity.magnitude;
            instance.pectoralRbs[LEFT].AddForce(gravityForce + additionalForce, ForceMode.Acceleration);
            instance.pectoralRbs[RIGHT].AddForce(gravityForce + additionalForce, ForceMode.Acceleration);
        }

        public static void ApplyForceToGlutes(Vector3 force)
        {
            instance.gluteRbs[LEFT].AddForce(force, ForceMode.Acceleration);
            instance.gluteRbs[RIGHT].AddForce(force, ForceMode.Acceleration);
        }

        public static void ApplyForceToBreasts(Vector3 force)
        {
            instance.pectoralRbs[LEFT].AddForce(force, ForceMode.Acceleration);
            instance.pectoralRbs[RIGHT].AddForce(force, ForceMode.Acceleration);
        }

        public PresetParams GetPosePresetParams()
        {
            return PresetParams.GetParamsFromControl(_posePmc);
        }

        public void SetPosePresetParams(PresetParams presetParams)
        {
            presetParams.SetParamsToControl(_posePmc);
        }

        public void SavePosePreset(
            string name,
            bool storePoseMorphsInPose = true,
            bool storeBonesInPose = true,
            bool storeRootNodeInPose = false
        )
        {
            new PresetParams(
                $"everlaster/{nameof(Naturalis)}/{name}",
                storeRootNodeInPose,
                storeBonesInPose,
                storePoseMorphsInPose
            ).SetParamsToControl(_posePmc);
            _posePmc.CallAction("StorePreset");
        }

        public void LoadPosePreset(string name, bool fromPackage)
        {
            string prefix = fromPackage ? _packagePrefix : "";
            _posePmc.SetStringParamValue("presetName", $"{prefix}everlaster/{nameof(Naturalis)}/{name}");
            _posePmc.CallAction("LoadPreset");
        }

        public void DeletePosePreset(string name)
        {
            _posePmc.SetStringParamValue("presetName", $"everlaster/{nameof(Naturalis)}/{name}");
            _posePm.DeletePreset();
        }

        readonly Dictionary<string, IEnumerable<string>> _morphNamePrefixes = new Dictionary<string, IEnumerable<string>>
        {
            { "BM_", UidPrefixesSet("BM_") },
            { "TM_", UidPrefixesSet("TM_") },
            { "TM3_", UidPrefixesSet("TM3_") },
            { "TM5_", UidPrefixesSet("TM5_") },
        };

        static IEnumerable<string> UidPrefixesSet(string prefix)
        {
            return new[]
            {
                $"{prefix}{MorphPrefix.DN}",
                $"{prefix}{MorphPrefix.UP}",
                $"{prefix}{MorphPrefix.FW}",
                $"{prefix}{MorphPrefix.BK}",
                $"{prefix}{MorphPrefix.LT}",
                $"{prefix}{MorphPrefix.RT}",
            };
        }

        void ZeroAllDuplicateMorphs()
        {
            if(Utils.BootyMagicExists())
            {
                ZeroDuplicateBootyMagicMorphs();
            }

            if(Utils.TittyMagicExists())
            {
                ZeroDuplicateTittyMagicMorphs();
            }
        }

        public void ZeroDuplicateBootyMagicMorphs()
        {
            ZeroDuplicateMorphs("BM_", new[] { "everlaster.Naturalis." });
        }

        public void ZeroDuplicateTittyMagicMorphs()
        {
            string[] packageUidPrefixes = { "everlaster.TittyMagic.", "everlaster.Naturalis." };
            ZeroDuplicateMorphs("TM_", packageUidPrefixes);
            ZeroDuplicateMorphs("TM3_", packageUidPrefixes);
            ZeroDuplicateMorphs("TM5_", packageUidPrefixes);
        }

        void ZeroDuplicateMorphs(string prefix, string[] packageUidPrefixes)
        {
            foreach(var morph in geometry.morphBank1.morphs)
            {
                if(IsDuplicateMorphByName(prefix, morph) || IsDuplicatePackageMorph(prefix, packageUidPrefixes, morph))
                {
                    morph.morphValue = 0;
                }
            }
        }

        bool IsDuplicateMorphByName(string prefix, DAZMorph morph)
        {
            return !morph.isInPackage &&
                morph.morphValue != 0 &&
                _morphNamePrefixes[prefix].Any(morph.morphName.StartsWith);
        }

        bool IsDuplicatePackageMorph(string prefix, IEnumerable<string> packageUidPrefixes, DAZMorph morph)
        {
            return morph.isInPackage &&
                morph.morphValue != 0 &&
                _morphNamePrefixes[prefix].Any(morph.morphName.StartsWith) &&
                morph.packageUid != Settings.packageId && // packageId is null when running plugin locally
                packageUidPrefixes.Any(morph.packageUid.StartsWith);
        }

        void RemovePosePresetLoadedListener()
        {
            _posePm.postLoadEvent.RemoveListener(OnPosePresetLoaded);
            _morphPm.postLoadEvent.RemoveListener(ZeroAllDuplicateMorphs);
        }

        public static void Destroy()
        {
            instance?.RemovePosePresetLoadedListener();
            /* ensure GC */
            instance = null;
        }
    }
}
