﻿using System;
using System.Collections;
using Naturalis.Common.Components;
using SimpleJSON;
using UnityEngine.UI;

namespace Naturalis.Common
{
    class ScriptBase : MVRScript
    {
        protected UnityEventsListener pluginUIEventsListener { get; private set; }
        public bool? initialized { get; protected set; }

        // Prevent ScriptBase from showing up as a plugin in Plugins tab.
        public override bool ShouldIgnore()
        {
            return true;
        }

        public override void InitUI()
        {
            base.InitUI();
            if(!UITransform || pluginUIEventsListener)
            {
                return;
            }

            StartCoroutine(InitUICo());
        }

        protected virtual Action OnUIEnabled()
        {
            return null;
        }

        protected virtual Action OnUIDisabled()
        {
            return null;
        }

        IEnumerator InitUICo()
        {
            pluginUIEventsListener = UITransform.gameObject.AddComponent<UnityEventsListener>();
            pluginUIEventsListener.onEnable.AddListener(() => StartCoroutine(OnUIEnabledCo(OnUIEnabled())));

            yield return WaitForSettingsInstance();

            if(Settings.instance.initialized == false)
            {
                enabledJSON.val = false;
                yield break;
            }

            var onUIDisabled = OnUIDisabled();
            if(onUIDisabled != null)
            {
                pluginUIEventsListener.onDisable.AddListener(() => StartCoroutine(OnUIDisabledCo(onUIDisabled)));
            }
        }

        protected static IEnumerator WaitForSettingsInstance()
        {
            while(!Settings.instance || Settings.instance.initialized == null)
            {
                yield return null;
            }
        }

        bool _inEnabledCo;

        IEnumerator OnUIEnabledCo(Action callback = null)
        {
            if(_inEnabledCo)
            {
                /* When VAM UI is toggled back on with the plugin UI already active, onEnable gets called twice and onDisable once.
                 * This ensures onEnable logic executes just once.
                 */
                yield break;
            }

            _inEnabledCo = true;
            SetGrayBackground();

            if(callback != null)
            {
                yield return null;
                yield return null;
                yield return null;

                while(initialized == null)
                {
                    yield return null;
                }

                if(initialized == false)
                {
                    yield break;
                }

                callback();
            }

            _inEnabledCo = false;
        }

        void SetGrayBackground()
        {
            var background = rightUIContent.parent.parent.parent.transform.GetComponent<Image>();
            background.color = Colors.backgroundGray;
        }

        IEnumerator OnUIDisabledCo(Action callback)
        {
            if(_inEnabledCo)
            {
                /* When VAM UI is toggled back on with the plugin UI already active, onEnable gets called twice and onDisable once.
                 * This ensures only onEnable logic executes.
                 */
                yield break;
            }

            callback();
        }

        public void AddTextFieldToJss(UIDynamicTextField textField, JSONStorableString jss)
        {
            jss.dynamicText = textField;
            textFieldToJSONStorableString.Add(textField, jss);
        }

        public void AddToggleToJsb(UIDynamicToggle toggle, JSONStorableBool jsb)
        {
            jsb.toggle = toggle.toggle;
            toggleToJSONStorableBool.Add(toggle, jsb);
        }

        public void AddSliderToJsf(UIDynamicSlider slider, JSONStorableFloat jsf)
        {
            jsf.slider = slider.slider;
            sliderToJSONStorableFloat.Add(slider, jsf);
        }

        public void DeferSetSaveFileBrowser(string fileExt)
        {
            StartCoroutine(SetSaveFileBrowser(fileExt));
        }

        static IEnumerator SetSaveFileBrowser(string fileExt)
        {
            yield return null;
            var browser = SuperController.singleton.mediaFileBrowserUI;
            browser.browseVarFilesAsDirectories = false;
            browser.SetTextEntry(true);
            browser.fileEntryField.text = $"{((int) (DateTime.UtcNow - new DateTime(1970, 1, 1)).TotalSeconds).ToString()}.{fileExt}";
            browser.ActivateFileNameField();
        }

        public virtual void RestoreFromPresetJSON(JSONClass jc)
        {
        }

        protected void OnDestroy()
        {
            if(pluginUIEventsListener)
            {
                DestroyImmediate(pluginUIEventsListener);
                pluginUIEventsListener = null;
            }
        }
    }
}
