﻿using UnityEngine;
using UnityEngine.Events;
using UnityEngine.UI;

namespace Naturalis.Common.UI
{
    sealed class SettingsWindow : WindowBase
    {
        readonly ScriptBase _owner;

        public SettingsWindow(ScriptBase owner, UnityAction onReturnToParent = null) : base(owner, nameof(SettingsWindow), onReturnToParent)
        {
            _owner = owner;
        }

        protected override void OnBuild()
        {
            BuildLeftSide();
            BuildRightSide();
            RefreshPauseSceneToggle();
        }

        void BuildLeftSide(bool rightSide = false)
        {
            AddHeaderTextField("User Preferences", rightSide);

            AddInfoTextField(
                "These settings are saved to file and apply to all Naturalis instances.",
                rightSide,
                75
            );

            AddElement(() =>
                script.CreateToggle(Settings.instance.calibrateOnUIClosedJsb, rightSide)
            );

            AddInfoTextField(
                "If a calibration is pending, it will trigger when you navigate away from the TittyMagic or BootyMagic plugin UI.",
                rightSide
            );

            AddElement(() =>
                script.CreateToggle(Settings.instance.calibrateOnPresetLoadedJsb, rightSide)
            );

            AddInfoTextField(
                "Calibrate immediately after loading a preset file. Puts to pending instead if disabled.",
                rightSide,
                75
            );

            AddElement(() =>
            {
                var toggle = script.CreateToggle(Settings.instance.checkForSizeChangesJsb, rightSide);
                toggle.label = "Monitor: Check For Size Changes";
                return toggle;
            });

            AddInfoTextField(
                "Monitor for size changes to breasts/glutes caused by non-pose morphs and atom scale." +
                " A change puts calibration to pending.",
                rightSide
            );

            AddHeaderTextField("Plugin Settings", rightSide);

            AddInfoTextField(
                "These settings are saved/loaded with the current Naturalis instance.",
                rightSide,
                75
            );

            AddElement(() =>
            {
                var toggle = script.CreateToggle(Settings.instance.freezeMotionSoundJsb, rightSide);
                toggle.toggle.onValueChanged.AddListener(_ => RefreshPauseSceneToggle());
                return toggle;
            });

            AddInfoTextField(
                "Enable the Freeze/Motion Sound toggle (in the below toolbar) during calibration." +
                " This prevents all animation from interfering with the calibration, but also disables audio.",
                rightSide,
                125
            );
        }

        void BuildRightSide(bool rightSide = true)
        {
            AddElement(() =>
            {
                var button = script.CreateButton("Open plugin data directory\nin Windows Explorer", rightSide);
                button.button.onClick.AddListener(FileUtils.OpenDataDirInExplorer);
                button.height = 100f;
                return button;
            });

            AddElement(() =>
            {
                var textField = script.CreateTextField(Settings.instance.savedPrefsInfoJss, rightSide);
                textField.UItext.fontSize = 26;
                textField.backgroundColor = Color.clear;
                var layout = textField.GetComponent<LayoutElement>();
                layout.preferredHeight = 50;
                layout.minHeight = 50;
                return textField;
            });

            AddLabelTextField("LOCAL DISTORTION PHYSICS (LDP)", rightSide);

            AddElement(() =>
                script.CreateToggle(Settings.instance.glutesLdpDefaultEnabledJsb, rightSide)
            );

            AddElement(() =>
                script.CreateToggle(Settings.instance.breastsLdpDefaultEnabledJsb, rightSide)
            );

            AddInfoTextField(
                "Enable LDP by default when plugin is added and when soft physics is enabled." +
                " LDP can cause a noticeable framerate drop.",
                rightSide,
                100
            );

            AddElement(Settings.instance.ldpUpdateRateJssc.name, () =>
                script.CreateScrollablePopup(Settings.instance.ldpUpdateRateJssc, rightSide)
            );

            AddInfoTextField(
                "A higher local distortion physics update rate is more responsive but has a higher performance impact.",
                rightSide
            );

            AddSpacer(190, rightSide);

            AddElement(Settings.instance.pauseSceneAnimationJsb.name, () =>
                script.CreateToggle(Settings.instance.pauseSceneAnimationJsb, rightSide)
            );

            AddInfoTextField(
                "Pause scene animation during calibration. Warning: this has no effect on other animation sources" +
                " (Animation Pattern, Timeline or other plugin, Cycle Forces, etc.).",
                rightSide,
                125
            );
        }

        void AddVersionTextField()
        {
            var versionJss = new JSONStorableString("version", "");
            var versionTextField = CreateVersionTextField(versionJss);
            AddElement(versionTextField);
            _owner.AddTextFieldToJss(versionTextField, versionJss);
        }

        void RefreshPauseSceneToggle()
        {
            var pauseSceneToggle = GetElement(Settings.instance.pauseSceneAnimationJsb.name);
            pauseSceneToggle.SetActiveStyle(!Settings.instance.freezeMotionSoundJsb.val, true);
        }
    }
}
