// ReSharper disable UnusedMember.Global
using UnityEngine;

namespace Naturalis.Common
{
    static class Calc
    {
        // value returned is smoothed (for better animation) i.e. no longer maps linearly to the actual rotation angle
        public static float SmoothStep(float val)
        {
            return val > 0
                ? Mathf.SmoothStep(0, 1, val)
                : -Mathf.SmoothStep(0, 1, -val);
        }

        public static float RoundToDecimals(float value, float roundFactor = 1000f)
        {
            return Mathf.Round(value * roundFactor) / roundFactor;
        }

        public static Vector3 RelativePosition(Rigidbody origin, Vector3 position)
        {
            return RelativeDiff(position - origin.position, origin.transform);
        }

        // TODO test with futa
        public static Vector3 RelativePosition(Transform transform, Vector3 position)
        {
            return RelativeDiff(position - transform.position, transform);
        }

        public static Vector3 RelativePosition(Vector3 originPosition, Transform originT, Vector3 position)
        {
            return RelativeDiff(position - originPosition, originT);
        }

        public static Vector3 RelativePosition(Vector3 originPosition, Transform originT1, Transform originT2, Vector3 position)
        {
            var difference = position - originPosition;
            return new Vector3(
                Vector3.Dot(difference, (originT1.right + originT2.right) / 2),
                Vector3.Dot(difference, (originT1.up + originT2.up) / 2),
                Vector3.Dot(difference, (originT1.forward + originT2.forward) / 2)
            );
        }

        static Vector3 RelativeDiff(Vector3 difference, Transform originTransform)
        {
            return new Vector3(
                Vector3.Dot(difference, originTransform.right),
                Vector3.Dot(difference, originTransform.up),
                Vector3.Dot(difference, originTransform.forward)
            );
        }

        public static Vector3 AveragePosition(Vector3[] positions)
        {
            var sum = Vector3.zero;
            // ReSharper disable once LoopCanBeConvertedToQuery
            foreach(var position in positions)
            {
                sum += position;
            }

            return sum / positions.Length;
        }

        public static Vector3 AverageRigidbodiesPosition(Rigidbody[] rigidbodies)
        {
            var positions = new Vector3[rigidbodies.Length];
            for(int i = 0; i < rigidbodies.Length; i++)
            {
                positions[i] = rigidbodies[i].position;
            }

            return AveragePosition(positions);
        }

        static bool IsEqualWithin(float diff, float v1, float v2)
        {
            return Mathf.Abs(v1 - v2) < diff;
        }

        // ReSharper disable once UnusedMember.Global
        public static bool DeviatesAtLeast(float v1, float v2, int percent)
        {
            bool result;
            if(v1 > v2)
            {
                result = (v1 - v2) / v1 > (float) percent / 100;
            }
            else
            {
                result = (v2 - v1) / v2 > (float) percent / 100;
            }

            return result;
        }

        public static bool VectorIsEqualWithin(float diff, Vector3 v1, Vector3 v2)
        {
            return IsEqualWithin(diff, v1.x, v2.x) &&
                IsEqualWithin(diff, v1.y, v2.y) &&
                IsEqualWithin(diff, v1.z, v2.z);
        }

        public static float[] ExponentialMovingAverage(float[] source, float k)
        {
            float[] result = new float[source.Length];
            result[source.Length - 1] = source[source.Length - 1];
            for(int i = source.Length - 2; i >= 0; i--)
            {
                result[i] = k * source[i] + (1 - k) * result[i + 1];
            }

            return result;
        }

        public static Vector3[] ExponentialMovingAverage(Vector3[] source, float k)
        {
            var result = new Vector3[source.Length];
            result[source.Length - 1] = source[source.Length - 1];
            for(int i = source.Length - 2; i >= 0; i--)
            {
                result[i] = new Vector3(
                    k * source[i].x + (1 - k) * result[i + 1].x,
                    k * source[i].y + (1 - k) * result[i + 1].y,
                    k * source[i].z + (1 - k) * result[i + 1].z
                );
            }

            return result;
        }

#region *** Curve Functions ***

        // https://www.desmos.com/calculator/ln9oecyv4y
        // dual: https://www.desmos.com/calculator/icmymkbfux
        public static float InverseSmoothStep(float value, float b, float curvature, float midPoint)
        {
            if(value < 0)
            {
                return 0;
            }

            if(value > b)
            {
                return 1;
            }

            float s = curvature < -2.99f ? -2.99f : curvature > 0.99f ? 0.99f : curvature;
            float p = midPoint * b;
            p = p < 0 ? 0 : p > b ? b : p;
            float c = 2 / (1 - s) - p / b;

            if(value < p)
            {
                return F1(value, b, p, c);
            }

            return 1 - F1(b - value, b, b - p, c);
        }

        static float F1(float value, float b, float n, float c)
        {
            return Mathf.Pow(value, c) / (b * Mathf.Pow(n, c - 1));
        }

        // https://www.desmos.com/calculator/7s7iak8rqb
        public static float QuadraticRegression(float value)
        {
            return -0.118f * value * value + 0.97f * value;
        }

        // ReSharper disable once UnusedMember.Local
        static float Polynomial(
            float x,
            float a = 1,
            float b = 1,
            float c = 1,
            float p = 1,
            float q = 1
        )
        {
            return a * Mathf.Pow(x, p) + b * Mathf.Pow(x, q) + c * x;
        }

        // https://www.desmos.com/calculator/okwggb6a3s
        // ReSharper disable once UnusedMember.Global
        public static float QuadraticOneToZero(float x, float a)
        {
            return a * Mathf.Pow(x - Mathf.Sqrt(1 / a), 2);
        }

        public static float SoftnessBaseCurve(float x)
        {
            return Exponential1(x, 6.44f, 1.27f, 1.15f);
        }

        public static float ForcePhysicsSoftnessCurve(float x)
        {
            return Exponential1(x, 1.60f, 2.29f, 1.00f);
        }

        public static float AngleTargetMassCurve(float x)
        {
            return Exponential1(x, 0.30f, 1.75f, 1.00f);
        }

        public static float AngleTargetSoftnessCurve(float x)
        {
            return Exponential1(x, 3.00f, 1.35f, 1.00f);
        }

        public static float InOutSpringSoftnessCurve(float x)
        {
            return Exponential1(x, 2.11f, 1.21f, 0.88f);
        }

        // https://www.desmos.com/calculator/ltyekxtlji
        public static float Exponential1(float x, float b, float p, float q, float a = 1, float s = 0)
        {
            return (1 - b) * Mathf.Pow(a * x + s, p) + b * Mathf.Pow(a * x + s, q);
        }

        // https://www.desmos.com/calculator/6pxtrvvyby
        public static float Exponential2(float x, float c, float s = 0)
        {
            float baseValue = (2 - c) * (x + s) - 1 + c / 2;
            return c / (1 + c / 3) + baseValue * baseValue;
        }

        // https://www.desmos.com/calculator/mpmm8lzvs2
        // ReSharper disable once MemberCanBePrivate.Global
        public static float Exponential3(float x, float c, float s, float l, float q, float p)
        {
            float xToLimitPow = Mathf.Pow(x / l, q);
            return c + s * Mathf.Pow(1 - xToLimitPow, p);
        }

#endregion
    }
}
