using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Text;
using System.Text.RegularExpressions;
using SimpleJSON;
using UnityEngine;
using UnityEngine.UI;

namespace Naturalis.Common
{
    static class Utils
    {
        public static bool TittyMagicExists()
        {
            return Type.GetType("Naturalis.TittyMagic") != null;
        }

        public static bool BootyMagicExists()
        {
            return Type.GetType("Naturalis.BootyMagic") != null;
        }

        // ReSharper disable once UnusedMember.Global
        public static MVRScript FindPluginOnAtom(Atom atom, string search)
        {
            string match = atom.GetStorableIDs().Find(id => id.Contains(search));
            return match == null ? null : atom.GetStorableByID(match) as MVRScript;
        }

        public static IEnumerable<Atom> GetAtomsByType(string type)
        {
            foreach(var atom in SuperController.singleton.GetAtoms())
            {
                if(atom.type == type)
                {
                    yield return atom;
                }
            }
        }

        public static float PhysicsRateMultiplier()
        {
            return 0.01666667f / Time.fixedDeltaTime;
        }

        public static string ParamName(string paramName, string side)
        {
            string result = paramName;
            if(side != Side.LEFT)
            {
                result += $" {side}";
            }

            return result;
        }

        // ReSharper disable once UnusedMember.Global
        public static string ObjectPropertiesString(object obj)
        {
            string result = "";
            foreach(PropertyDescriptor descriptor in TypeDescriptor.GetProperties(obj))
            {
                result += $"{descriptor.Name} = {descriptor.GetValue(obj)}\n";
            }

            return result;
        }

        /// <summary>
        /// Get a string containing a visual representation of a Transform's children hierarchy.
        /// </summary>
        /// <param name="root">The parent Transform (GameObject)</param>
        /// <param name="maxDepth">Maximum depth to traverse the hierarchy.</param>
        /// <param name="propertyDel">A delegate function that takes a Transform and returns the string to print as an entry in the hierarchy.
        /// For example, to print a hierarchy of each transform's name: <code>thisTransform => thisTransform.name</code> A null value will print names.</param>
        /// <returns>A string containing a visual hierarchy of all child transforms, including the root</returns>
        // ReSharper disable once UnusedMember.Global
        public static string ObjectHierarchyToString(Transform root, int? maxDepth = null, Func<Transform, string> propertyDel = null)
        {
            if(propertyDel == null)
            {
                propertyDel = t => t.name;
            }

            var builder = new StringBuilder();
            ObjectHierarchyToString(root, propertyDel, builder, maxDepth);

            if(builder.Length < 1024)
            {
                return builder.ToString();
            }

            return
                $"Output string length {builder.Length.ToString()} may be too large for viewing in VAM. " +
                $"See %userprofile%/AppData/LocalLow/MeshedVR/VaM/output_log.txt for the full output.\n{builder}";
        }

        static void ObjectHierarchyToString(
            Transform root,
            Func<Transform, string> propertyDel,
            StringBuilder builder,
            int? maxDepth,
            int currentDepth = 0
        )
        {
            if(currentDepth > maxDepth)
            {
                return;
            }

            for(int i = 0; i < currentDepth; i++)
            {
                builder.Append("|   ");
            }

            builder.Append(propertyDel(root) + "\n");
            foreach(Transform child in root)
            {
                ObjectHierarchyToString(child, propertyDel, builder, maxDepth, currentDepth + 1);
            }
        }

        public static bool GlobalAnimationFrozen()
        {
            bool mainToggleFrozen =
                SuperController.singleton.freezeAnimationToggle &&
                SuperController.singleton.freezeAnimationToggle.isOn;
            bool altToggleFrozen =
                SuperController.singleton.freezeAnimationToggleAlt &&
                SuperController.singleton.freezeAnimationToggleAlt.isOn;
            return mainToggleFrozen || altToggleFrozen;
        }

        public static Transform DestroyLayout(Transform transform)
        {
            UnityEngine.Object.Destroy(transform.GetComponent<LayoutElement>());
            return transform;
        }

        public static HorizontalLayoutGroup CreateHorizontalLayoutGroup(RectTransform uiContent)
        {
            var verticalLayoutGroup = uiContent.GetComponent<VerticalLayoutGroup>();

            var gameObj = new GameObject();
            gameObj.transform.SetParent(verticalLayoutGroup.transform, false);

            var horizontalLayoutGroup = gameObj.AddComponent<HorizontalLayoutGroup>();
            horizontalLayoutGroup.spacing = 16f;
            horizontalLayoutGroup.childForceExpandWidth = true;
            horizontalLayoutGroup.childControlHeight = false;

            return horizontalLayoutGroup;
        }

        public static Rigidbody FindRigidbody(Atom atom, string name)
        {
            foreach(var rigidbody in atom.GetComponentsInChildren<Rigidbody>())
            {
                if(rigidbody.name == name)
                {
                    return rigidbody;
                }
            }

            return null;
        }

        public static IEnumerator WaitAndRepeat(Action callback, int times, float intervalWait)
        {
            int count = 0;
            while(count < times)
            {
                count++;
                yield return new WaitForSeconds(intervalWait);
                callback.Invoke();
            }
        }

        public static Regex NewRegex(string regexStr)
        {
            return new Regex(regexStr, RegexOptions.Compiled);
        }

        // ReSharper disable once UnusedMember.Global
        public static string ToValueFormat(int roundFactor)
        {
            switch(roundFactor)
            {
                case 1: return "F0";
                case 10: return "F1";
                case 100: return "F2";
                case 1000: return "F3";
                case 10000: return "F4";
                case 100000: return "F5";
                default: throw new Exception($"Invalid roundFactor {roundFactor}");
            }
        }

        public static int ToRoundFactor(string valueFormat)
        {
            switch(valueFormat)
            {
                case "F0": return 1;
                case "F1": return 10;
                case "F2": return 100;
                case "F3": return 1000;
                case "F4": return 10000;
                case "F5": return 100000;
                default: throw new Exception($"Invalid valueFormat {valueFormat}");
            }
        }

        public static Vector3 DirectionVector(Rigidbody rb, string direction)
        {
            switch(direction)
            {
                case Direction.UP: return rb.transform.up;
                case Direction.DOWN: return -rb.transform.up;
                case Direction.FORWARD: return rb.transform.forward;
                case Direction.BACK: return -rb.transform.forward;
                case Direction.LEFT: return -rb.transform.right;
                case Direction.RIGHT: return rb.transform.right;
                default: throw new ArgumentException($"Invalid direction {direction}");
            }
        }

        public static void RemoveInternalKeys(JSONClass jc)
        {
            var keysToRemove = new List<string>();
            foreach(string key in jc.Keys)
            {
                if(key.StartsWith("internal_"))
                {
                    keysToRemove.Add(key);
                }
            }

            foreach(string key in keysToRemove)
            {
                jc.Remove(key);
            }
        }
    }
}
