﻿using System.Collections;
using Naturalis.Common;
using UnityEngine;

namespace Naturalis.SET
{
    sealed class PersonMonitor : MonoBehaviour
    {
        Person _person;
        FrequencyRunner _runner;
        JSONStorable _breastInOut;
        DAZCharacter _selectedCharacter;

        float _fixedDeltaTime;

        public bool BreastSoftPhysicsIsEnabled()
        {
            return _globalSoftPhysicsOn && _atomSoftPhysicsOn && _breastSoftPhysicsOn;
        }

        public bool GluteSoftPhysicsIsEnabled()
        {
            return _globalSoftPhysicsOn && _atomSoftPhysicsOn && _gluteSoftPhysicsOn;
        }

        bool _globalSoftPhysicsOn;
        bool _atomSoftPhysicsOn;
        bool _breastSoftPhysicsOn;
        bool _gluteSoftPhysicsOn;

        public bool initialized { get; private set; }

        public void Init()
        {
            enabled = false; // will be enabled after first-time calibration
            _person = Person.instance;
            _runner = new FrequencyRunner(new JSONStorableFloat("Settings Check Frequency", 1, 0.2f, 5, false));
            _selectedCharacter = _person.geometry.selectedCharacter;
            _fixedDeltaTime = Time.fixedDeltaTime;

            if(Person.IsFemale())
            {
                _breastInOut = Person.instance.atom.GetStorableByID("BreastInOut");
                _atomSoftPhysicsOn = _person.softBodyPhysicsEnabler.GetBoolParamValue("enabled");
                _globalSoftPhysicsOn = UserPreferences.singleton.softPhysics;

                if(Settings.tittyMagic.IsEnabledNullSafe())
                {
                    /* prevent breasts being flattened due to breastInOut morphs on scene load with plugin already present */
                    _breastInOut.SetBoolParamValue("enabled", true);
                    _breastInOut.SetBoolParamValue("enabled", false);
                }

                StartCoroutine(InitSoftPhysicsValuesCo());
            }
            else
            {
                initialized = true;
            }
        }

        IEnumerator InitSoftPhysicsValuesCo()
        {
            if(Settings.bootyMagic.IsEnabledNullSafe())
            {
                while(!Settings.bootyMagic || !Settings.bootyMagic.IsBoolJSONParam(Constant.GLUTE_SOFT_PHYSICS_ENABLED))
                {
                    yield return null;
                }

                _gluteSoftPhysicsOn = Settings.bootyMagic.GetBoolParamValue(Constant.GLUTE_SOFT_PHYSICS_ENABLED);
            }

            if(Settings.tittyMagic.IsEnabledNullSafe())
            {
                while(!Settings.tittyMagic || !Settings.tittyMagic.IsBoolJSONParam(Constant.BREAST_SOFT_PHYSICS_ENABLED))
                {
                    yield return null;
                }

                _breastSoftPhysicsOn = Settings.tittyMagic.GetBoolParamValue(Constant.BREAST_SOFT_PHYSICS_ENABLED);
            }

            initialized = true;
        }

        public void Check()
        {
            if(!initialized || !enabled)
            {
                return;
            }

            if(Settings.tittyMagic.IsEnabledNullSafe())
            {
                CheckBreasts();
            }

            if(Settings.tittyMagic.IsEnabledNullSafe() || Settings.bootyMagic.IsEnabledNullSafe())
            {
                CheckCommon();

                if(_selectedCharacter != _person.geometry.selectedCharacter)
                {
                    StartCoroutine(OnCharacterChangedCo());
                }
            }

            if(Person.IsFemale())
            {
                CheckSoftPhysicsToggled();
            }

            CheckPhysicsRateChanged();
        }

        void CheckBreasts()
        {
            if(Person.IsFemale())
            {
                if(_breastInOut.GetBoolParamValue("enabled"))
                {
                    _breastInOut.SetBoolParamValue("enabled", false);
                    Loggr.Message("Auto Breast In/Out Morphs disabled - directional force morphing works better without them.");
                }

                if(!_person.geometry.useAuxBreastColliders)
                {
                    _person.geometry.useAuxBreastColliders = true;
                    Loggr.Message("Breast Hard Colliders enabled.");
                }

                /* Disable pectoral joint rb's collisions if enabled by e.g. person atom collisions being toggled off/on */
                if(Person.PectoralCollisionsEnabled(true))
                {
                    Person.SetPectoralCollisions(false);
                }
            }
            else
            {
                /* Force enable pectoral joint rb's collisions for futa */
                if(Person.PectoralCollisionsEnabled(false))
                {
                    Person.SetPectoralCollisions(true);
                }
            }
        }

        void CheckCommon()
        {
            if(_person.isFemale)
            {
                if(!_person.geometry.useAdvancedColliders)
                {
                    _person.geometry.useAdvancedColliders = true;
                    Loggr.Message(
                        "Advanced Colliders enabled - they are necessary for directional force morphing and hard colliders to work."
                    );
                }
            }
        }

        void CheckSoftPhysicsToggled()
        {
            bool atomSoftPhysicsOn = _person.softBodyPhysicsEnabler.GetBoolParamValue("enabled");
            bool globalSoftPhysicsOn = UserPreferences.singleton.softPhysics;

            if(Settings.bootyMagic.IsEnabledNullSafe())
            {
                bool gluteSoftPhysicsOn = _person.glutePhysicsMesh.on;
                string location = LocationWhereStillDisabled(gluteSoftPhysicsOn, _gluteSoftPhysicsOn, atomSoftPhysicsOn, globalSoftPhysicsOn);
                if(!string.IsNullOrEmpty(location))
                {
                    Loggr.Message($"Soft physics is still disabled in {location}");
                }

                bool enabledNow = globalSoftPhysicsOn && atomSoftPhysicsOn && gluteSoftPhysicsOn;
                if(enabledNow != GluteSoftPhysicsIsEnabled())
                {
                    Settings.bootyMagic.SetBoolParamValue(Constant.INTERNAL_SYNC_SOFT_PHYSICS, enabledNow);
                }

                _gluteSoftPhysicsOn = gluteSoftPhysicsOn;
            }

            if(Settings.tittyMagic.IsEnabledNullSafe())
            {
                bool breastSoftPhysicsOn = _person.breastPhysicsMesh.on;
                string location = LocationWhereStillDisabled(breastSoftPhysicsOn, _breastSoftPhysicsOn, atomSoftPhysicsOn, globalSoftPhysicsOn);
                if(!string.IsNullOrEmpty(location))
                {
                    Loggr.Message($"Soft physics is still disabled in {location}");
                }

                bool enabledNow = globalSoftPhysicsOn && atomSoftPhysicsOn && breastSoftPhysicsOn;
                if(enabledNow != BreastSoftPhysicsIsEnabled())
                {
                    Settings.tittyMagic.SetBoolParamValue(Constant.INTERNAL_SYNC_SOFT_PHYSICS, enabledNow);
                }

                _breastSoftPhysicsOn = breastSoftPhysicsOn;
            }

            _atomSoftPhysicsOn = atomSoftPhysicsOn;
            _globalSoftPhysicsOn = globalSoftPhysicsOn;
        }

        void CheckPhysicsRateChanged()
        {
            float value = Time.fixedDeltaTime;
            if(Mathf.Abs(value - _fixedDeltaTime) > 0.001f)
            {
                Settings.tittyMagic.CallActionNullSafe(Constant.INTERNAL_ON_PHYSICS_RATE_CHANGED);
                Settings.bootyMagic.CallActionNullSafe(Constant.INTERNAL_ON_PHYSICS_RATE_CHANGED);
            }

            _fixedDeltaTime = value;
        }

        bool _inOnCharacterChangedCo;

        IEnumerator OnCharacterChangedCo()
        {
            if(_inOnCharacterChangedCo)
            {
                yield break;
            }

            _inOnCharacterChangedCo = true;
            while(!_person.SelectedCharacterIsReady())
            {
                yield return null;
            }

            if(_selectedCharacter.isMale == _person.geometry.selectedCharacter.isMale)
            {
                yield return _person.RefreshSkin();
                Settings.tittyMagic.CallActionNullSafe(Constant.INTERNAL_ON_CHARACTER_CHANGED);
                Settings.bootyMagic.CallActionNullSafe(Constant.INTERNAL_ON_CHARACTER_CHANGED);
                _selectedCharacter = _person.geometry.selectedCharacter;
            }
            else
            {
                Loggr.Message(
                    "Changing gender while the plugin is active is not supported. " +
                    "Disable the plugin, change the gender and then reload the plugin."
                );
                _person.geometry.selectedCharacter = _selectedCharacter;
            }

            _inOnCharacterChangedCo = false;
        }

        string LocationWhereStillDisabled(bool currEnabled, bool prevEnabled, bool atomSoftPhysicsOn, bool globalSoftPhysicsOn)
        {
            if(currEnabled && !prevEnabled)
            {
                if(!atomSoftPhysicsOn && !globalSoftPhysicsOn)
                {
                    return "Control & Physics 1 and in User Preferences";
                }

                if(!atomSoftPhysicsOn)
                {
                    return "Control & Physics 1";
                }

                if(!globalSoftPhysicsOn)
                {
                    return "User Preferences";
                }
            }
            else if(atomSoftPhysicsOn && !_atomSoftPhysicsOn)
            {
                if(!currEnabled && !globalSoftPhysicsOn)
                {
                    return "the plugin UI and in User Preferences";
                }

                if(!currEnabled)
                {
                    return "the plugin UI";
                }

                if(!globalSoftPhysicsOn)
                {
                    return "User Preferences";
                }
            }
            else if(globalSoftPhysicsOn && !_globalSoftPhysicsOn)
            {
                if(!currEnabled && !atomSoftPhysicsOn)
                {
                    return "the plugin UI and in Control & Physics 1";
                }

                if(!currEnabled)
                {
                    return "the plugin UI";
                }

                if(!atomSoftPhysicsOn)
                {
                    return "Control & Physics 1";
                }
            }

            return null;
        }

        void Update()
        {
            _runner.Run(Check);
        }
    }
}
