﻿//
using System;
using System.Collections;
using System.Collections.Generic;
using Naturalis.Common;
using Naturalis.SET;
using SimpleJSON;
using System.Linq;
using UnityEngine;

namespace Naturalis
{
    /* Provides a global settings UI and bootstraps common variables */
    sealed class Settings : MonoBehaviour
    {
        public const string NATURALIS_VERSION = "1.0.2";
        public static bool envIsDevelopment { get; private set; }
        public static Settings instance { get; set; }
        public static string packageId { get; private set; }
        public static ScriptBase initScript { get; private set; }
        public static JSONStorable bootyMagic { get; private set; }
        public static JSONStorable tittyMagic { get; private set; }
        public static PersonMonitor personMonitor { get; private set; }

        const string PREFS_FILE = "preferences.json";

#region *** Init ***

        public bool? initialized { get; private set; }

        public static bool IsInitialized()
        {
            return instance.initialized == true;
        }

        bool _setupCalled;

        /// <param name="script">
        /// ScriptBase instance used to create common storables and provide access to ScriptBase and MVRScript methods and properties.
        /// </param>
        public void Setup(ScriptBase script)
        {
            if(_setupCalled)
            {
                return;
            }

            _setupCalled = true;
            initScript = script;
            packageId = initScript.GetPackageId();

            #if ENV_DEVELOPMENT
            {
                envIsDevelopment = true;
            }
            #endif

            try
            {
                var atom = initScript.containingAtom;
                if(atom.type != "Person")
                {
                    Loggr.Message($"Add to a Person atom, not {atom.type}.");
                    initialized = false;
                    return;
                }

                var instances = atom.FindStorablesByRegexMatch(Utils.NewRegex($@"^plugin#\d+_{nameof(Naturalis)}.(BootyMagic|TittyMagic)"));
                if(instances.Any(x => x.GetPluginId() != initScript.GetPluginId()))
                {
                    Loggr.Message($"{nameof(Naturalis)} is already added.");
                    initialized = false;
                    return;
                }

                if(atom.FindStorablesByRegexMatch(Utils.NewRegex(@"^plugin#\d+_TittyMagic")).Count > 0)
                {
                    Loggr.Message($"Person already has an instance of TittyMagic - remove it and then reload {nameof(Naturalis)}.");
                    initialized = false;
                    return;
                }

                StartCoroutine(InitCo());
            }
            catch(Exception e)
            {
                initialized = false;
                Loggr.Error($"Init: {e}");
            }
        }

        public JSONStorableString savedPrefsInfoJss { get; private set; }
        public JSONStorableBool pauseFemaleMorphUpdatesJsb { get; private set; }
        public JSONStorableBool pauseMaleMorphUpdatesJsb { get; private set; }
        public JSONStorableBool freezeMotionSoundJsb { get; private set; }
        public JSONStorableBool pauseSceneAnimationJsb { get; private set; }
        public JSONStorableBool glutesLdpDefaultEnabledJsb { get; private set; }
        public JSONStorableBool breastsLdpDefaultEnabledJsb { get; private set; }
        public JSONStorableStringChooser ldpUpdateRateJssc { get; private set; }
        public JSONStorableBool checkForSizeChangesJsb { get; private set; }
        public JSONStorableBool calibrateOnUIClosedJsb { get; private set; }
        public JSONStorableBool calibrateOnPresetLoadedJsb { get; private set; }
        public UIModManager uiModManager { get; private set; }

        public bool MorphUpdatesPaused()
        {
            return uiModManager.MorphsTabIsVisible() && instance.pauseFemaleMorphUpdatesJsb.val;
        }

        IEnumerator InitCo()
        {
            yield return new WaitForEndOfFrame();
            while(SuperController.singleton.isLoading)
            {
                yield return null;
            }

            yield return SetupPerson(initScript.containingAtom);
            if(initialized == false)
            {
                yield break;
            }

            try
            {
                bootyMagic = Person.IsFemale() ? initScript.GetOwnScript("BootyMagic") : null;
                tittyMagic = initScript.GetOwnScript("TittyMagic");
                SetupStorables();
                ReadUserPreferencesFromFile();
                personMonitor = gameObject.AddComponent<PersonMonitor>();
                personMonitor.Init();
                uiModManager = gameObject.AddComponent<UIModManager>();
                uiModManager.ModifyAtomUI();
                initialized = true;
            }
            catch(Exception e)
            {
                Loggr.Error($"Init error: {e}");
                initialized = false;
            }
        }

        IEnumerator SetupPerson(Atom atom)
        {
            Person.instance.atom = atom;
            const int limit = 15;
            yield return Person.instance.WaitForGeometryAndSkinReady(limit);
            if(!Person.instance.characterReady)
            {
                Loggr.Error(
                    $"Selected character {Person.instance.geometry.selectedCharacter.name} was not " +
                    $"ready after {limit.ToString()} seconds of waiting"
                );
                initialized = false;
                yield break;
            }

            if(!Person.instance.skinReady)
            {
                Loggr.Error($"Person skin materials not found after {limit.ToString()} seconds of waiting");
                initialized = false;
                yield break;
            }

            try
            {
                Person.instance.Setup();
            }
            catch(Exception e)
            {
                Loggr.Error($"Person setup error: {e}");
            }
        }

        void SetupStorables()
        {
            savedPrefsInfoJss = new JSONStorableString("savedPrefsInfo", "");
            pauseFemaleMorphUpdatesJsb = new JSONStorableBool("Pause Morph Updates", true);
            pauseMaleMorphUpdatesJsb = new JSONStorableBool("Pause Morph Updates", true);
            pauseFemaleMorphUpdatesJsb.setCallbackFunction = value => pauseMaleMorphUpdatesJsb.valNoCallback = value;
            pauseMaleMorphUpdatesJsb.setCallbackFunction = value => pauseFemaleMorphUpdatesJsb.valNoCallback = value;
            SetupPluginInstanceStorables();
            SetupUserPreferencesStorables();
        }

        void SetupPluginInstanceStorables()
        {
            freezeMotionSoundJsb = initScript.NewStorableBool("Calibration: Freeze Motion/Sound", true);
            freezeMotionSoundJsb.setCallbackFunction = value =>
            {
                if(value)
                {
                    pauseSceneAnimationJsb.val = false;
                }
            };

            pauseSceneAnimationJsb = initScript.NewStorableBool("Calibration: Pause Scene Animation", false);
            pauseSceneAnimationJsb.setCallbackFunction = value =>
            {
                if(value && freezeMotionSoundJsb.val)
                {
                    pauseSceneAnimationJsb.valNoCallback = false;
                }
            };
        }

        void SetupUserPreferencesStorables()
        {
            calibrateOnUIClosedJsb = new JSONStorableBool("Calibrate On Plugin UI Closed", false, (bool _) => SyncPreferencesFile());
            calibrateOnPresetLoadedJsb = new JSONStorableBool("Calibrate On Preset Loaded", true, (bool _) => SyncPreferencesFile());
            glutesLdpDefaultEnabledJsb = new JSONStorableBool("Glutes LDP Default Enabled", false, (bool _) => SyncPreferencesFile());
            breastsLdpDefaultEnabledJsb = new JSONStorableBool("Breasts LDP Default Enabled", false, (bool _) => SyncPreferencesFile());

            var options = new List<string> { "20 Hz", "30 Hz", "45 Hz", "60 Hz", "Use Physics Rate" };
            ldpUpdateRateJssc = new JSONStorableStringChooser(
                "Local Distortion Physics Update Rate",
                options,
                "30 Hz",
                "LDP Update Rate"
            );
            ldpUpdateRateJssc.setCallbackFunction += _ => SyncPreferencesFile();

            checkForSizeChangesJsb = new JSONStorableBool("Monitor: Check For Size Changes", true, value =>
            {
                if(!value)
                {
                    bootyMagic.SetStringParamValueNullSafe(Constant.INTERNAL_UNSET_CALIBRATION_PENDING, Constant.SIZE_CHANGED);
                    tittyMagic.SetStringParamValueNullSafe(Constant.INTERNAL_UNSET_CALIBRATION_PENDING, Constant.SIZE_CHANGED);
                }

                SyncPreferencesFile();
            });
        }

        void ReadUserPreferencesFromFile()
        {
            var prefsJson = FileUtils.ReadDataJSON(PREFS_FILE);
            if(prefsJson != null)
            {
                SetValueFromPrefsJson(prefsJson, calibrateOnUIClosedJsb);
                SetValueFromPrefsJson(prefsJson, calibrateOnPresetLoadedJsb);
                SetValueFromPrefsJson(prefsJson, ldpUpdateRateJssc);
                SetValueFromPrefsJson(prefsJson, checkForSizeChangesJsb);

                if(prefsJson.HasKey("Local Distortion Physics Default On")) // legacy setting
                {
                    bool ldpDefaultOn = prefsJson["Local Distortion Physics Default On"].AsBool;
                    glutesLdpDefaultEnabledJsb.val = glutesLdpDefaultEnabledJsb.defaultVal = ldpDefaultOn;
                    breastsLdpDefaultEnabledJsb.val = breastsLdpDefaultEnabledJsb.defaultVal = ldpDefaultOn;
                }
                else
                {
                    SetValueFromPrefsJson(prefsJson, glutesLdpDefaultEnabledJsb);
                    SetValueFromPrefsJson(prefsJson, breastsLdpDefaultEnabledJsb);
                }
            }
            else
            {
                SyncPreferencesFile();
            }
        }

        static void SetValueFromPrefsJson(JSONClass prefsJson, JSONStorableParam storable)
        {
            if(!prefsJson.HasKey(storable.name))
            {
                return;
            }

            var jsb = storable as JSONStorableBool;
            if(jsb != null)
            {
                jsb.val = jsb.defaultVal = prefsJson[jsb.name].AsBool;
                return;
            }

            var jssc = storable as JSONStorableStringChooser;
            if(jssc != null)
            {
                jssc.val = jssc.defaultVal = prefsJson[jssc.name];
            }
        }

#endregion

        void SyncPreferencesFile()
        {
            var jc = FileUtils.ReadDataJSON(PREFS_FILE) ?? new JSONClass();
            jc.Remove("Local Distortion Physics Default On"); // legacy setting
            jc[calibrateOnUIClosedJsb.name].AsBool = calibrateOnUIClosedJsb.val;
            jc[calibrateOnPresetLoadedJsb.name].AsBool = calibrateOnPresetLoadedJsb.val;
            jc[glutesLdpDefaultEnabledJsb.name].AsBool = glutesLdpDefaultEnabledJsb.val;
            jc[breastsLdpDefaultEnabledJsb.name].AsBool = breastsLdpDefaultEnabledJsb.val;
            jc[ldpUpdateRateJssc.name] = ldpUpdateRateJssc.val;
            jc[checkForSizeChangesJsb.name].AsBool = checkForSizeChangesJsb.val;
            FileUtils.WriteDataJSON(
                PREFS_FILE,
                jc,
                () => StartCoroutine(ShowSavedInfoTextCo())
            );
        }

        bool _inShowSavedInfoTextCo;
        bool _doClearSavedInfoText = true;

        IEnumerator ShowSavedInfoTextCo()
        {
            if(!IsInitialized())
            {
                yield break;
            }

            if(_inShowSavedInfoTextCo)
            {
                _doClearSavedInfoText = false;
            }

            _inShowSavedInfoTextCo = true;
            savedPrefsInfoJss.val = "Changes saved";
            yield return new WaitForSeconds(1.0f);
            if(_doClearSavedInfoText)
            {
                savedPrefsInfoJss.val = "";
            }

            _doClearSavedInfoText = true;
            _inShowSavedInfoTextCo = false;
        }

        void OnDestroy()
        {
            try
            {
                Destroy(personMonitor);
                Person.RestoreOriginalUseColliders();
                instance = null;
                personMonitor = null;
                initScript = null;
                bootyMagic = null;
                tittyMagic = null;
                Person.Destroy();
                Destroy(uiModManager);
            }
            catch(Exception e)
            {
                if(initialized == true)
                {
                    SuperController.LogError($"{nameof(OnDestroy)}: {e}");
                }
                else
                {
                    Debug.LogError($"{nameof(OnDestroy)}: {e}");
                }
            }
        }
    }
}
