﻿using System.Collections.Generic;
using Naturalis.Common;
using Naturalis.Common.Components;
using Naturalis.Common.Models;
using Naturalis.TM.Components;
using UnityEngine;
using static Direction;

namespace Naturalis.TM.Handlers
{
    static class ForceMorphHandler
    {
        static Breast _leftBreast;
        static Breast _rightBreast;

        public static Dictionary<string, List<ConfigurableMorph>> morphSets { get; private set; }

        public static StorableFloat baseJsf { get; private set; }
        public static StorableFloat upJsf { get; private set; }
        public static StorableFloat depthInJsf { get; private set; }
        public static StorableFloat depthOutJsf { get; private set; }
        public static StorableFloat sidewaysInJsf { get; private set; }
        public static StorableFloat sidewaysOutJsf { get; private set; }

        static float UserUpMultiplier()
        {
            return Calc.QuadraticRegression(baseJsf.val * upJsf.val);
        }

        static float UserDepthInMultiplier()
        {
            return Calc.QuadraticRegression(baseJsf.val * depthInJsf.val);
        }

        static float UserDepthOutMultiplier()
        {
            return Calc.QuadraticRegression(baseJsf.val * depthOutJsf.val);
        }

        static float UserSidewaysInMultiplier()
        {
            return Calc.QuadraticRegression(baseJsf.val * sidewaysInJsf.val);
        }

        static float UserSidewaysOutMultiplier()
        {
            return Calc.QuadraticRegression(baseJsf.val * sidewaysOutJsf.val);
        }

        public static JSONStorableStringChooser directionJssc { get; private set; }

        public static void Init()
        {
            _leftBreast = TittyMagic.script.leftBreast;
            _rightBreast = TittyMagic.script.rightBreast;

            baseJsf = TittyMagic.script.NewStorableFloat("Force Morphing Base Multiplier", 1.00f, 0.00f, 2.00f);
            upJsf = TittyMagic.script.NewStorableFloat("Force Morphing Up", 1.00f, 0.00f, 2.00f);
            depthInJsf = TittyMagic.script.NewStorableFloat("Force Morphing Depth In", 1.00f, 0.00f, 2.00f);
            depthOutJsf = TittyMagic.script.NewStorableFloat("Force Morphing Depth Out", 1.00f, 0.00f, 2.00f);
            sidewaysInJsf = TittyMagic.script.NewStorableFloat("Force Morphing Sideways In", 1.00f, 0.00f, 2.00f);
            sidewaysOutJsf = TittyMagic.script.NewStorableFloat("Force Morphing Sideways Out", 1.00f, 0.00f, 2.00f);

            baseJsf.setJSONCallbackFunction = OnMorphMultiplierValueChanged;
            upJsf.setJSONCallbackFunction = OnMorphMultiplierValueChanged;
            depthInJsf.setJSONCallbackFunction = OnMorphMultiplierValueChanged;
            depthOutJsf.setJSONCallbackFunction = OnMorphMultiplierValueChanged;
            sidewaysInJsf.setJSONCallbackFunction = OnMorphMultiplierValueChanged;
            sidewaysOutJsf.setJSONCallbackFunction = OnMorphMultiplierValueChanged;

            LoadSettings();
        }

        static void OnMorphMultiplierValueChanged(JSONStorableFloat jsf)
        {
            var storable = (StorableFloat) jsf;
            if(SoftPhysicsHandler.LdpUpdateIsEnabled())
            {
                TittyMagic.script.SetCalibrationPending(storable);
            }
        }

        static void LoadSettings()
        {
            morphSets = new Dictionary<string, List<ConfigurableMorph>>
            {
                { UP_L, UpForceMorphs("L") },
                { UP_R, UpForceMorphs("R") },
                { UP_C, UpForceCenterMorphs() },
                { BACK_L, DepthInForceMorphs("L") },
                { BACK_R, DepthInForceMorphs("R") },
                { BACK_C, DepthInForceCenterMorphs() },
                { FORWARD_L, DepthOutForceMorphs("L") },
                { FORWARD_R, DepthOutForceMorphs("R") },
                { FORWARD_C, DepthOutForceCenterMorphs() },
                { LEFT_L, LeftForceLeftMorphs() },
                { LEFT_R, LeftForceRightMorphs() },
                { RIGHT_L, RightForceLeftMorphs() },
                { RIGHT_R, RightForceRightMorphs() },
            };

            var directions = new List<string>
            {
                UP,
                UP_C,
                BACK,
                BACK_C,
                FORWARD,
                FORWARD_C,
                LEFT_L,
                RIGHT_L,
                LEFT_R,
                RIGHT_R,
            };
            directionJssc = new JSONStorableStringChooser("direction", directions, directions[0], "Direction");
        }

#region *** Morph Configuration ***

        static ConfigurableMorph NewConfigurableMorph(string morphName, bool isNegative, float softMultiplier, float massMultiplier)
        {
            return new ConfigurableMorph(
                Person.GetBreastMorph(morphName),
                isNegative,
                new JSONStorableFloat("softMultiplier", softMultiplier, -3.00f, 3.00f),
                new JSONStorableFloat("massMultiplier", massMultiplier, -3.00f, 3.00f)
            );
        }

        static List<ConfigurableMorph> UpForceMorphs(string side)
        {
            string prefix = $"{MorphPrefix.UP}/{MorphPrefix.UP}";
            return new List<ConfigurableMorph>
            {
                NewConfigurableMorph($"{prefix} Breast Height {side}", false, 0.70f, 0.53f),
                NewConfigurableMorph($"{prefix} Breast Sag1 {side}", true, -0.75f, -0.38f),
                NewConfigurableMorph($"{prefix} Breast Sag2 {side}", true, -1.00f, -0.50f),
                NewConfigurableMorph($"{prefix} Breasts Natural {side}", true, -0.75f, -0.38f),
                NewConfigurableMorph($"{prefix} Breast Rotate Up {side}", false, 0.80f, 0.75f),
                NewConfigurableMorph($"{prefix} Breast Top Curve2 {side}", true, -1.20f, -0.90f),
                NewConfigurableMorph($"{prefix} Breasts Implants {side}", false, 0.36f, 0.18f),
                NewConfigurableMorph($"{prefix} Breast Diameter {side}", false, 2.00f, 0.50f),
                NewConfigurableMorph($"{prefix} Breast Diameter(Pose) {side}", false, 1.50f, 0.38f),
                NewConfigurableMorph($"{prefix} Breast Width {side}", false, 0.21f, 0.14f),
                NewConfigurableMorph($"{prefix} Breast Zero {side}", false, 0.24f, 0.12f),
                NewConfigurableMorph($"{prefix} Breast flat(Fixed) {side}", false, 0.65f, 0.15f),
                NewConfigurableMorph($"{prefix} BreastsShape2 {side}", false, 1.80f, 0.53f),
                NewConfigurableMorph($"{prefix} Breast Height Lower {side}", true, -1.50f, -0.60f),
                NewConfigurableMorph($"{prefix} Breast Move Up {side}", false, 1.60f, 0.60f),
                NewConfigurableMorph($"{prefix} Breasts Flatten {side}", false, 0.60f, 0.90f),
                NewConfigurableMorph($"{prefix} Breasts TogetherApart {side}", false, 0.90f, 0.90f),
                NewConfigurableMorph($"{prefix} Breast Top Curve1 {side}", true, -0.75f, -0.38f),
                NewConfigurableMorph($"{prefix} Breasts Under Curve {side}", true, -1.50f, -0.75f),
                NewConfigurableMorph($"{prefix} ChestUnderBreast {side}", false, 0.80f, 0.375f),
                NewConfigurableMorph($"{prefix} Breast Under Smoother1 {side}", false, 0.25f, 0.00f),
                NewConfigurableMorph($"{prefix} Breast Under Smoother2 {side}", false, 0.40f, 0.30f),
                NewConfigurableMorph($"{prefix} Breast Under Smoother3 {side}", false, 0.15f, 0.23f),
                NewConfigurableMorph($"{prefix} Breast Under Smoother4 {side}", false, 0.30f, 0.15f),
                NewConfigurableMorph($"{prefix} Breast Height Upper {side}", false, 0.50f, 0.75f),
                NewConfigurableMorph($"{prefix} Breasts Upward Slope {side}", false, 1.30f, 3.00f),
                NewConfigurableMorph($"{prefix} Breast upper down {side}", false, 0.15f, 0.08f),
                NewConfigurableMorph($"{prefix} Breasts Small Top Slope {side}", false, 0.06f, 0.03f),
                NewConfigurableMorph($"{prefix} Areolae Depth {side}", false, 0.30f, 0.15f),
            };
        }

        static List<ConfigurableMorph> UpForceCenterMorphs()
        {
            string prefix = $"{MorphPrefix.UP}/{MorphPrefix.UP}";
            return new List<ConfigurableMorph>
            {
                NewConfigurableMorph($"{prefix} Center Gap Smooth", false, 1.70f, 0.30f),
                NewConfigurableMorph($"{prefix} Center Gap Depth", true, -0.50f, -1.20f),
                NewConfigurableMorph($"{prefix} Center Gap Height", false, 1.20f, 0.75f),
                NewConfigurableMorph($"{prefix} Centre Gap Wide", false, 0.50f, 0.38f),
                NewConfigurableMorph($"{prefix} Center Gap UpDown", false, 2.00f, -0.45f),
                NewConfigurableMorph($"{prefix} Chest Smoother", false, 3.00f, 1.13f),
                NewConfigurableMorph($"{prefix} ChestSmoothCenter", false, 3.00f, 1.26f),
            };
        }

        static List<ConfigurableMorph> DepthInForceMorphs(string side)
        {
            string prefix = $"{MorphPrefix.BK}/{MorphPrefix.BK}";
            return new List<ConfigurableMorph>
            {
                NewConfigurableMorph($"{prefix} Breast Diameter {side}", false, 1.20f, 0.40f),
                NewConfigurableMorph($"{prefix} Breast flat(Fixed) {side}", false, 0.54f, 0.27f),
                NewConfigurableMorph($"{prefix} Breast Pointed {side}", true, -0.40f, -0.20f),
                NewConfigurableMorph($"{prefix} Breast Top Curve1 {side}", true, -0.66f, -0.33f),
                NewConfigurableMorph($"{prefix} Breast Top Curve2 {side}", false, 0.30f, 0.15f),
                NewConfigurableMorph($"{prefix} Breast Under Smoother1 {side}", false, 0.70f, 0.53f),
                NewConfigurableMorph($"{prefix} Breast Under Smoother2 {side}", false, 0.07f, 0.32f),
                NewConfigurableMorph($"{prefix} Breast Under Smoother3 {side}", false, 0.60f, 0.30f),
                NewConfigurableMorph($"{prefix} Breast Zero {side}", false, 0.22f, 0.12f),
                NewConfigurableMorph($"{prefix} Breasts Flatten {side}", false, 1.00f, 0.80f),
                NewConfigurableMorph($"{prefix} Breasts Hang Forward {side}", true, -0.40f, -0.15f),
                NewConfigurableMorph($"{prefix} Breasts Implants {side}", false, 0.12f, 0.00f),
                NewConfigurableMorph($"{prefix} Breasts TogetherApart {side}", false, 0.15f, 0.23f),
                NewConfigurableMorph($"{prefix} Depth Squash {side}", false, 1.60f, 0.75f),
            };
        }

        static List<ConfigurableMorph> DepthInForceCenterMorphs()
        {
            string prefix = $"{MorphPrefix.BK}/{MorphPrefix.BK}";
            return new List<ConfigurableMorph>
            {
                NewConfigurableMorph($"{prefix} Sternum Depth", true, 0.10f, -0.60f),
                NewConfigurableMorph($"{prefix} Sternum Width", true, -0.30f, -0.45f),
                NewConfigurableMorph($"{prefix} Center Gap Depth", true, 0.00f, -0.60f),
            };
        }

        static List<ConfigurableMorph> DepthOutForceMorphs(string side)
        {
            string prefix = $"{MorphPrefix.FW}/{MorphPrefix.FW}";
            return new List<ConfigurableMorph>
            {
                NewConfigurableMorph($"{prefix} Breast Diameter(Pose) {side}", false, 0.40f, 0.30f),
                NewConfigurableMorph($"{prefix} Breast Height2 {side}", true, -0.40f, -0.15f),
                NewConfigurableMorph($"{prefix} Breast Rotate Down {side}", true, -0.40f, -0.15f),
                NewConfigurableMorph($"{prefix} Breast Width {side}", true, -0.24f, -0.12f),
                NewConfigurableMorph($"{prefix} Breast Depth {side}", false, 1.60f, -0.12f),
                NewConfigurableMorph($"{prefix} Breast Small {side}", false, 0.64f, 0.27f),
                NewConfigurableMorph($"{prefix} Breast Round {side}", true, -1.80f, -0.90f),
                NewConfigurableMorph($"{prefix} Breasts Hang Forward {side}", true, -0.36f, -0.24f),
            };
        }

        static List<ConfigurableMorph> DepthOutForceCenterMorphs()
        {
            string prefix = $"{MorphPrefix.FW}/{MorphPrefix.FW}";
            return new List<ConfigurableMorph>
            {
                NewConfigurableMorph($"{prefix} Sternum Width", false, 0.41f, 0.21f),
                NewConfigurableMorph($"{prefix} ChestSeparateBreasts", false, 0.00f, 0.45f),
            };
        }

        static List<ConfigurableMorph> LeftForceLeftMorphs()
        {
            string prefix = $"{MorphPrefix.LT}/{MorphPrefix.LT}";
            return new List<ConfigurableMorph>
            {
                NewConfigurableMorph($"{prefix} Breasts Shift S2S Left L", false, 1.50f, 0.75f),
                NewConfigurableMorph($"{prefix} Breasts Hang Forward L", true, -1.00f, -1.20f),
                NewConfigurableMorph($"{prefix} Breast Width L", true, -0.60f, -0.30f),
                NewConfigurableMorph($"{prefix} Breast Rotate X Out L", true, -0.60f, 0.30f),
                NewConfigurableMorph($"{prefix} Breast Rotate X In L", true, -0.30f, -0.15f),
                NewConfigurableMorph($"{prefix} Breast Diameter L", true, -0.20f, -0.30f),
                NewConfigurableMorph($"{prefix} Breast Under Smoother2 L", false, 0.10f, 0.60f),
                NewConfigurableMorph($"{prefix} Breast Under Smoother3 L", false, 0.40f, 0.60f),
            };
        }

        static List<ConfigurableMorph> LeftForceRightMorphs()
        {
            string prefix = $"{MorphPrefix.LT}/{MorphPrefix.LT}";
            return new List<ConfigurableMorph>
            {
                NewConfigurableMorph($"{prefix} Breasts Shift S2S Left R", false, 1.50f, 0.45f),
                NewConfigurableMorph($"{prefix} Breast Rotate X Out R", true, -0.84f, -0.42f),
                NewConfigurableMorph($"{prefix} Breasts Flatten R", true, -0.40f, -0.30f),
                NewConfigurableMorph($"{prefix} Breast Width R", false, 0.90f, -0.75f),
                NewConfigurableMorph($"{prefix} Breast Rotate X In R", false, 0.20f, 0.00f),
                NewConfigurableMorph($"{prefix} Breasts Hang Forward R", false, 0.20f, 0.11f),
                NewConfigurableMorph($"{prefix} Areola S2S R", false, 0.60f, 2.70f),
                NewConfigurableMorph($"{prefix} Breast Diameter R", true, -0.20f, -0.30f),
                NewConfigurableMorph($"{prefix} Breast Depth Squash R", false, 0.20f, 3.00f),
                NewConfigurableMorph($"{prefix} Breast Move S2S In R", false, 0.00f, 0.45f),
                NewConfigurableMorph($"{prefix} Breasts Implants R", true, 0.10f, -0.15f),
                NewConfigurableMorph($"{prefix} Breast Under Smoother2 R", false, 0.10f, 0.60f),
                NewConfigurableMorph($"{prefix} Breast Under Smoother3 R", false, 0.40f, 0.60f),
            };
        }

        static List<ConfigurableMorph> RightForceLeftMorphs()
        {
            string prefix = $"{MorphPrefix.RT}/{MorphPrefix.RT}";
            return new List<ConfigurableMorph>
            {
                NewConfigurableMorph($"{prefix} Breasts Shift S2S Right L", false, 1.50f, 0.45f),
                NewConfigurableMorph($"{prefix} Breast Rotate X Out L", true, -0.84f, -0.42f),
                NewConfigurableMorph($"{prefix} Breasts Flatten L", true, -0.40f, -0.30f),
                NewConfigurableMorph($"{prefix} Breast Width L", false, 0.90f, -0.75f),
                NewConfigurableMorph($"{prefix} Breast Rotate X In L", false, 0.20f, 0.00f),
                NewConfigurableMorph($"{prefix} Breasts Hang Forward L", false, 0.20f, 0.11f),
                NewConfigurableMorph($"{prefix} Areola S2S L", false, 0.60f, 2.70f),
                NewConfigurableMorph($"{prefix} Breast Diameter L", true, -0.20f, -0.30f),
                NewConfigurableMorph($"{prefix} Breast Depth Squash L", false, 0.20f, 3.00f),
                NewConfigurableMorph($"{prefix} Breast Move S2S In L", false, 0.00f, 0.45f),
                NewConfigurableMorph($"{prefix} Breasts Implants L", true, 0.10f, -0.15f),
                NewConfigurableMorph($"{prefix} Breast Under Smoother2 L", false, 0.10f, 0.60f),
                NewConfigurableMorph($"{prefix} Breast Under Smoother3 L", false, 0.40f, 0.60f),
            };
        }

        static List<ConfigurableMorph> RightForceRightMorphs()
        {
            string prefix = $"{MorphPrefix.RT}/{MorphPrefix.RT}";
            return new List<ConfigurableMorph>
            {
                NewConfigurableMorph($"{prefix} Breasts Shift S2S Right R", false, 1.50f, 0.75f),
                NewConfigurableMorph($"{prefix} Breasts Hang Forward R", true, -1.00f, -1.20f),
                NewConfigurableMorph($"{prefix} Breast Width R", true, -0.40f, -0.20f),
                NewConfigurableMorph($"{prefix} Breast Rotate X Out R", true, -0.60f, 0.30f),
                NewConfigurableMorph($"{prefix} Breast Rotate X In R", true, -0.30f, -0.15f),
                NewConfigurableMorph($"{prefix} Breast Diameter R", true, -0.20f, -0.30f),
                NewConfigurableMorph($"{prefix} Breast Under Smoother2 R", false, 0.10f, 0.60f),
                NewConfigurableMorph($"{prefix} Breast Under Smoother3 R", false, 0.40f, 0.60f),
            };
        }

#endregion

        static bool _softPhysicsOn;
        static float _mass;
        static float _softness;

        static float _upLimit;
        static float _depthOutLimit;
        static float _depthInLimit;
        static float _sidewaysLimit;

        static float _upMultiplier;
        static float _depthOutMultiplier;
        static float _depthInMultiplier;
        static float _sidewaysMultiplier;

        static float SoftnessMultiplier(float x)
        {
            return Calc.InverseSmoothStep(x, 1.00f, 0.80f, 0.03f);
        }

        // https://www.desmos.com/calculator/ychmmr0gqu
        static float MassMultiplier(float x)
        {
            return Calc.Exponential3(x, 2.17f, 1.56f, 1.00f, 1.30f, 1.70f);
        }

        public static void SetMultipliers()
        {
            _softPhysicsOn = SoftPhysicsHandler.softPhysicsOn;
            _mass = MainPhysicsHandler.normalizedRealMass;
            _softness = SoftnessMultiplier(TittyMagic.softness);
            SetUpMultiplier();
            SetDepthOutMultiplier();
            SetDepthInMultiplier();
            SetSidewaysMultiplier();
        }

        static void SetUpMultiplier()
        {
            float softnessMultiplier = Mathf.Lerp(0.65f, 1.33f, _softness);
            _upLimit = softnessMultiplier * UpLimit();
            _upMultiplier = 0.0070f * softnessMultiplier * _upLimit * MassMultiplier(_mass);
        }

        static float UpLimit()
        {
            const float cutoff1 = 0.12f;
            const float cutoff2 = 0.43f;

            if(_mass < cutoff1)
            {
                float ratioToCutoff1 = _mass / cutoff1;
                return Mathf.Lerp(37.0f, 50.0f, ratioToCutoff1);
            }

            if(_mass < cutoff2)
            {
                float ratioFromCutoff1To2 = (_mass - cutoff1) / (cutoff2 - cutoff1);
                return Mathf.Lerp(50.0f, 55.4f, Calc.Exponential1(ratioFromCutoff1To2, 1.76f, 1.43f, 0.67f));
            }

            float ratioFromCutoff2 = (_mass - cutoff2) / cutoff2;
            return Mathf.Lerp(47.10f, 55.40f, 1 - ratioFromCutoff2);
        }

        static float DepthOutMassMultiplier()
        {
            return _softPhysicsOn
                ? Mathf.Lerp(4.00f, 2.80f, 1 - _mass)
                : Mathf.Lerp(5.70f, 1.20f, 1 - _mass);
        }

        static void SetDepthOutMultiplier()
        {
            float softnessMultiplier = Mathf.Lerp(0.60f, 1.20f, _softness);
            _depthOutLimit = softnessMultiplier * DepthOutLimit();
            _depthOutMultiplier = softnessMultiplier * (12.50f / _depthOutLimit) * DepthOutMassMultiplier();
        }

        static float DepthOutLimit()
        {
            const float cutoff1 = 0.12f;
            const float cutoff2 = 0.25f;

            if(_mass < cutoff1)
            {
                float ratioToCutoff1 = _mass / cutoff1;
                return Mathf.Lerp(16f, 31.0f, 1 - ratioToCutoff1);
            }

            if(_mass > cutoff2)
            {
                float ratioFromCutoff2 = (_mass - cutoff2) / cutoff2;
                return Mathf.Lerp(16f, 24f, Mathf.Pow(ratioFromCutoff2, 1.75f));
            }

            return 16f;
        }

        // https://www.desmos.com/calculator/ntxyujsadd
        static float DepthInMassMultiplier()
        {
            float value = _softPhysicsOn
                ? 9.50f / Mathf.Sqrt(3 * _mass + 3.40f) - 2.40f
                : 9.80f / Mathf.Sqrt(2.30f * _mass + 10.00f) - 1.63f;
            return 2.00f * value;
        }

        static void SetDepthInMultiplier()
        {
            float softnessMultiplier = Mathf.Lerp(
                Mathf.Lerp(0.55f, 1.10f, Calc.InverseSmoothStep(_mass, 1, 0.50f, 0.05f)),
                1.100f,
                _softness
            );
            _depthInLimit = softnessMultiplier * DepthInLimit();
            _depthInMultiplier = softnessMultiplier * 12.50f / _depthInLimit * DepthInMassMultiplier();
        }

        static float DepthInLimit()
        {
            const float cutoff1 = 0.12f;
            const float cutoff2 = 0.25f;

            if(_mass < cutoff1)
            {
                float ratioToCutoff1 = _mass / cutoff1;
                return Mathf.Lerp(30f, 39f, 1 - ratioToCutoff1);
            }

            if(_mass > cutoff2)
            {
                float ratioFromCutoff2 = (_mass - cutoff2) / cutoff2;
                return Mathf.Lerp(30f, 10.0f, Mathf.Pow(ratioFromCutoff2, 1.75f));
            }

            return 30f;
        }

        static void SetSidewaysMultiplier()
        {
            float softnessMultiplier = Mathf.Lerp(0.90f, 1.20f, _softness);
            _sidewaysLimit = softnessMultiplier * SidewaysLimit();
            _sidewaysMultiplier = softnessMultiplier * 0.0109f * _sidewaysLimit * MassMultiplier(_mass);
        }

        // https://www.desmos.com/calculator/fusmae0li0
        static float SidewaysLimit()
        {
            const float cutoff1 = 0.25f;
            if(_mass < cutoff1)
            {
                float ratioToCutoff1 = _mass / cutoff1;
                return Mathf.Lerp(33.4f, 40.0f, Calc.InverseSmoothStep(ratioToCutoff1, 1, -0.16f, 0));
            }

            float ratioFromCutoff1 = (_mass - cutoff1) / cutoff1;
            return Mathf.Lerp(27.8f, 40.0f, 1 - Calc.InverseSmoothStep(ratioFromCutoff1, 1, -0.33f, 0));
        }

        public static void SaveCalibrated()
        {
            baseJsf.SaveCalibrated();
            upJsf.SaveCalibrated();
            depthInJsf.SaveCalibrated();
            depthOutJsf.SaveCalibrated();
            sidewaysInJsf.SaveCalibrated();
            sidewaysOutJsf.SaveCalibrated();
        }

        static float _rollL;
        static float _rollR;
        static float _pitchL;
        static float _pitchR;
        static float _rollAngleMultiL;
        static float _rollAngleMultiR;
        static float _leanBackFixerMultiplierLeft;
        static float _leanBackFixerMultiplierRight;

        public static void Update()
        {
            _rollL = GravityEffectCalc.pectoralRollL;
            _rollR = -GravityEffectCalc.pectoralRollR;
            _pitchL = GravityEffectCalc.pectoralPitchL;
            _pitchR = GravityEffectCalc.pectoralPitchR;
            _rollAngleMultiL = Mathf.Lerp(1.25f, 1f, Mathf.Abs(_rollL));
            _rollAngleMultiR = Mathf.Lerp(1.25f, 1f, Mathf.Abs(_rollR));
            _leanBackFixerMultiplierLeft = CalculateLeanBackFixerMultiplier(_pitchL, _rollL);
            _leanBackFixerMultiplierRight = CalculateLeanBackFixerMultiplier(_pitchR, _rollR);

            AdjustUpMorphs();
            AdjustDepthOutMorphs();
            AdjustDepthInMorphs();
            AdjustSidewaysMorphs();
        }

        static float UpForceEffectCurve(float x)
        {
            return _softPhysicsOn
                ? Calc.Exponential1(x, 1.60f, 2.26f, 1.23f, 0.96f)
                : Calc.Exponential1(x, 1.60f, 2.14f, 1.18f, 0.96f);
        }

        static float UpEffect(float angle, float pitchMultiplier)
        {
            return pitchMultiplier *
                _upMultiplier *
                UserUpMultiplier() *
                UpForceEffectCurve(Mathf.Abs(angle) / _upLimit);
        }

        static void AdjustUpMorphs()
        {
            float pitchMultiplierLeft = 0;
            float pitchMultiplierRight = 0;

            if(_leftBreast.angleY >= 0)
            {
                // up force on left breast
                pitchMultiplierLeft = Mathf.Lerp(0.63f, 1f, _leanBackFixerMultiplierLeft * GravityEffectCalc.diffFromHorizontalL);
                UpdateMorphs(UP_L, UpEffect(_leftBreast.angleY, pitchMultiplierLeft));
            }
            else
            {
                // down force on left breast
                ResetMorphs(UP_L);
            }

            if(_rightBreast.angleY >= 0)
            {
                // up force on right breast
                pitchMultiplierRight = Mathf.Lerp(0.63f, 1f, _leanBackFixerMultiplierRight * GravityEffectCalc.diffFromHorizontalR);
                UpdateMorphs(UP_R, UpEffect(_rightBreast.angleY, pitchMultiplierRight));
            }
            else
            {
                // down force on right breast
                ResetMorphs(UP_R);
            }

            float angleYCenter = (_rightBreast.angleY + _leftBreast.angleY) / 2;
            if(angleYCenter >= 0)
            {
                // up force on average of left and right breast
                float pitchMultiplierCenter = (pitchMultiplierLeft + pitchMultiplierRight) / 2;
                UpdateMorphs(UP_C, UpEffect(angleYCenter, pitchMultiplierCenter));
            }
            else
            {
                ResetMorphs(UP_C);
            }
        }

        static float CalculateUpsideDownFixerMultiplier(float pitch, float roll)
        {
            if(pitch > -1 && pitch < 1)
            {
                return 1;
            }

            // return 1;
            float effect = GravityEffectCalc.UpDownAdjustByAngle(pitch) * GravityEffectCalc.RollMultiplier(roll) / 2;
            return 1 - effect;
        }

        static float CalculateLeanBackFixerMultiplier(float pitch, float roll)
        {
            if(pitch < -0.5f || pitch > 0)
            {
                return 1;
            }

            float diff = 4 * Mathf.Abs(-0.25f - pitch);
            float minTarget1 = Mathf.Lerp(0.25f, 1.00f, 1 - MainPhysicsHandler.normalizedMass);
            float minTarget2 = Mathf.Lerp(minTarget1, 1.00f, Mathf.Abs(roll * roll));
            return Mathf.Lerp(minTarget2, 1.00f, diff);
        }

        static float DepthInForceEffectCurve(float x)
        {
            return _softPhysicsOn
                ? Calc.Exponential1(x, 1.66f, 1.70f, 1.06f)
                : Calc.Exponential1(x, 1.60f, 1.70f, 1.07f, 0.96f);
        }

        static float DepthInEffect(float depthDiff, float leanBackFixerMultiplier)
        {
            return leanBackFixerMultiplier *
                _depthInMultiplier *
                UserDepthInMultiplier() *
                DepthInForceEffectCurve(Mathf.Abs(depthDiff) * _depthInLimit);
        }

        static void AdjustDepthInMorphs()
        {
            if(_leftBreast.depthDiff < 0)
            {
                // in force on left breast
                ResetMorphs(BACK_L);
            }
            else
            {
                // out force on left breast
                UpdateMorphs(BACK_L, DepthInEffect(_leftBreast.depthDiff, _leanBackFixerMultiplierLeft));
            }

            if(_rightBreast.depthDiff < 0)
            {
                // in force on right breast
                ResetMorphs(BACK_R);
            }
            else
            {
                // out force on right breast
                UpdateMorphs(BACK_R, DepthInEffect(_rightBreast.depthDiff, _leanBackFixerMultiplierRight));
            }

            float depthDiffCenter = (_leftBreast.depthDiff + _rightBreast.depthDiff) / 2;
            float leanBackFixerMultiplierCenter = (_leanBackFixerMultiplierLeft + _leanBackFixerMultiplierRight) / 2;
            if(depthDiffCenter < 0)
            {
                // in force on average of left and right breast
                ResetMorphs(BACK_C);
            }
            else
            {
                // out force on average of left and right breast
                UpdateMorphs(BACK_C, DepthInEffect(depthDiffCenter, leanBackFixerMultiplierCenter));
            }
        }

        static float DepthOutEffect(float depthDiff, float upsideDownFixerMultiplier)
        {
            return upsideDownFixerMultiplier *
                _depthOutMultiplier *
                UserDepthOutMultiplier() *
                Calc.Exponential1(Mathf.Abs(depthDiff) * _depthOutLimit, 1.42f, 3.20f, 1.15f);
        }

        static void AdjustDepthOutMorphs()
        {
            float upsideDownFixerMultiLeft = CalculateUpsideDownFixerMultiplier(_pitchL, _rollL);
            float upsideDownFixerMultiRight = CalculateUpsideDownFixerMultiplier(_pitchR, _rollR);

            if(_leftBreast.depthDiff < 0)
            {
                // in force on left breast
                UpdateMorphs(FORWARD_L, DepthOutEffect(_leftBreast.depthDiff, upsideDownFixerMultiLeft));
            }
            else
            {
                // out force on left breast
                ResetMorphs(FORWARD_L);
            }

            if(_rightBreast.depthDiff < 0)
            {
                // in force on right breast
                UpdateMorphs(FORWARD_R, DepthOutEffect(_rightBreast.depthDiff, upsideDownFixerMultiRight));
            }
            else
            {
                // out force on right breast
                ResetMorphs(FORWARD_R);
            }

            float depthDiffCenter = (_leftBreast.depthDiff + _rightBreast.depthDiff) / 2;
            float upsideDownFixerMultiCenter = (upsideDownFixerMultiLeft + upsideDownFixerMultiRight) / 2;
            if(depthDiffCenter < 0)
            {
                // in force on average of left and right breast
                UpdateMorphs(FORWARD_C, DepthOutEffect(depthDiffCenter, upsideDownFixerMultiCenter));
            }
            else
            {
                // out force on average of left and right breast
                ResetMorphs(FORWARD_C);
            }
        }

        static float SidewaysForceEffectCurve(float x)
        {
            return _softPhysicsOn
                ? Calc.Exponential1(x, 1.60f, 2.26f, 1.23f, 0.96f)
                : Calc.Exponential1(x, 1.60f, 2.14f, 1.18f, 0.96f);
        }

        static float SidewaysEffect(float angle, float userMultiplier, float rollAngleMultiplier)
        {
            return _sidewaysMultiplier *
                userMultiplier *
                SidewaysForceEffectCurve(rollAngleMultiplier * Mathf.Abs(angle) / _sidewaysLimit);
        }

        static void AdjustSidewaysMorphs()
        {
            if(_leftBreast.angleX >= 0)
            {
                // right force on left breast
                ResetMorphs(LEFT_L);
                UpdateMorphs(RIGHT_L, SidewaysEffect(_leftBreast.angleX, UserSidewaysInMultiplier(), _rollAngleMultiL));
            }
            else
            {
                // left force on left breast
                ResetMorphs(RIGHT_L);
                UpdateMorphs(LEFT_L, SidewaysEffect(_leftBreast.angleX, UserSidewaysOutMultiplier(), _rollAngleMultiL));
            }

            if(_rightBreast.angleX >= 0)
            {
                // right force on right breast
                ResetMorphs(LEFT_R);
                UpdateMorphs(RIGHT_R, SidewaysEffect(_rightBreast.angleX, UserSidewaysOutMultiplier(), _rollAngleMultiR));
            }
            else
            {
                // left force on right breast
                ResetMorphs(RIGHT_R);
                UpdateMorphs(LEFT_R, SidewaysEffect(_rightBreast.angleX, UserSidewaysInMultiplier(), _rollAngleMultiR));
            }
        }

        static void UpdateMorphs(string direction, float effect)
        {
            float halfEffect = effect / 2;
            foreach(var configurableMorph in morphSets[direction])
            {
                // softness fixed to 0.62f due to historical reasons
                configurableMorph.Update(halfEffect, _mass, 0.62f);
            }
        }

        public static void ResetAll()
        {
            if(morphSets == null)
            {
                return;
            }

            foreach(var kvp in morphSets)
            {
                foreach(var configurableMorph in kvp.Value)
                {
                    configurableMorph.Reset();
                }
            }
        }

        static void ResetMorphs(string direction)
        {
            foreach(var configurableMorph in morphSets[direction])
            {
                configurableMorph.Reset();
            }
        }

        public static void Destroy()
        {
            _leftBreast = null;
            _rightBreast = null;
            morphSets = null;
            baseJsf = null;
            upJsf = null;
            depthOutJsf = null;
            depthInJsf = null;
            sidewaysInJsf = null;
            sidewaysOutJsf = null;
            directionJssc = null;
        }
    }
}
