﻿using System.Collections.Generic;
using Naturalis.Common;
using Naturalis.Common.Handlers.Configs;
using Naturalis.TM.Components;
using UnityEngine;
using static Direction;

namespace Naturalis.TM.Handlers
{
    static class ForcePhysicsHandler
    {
        static Breast _leftBreast;
        static Breast _rightBreast;

        static JSONStorableFloat _baseJsf;
        static JSONStorableFloat _depthOutJsf;
        static JSONStorableFloat _depthInJsf;

        static float DepthInMultiplier()
        {
            return _baseJsf.val * _depthInJsf.val;
        }

        static float DepthOutMultiplier()
        {
            return _baseJsf.val * _depthOutJsf.val;
        }

        public static void Init()
        {
            _leftBreast = TittyMagic.script.leftBreast;
            _rightBreast = TittyMagic.script.rightBreast;
            _baseJsf = TittyMagic.script.NewStorableFloat("Force Physics Base Multiplier", 1.00f, 0.00f, 2.00f);
            _depthInJsf = TittyMagic.script.NewStorableFloat("Force Physics Depth In", 1.00f, 0.00f, 2.00f);
            _depthOutJsf = TittyMagic.script.NewStorableFloat("Force Physics Depth Out", 1.00f, 0.00f, 2.00f);
            LoadSettings();
        }

        static void LoadSettings()
        {
            MainPhysicsHandler.centerOfGravity.SetForceCalculators(CenterOfGravityCalculators());
            MainPhysicsHandler.inOutDamper.SetForceCalculators(InOutDamperCalculators());
        }

        static Dictionary<string, Dictionary<string, PhysicsCalculator>> CenterOfGravityCalculators()
        {
            var backConfig = new PhysicsCalculator
            {
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 0,
                    multiplierA = -0.195f,
                    multiplierB = -0.130f,
                    curveA = x => 1 - x,
                    curveB = Calc.ForcePhysicsSoftnessCurve,
                },
                secondaryConfig = new MultiplierConfig
                {
                    baseValue = 0,
                    multiplierA = -0.100f,
                    multiplierB = -0.070f,
                    curveA = x => 1 - x,
                    curveB = Calc.ForcePhysicsSoftnessCurve,
                },
            };
            var forwardConfig = new PhysicsCalculator
            {
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 0,
                    multiplierA = 0.195f,
                    multiplierB = 0.130f,
                    curveA = x => 1 - x,
                    curveB = Calc.ForcePhysicsSoftnessCurve,
                },
                secondaryConfig = new MultiplierConfig
                {
                    baseValue = 0,
                    multiplierA = 0.100f,
                    multiplierB = 0.070f,
                    curveA = x => 1 - x,
                    curveB = Calc.ForcePhysicsSoftnessCurve,
                },
            };
            var leftBreast = new Dictionary<string, PhysicsCalculator>
            {
                { BACK, backConfig },
                { FORWARD, forwardConfig },
            };
            var rightBreast = new Dictionary<string, PhysicsCalculator>
            {
                { BACK, backConfig },
                { FORWARD, forwardConfig },
            };
            return new Dictionary<string, Dictionary<string, PhysicsCalculator>>
            {
                { Side.LEFT, leftBreast },
                { Side.RIGHT, rightBreast },
            };
        }

        static Dictionary<string, Dictionary<string, PhysicsCalculator>> InOutDamperCalculators()
        {
            var depthInConfig = new PhysicsCalculator
            {
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 0,
                    multiplierA = 6f,
                    multiplierB = -8f,
                    curveB = Calc.ForcePhysicsSoftnessCurve,
                },
                secondaryConfig = new MultiplierConfig
                {
                    baseValue = 0,
                    multiplierA = 6f,
                    multiplierB = -8f,
                    curveB = Calc.ForcePhysicsSoftnessCurve,
                },
            };
            var depthOutConfig = new PhysicsCalculator
            {
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 0,
                    multiplierA = 6f,
                    multiplierB = -8f,
                    curveB = Calc.ForcePhysicsSoftnessCurve,
                },
                secondaryConfig = new MultiplierConfig
                {
                    baseValue = 0,
                    multiplierA = 6f,
                    multiplierB = -8f,
                    curveB = Calc.ForcePhysicsSoftnessCurve,
                },
            };
            var leftBreast = new Dictionary<string, PhysicsCalculator>
            {
                { BACK, depthInConfig },
                { FORWARD, depthOutConfig },
            };
            var rightBreast = new Dictionary<string, PhysicsCalculator>
            {
                { BACK, depthInConfig },
                { FORWARD, depthOutConfig },
            };
            return new Dictionary<string, Dictionary<string, PhysicsCalculator>>
            {
                { Side.LEFT, leftBreast },
                { Side.RIGHT, rightBreast },
            };
        }

        static bool _softPhysicsOn;
        static float _depthInMultiplier;
        static float _depthOutMultiplier;

        public static void SetMultipliers()
        {
            _softPhysicsOn = SoftPhysicsHandler.softPhysicsOn;
            _depthInMultiplier = 0.90f;
            _depthOutMultiplier = 0.90f;
        }

        static float DepthForceEffectCurve(float x)
        {
            return Calc.Exponential1(x, 1.42f, 3.20f, 1.15f);
        }

        static float _mass;
        static float _softness;

        public static void Update()
        {
            _mass = MainPhysicsHandler.normalizedMass;
            _softness = TittyMagic.softness;
            AdjustDepthInPhysics();
            AdjustDepthOutPhysics();
        }

        static float DepthOutEffect(float distance)
        {
            return _depthOutMultiplier * Calc.QuadraticRegression(DepthOutMultiplier()) * DepthForceEffectCurve(Mathf.Abs(distance) * 27f);
        }

        static void AdjustDepthOutPhysics()
        {
            if(_leftBreast.depthDiff <= 0)
            {
                // forward force on left breast
                UpdateLeftBreast(FORWARD, DepthOutEffect(_leftBreast.depthDiff));
            }
            else
            {
                ResetLeftBreast(FORWARD);
            }

            if(_rightBreast.depthDiff <= 0)
            {
                // forward force on right breast
                UpdateRightBreast(FORWARD, DepthOutEffect(_rightBreast.depthDiff));
            }
            else
            {
                ResetRightBreast(FORWARD);
            }
        }

        static float DepthInEffect(float distance)
        {
            return _depthInMultiplier * Calc.QuadraticRegression(DepthInMultiplier()) * DepthForceEffectCurve(Mathf.Abs(distance) * 20.50f);
        }

        static void AdjustDepthInPhysics()
        {
            if(_leftBreast.depthDiff > 0)
            {
                // back force on left breast
                UpdateLeftBreast(BACK, DepthInEffect(_leftBreast.depthDiff));
            }
            else
            {
                ResetLeftBreast(BACK);
            }

            if(_rightBreast.depthDiff > 0)
            {
                // back force on right breast
                UpdateRightBreast(BACK, DepthInEffect(_rightBreast.depthDiff));
            }
            else
            {
                ResetRightBreast(BACK);
            }
        }

        static void UpdateLeftBreast(string direction, float effect)
        {
            MainPhysicsHandler.centerOfGravity.left.UpdateForceValue(direction, effect, _mass, _softness, _softPhysicsOn);
            MainPhysicsHandler.inOutDamper.left.UpdateForceValue(direction, effect, _mass, _softness, _softPhysicsOn);
        }

        static void UpdateRightBreast(string direction, float effect)
        {
            MainPhysicsHandler.centerOfGravity.right.UpdateForceValue(direction, effect, _mass, _softness, _softPhysicsOn);
            MainPhysicsHandler.inOutDamper.right.UpdateForceValue(direction, effect, _mass, _softness, _softPhysicsOn);
        }

        public static void ResetAll()
        {
            foreach(string direction in new[]
            {
                LEFT,
                RIGHT,
                UP,
                DOWN,
                FORWARD,
                BACK,
            })
            {
                ResetLeftBreast(direction);
                ResetRightBreast(direction);
            }
        }

        static void ResetLeftBreast(string direction)
        {
            MainPhysicsHandler.centerOfGravity.left.ResetForceValue(direction);
            MainPhysicsHandler.inOutDamper.left.ResetForceValue(direction);
        }

        static void ResetRightBreast(string direction)
        {
            MainPhysicsHandler.centerOfGravity.right.ResetForceValue(direction);
            MainPhysicsHandler.inOutDamper.right.ResetForceValue(direction);
        }

        public static void Destroy()
        {
            _leftBreast = null;
            _rightBreast = null;
            _baseJsf = null;
            _depthInJsf = null;
            _depthOutJsf = null;
        }
    }
}
