using System;
using Naturalis.Common;
using Naturalis.Common.Handlers.Configs;
using UnityEngine;

namespace Naturalis.TM.Handlers
{
    sealed class FrictionHandler
    {
        public static FrictionHandler instance { get; private set; } = new FrictionHandler();

        JSONStorable _skinMaterialsStorable;
        public JSONStorableFloat frictionOffsetJsf { get; private set; }
        public JSONStorableFloat frictionValueJsf { get; private set; }
        public JSONStorableBool adaptiveFrictionJsb { get; private set; }
        public JSONStorableFloat drySkinFrictionJsf { get; private set; }
        public static float maxHardColliderFriction { get; private set; }

        JSONStorableFloat _glossJsf;
        JSONStorableFloat _specularBumpinessJsf;

        public static bool enabled { get; private set; }

        public void Init()
        {
            if(!Person.IsFemale())
            {
                return;
            }

            frictionOffsetJsf = TittyMagic.script.NewStorableFloat("Friction Offset", 0, -0.5f, 0.5f);
            frictionOffsetJsf.setCallbackFunction = _ => CalculateFriction();

            frictionValueJsf = new JSONStorableFloat("Friction Value", 1.00f, 0.00f, 1.00f)
            {
                setCallbackFunction = UpdateSoftColliders,
            };

            adaptiveFrictionJsb = TittyMagic.script.NewStorableBool("Use Adaptive Friction", false);
            adaptiveFrictionJsb.setCallbackFunction = val =>
            {
                TittyMagic.script.uiModManager.drySkinFrictionSlider.SetActiveStyle(val, true);
                CalculateFriction();
            };

            drySkinFrictionJsf = TittyMagic.script.NewStorableFloat("Dry Skin Friction", 1.000f, 0.000f, 1.000f);
            drySkinFrictionJsf.setCallbackFunction = _ => CalculateFriction();

            TittyMagic.script.NewJSONStorableAction(Constant.INTERNAL_CALC_FRICTION, CalculateFriction);

            _skinMaterialsStorable = TittyMagic.script.containingAtom.GetStorableByID("skin");
            if(_skinMaterialsStorable)
            {
                SetGlossAndSpecularStorables();
                enabled = true;
            }
            else
            {
                enabled = false;
            }

            LoadSettings();
        }

        public void Refresh()
        {
            _skinMaterialsStorable = TittyMagic.script.containingAtom.GetStorableByID("skin");
            if(Person.IsFemale() && _skinMaterialsStorable)
            {
                enabled = true;
                SetGlossAndSpecularStorables();
                LoadSettings();
                CalculateFriction();
            }
            else
            {
                enabled = false;
            }
        }

        static PhysicsCalculator FatColliderRadiusCalculator()
        {
            return new PhysicsCalculator
            {
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 0,
                    multiplierA = 0.0090f,
                    multiplierB = 0,
                    curveA = x => 1.75f * Calc.InverseSmoothStep(x, 1, 0.4f, 0.7f),
                },
            };
        }

        static void LoadSettings()
        {
            if(!enabled)
            {
                return;
            }

            SoftPhysicsHandler.fatColliderRadius.SetFrictionCalculators(
                FatColliderRadiusCalculator(),
                FatColliderRadiusCalculator()
            );
        }

        void UpdateSoftColliders(float friction)
        {
            if(!SoftPhysicsHandler.softPhysicsOn)
            {
                return;
            }

            if(adaptiveFrictionJsb.val)
            {
                SoftPhysicsHandler.fatColliderRadius.UpdateInverseFrictionValue(
                    1 - Mathf.InverseLerp(0, drySkinFrictionJsf.val, friction),
                    SoftPhysicsHandler.fatColliderRadius.left.usesRealMass
                        ? MainPhysicsHandler.normalizedRealMass
                        : MainPhysicsHandler.normalizedMass,
                    TittyMagic.softness
                );
            }
            else
            {
                SoftPhysicsHandler.fatColliderRadius.ResetInverseFrictionValue();
            }

            /* Update collider friction */
            SoftPhysicsHandler.SyncFriction(Side.LEFT, friction);
            SoftPhysicsHandler.SyncFriction(Side.RIGHT, friction);
        }

        void SetGlossAndSpecularStorables()
        {
            _glossJsf = _skinMaterialsStorable.GetFloatJSONParam("Gloss");
            _glossJsf.setJSONCallbackFunction = _ =>
            {
                Settings.bootyMagic.CallActionNullSafe(Constant.INTERNAL_CALC_FRICTION);
                CalculateFriction();
            };

            _specularBumpinessJsf = _skinMaterialsStorable.GetFloatJSONParam("Specular Bumpiness");
            _specularBumpinessJsf.setJSONCallbackFunction = _ =>
            {
                Settings.bootyMagic.CallActionNullSafe(Constant.INTERNAL_CALC_FRICTION);
                CalculateFriction();
            };
        }

        /* Maximum friction that a collider can have, drops off dynamically with distance from collider's normal position */
        public void CalculateFriction()
        {
            try
            {
                if(!enabled || !TittyMagic.script.IsEnabledNullSafe() || _glossJsf == null || _specularBumpinessJsf == null)
                {
                    return;
                }

                if(!adaptiveFrictionJsb.val)
                {
                    maxHardColliderFriction = Mathf.Clamp(0.60f + frictionOffsetJsf.val, 0.00f, 1.00f);
                    frictionValueJsf.valNoCallback = Mathf.Clamp(0.80f + frictionOffsetJsf.val, 0.00f, 1.00f);
                }
                else
                {
                    float normalizedLinearGloss = Mathf.InverseLerp(3.000f, 7.500f, _glossJsf.val);
                    float normalizedSpecBump = Mathf.InverseLerp(0.200f, 3.000f, _specularBumpinessJsf.val);

                    /* Gloss based multiplier for how much effect bumpiness has on friction */
                    float bumpinessEffectMultiplier = Calc.InverseSmoothStep(normalizedLinearGloss, 1, 0.50f, 0.90f);

                    /* Inverse of gloss determines the minimum value for friction along a curve. */
                    float minFriction = drySkinFrictionJsf.val * (1 - Calc.InverseSmoothStep(normalizedLinearGloss, 1, 0.50f, 0.90f));

                    /* Hard colliders */
                    {
                        float specBumpComponent = 0.350f * bumpinessEffectMultiplier * normalizedSpecBump;
                        float friction = 0.80f * Mathf.Lerp(minFriction, 1.000f, specBumpComponent);
                        maxHardColliderFriction = Mathf.Clamp(friction + frictionOffsetJsf.val, 0.00f, 1.00f);
                    }

                    /* Soft colliders */
                    {
                        float specBumpComponent = 0.350f * bumpinessEffectMultiplier * normalizedSpecBump;
                        float friction = Mathf.Lerp(minFriction, 1.000f, specBumpComponent);
                        frictionValueJsf.valNoCallback = Mathf.Clamp(friction + frictionOffsetJsf.val, 0.00f, 1.00f);
                    }
                }

                UpdateSoftColliders(frictionValueJsf.val);
            }
            catch(Exception e)
            {
                TittyMagic.logger.LogError($"{nameof(CalculateFriction)} error {e}");
            }
        }

        public static void Destroy()
        {
            if(instance._glossJsf != null)
            {
                instance._glossJsf.setJSONCallbackFunction = null;
            }

            if(instance._specularBumpinessJsf != null)
            {
                instance._specularBumpinessJsf.setJSONCallbackFunction = null;
            }

            /* ensure GC */
            instance = null;
        }
    }
}
