using System;
using Naturalis.Common;
using UnityEngine;

namespace Naturalis.TM.Handlers
{
    static class GravityEffectCalc
    {
        public static float pectoralRollL { get; private set; }
        public static float pectoralPitchL { get; private set; }
        public static float pectoralRollR { get; private set; }
        public static float pectoralPitchR { get; private set; }
        public static float chestRoll { get; private set; }
        public static float chestPitch { get; private set; }
        public static float chestPitch2X { get; private set; }
        public static float diffFromHorizontalL { get; private set; }
        public static float diffFromHorizontalR { get; private set; }

        public static void CalculateAngles()
        {
            var lPectoralRotation = Person.GetPectoralRotation(Side.LEFT);
            var rPectoralRotation = Person.GetPectoralRotation(Side.RIGHT);
            var chestRotation = Person.GetChestRotation();

            pectoralRollL = Roll(lPectoralRotation);
            pectoralPitchL = Pitch(lPectoralRotation);
            pectoralRollR = Roll(rPectoralRotation);
            pectoralPitchR = Pitch(rPectoralRotation);
            chestRoll = Roll(chestRotation);
            chestPitch = Pitch(chestRotation);
            chestPitch2X = 2 * chestPitch;
            diffFromHorizontalL = DiffFromHorizontal(pectoralPitchL, pectoralRollL);
            diffFromHorizontalR = DiffFromHorizontal(pectoralPitchR, pectoralRollR);
        }

        static float Roll(Quaternion q)
        {
            return 2 * InverseLerpToPi(Mathf.Asin(2 * q.x * q.y + 2 * q.z * q.w));
        }

        static float Pitch(Quaternion q)
        {
            return InverseLerpToPi(Mathf.Atan2(2 * q.x * q.w - 2 * q.y * q.z, 1 - 2 * q.x * q.x - 2 * q.z * q.z));
        }

        static float InverseLerpToPi(float val)
        {
            return val > 0
                ? Mathf.InverseLerp(0, Mathf.PI, val)
                : -Mathf.InverseLerp(0, Mathf.PI, -val);
        }

        static float DiffFromHorizontal(float pitch, float roll)
        {
            return Mathf.Abs(Mathf.Abs(2 * pitch) - 1) * RollMultiplier(roll);
        }

        // ReSharper disable once UnusedMember.Global
        public static float UpEffect(
            float pitch,
            float roll,
            float multiplier,
            float additionalRollEffect
        )
        {
            float effect = Mathf.Abs(pitch) * RollMultiplier(roll) / 2;
            return (effect + additionalRollEffect) * multiplier / 2;
        }

        // ReSharper disable once UnusedMember.Global
        public static float DownEffect(float pitch, float roll, float multiplier)
        {
            return (2 - Mathf.Abs(pitch) / 2) * RollMultiplier(roll) * multiplier / 2;
        }

        public static float DepthAdjustByAngle(float pitch)
        {
            // leaning forward
            if(pitch >= 0)
            {
                // upright
                if(pitch < 1)
                {
                    return pitch;
                }

                // upside down
                return 2 - pitch;
            }

            // leaning back
            // upright
            if(pitch >= -1)
            {
                return -pitch;
            }

            // upside down
            return 2 + pitch;
        }

        public static float UpDownAdjustByAngle(float pitch)
        {
            // leaning forward
            if(pitch >= 0)
            {
                // upright
                if(pitch < 1)
                {
                    return 1 - pitch;
                }

                // upside down
                return pitch - 1;
            }

            // leaning back
            // upright
            if(pitch >= -1)
            {
                return 1 + pitch;
            }

            // upside down
            return -pitch - 1;
        }

        public static float RollMultiplier(float roll)
        {
            return 1 - Mathf.Abs(roll);
        }

        public static void OverrideAngles(string direction)
        {
            switch(direction)
            {
                case Direction.UP:
                {
                    chestPitch = 2;
                    pectoralPitchL = 2;
                    pectoralPitchR = 2;
                    chestRoll = 0;
                    pectoralRollL = 0;
                    pectoralRollR = 0;
                    break;
                }
                case Direction.DOWN:
                {
                    chestPitch = 0;
                    pectoralPitchL = 0;
                    pectoralPitchR = 0;
                    chestRoll = 0;
                    pectoralRollL = 0;
                    pectoralRollR = 0;
                    break;
                }
                case Direction.FORWARD:
                {
                    chestPitch = 1.0f;
                    pectoralPitchL = 1.0f;
                    pectoralPitchR = 1.0f;
                    chestRoll = 0;
                    pectoralRollL = -0.2f;
                    pectoralRollR = 0.2f;
                    break;
                }
                case Direction.BACK:
                {
                    chestPitch = -1.0f;
                    pectoralPitchL = -1.0f;
                    pectoralPitchR = -1.0f;
                    chestRoll = 0;
                    pectoralRollL = 0.25f;
                    pectoralRollR = -0.25f;
                    break;
                }
                case Direction.LEFT:
                {
                    chestPitch = 2;
                    pectoralPitchL = 2;
                    pectoralPitchR = 2;
                    chestRoll = 1;
                    pectoralRollL = 0.65f;
                    pectoralRollR = 0.90f;
                    break;
                }
                case Direction.RIGHT:
                {
                    chestPitch = 2;
                    pectoralPitchL = 2;
                    pectoralPitchR = 2;
                    chestRoll = -1;
                    pectoralRollL = -0.90f;
                    pectoralRollR = -0.65f;
                    break;
                }
                default: throw new ArgumentException($"Invalid direction {direction}");
            }
        }
    }
}
