using System.Collections.Generic;
using Naturalis.Common;
using Naturalis.Common.Components;
using Naturalis.Common.Models;
using static Direction;

namespace Naturalis.TM.Handlers
{
    static class GravityOffsetMorphHandler
    {
        static Dictionary<string, List<ConfigurableMorph>> _morphSets;

        public static StorableFloat offsetMorphingJsf { get; private set; }

        public static void Init()
        {
            offsetMorphingJsf = TittyMagic.script.NewStorableFloat("Gravity Offset Morphing", 1.00f, 0.00f, 2.00f);
            offsetMorphingJsf.setJSONCallbackFunction = jsf =>
            {
                var storable = (StorableFloat) jsf;
                TittyMagic.script.SetCalibrationPending(storable);
            };
            LoadSettings();
        }

        static void LoadSettings()
        {
            _morphSets = new Dictionary<string, List<ConfigurableMorph>>
            {
                {
                    DOWN, new List<ConfigurableMorph>
                    {
                        new ConfigurableMorph(
                            Person.GetBreastMorph("DN/DN Breast Rotate Up L"),
                            false,
                            new JSONStorableFloat("softnessMultiplier", 0.76f, 0, 3),
                            new JSONStorableFloat("massMultiplier", 0.855f, 0, 3)
                        ),
                        new ConfigurableMorph(
                            Person.GetBreastMorph("DN/DN Breast Rotate Up R"),
                            false,
                            new JSONStorableFloat("softnessMultiplier", 0.76f, 0, 3),
                            new JSONStorableFloat("massMultiplier", 0.855f, 0, 3)
                        ),
                        new ConfigurableMorph(
                            Person.GetBreastMorph("DN/DN Breasts Natural Reverse L"),
                            false,
                            new JSONStorableFloat("softnessMultiplier", 0.36f, 0, 3),
                            new JSONStorableFloat("massMultiplier", 0.405f, 0, 3)
                        ),
                        new ConfigurableMorph(
                            Person.GetBreastMorph("DN/DN Breasts Natural Reverse R"),
                            false,
                            new JSONStorableFloat("softnessMultiplier", 0.37f, 0, 3),
                            new JSONStorableFloat("massMultiplier", 0.405f, 0, 3)
                        ),
                    }
                },
            };
        }

        static float _upDownExtraMultiplier;

        public static void SetMultipliers()
        {
            _upDownExtraMultiplier = 0.50f *
                Calc.Exponential1(
                    1 - MainPhysicsHandler.normalizedRealMass,
                    0.26f,
                    2.60f,
                    5.38f,
                    0.84f,
                    0.33f
                );
        }

        public static void SaveCalibrated()
        {
            offsetMorphingJsf.SaveCalibrated();
        }

        public static void Update()
        {
            float roll = Calc.SmoothStep(GravityEffectCalc.chestRoll);
            float pitch = 2 * Calc.SmoothStep(GravityEffectCalc.chestPitch);
            AdjustUpDownMorphs(roll, pitch);
        }

        static void AdjustUpDownMorphs(float roll, float pitch)
        {
            float effect =
                offsetMorphingJsf.val *
                _upDownExtraMultiplier *
                MainPhysicsHandler.upDownAngleTarget.left.gravityCalculators[DOWN].multiplierJsf.val *
                GravityEffectCalc.UpDownAdjustByAngle(pitch) *
                GravityEffectCalc.RollMultiplier(roll) /
                2;

            if(pitch > 0)
            {
                // leaning forward
                if(pitch < 1)
                {
                    // upright
                    UpdateMorphs(DOWN, effect);
                }
                else
                {
                    // upside down
                    ResetMorphs(DOWN);
                }
            }
            else
            {
                // leaning back
                if(pitch > -1)
                {
                    // upright
                    UpdateMorphs(DOWN, effect);
                }
                else
                {
                    // upside down
                    ResetMorphs(DOWN);
                }
            }
        }

        static void UpdateMorphs(string direction, float effect)
        {
            foreach(var configurableMorph in _morphSets[direction])
            {
                configurableMorph.Update(effect, TittyMagic.softness, MainPhysicsHandler.normalizedRealMass);
            }
        }

        public static void ResetAll()
        {
            if(_morphSets == null)
            {
                return;
            }

            foreach(var kvp in _morphSets)
            {
                foreach(var configurableMorph in kvp.Value)
                {
                    configurableMorph.Reset();
                }
            }
        }

        static void ResetMorphs(string direction)
        {
            foreach(var configurableMorph in _morphSets[direction])
            {
                configurableMorph.Reset();
            }
        }

        public static void Destroy()
        {
            _morphSets = null;
            offsetMorphingJsf = null;
        }
    }
}
