﻿using System.Collections.Generic;
using Naturalis.Common;
using Naturalis.Common.Components;
using Naturalis.Common.Handlers.Configs;
using Naturalis.TM.UI;
using UnityEngine;
using static Direction;

namespace Naturalis.TM.Handlers
{
    static class GravityPhysicsHandler
    {
        public static StorableFloat twistAngleTargetJsf { get; private set; }
        public static StorableFloat gravityCounterforceJsf { get; private set; }
        static PhysicsCalculator _gravityCounterForceCalculator;
        public static JSONStorableBool autoAdjustGravityCounterforceJsb { get; private set; }

        static bool _useGravityCounterforce;
        static Vector3 _gravityCounterForce;

        public static void Init()
        {
            twistAngleTargetJsf = TittyMagic.script.NewStorableFloat(TWIST_ANGLE_TARGET.NAME, 0.00f, -30.00f, 30.00f);
            gravityCounterforceJsf = TittyMagic.script.NewStorableFloat("Gravity Counterforce (%)", 0f, 0f, 100f);
            gravityCounterforceJsf.setCallbackFunction = OnGravityCounterforceAdjusted;
            autoAdjustGravityCounterforceJsb = TittyMagic.script.NewStorableBool("Auto-Adjust Gravity Counterforce", true);
            autoAdjustGravityCounterforceJsb.setCallbackFunction = OnAutoAdjustGravityCounterforceToggled;
            LoadSettings();
        }

        static void OnGravityCounterforceAdjusted(float value)
        {
            // will still change value if called via trigger, but will not trigger calculation
            if(!autoAdjustGravityCounterforceJsb.val)
            {
                CalculateGravityCounterForce(value);
                TittyMagic.script.SetCalibrationPending(gravityCounterforceJsf);
            }
        }

        static void CalculateGravityCounterForce(float value)
        {
            int counterForcePercent = Mathf.RoundToInt(value);

            gravityCounterforceJsf.valNoCallback = counterForcePercent;
            if(counterForcePercent > 0)
            {
                _useGravityCounterforce = true;
                _gravityCounterForce = -Physics.gravity * counterForcePercent / 100;
            }
            else
            {
                _useGravityCounterforce = false;
                _gravityCounterForce = Vector3.zero;
            }
        }

        static void OnAutoAdjustGravityCounterforceToggled(bool value)
        {
            if(TittyMagic.restoringFromJson)
            {
                return;
            }

            if(value)
            {
                AutoCalculateGravityCounterforcePct();
            }

            var parameterWindow = TittyMagic.tabs.GetActiveNestedWindow() as ParameterWindow;
            if(parameterWindow != null && parameterWindow.GetId() == BREAST_JOINT_MASS.NAME)
            {
                parameterWindow.Refresh();
            }
        }

        public static void AutoCalculateGravityCounterforcePct(bool setPending = true)
        {
            float calculatedValue = _gravityCounterForceCalculator.CalculateDynamic(
                1,
                MainPhysicsHandler.normalizedMass,
                TittyMagic.softness,
                SoftPhysicsHandler.softPhysicsOn
            );
            float rounded = Mathf.Round(calculatedValue);
            gravityCounterforceJsf.defaultVal = rounded;
            if(autoAdjustGravityCounterforceJsb.val)
            {
                CalculateGravityCounterForce(calculatedValue);
                if(setPending)
                {
                    TittyMagic.script.SetCalibrationPending(gravityCounterforceJsf);
                }
            }
        }

        static void LoadSettings()
        {
            _gravityCounterForceCalculator = new PhysicsCalculator
            {
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 10.00f,
                    multiplierA = -8.00f,
                    multiplierB = 8.00f,
                    curveA = x => Calc.InverseSmoothStep(x, 1, 0.28f, 0.46f),
                    curveB = x => Calc.InverseSmoothStep(1 - x, 1, 0.39f, 0.80f),
                },
                secondaryConfig = new MultiplierConfig
                {
                    baseValue = 25.00f,
                    multiplierA = 6.00f,
                    multiplierB = 8.00f,
                    curveA = x => Calc.InverseSmoothStep(x, 1, 0.28f, 0.46f),
                    curveB = x => Calc.InverseSmoothStep(1 - x, 1, 0.39f, 0.80f),
                },
            };

            var spring = MainPhysicsHandler.spring;
            var damper = MainPhysicsHandler.damper;
            var inOutSpring = MainPhysicsHandler.inOutSpring;
            var upDownAngleTarget = MainPhysicsHandler.upDownAngleTarget;
            var rightLeftAngleTarget = MainPhysicsHandler.rightLeftAngleTarget;
            var twistAngleTarget = MainPhysicsHandler.twistAngleTarget;

            spring.SetGravityCalculators(SpringCalculators());
            damper.SetGravityCalculators(DamperCalculators());
            inOutSpring.SetGravityCalculators(InOutSpringCalculators());
            upDownAngleTarget.SetGravityCalculators(UpDownAngleTargetCalculators());
            rightLeftAngleTarget.SetGravityCalculators(RightLeftAngleTargetCalculators());
            twistAngleTarget.SetGravityCalculators(TwistAngleTargetCalculators());

            spring.SetGravityMultiplierCallbacks();
            damper.SetGravityMultiplierCallbacks();
            inOutSpring.SetGravityMultiplierCallbacks();
            upDownAngleTarget.SetGravityMultiplierCallbacks();
            rightLeftAngleTarget.SetGravityMultiplierCallbacks();

            spring.gravityInfoText =
                "Spring decreases when leaning in any direction. This makes breasts move" +
                " and morph more easily due to gravity.";
            damper.gravityInfoText =
                "Damper increases when leaning back. When leaning to the side," +
                " damper increases for the breast that rests against the chest.";
            inOutSpring.gravityInfoText =
                "In/Out Spring decreases when leaning forward or back in order to make depth" +
                " morphing take effect more easily due to gravity.";
            upDownAngleTarget.gravityInfoText =
                "Up/Down Angle Target is negative when upright (pulling breasts down)," +
                " and positive when upside down (pulling breasts up).";
            rightLeftAngleTarget.gravityInfoText =
                "Right/Left Angle Target is positive when leaning to the right and negative" +
                " when leaning to the left (the other way around for the right breast).";
        }

        static PhysicsCalculator SpringCalculator(string side, string direction)
        {
            return new PhysicsCalculator
            {
                multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(SPRING.NAME, side), direction, side == Side.LEFT),
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 0,
                    multiplierA = -2.5f,
                    multiplierB = -7.5f,
                    curveA = x => 1 - x,
                },
                secondaryConfig = new MultiplierConfig
                {
                    baseValue = 0,
                    multiplierA = -2.5f,
                    multiplierB = -7.5f,
                    curveA = x => 1 - x,
                },
            };
        }

        static PhysicsCalculator BackSpringCalculator(string side)
        {
            return new PhysicsCalculator
            {
                multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(SPRING.NAME, side), BACK, side == Side.LEFT),
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 0,
                    multiplierA = 22.5f,
                    multiplierB = 7.5f,
                    curveA = x => 1 - x,
                },
                secondaryConfig = new MultiplierConfig
                {
                    baseValue = 0,
                    multiplierA = 22.5f,
                    multiplierB = 7.5f,
                    curveA = x => 1 - x,
                },
            };
        }

        static Dictionary<string, Dictionary<string, PhysicsCalculator>> SpringCalculators()
        {
            return new Dictionary<string, Dictionary<string, PhysicsCalculator>>
            {
                {
                    Side.LEFT, new Dictionary<string, PhysicsCalculator>
                    {
                        { UP, SpringCalculator(Side.LEFT, UP) },
                        { BACK, BackSpringCalculator(Side.LEFT) },
                        { FORWARD, SpringCalculator(Side.LEFT, FORWARD) },
                        { SIDEWAYS_IN, SpringCalculator(Side.LEFT, SIDEWAYS_IN) },
                        { SIDEWAYS_OUT, SpringCalculator(Side.LEFT, SIDEWAYS_OUT) },
                    }
                },
                {
                    Side.RIGHT, new Dictionary<string, PhysicsCalculator>
                    {
                        { UP, SpringCalculator(Side.RIGHT, UP) },
                        { BACK, BackSpringCalculator(Side.RIGHT) },
                        { FORWARD, SpringCalculator(Side.RIGHT, FORWARD) },
                        { SIDEWAYS_IN, SpringCalculator(Side.RIGHT, SIDEWAYS_IN) },
                        { SIDEWAYS_OUT, SpringCalculator(Side.RIGHT, SIDEWAYS_OUT) },
                    }
                },
            };
        }

        static PhysicsCalculator BackDamperCalculator(string side, string direction)
        {
            return new PhysicsCalculator
            {
                multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(DAMPER.NAME, side), direction, side == Side.LEFT),
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 0,
                    multiplierA = 0.50f,
                    multiplierB = 1.20f,
                    curveA = x => 1 - x,
                },
                secondaryConfig = new MultiplierConfig
                {
                    baseValue = 0,
                    multiplierA = 0.50f,
                    multiplierB = 1.20f,
                    curveA = x => 1 - x,
                },
            };
        }

        static PhysicsCalculator ForwardDamperCalculator(string side, string direction)
        {
            return new PhysicsCalculator
            {
                multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(DAMPER.NAME, side), direction, side == Side.LEFT),
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 0,
                    multiplierA = 0.50f,
                    multiplierB = 1.20f,
                    curveA = x => 1 - x,
                },
                secondaryConfig = new MultiplierConfig
                {
                    baseValue = 0,
                    multiplierA = 0.50f,
                    multiplierB = 1.20f,
                    curveA = x => 1 - x,
                },
            };
        }

        static PhysicsCalculator SidewaysInDamperCalculator(string side, string direction)
        {
            return new PhysicsCalculator
            {
                multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(DAMPER.NAME, side), direction, side == Side.LEFT),
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 0,
                    multiplierA = 0.23f,
                    multiplierB = 0.60f,
                    curveA = x => 1 - x,
                },
                secondaryConfig = new MultiplierConfig
                {
                    baseValue = 0,
                    multiplierA = 0.23f,
                    multiplierB = 0.60f,
                    curveA = x => 1 - x,
                },
            };
        }

        static Dictionary<string, Dictionary<string, PhysicsCalculator>> DamperCalculators()
        {
            return new Dictionary<string, Dictionary<string, PhysicsCalculator>>
            {
                {
                    Side.LEFT, new Dictionary<string, PhysicsCalculator>
                    {
                        { BACK, BackDamperCalculator(Side.LEFT, BACK) },
                        { FORWARD, ForwardDamperCalculator(Side.LEFT, FORWARD) },
                        { SIDEWAYS_IN, SidewaysInDamperCalculator(Side.LEFT, SIDEWAYS_IN) },
                    }
                },
                {
                    Side.RIGHT, new Dictionary<string, PhysicsCalculator>
                    {
                        { BACK, BackDamperCalculator(Side.RIGHT, BACK) },
                        { FORWARD, ForwardDamperCalculator(Side.RIGHT, FORWARD) },
                        { SIDEWAYS_IN, SidewaysInDamperCalculator(Side.RIGHT, SIDEWAYS_IN) },
                    }
                },
            };
        }

        static PhysicsCalculator InOutSpringCalculator(string side, string direction)
        {
            return new PhysicsCalculator
            {
                multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(IN_OUT_SPRING.NAME, side), direction, side == Side.LEFT),
                primaryConfig = new MultiplierConfig
                {
                    baseValue = -100f,
                    multiplierA = -100f,
                    multiplierB = -125f,
                    curveA = x => Calc.Exponential1(1 - x, 0.3f, 10f, 0.84f, 1.01f),
                    curveB = Calc.InOutSpringSoftnessCurve,
                },
                secondaryConfig = new MultiplierConfig
                {
                    baseValue = -100f,
                    multiplierA = -100f,
                    multiplierB = -125f,
                    curveA = x => Calc.Exponential1(1 - x, 0.3f, 10f, 0.84f, 1.01f),
                    curveB = Calc.InOutSpringSoftnessCurve,
                },
            };
        }

        static Dictionary<string, Dictionary<string, PhysicsCalculator>> InOutSpringCalculators()
        {
            return new Dictionary<string, Dictionary<string, PhysicsCalculator>>
            {
                {
                    Side.LEFT, new Dictionary<string, PhysicsCalculator>
                    {
                        { BACK, InOutSpringCalculator(Side.LEFT, BACK) },
                        { FORWARD, InOutSpringCalculator(Side.LEFT, FORWARD) },
                    }
                },
                {
                    Side.RIGHT, new Dictionary<string, PhysicsCalculator>
                    {
                        { BACK, InOutSpringCalculator(Side.RIGHT, BACK) },
                        { FORWARD, InOutSpringCalculator(Side.RIGHT, FORWARD) },
                    }
                },
            };
        }

        static PhysicsCalculator UpAngleTargetCalculator(string side)
        {
            return new PhysicsCalculator
            {
                multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(UP_DOWN_ANGLE_TARGET.NAME, side), UP, side == Side.LEFT),
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 1.60f,
                    multiplierA = 7.60f,
                    multiplierB = 8.50f,
                    curveA = Calc.AngleTargetMassCurve,
                    curveB = Calc.AngleTargetSoftnessCurve,
                },
                secondaryConfig = new MultiplierConfig
                {
                    baseValue = 1.60f,
                    multiplierA = 7.60f,
                    multiplierB = 11.00f,
                    curveA = Calc.AngleTargetMassCurve,
                    curveB = Calc.AngleTargetSoftnessCurve,
                },
            };
        }

        static PhysicsCalculator DownAngleTargetCalculator(string side)
        {
            return new PhysicsCalculator
            {
                multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(UP_DOWN_ANGLE_TARGET.NAME, side), DOWN, side == Side.LEFT),
                primaryConfig = new MultiplierConfig
                {
                    baseValue = -1.60f,
                    multiplierA = -13.50f,
                    multiplierB = -8.50f,
                    curveA = Calc.AngleTargetMassCurve,
                    curveB = Calc.AngleTargetSoftnessCurve,
                },
                secondaryConfig = new MultiplierConfig
                {
                    baseValue = -1.60f,
                    multiplierA = -13.50f,
                    multiplierB = -11.00f,
                    curveA = Calc.AngleTargetMassCurve,
                    curveB = Calc.AngleTargetSoftnessCurve,
                },
            };
        }

        static Dictionary<string, Dictionary<string, PhysicsCalculator>> UpDownAngleTargetCalculators()
        {
            return new Dictionary<string, Dictionary<string, PhysicsCalculator>>
            {
                {
                    Side.LEFT, new Dictionary<string, PhysicsCalculator>
                    {
                        { UP, UpAngleTargetCalculator(Side.LEFT) },
                        { DOWN, DownAngleTargetCalculator(Side.LEFT) },
                    }
                },
                {
                    Side.RIGHT, new Dictionary<string, PhysicsCalculator>
                    {
                        { UP, UpAngleTargetCalculator(Side.RIGHT) },
                        { DOWN, DownAngleTargetCalculator(Side.RIGHT) },
                    }
                },
            };
        }

        static Dictionary<string, Dictionary<string, PhysicsCalculator>> RightLeftAngleTargetCalculators()
        {
            return new Dictionary<string, Dictionary<string, PhysicsCalculator>>
            {
                {
                    Side.LEFT, new Dictionary<string, PhysicsCalculator>
                    {
                        {
                            SIDEWAYS_IN, new PhysicsCalculator
                            {
                                multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(RIGHT_LEFT_ANGLE_TARGET.NAME, Side.LEFT), SIDEWAYS_IN, true),
                                primaryConfig = new MultiplierConfig
                                {
                                    baseValue = 1.60f,
                                    multiplierA = 8.00f,
                                    multiplierB = 9.30f,
                                    curveA = Calc.AngleTargetMassCurve,
                                    curveB = Calc.AngleTargetSoftnessCurve,
                                },
                                secondaryConfig = new MultiplierConfig
                                {
                                    baseValue = 1.60f,
                                    multiplierA = 8.00f,
                                    multiplierB = 12.00f,
                                    curveA = Calc.AngleTargetMassCurve,
                                    curveB = Calc.AngleTargetSoftnessCurve,
                                },
                            }
                        },
                        {
                            SIDEWAYS_OUT, new PhysicsCalculator
                            {
                                multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(RIGHT_LEFT_ANGLE_TARGET.NAME, Side.LEFT), SIDEWAYS_OUT, true),
                                primaryConfig = new MultiplierConfig
                                {
                                    baseValue = -1.60f,
                                    multiplierA = -6.00f,
                                    multiplierB = -7.00f,
                                    curveA = Calc.AngleTargetMassCurve,
                                    curveB = Calc.AngleTargetSoftnessCurve,
                                },
                                secondaryConfig = new MultiplierConfig
                                {
                                    baseValue = -1.60f,
                                    multiplierA = -6.00f,
                                    multiplierB = -9.00f,
                                    curveA = Calc.AngleTargetMassCurve,
                                    curveB = Calc.AngleTargetSoftnessCurve,
                                },
                            }
                        },
                    }
                },
                {
                    Side.RIGHT, new Dictionary<string, PhysicsCalculator>
                    {
                        {
                            SIDEWAYS_IN, new PhysicsCalculator
                            {
                                multiplierJsf =
                                    NewGravityMultiplierJsf(Utils.ParamName(RIGHT_LEFT_ANGLE_TARGET.NAME, Side.RIGHT), SIDEWAYS_IN, false),
                                primaryConfig = new MultiplierConfig
                                {
                                    baseValue = -1.60f,
                                    multiplierA = -8.00f,
                                    multiplierB = -12.00f,
                                    curveA = Calc.AngleTargetMassCurve,
                                    curveB = Calc.AngleTargetSoftnessCurve,
                                },
                                secondaryConfig = new MultiplierConfig
                                {
                                    baseValue = -1.60f,
                                    multiplierA = -8.00f,
                                    multiplierB = -12.00f,
                                    curveA = Calc.AngleTargetMassCurve,
                                    curveB = Calc.AngleTargetSoftnessCurve,
                                },
                            }
                        },
                        {
                            SIDEWAYS_OUT, new PhysicsCalculator
                            {
                                multiplierJsf = NewGravityMultiplierJsf(
                                    Utils.ParamName(RIGHT_LEFT_ANGLE_TARGET.NAME, Side.RIGHT),
                                    SIDEWAYS_OUT,
                                    false
                                ),
                                primaryConfig = new MultiplierConfig
                                {
                                    baseValue = 1.60f,
                                    multiplierA = 6.00f,
                                    multiplierB = 9.00f,
                                    curveA = Calc.AngleTargetMassCurve,
                                    curveB = Calc.AngleTargetSoftnessCurve,
                                },
                                secondaryConfig = new MultiplierConfig
                                {
                                    baseValue = 1.60f,
                                    multiplierA = 6.00f,
                                    multiplierB = 9.00f,
                                    curveA = Calc.AngleTargetMassCurve,
                                    curveB = Calc.AngleTargetSoftnessCurve,
                                },
                            }
                        },
                    }
                },
            };
        }

        static StorableFloat NewGravityMultiplierJsf(string paramName, string direction, bool register)
        {
            return TittyMagic.script.NewStorableFloat($"{paramName} {direction} Gravity", 1, 0, 2, register);
        }

        static Dictionary<string, PhysicsCalculator> TwistAngleTargetCalculator()
        {
            return new Dictionary<string, PhysicsCalculator>
            {
                {
                    UP, new PhysicsCalculator
                    {
                        primaryConfig = new MultiplierConfig(),
                        secondaryConfig = new MultiplierConfig(),
                    }
                },
                {
                    DOWN, new PhysicsCalculator
                    {
                        primaryConfig = new MultiplierConfig(),
                        secondaryConfig = new MultiplierConfig(),
                    }
                },
            };
        }

        static Dictionary<string, Dictionary<string, PhysicsCalculator>> TwistAngleTargetCalculators()
        {
            var left = TwistAngleTargetCalculator();
            var right = TwistAngleTargetCalculator();

            twistAngleTargetJsf.setJSONCallbackFunction = jsf =>
            {
                var storable = (StorableFloat) jsf;
                float rounded = Calc.RoundToDecimals(2 * storable.val);
                left[UP].primaryConfig.baseValue = -rounded;
                left[UP].secondaryConfig.baseValue = -rounded;
                left[DOWN].primaryConfig.baseValue = rounded;
                left[DOWN].secondaryConfig.baseValue = rounded;
                right[UP].primaryConfig.baseValue = rounded;
                right[UP].secondaryConfig.baseValue = rounded;
                right[DOWN].primaryConfig.baseValue = -rounded;
                right[DOWN].secondaryConfig.baseValue = -rounded;
                TittyMagic.script.SetCalibrationPending(storable);
                TittyMagic.tabs.physicsWindow.OnMainParamAdjusted(MainPhysicsHandler.twistAngleTarget);
            };

            return new Dictionary<string, Dictionary<string, PhysicsCalculator>>
            {
                { Side.LEFT, left },
                { Side.RIGHT, right },
            };
        }

        public static void SaveCalibrated()
        {
            gravityCounterforceJsf.SaveCalibrated();
        }

        static bool _simulatingGravity;
        static bool _softPhysicsOn;
        static float _mass;
        static float _softness;
        static float _rollL;
        static float _rollR;
        static float _pitchL;
        static float _pitchR;

        public static void Update(bool simulatingGravity = false)
        {
            _simulatingGravity = simulatingGravity;
            _softPhysicsOn = SoftPhysicsHandler.softPhysicsOn;
            _mass = MainPhysicsHandler.normalizedMass;
            _softness = TittyMagic.softness;
            _rollL = Calc.SmoothStep((GravityEffectCalc.pectoralRollL + GravityEffectCalc.chestRoll) / 2);
            _rollR = Calc.SmoothStep((GravityEffectCalc.pectoralRollR + GravityEffectCalc.chestRoll) / 2);
            _pitchL = 2 * Calc.SmoothStep(GravityEffectCalc.pectoralPitchL);
            _pitchR = 2 * Calc.SmoothStep(GravityEffectCalc.pectoralPitchR);
            if(!_simulatingGravity)
            {
                ApplyGravityCounterforce();
            }

            AdjustSidewaysPhysics();
            AdjustVerticalPhysics();
            AdjustHorizontalPhysics();
        }

        static void AdjustSidewaysPhysics()
        {
            float effectL = Mathf.Abs(_rollL) / 2;
            if(_rollL > 0)
            {
                ResetLeftBreast(SIDEWAYS_IN);
                UpdateLeftBreast(SIDEWAYS_OUT, effectL);
            }
            else
            {
                ResetLeftBreast(SIDEWAYS_OUT);
                UpdateLeftBreast(SIDEWAYS_IN, effectL);
            }

            float effectR = Mathf.Abs(_rollR) / 2;
            if(_rollR > 0)
            {
                ResetRightBreast(SIDEWAYS_OUT);
                UpdateRightBreast(SIDEWAYS_IN, effectR);
            }
            else
            {
                ResetRightBreast(SIDEWAYS_IN);
                UpdateRightBreast(SIDEWAYS_OUT, effectR);
            }
        }

        static void AdjustVerticalPhysics()
        {
            float effectL = GravityEffectCalc.UpDownAdjustByAngle(_pitchL) * GravityEffectCalc.RollMultiplier(_rollL) / 2;
            float effectR = GravityEffectCalc.UpDownAdjustByAngle(_pitchR) * GravityEffectCalc.RollMultiplier(_rollR) / 2;

            if(_pitchL > 1 || _pitchL < -1)
            {
                // upside down, leaning forward or back
                ResetLeftBreast(DOWN);
                UpdateLeftBreast(UP, effectL);
            }
            else
            {
                ResetLeftBreast(UP);
                UpdateLeftBreast(DOWN, effectL);
            }

            if(_pitchR > 1 || _pitchR < -1)
            {
                // upside down, leaning forward or back
                ResetRightBreast(DOWN);
                UpdateRightBreast(UP, effectR);
            }
            else
            {
                ResetRightBreast(UP);
                UpdateRightBreast(DOWN, effectR);
            }
        }

        static void AdjustHorizontalPhysics()
        {
            float effectL = GravityEffectCalc.DepthAdjustByAngle(_pitchL) * GravityEffectCalc.RollMultiplier(_rollL) / 2;
            float effectR = GravityEffectCalc.DepthAdjustByAngle(_pitchR) * GravityEffectCalc.RollMultiplier(_rollR) / 2;

            if(_pitchL > 0)
            {
                // leaning forward, upright or upside down
                ResetLeftBreast(BACK);
                UpdateLeftBreast(FORWARD, effectL);
            }
            else
            {
                // leaning back
                ResetLeftBreast(FORWARD);
                UpdateLeftBreast(BACK, effectL);
            }

            if(_pitchR > 0)
            {
                // leaning forward, upright or upside down
                ResetRightBreast(BACK);
                UpdateRightBreast(FORWARD, effectR);
            }
            else
            {
                // leaning back
                UpdateRightBreast(BACK, effectR);
                ResetRightBreast(FORWARD);
            }
        }

        public static void ApplyGravitySimulationForce(string direction)
        {
            Person.ApplyGravityToBreasts(direction, _gravityCounterForce);
        }

        static void ApplyGravityCounterforce()
        {
            if(!_useGravityCounterforce)
            {
                return;
            }

            Person.ApplyForceToBreasts(_gravityCounterForce);
        }

        static void UpdateLeftBreast(string direction, float effect)
        {
            MainPhysicsHandler.spring.left.UpdateGravityValue(direction, effect, _mass, _softness, _softPhysicsOn);
            if(!_simulatingGravity)
            {
                MainPhysicsHandler.damper.left.UpdateGravityValue(direction, effect, _mass, _softness, _softPhysicsOn);
            }

            MainPhysicsHandler.inOutSpring.left.UpdateGravityValue(direction, effect, _mass, _softness, _softPhysicsOn);
            MainPhysicsHandler.upDownAngleTarget.left.UpdateGravityValue(direction, effect, _mass, _softness, _softPhysicsOn);
            MainPhysicsHandler.rightLeftAngleTarget.left.UpdateGravityValue(direction, effect, _mass, _softness, _softPhysicsOn);
            MainPhysicsHandler.twistAngleTarget.left.UpdateGravityValue(direction, effect, _mass, _softness, _softPhysicsOn);
        }

        static void UpdateRightBreast(string direction, float effect)
        {
            MainPhysicsHandler.spring.right.UpdateGravityValue(direction, effect, _mass, _softness, _softPhysicsOn);
            if(!_simulatingGravity)
            {
                MainPhysicsHandler.damper.right.UpdateGravityValue(direction, effect, _mass, _softness, _softPhysicsOn);
            }

            MainPhysicsHandler.inOutSpring.right.UpdateGravityValue(direction, effect, _mass, _softness, _softPhysicsOn);
            MainPhysicsHandler.upDownAngleTarget.right.UpdateGravityValue(direction, effect, _mass, _softness, _softPhysicsOn);
            MainPhysicsHandler.rightLeftAngleTarget.right.UpdateGravityValue(direction, effect, _mass, _softness, _softPhysicsOn);
            MainPhysicsHandler.twistAngleTarget.right.UpdateGravityValue(direction, effect, _mass, _softness, _softPhysicsOn);
        }

        public static void ResetAll()
        {
            foreach(string direction in new[]
            {
                SIDEWAYS_IN,
                SIDEWAYS_OUT,
                UP,
                DOWN,
                FORWARD,
                BACK,
            })
            {
                ResetLeftBreast(direction);
                ResetRightBreast(direction);
            }
        }

        static void ResetLeftBreast(string direction)
        {
            MainPhysicsHandler.spring.left.ResetGravityValue(direction);
            if(!_simulatingGravity)
            {
                MainPhysicsHandler.damper.left.ResetGravityValue(direction);
            }

            MainPhysicsHandler.inOutSpring.left.ResetGravityValue(direction);
            MainPhysicsHandler.upDownAngleTarget.left.ResetGravityValue(direction);
            MainPhysicsHandler.rightLeftAngleTarget.left.ResetGravityValue(direction);
            MainPhysicsHandler.twistAngleTarget.left.ResetGravityValue(direction);
        }

        static void ResetRightBreast(string direction)
        {
            MainPhysicsHandler.spring.right.ResetGravityValue(direction);
            if(!_simulatingGravity)
            {
                MainPhysicsHandler.damper.right.ResetGravityValue(direction);
            }

            MainPhysicsHandler.inOutSpring.right.ResetGravityValue(direction);
            MainPhysicsHandler.upDownAngleTarget.right.ResetGravityValue(direction);
            MainPhysicsHandler.rightLeftAngleTarget.right.ResetGravityValue(direction);
            MainPhysicsHandler.twistAngleTarget.right.ResetGravityValue(direction);
        }

        public static void Destroy()
        {
            twistAngleTargetJsf = null;
            gravityCounterforceJsf = null;
            autoAdjustGravityCounterforceJsb = null;
        }
    }
}
