using System;
using System.Collections.Generic;
using System.Linq;
using Naturalis.Common;
using Naturalis.Common.Components;
using Naturalis.TM.Models;
using SimpleJSON;
using UnityEngine;

namespace Naturalis.TM.Handlers
{
    static class HardColliderHandler
    {
        public static ColliderVisualizer colliderVisualizer { get; private set; }
        public static JSONStorableStringChooser colliderGroupsJssc { get; private set; }
        public static List<HardColliderGroup> hardColliderGroups { get; private set; }
        public static JSONStorableFloat baseMassMultiplierJsf { get; private set; }
        public static JSONStorableString allCollidersMassDisplayValueJss { get; private set; }
        public static JSONStorableBool highlightAllJsb { get; private set; }

        static float _frictionSizeMultiplierLeft;
        static float _frictionSizeMultiplierRight;

        public static void Init()
        {
            if(!Person.IsFemale())
            {
                return;
            }

            InitColliderVisualizer();

            var autoColliders = TittyMagic.script.containingAtom.GetComponentInChildren<AutoColliderBatchUpdater>().autoColliders;
            hardColliderGroups = new List<HardColliderGroup>
            {
                NewHardColliderGroup("Pectoral1", autoColliders),
                NewHardColliderGroup("Pectoral2", autoColliders),
                NewHardColliderGroup("Pectoral3", autoColliders),
                NewHardColliderGroup("Pectoral4", autoColliders),
                NewHardColliderGroup("Pectoral5", autoColliders),
            };
            EnableMultiplyFriction();

            var options = hardColliderGroups.Select(c => c.visualizerId).ToList();
            var displayOptions = hardColliderGroups.Select(c => c.id).ToList();
            colliderGroupsJssc = new JSONStorableStringChooser(
                "colliderGroup",
                options,
                displayOptions,
                "",
                "Collider"
            )
            {
                setCallbackFunction = value =>
                {
                    colliderVisualizer.SelectablesJSON.val = value;
                    SyncSizeAuto();
                },
                val = options[0],
            };

            colliderVisualizer.GroupsJSON.setJSONCallbackFunction = jssc =>
            {
                colliderVisualizer.SelectedPreviewOpacityJSON.val = jssc.val == "Off" ? 0 : 1;
                if(jssc.val != "Off")
                {
                    colliderVisualizer.SelectablesJSON.val = colliderGroupsJssc.val;
                }

                SyncSizeAuto();
            };
            colliderVisualizer.XRayPreviewsJSON.setJSONCallbackFunction = _ => SyncSizeAuto();

            baseMassMultiplierJsf = TittyMagic.script.NewStorableFloat($"Hard Colliders Base {Constant.MASS_MULTIPLIER}", 1.00f, 0.01f, 2.00f);
            baseMassMultiplierJsf.setCallbackFunction = _ => SyncCollidersMass();
            allCollidersMassDisplayValueJss = new JSONStorableString("allCollidersMassDisplayValue", "");
            highlightAllJsb = new JSONStorableBool(
                "Highlight All Colliders",
                false,
                value => colliderVisualizer.PreviewOpacityJSON.val = value ? 1.00f : 0.67f
            );
        }

        static void InitColliderVisualizer()
        {
            colliderVisualizer = TittyMagic.script.gameObject.AddComponent<ColliderVisualizer>();
            var groups = new List<Group>
            {
                new Group("Off", @"$off"), //match nothing
                new Group("Both breasts", @"[lr](Pectoral\d)"),
                new Group("Left breast", @"lPectoral\d"),
            };
            colliderVisualizer.Init(TittyMagic.script, groups);
            colliderVisualizer.PreviewOpacityJSON.val = 0.67f;
            colliderVisualizer.PreviewOpacityJSON.defaultVal = 0.67f;
            colliderVisualizer.SelectedPreviewOpacityJSON.val = 1;
            colliderVisualizer.SelectedPreviewOpacityJSON.defaultVal = 1;
            colliderVisualizer.GroupsJSON.val = "Left breast";
            colliderVisualizer.GroupsJSON.defaultVal = "Left breast";
            colliderVisualizer.HighlightMirrorJSON.val = true;

            foreach(string option in new[] { "Select...", "Other", "All" })
            {
                colliderVisualizer.GroupsJSON.choices.Remove(option);
            }
        }

        static float ColliderMassJointMassCurve(float x)
        {
            return Calc.InverseSmoothStep(x, 1.00f, 0.50f, 0.80f);
        }

        static float ColliderMassJointMassSoftnessCurve(float x)
        {
            return Calc.InverseSmoothStep(x, 1.00f, 0.24f, 0.70f);
        }

        static float ColliderRadiusAndPositionSizeCurve(float x)
        {
            return Calc.InverseSmoothStep(x, 1.00f, 0.35f, 0.06f);
        }

        static float ColliderRadiusAndPositionSizeCurve2(float x)
        {
            return Calc.InverseSmoothStep(x, 1.00f, 0.07f, 0.04f);
        }

        static Dictionary<string, Scaler> GetScalingConfig(string id)
        {
            switch(id)
            {
                case "Pectoral1":
                    return new Dictionary<string, Scaler>
                    {
                        {
                            Constant.RADIUS_OFFSET, new Scaler
                            {
                                offset = -0.26f,
                                max = 40,
                                massCurve = x => 0.27f * ColliderRadiusAndPositionSizeCurve(x),
                            }
                        },
                        {
                            Constant.X_OFFSET, new Scaler
                            {
                                offset = -0.09f,
                                max = 40,
                                massCurve = x => -0.32f * ColliderRadiusAndPositionSizeCurve(x),
                            }
                        },
                        {
                            Constant.Y_OFFSET, new Scaler
                            {
                                offset = 0.35f,
                                max = 40,
                                massCurve = x => 0.38f * ColliderRadiusAndPositionSizeCurve(x),
                            }
                        },
                        {
                            Constant.Z_OFFSET, new Scaler
                            {
                                offset = -0.12f,
                                max = 40,
                            }
                        },
                    };
                case "Pectoral2":
                    return new Dictionary<string, Scaler>
                    {
                        {
                            Constant.RADIUS_OFFSET, new Scaler
                            {
                                offset = -0.56f,
                                max = 40,
                                massCurve = x => -0.15f * ColliderRadiusAndPositionSizeCurve(x),
                            }
                        },
                        {
                            Constant.X_OFFSET, new Scaler
                            {
                                offset = 0.88f,
                                max = 40,
                                massCurve = x => 0.48f * ColliderRadiusAndPositionSizeCurve(x),
                            }
                        },
                        {
                            Constant.Y_OFFSET, new Scaler
                            {
                                offset = -0.08f,
                                max = 40,
                                massCurve = x => -0.11f * ColliderRadiusAndPositionSizeCurve2(x),
                            }
                        },
                        {
                            Constant.Z_OFFSET, new Scaler
                            {
                                offset = -0.43f,
                                max = 40,
                                massCurve = x => -0.07f * ColliderRadiusAndPositionSizeCurve2(x),
                            }
                        },
                    };
                case "Pectoral3":
                    return new Dictionary<string, Scaler>
                    {
                        {
                            Constant.RADIUS_OFFSET, new Scaler
                            {
                                offset = -0.19f,
                                max = 40,
                                massCurve = x => -0.94f * ColliderRadiusAndPositionSizeCurve2(x),
                            }
                        },
                        {
                            Constant.X_OFFSET, new Scaler
                            {
                                offset = -0.32f,
                                max = 40,
                                massCurve = x => -0.12f * ColliderRadiusAndPositionSizeCurve2(x),
                            }
                        },
                        {
                            Constant.Y_OFFSET, new Scaler
                            {
                                offset = 0.21f,
                                max = 40,
                            }
                        },
                        {
                            Constant.Z_OFFSET, new Scaler
                            {
                                offset = -0.19f,
                                max = 40,
                                massCurve = x => -0.42f * ColliderRadiusAndPositionSizeCurve2(x),
                            }
                        },
                    };
                case "Pectoral4":
                    return new Dictionary<string, Scaler>
                    {
                        {
                            Constant.RADIUS_OFFSET, new Scaler
                            {
                                offset = -0.28f,
                                max = 40,
                                massCurve = x => -0.62f * ColliderRadiusAndPositionSizeCurve2(x),
                            }
                        },
                        {
                            Constant.X_OFFSET, new Scaler
                            {
                                offset = -0.32f,
                                max = 40,
                                massCurve = x => -0.42f * ColliderRadiusAndPositionSizeCurve2(x),
                            }
                        },
                        {
                            Constant.Y_OFFSET, new Scaler
                            {
                                offset = -0.05f,
                                max = 40,
                                massCurve = x => 0.35f * ColliderRadiusAndPositionSizeCurve(x),
                            }
                        },
                        {
                            Constant.Z_OFFSET, new Scaler
                            {
                                offset = -0.13f,
                                max = 40,
                                massCurve = x => -0.12f * ColliderRadiusAndPositionSizeCurve(x),
                            }
                        },
                    };
                case "Pectoral5":
                    return new Dictionary<string, Scaler>
                    {
                        {
                            Constant.RADIUS_OFFSET, new Scaler
                            {
                                offset = -0.41f,
                                max = 40,
                                massCurve = x => -1.24f * ColliderRadiusAndPositionSizeCurve2(x),
                            }
                        },
                        {
                            Constant.X_OFFSET, new Scaler
                            {
                                offset = -0.37f,
                                max = 40,
                                massCurve = x => -0.84f * ColliderRadiusAndPositionSizeCurve2(x),
                            }
                        },
                        {
                            Constant.Y_OFFSET, new Scaler
                            {
                                offset = 0.76f,
                                max = 40,
                                massCurve = x => 0.65f * ColliderRadiusAndPositionSizeCurve2(x),
                            }
                        },
                        {
                            Constant.Z_OFFSET, new Scaler
                            {
                                offset = -0.29f,
                                max = 40,
                                massCurve = x => -0.29f * ColliderRadiusAndPositionSizeCurve2(x),
                            }
                        },
                    };
                default: throw new ArgumentException($"Invalid id {id}");
            }
        }

        static HardColliderGroup NewHardColliderGroup(string id, AutoCollider[] autoColliders)
        {
            var leftCollider = NewHardCollider($"l{id}", autoColliders);
            var rightCollider = NewHardCollider($"r{id}", autoColliders);
            var scalingConfig = GetScalingConfig(id);

            var frictionMultipliers = new Dictionary<string, float>
            {
                { "Pectoral1", 0.5f },
                { "Pectoral2", 1.0f },
                { "Pectoral3", 1.0f },
                { "Pectoral4", 1.0f },
                { "Pectoral5", 1.0f },
            };

            string visualizerId = colliderVisualizer.SelectablesJSON.choices.Find(option => option.EndsWith($"l{id}"));
            var hardColliderGroup = new HardColliderGroup(
                id,
                visualizerId,
                leftCollider,
                rightCollider,
                new Scaler
                {
                    offset = 0.040f,
                    max = 1,
                    massCurve = x => 0.040f * ColliderMassJointMassCurve(x),
                    softnessCurve = x =>
                        Mathf.Lerp(-0.067f, -0.077f, ColliderMassJointMassSoftnessCurve(MainPhysicsHandler.normalizedMass)) *
                        SoftPhysicsHandler.FatSpringSoftnessCurve(x),
                },
                scalingConfig[Constant.RADIUS_OFFSET],
                scalingConfig[Constant.X_OFFSET],
                scalingConfig[Constant.Y_OFFSET],
                scalingConfig[Constant.Z_OFFSET],
                frictionMultipliers[id]
            )
            {
                massMultiplierJsf = TittyMagic.script.NewStorableFloat($"{id} {Constant.MASS_MULTIPLIER}", 0.50f, 0.01f, 1.00f),
                radiusJsf = TittyMagic.script.NewStorableFloat($"{id} {Constant.RADIUS_OFFSET}", 0, -1f, 1f),
                rightJsf = TittyMagic.script.NewStorableFloat($"{id} {Constant.X_OFFSET}", 0, -1f, 1f),
                upJsf = TittyMagic.script.NewStorableFloat($"{id} {Constant.Y_OFFSET}", 0, -1f, 1f),
                lookJsf = TittyMagic.script.NewStorableFloat($"{id} {Constant.Z_OFFSET}", 0, -1f, 1f),
            };

            hardColliderGroup.massMultiplierJsf.setCallbackFunction = _ => SyncColliderMass(hardColliderGroup, true);
            hardColliderGroup.radiusJsf.setCallbackFunction = _ => SyncRadius(hardColliderGroup);
            hardColliderGroup.rightJsf.setCallbackFunction = _ => SyncPosition(hardColliderGroup);
            hardColliderGroup.upJsf.setCallbackFunction = _ => SyncPosition(hardColliderGroup);
            hardColliderGroup.lookJsf.setCallbackFunction = _ => SyncPosition(hardColliderGroup);

            return hardColliderGroup;
        }

        static HardCollider NewHardCollider(string id, IEnumerable<AutoCollider> autoColliders)
        {
            var collider = Person.GetBreastHardCollider(id);
            var autoCollider = autoColliders.First(ac => ac.jointCollider && ac.jointCollider.name == collider.name);
            return new HardCollider(autoCollider);
        }

        static void SyncRadius(HardColliderGroup group)
        {
            if(!Person.IsFemale())
            {
                return;
            }

            group.UpdateDimensions(MainPhysicsHandler.normalizedMass);
            SyncSizeAuto();
        }

        static void SyncPosition(HardColliderGroup group)
        {
            if(!Person.IsFemale())
            {
                return;
            }

            group.UpdatePosition(MainPhysicsHandler.normalizedMass);
            SyncSizeAuto();
        }

        public static void SyncAllOffsets()
        {
            if(!Person.IsFemale())
            {
                return;
            }

            float mass = MainPhysicsHandler.normalizedRealMass;
            foreach(var group in hardColliderGroups)
            {
                group.UpdateDimensions(mass);
                group.UpdatePosition(mass);
            }

            SyncSizeAuto();
        }

        static void SyncSizeAuto()
        {
            foreach(var group in hardColliderGroups)
            {
                group.AutoColliderSizeSet();
            }

            float averageRadius = hardColliderGroups.Average(group => group.left.autoCollider.colliderRadius);
            foreach(var group in hardColliderGroups)
            {
                group.UpdateMaxFrictionalDistance(
                    _frictionSizeMultiplierLeft,
                    _frictionSizeMultiplierRight,
                    averageRadius
                );
            }

            colliderVisualizer.SyncPreviews();
        }

        public static void UpdateFrictionSizeMultipliers()
        {
            /* int[]: left and right skin width markers */
            _frictionSizeMultiplierLeft = FrictionSizeMultiplier(new[] { 8956, 8967 });
            _frictionSizeMultiplierRight = FrictionSizeMultiplier(new[] { 19622, 19633 });
        }

        static float FrictionSizeMultiplier(int[] indices)
        {
            /* experimentally determined with 3kg breasts, is slightly different for different shapes */
            const float maxWidth = 0.17f;
            var skin = Person.instance.skin;
            float width = (skin.rawSkinnedVerts[indices[0]] - skin.rawSkinnedVerts[indices[1]]).magnitude;
            float multiplier = Mathf.InverseLerp(0, maxWidth, width);
            return Calc.InverseSmoothStep(multiplier, 1, 0.55f, 0.42f);
        }

        public static void UpdateFriction()
        {
            if(!Person.IsFemale())
            {
                return;
            }

            foreach(var group in hardColliderGroups)
            {
                group.UpdateFriction(FrictionHandler.maxHardColliderFriction);
            }
        }

        public static void SyncCollidersMass()
        {
            if(hardColliderGroups == null)
            {
                return;
            }

            foreach(var group in hardColliderGroups)
            {
                SyncColliderMass(group);
            }

            UpdateAllCollidersMassDisplayValue();
        }

        static void SyncColliderMass(HardColliderGroup group, bool updateDisplayValue = false)
        {
            if(!Person.IsFemale())
            {
                return;
            }

            float combinedMultiplier = baseMassMultiplierJsf.val * group.massMultiplierJsf.val;
            if(SoftPhysicsHandler.LdpUpdateIsEnabled())
            {
                combinedMultiplier *= Mathf.Lerp(1.00f, 1.44f, TittyMagic.softness);
            }

            group.UpdateRigidbodyMass(
                combinedMultiplier,
                MainPhysicsHandler.normalizedMass,
                TittyMagic.softness,
                SoftPhysicsHandler.softPhysicsOn
            );

            if(updateDisplayValue)
            {
                UpdateAllCollidersMassDisplayValue();
            }
        }

        static void UpdateAllCollidersMassDisplayValue()
        {
            float total = 0;
            foreach(var group in hardColliderGroups)
            {
                total += group.left.GetRigidbodyMass();
            }

            string displayValue = $"{Calc.RoundToDecimals(total)} kg".Bold();
            allCollidersMassDisplayValueJss.val = $"All Colliders Mass: {displayValue}";
        }

        public static void CalibrateColliders()
        {
            if(!Person.IsFemale())
            {
                return;
            }

            var breastCenterLeft = BreastCenter(MainPhysicsHandler.leftBreastIndices);
            var breastCenterRight = BreastCenter(MainPhysicsHandler.rightBreastIndices);
            foreach(var group in hardColliderGroups)
            {
                group.Calibrate(breastCenterLeft, breastCenterRight, Person.instance.chestRb);
            }
        }

        public static void UpdateDistanceDiffs()
        {
            if(!Person.IsFemale())
            {
                return;
            }

            var breastCenterLeft = BreastCenter(MainPhysicsHandler.leftBreastIndices);
            var breastCenterRight = BreastCenter(MainPhysicsHandler.rightBreastIndices);
            foreach(var group in hardColliderGroups)
            {
                group.UpdateDistanceDiffs(breastCenterLeft, breastCenterRight, Person.instance.chestRb);
            }
        }

        static Vector3 BreastCenter(int[] vertexIndices)
        {
            var rawSkinnedVerts = Person.instance.skin.rawSkinnedVerts;
            var vertices = new Vector3[vertexIndices.Length];
            for(int i = 0; i < vertexIndices.Length; i++)
            {
                vertices[i] = rawSkinnedVerts[vertexIndices[i]];
            }

            return Calc.RelativePosition(Person.instance.chestRb, Calc.AveragePosition(vertices));
        }

        public static void ResetDistanceDiffs()
        {
            if(!Person.IsFemale())
            {
                return;
            }

            foreach(var group in hardColliderGroups)
            {
                group.ResetDistanceDiffs();
            }
        }

        public static void EnableMultiplyFriction()
        {
            if(!Person.IsFemale())
            {
                return;
            }

            foreach(var group in hardColliderGroups)
            {
                group.EnableMultiplyFriction();
            }
        }

        public static JSONClass SerializeCalibrationData()
        {
            var jc = new JSONClass();
            foreach(var group in hardColliderGroups)
            {
                var groupJson = new JSONClass
                {
                    ["left_dist"] = { AsFloat = group.left.sqrNeutralDistance },
                    ["right_dist"] = { AsFloat = group.right.sqrNeutralDistance },
                };
                jc[group.GetJsonId()] = groupJson;
            }

            return jc;
        }

        public static void CalibrateFromJson(JSONClass jc)
        {
            foreach(var group in hardColliderGroups)
            {
                var groupJson = jc[group.GetJsonId()].AsObject;
                group.left.sqrNeutralDistance = groupJson["left_dist"].AsFloat;
                group.right.sqrNeutralDistance = groupJson["right_dist"].AsFloat;
            }
        }

        public static void RestoreOriginalPhysics()
        {
            if(!Person.IsFemale())
            {
                return;
            }

            /* Required for restoring default collider mass. Enabled in case disabled
             * programmatically, and not yet restored back on by PersonMonitor.
             */
            Person.instance.EnableUseColliders();
            foreach(var group in hardColliderGroups)
            {
                group.RestoreDefaults();
            }
        }

        public static void Destroy()
        {
            UnityEngine.Object.Destroy(colliderVisualizer);
            colliderVisualizer = null;
            colliderGroupsJssc = null;
            hardColliderGroups = null;
            baseMassMultiplierJsf = null;
            highlightAllJsb = null;
            allCollidersMassDisplayValueJss = null;
        }

        public static bool RigidbodiesFound()
        {
            return !Person.IsFemale() || hardColliderGroups.All(group => group.HasRigidbodies());
        }
    }
}
