﻿using System.Collections.Generic;
using Naturalis.Common;
using Naturalis.Common.Handlers.Configs;
using Naturalis.Common.Models;
using Naturalis.TM.Models;
using UnityEngine;
using static Side;

namespace Naturalis.TM.Handlers
{
    static class MainPhysicsHandler
    {
        // AdjustJoints.joint1.slerpDrive.maximumForce
        const float DEFAULT_SLERP_MAX_FORCE = 500;

        public static float calibratedMass { get; private set; }
        public static float normalizedMass { get; private set; }
        public static float normalizedRealMass { get; private set; }

        public static BreastPhysicsParameterGroup[] parameterGroups { get; private set; }

        public static BreastPhysicsParameterGroup breastMass { get; private set; }
        public static BreastPhysicsParameterGroup centerOfGravity { get; private set; }
        public static BreastPhysicsParameterGroup spring { get; private set; }
        public static BreastPhysicsParameterGroup damper { get; private set; }
        public static BreastPhysicsParameterGroup inOutSpring { get; private set; }
        public static BreastPhysicsParameterGroup inOutDamper { get; private set; }
        public static BreastPhysicsParameterGroup upDownAngleTarget { get; private set; }
        public static BreastPhysicsParameterGroup rightLeftAngleTarget { get; private set; }
        public static BreastPhysicsParameterGroup twistAngleTarget { get; private set; }

        public static void Init()
        {
            _rotationZZeroCount = new Dictionary<string, int>
            {
                { LEFT, 0 },
                { RIGHT, 0 },
            };

            LoadSettings();
        }

        /* Subsets of all breast indices, only ones that actually move */
        public static int[] leftBreastIndices =
        {
            2403, 2411, 2592, 7197, 7208, 8902, 8904, 8905, 8908, 8909, 8911, 8915, 8917, 8922, 8924, 8925, 8926, 8928,
            8930, 8933, 8934, 8935, 8939, 8941, 8942, 8943, 8945, 8946, 8947, 8948, 8949, 8951, 8952, 8954, 8959, 8969,
            8970,
        };

        public static int[] rightBreastIndices =
        {
            13233, 13241, 13410, 17904, 17915, 19568, 19570, 19571, 19574, 19575, 19577, 19581, 19583, 19588, 19590,
            19591, 19592, 19594, 19596, 19599, 19600, 19601, 19605, 19607, 19608, 19609, 19611, 19612, 19613, 19614,
            19615, 19617, 19618, 19620, 19625, 19635, 19636,
        };

        static float MinMass()
        {
            return SoftPhysicsHandler.softPhysicsOn ? BREAST_JOINT_MASS.SOFT_MIN : BREAST_JOINT_MASS.HARD_MIN;
        }

        static float MaxMass()
        {
            return SoftPhysicsHandler.softPhysicsOn ? BREAST_JOINT_MASS.SOFT_MAX : BREAST_JOINT_MASS.HARD_MAX;
        }

        public static void CalculateMass()
        {
            float leftVolume = CalculateVolume(leftBreastIndices);
            float rightVolume = CalculateVolume(rightBreastIndices);
            float volume = (leftVolume + rightVolume) / 2;

            breastMass.UpdateMinMax(MinMass(), MaxMass());

            // https://www.desmos.com/calculator/aw68taob0a
            float massWhenSoftPhysicsOn = Mathf.Pow(0.614f * volume, 1.290f); // 0.2kg to 3.0kg
            float massWhenSoftPhysicsOff = Mathf.Pow(0.903f * volume, 1.121f); // 0.3kg to 4.0kg;
            float massBaseValue = Mathf.Clamp(
                SoftPhysicsHandler.softPhysicsOn ? massWhenSoftPhysicsOn : massWhenSoftPhysicsOff,
                MinMass(),
                MaxMass()
            );

            UpdateMassValueAndAmounts(massBaseValue);
        }

        public static void UpdateMassValueAndAmounts(float mass)
        {
            calibratedMass = mass;
            breastMass.UpdateMassValue(mass);
            breastMass.UpdateOffsetMinMax();
            normalizedMass = Mathf.InverseLerp(MinMass(), MaxMass(), breastMass.left.activeValueJsf.val);
            normalizedRealMass = Mathf.InverseLerp(MinMass(), MaxMass(), breastMass.left.baseValueJsf.val);
        }

        static float CalculateVolume(int[] vertexIndices)
        {
            var rawSkinnedVerts = Person.instance.skin.rawSkinnedVerts;
            var positions = new Vector3[vertexIndices.Length];
            for(int i = 0; i < vertexIndices.Length; i++)
            {
                positions[i] = Calc.RelativePosition(Person.instance.chestRb, rawSkinnedVerts[vertexIndices[i]]);
            }

            return 1000 * VolumeCalculator.CupVolume(positions);
        }

        static void LoadSettings()
        {
            breastMass = new BreastPhysicsParameterGroup
            {
                name = BREAST_JOINT_MASS.NAME,
                valueFormat = "F3",
                left = BreastMass(LEFT),
                right = BreastMass(RIGHT),
                requiresCalibration = true,
                infoText = BREAST_JOINT_MASS.NAME.Bold() +
                    " is the mass of the pectoral joint." +
                    "\n\nSince mass represents breast size, other physics parameters are adjusted based on its value.",
            };

            centerOfGravity = new BreastPhysicsParameterGroup
            {
                name = CENTER_OF_GRAVITY.NAME,
                valueFormat = "F2",
                left = CenterOfGravity(LEFT),
                right = CenterOfGravity(RIGHT),
                requiresCalibration = true,
                infoText = CENTER_OF_GRAVITY.NAME.Bold() +
                    " is the position of the pectoral joint's center of mass." +
                    "\n\nAt 0, the center of mass is inside the chest at the pectoral joint. At 1, it is at the nipple.",
            };

            spring = new BreastPhysicsParameterGroup
            {
                name = SPRING.NAME,
                valueFormat = "F0",
                left = Spring(LEFT),
                right = Spring(RIGHT),
                requiresCalibration = true,
                infoText = SPRING.NAME.Bold() +
                    " pushes the pectoral joint towards its angle target." +
                    "\n\nThe angle target is defined by the Up/Down, Left/Right and Twist Angle Target parameters.",
            };

            damper = new BreastPhysicsParameterGroup
            {
                name = DAMPER.NAME,
                valueFormat = "F2",
                left = Damper(LEFT),
                right = Damper(RIGHT),
                dependsOnPhysicsRate = true,
                infoText = DAMPER.NAME.Bold() +
                    " reduces oscillation around the joint's angle target." +
                    "\n\nThe higher the damper, the quicker breasts will stop swinging.",
            };

            inOutSpring = new BreastPhysicsParameterGroup
            {
                name = IN_OUT_SPRING.NAME,
                valueFormat = "F0",
                left = InOutSpring(LEFT),
                right = InOutSpring(RIGHT),
                requiresCalibration = true,
                infoText = IN_OUT_SPRING.NAME.Bold() +
                    " pushes the pectoral joint towards its position target along the Z axis." +
                    "\n\nDirectional force morphing along the forward-back axis depends on In/Out Spring being suitably low" +
                    " for the given breast mass.",
            };

            inOutDamper = new BreastPhysicsParameterGroup
            {
                name = IN_OUT_DAMPER.NAME,
                valueFormat = "F0",
                left = InOutDamper(LEFT),
                right = InOutDamper(RIGHT),
                dependsOnPhysicsRate = true,
                infoText = IN_OUT_DAMPER.NAME.Bold() + " reduces oscillation around the joint position target along the Z axis.",
            };

            var upDownTargetL = UpDownAngleTarget(LEFT);
            var rightLeftTargetL = RightLeftAngleTarget(LEFT);
            var twistTargetL = TwistAngleTarget(LEFT);

            var upDownTargetR = UpDownAngleTarget(RIGHT);
            var rightLeftTargetR = RightLeftAngleTarget(RIGHT);
            var twistTargetR = TwistAngleTarget(RIGHT);

            upDownTargetL.sync = value => SyncAngleTarget(LEFT, value, rightLeftTargetL.activeValueJsf.val, twistTargetL.activeValueJsf.val);
            upDownTargetR.sync = value => SyncAngleTarget(RIGHT, value, rightLeftTargetR.activeValueJsf.val, twistTargetR.activeValueJsf.val);

            upDownAngleTarget = new BreastPhysicsParameterGroup
            {
                name = UP_DOWN_ANGLE_TARGET.NAME,
                valueFormat = "F2",
                left = upDownTargetL,
                right = upDownTargetR,
                requiresCalibration = true,
                infoText = UP_DOWN_ANGLE_TARGET.NAME.Bold() +
                    " is the vertical target angle of the pectoral joint." +
                    " Negative values pull breasts down, positive values push them up." +
                    "\n\nThe offset shifts the center around which the final angle is calculated based on the person's pose" +
                    " (see Gravity Physics below).",
            };

            rightLeftAngleTarget = new BreastPhysicsParameterGroup
            {
                name = RIGHT_LEFT_ANGLE_TARGET.NAME,
                valueFormat = "F2",
                left = rightLeftTargetL,
                right = rightLeftTargetR,
                requiresCalibration = true,
                rightInverted = true,
                infoText = RIGHT_LEFT_ANGLE_TARGET.NAME.Bold() +
                    " is the horizontal target angle of the pectoral joint." +
                    " A negative offset pulls breasts apart, while a positive offset pushes them together.",
            };

            twistAngleTarget = new BreastPhysicsParameterGroup
            {
                name = TWIST_ANGLE_TARGET.NAME,
                valueFormat = "F2",
                left = twistTargetL,
                right = twistTargetR,
                requiresCalibration = true,
                rightInverted = true, // correct but has no effect because offset not adjusted directly
                allowOffsetOnlyLeftParam = false,
                infoText = TWIST_ANGLE_TARGET.NAME.Bold() +
                    " is the forward axis target angle of the pectoral joint." +
                    "\n\nThe final value depends on the person's pose. The offset determines the max angle" +
                    " when the person is upright. The angle is inverted when upside down and zero when" +
                    " horizontal.",
            };

            parameterGroups = new[]
            {
                centerOfGravity,
                spring,
                damper,
                inOutSpring,
                inOutDamper,
                upDownAngleTarget,
                rightLeftAngleTarget,
                twistAngleTarget,
            };

            breastMass.SetCallbackFunctions();
            foreach(var group in parameterGroups)
            {
                group.SetCallbackFunctions();
            }
        }

#region *** Parameter Setup ***

        static PhysicsParameter BreastMass(string side)
        {
            var parameter = new PhysicsParameter(
                TittyMagic.script,
                side,
                BREAST_JOINT_MASS.NAME,
                BREAST_JOINT_MASS.SOFT_MIN,
                BREAST_JOINT_MASS.SOFT_MAX
            )
            {
                displayValueJss = new JSONStorableString($"{BREAST_JOINT_MASS.NAME} {side} Value Alt", ""),
                buildDisplayValueString = value => $"{BREAST_JOINT_MASS.NAME}: {$"{Calc.RoundToDecimals(value, breastMass.roundFactor)} kg".Bold()}",
                sync = value => SyncBreastMass(Person.instance.pectoralRbs[side], value),
            };
            return parameter;
        }

        static PhysicsParameter CenterOfGravity(string side)
        {
            var parameter = new PhysicsParameter(
                TittyMagic.script,
                side,
                CENTER_OF_GRAVITY.NAME,
                CENTER_OF_GRAVITY.MIN,
                CENTER_OF_GRAVITY.MAX
            )
            {
                sync = value => SyncCenterOfGravity(Person.instance.pectoralRbs[side], value),
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 0.51f,
                    multiplierA = 0.30f,
                    multiplierB = 0.00f,
                    // TODO curveA
                },
                secondaryConfig = new MultiplierConfig
                {
                    baseValue = 0.46f,
                    multiplierA = 0.30f,
                    multiplierB = 0.00f,
                },
                quicknessConfig = new MultiplierConfig
                {
                    baseValue = -0.06f,
                    multiplierA = 0.24f,
                    multiplierB = 0.00f,
                },
                slownessConfig = new MultiplierConfig
                {
                    baseValue = 0.06f,
                    multiplierA = 0.24f,
                    multiplierB = 0.00f,
                },
            };
            return parameter;
        }

        static PhysicsParameter Spring(string side)
        {
            var parameter = new PhysicsParameter(
                TittyMagic.script,
                side,
                SPRING.NAME,
                SPRING.MIN,
                SPRING.MAX
            )
            {
                sync = value => SyncSpring(Person.instance.pectoralJoints[side], Person.instance.pectoralRbs[side], value),
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 83f,
                    multiplierA = 0.19f,
                    multiplierB = -0.71f,
                    curveA = x => Calc.InverseSmoothStep(x, 1.00f, 0.23f, 0.30f),
                    curveB = x => Calc.InverseSmoothStep(x, 1.00f, 0.10f, 0.13f),
                },
                secondaryConfig = new MultiplierConfig
                {
                    baseValue = 78f,
                    multiplierA = 0.18f,
                    multiplierB = -0.74f,
                    curveA = x => Calc.InverseSmoothStep(x, 1.00f, 0.23f, 0.30f),
                    curveB = x => Calc.InverseSmoothStep(x, 1.00f, 0.10f, 0.13f),
                },
                quicknessConfig = new MultiplierConfig
                {
                    baseValue = 6f,
                    multiplierA = 0.67f,
                    multiplierB = 0.00f,
                    curveA = x => Calc.InverseSmoothStep(x, 1.00f, 0.23f, 0.30f),
                },
                slownessConfig = new MultiplierConfig
                {
                    baseValue = -9f,
                    multiplierA = 0.67f,
                    multiplierB = 0.00f,
                    curveA = x => Calc.InverseSmoothStep(x, 1.00f, 0.23f, 0.30f),
                },
            };

            return parameter;
        }

        static PhysicsParameter Damper(string side)
        {
            var parameter = new PhysicsParameter(
                TittyMagic.script,
                side,
                DAMPER.NAME,
                DAMPER.MIN,
                DAMPER.MAX
            )
            {
                sync = value => SyncDamper(Person.instance.pectoralJoints[side], Person.instance.pectoralRbs[side], value),
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 1.20f,
                    multiplierA = 1.35f,
                    multiplierB = -0.79f,
                    curveA = x => Calc.InverseSmoothStep(x, 1.00f, 0.31f, 0.80f),
                    curveB = x => Calc.InverseSmoothStep(x, 1.00f, 0.24f, 0.61f),
                },
                secondaryConfig = new MultiplierConfig
                {
                    baseValue = 1.80f,
                    multiplierA = 1.35f,
                    multiplierB = -0.88f,
                    curveA = x => Calc.InverseSmoothStep(x, 1.00f, 0.31f, 0.80f),
                    curveB = x => Calc.InverseSmoothStep(x, 1.00f, 0.24f, 0.61f),
                },
                quicknessConfig = new MultiplierConfig
                {
                    baseValue = -0.08f,
                    multiplierA = 8.375f,
                    multiplierB = 0.00f,
                    curveA = x => Calc.InverseSmoothStep(x, 1.00f, 0.31f, 0.80f),
                },
                slownessConfig = new MultiplierConfig
                {
                    baseValue = 0.04f,
                    multiplierA = 8.375f,
                    multiplierB = 0.00f,
                    curveA = x => Calc.InverseSmoothStep(x, 1.00f, 0.31f, 0.80f),
                },
            };
            return parameter;
        }

        static PhysicsParameter InOutSpring(string side)
        {
            var parameter = new PhysicsParameter(
                TittyMagic.script,
                side,
                IN_OUT_SPRING.NAME,
                IN_OUT_SPRING.MIN,
                IN_OUT_SPRING.MAX
            )
            {
                sync = value => SyncInOutSpring(Person.instance.pectoralJoints[side], Person.instance.pectoralRbs[side], value),
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 900f,
                    multiplierA = -0.15f,
                    multiplierB = -0.47f,
                    curveA = x => 1 - x, // TODO smoothstep
                },
                secondaryConfig = new MultiplierConfig
                {
                    baseValue = 900f,
                    multiplierA = -0.15f,
                    multiplierB = -0.47f,
                    curveA = x => 1 - x,
                },
                quicknessConfig = new MultiplierConfig
                {
                    baseValue = 42f,
                    multiplierA = -0.28f,
                    multiplierB = 0.00f,
                    curveA = x => 1 - x,
                },
                slownessConfig = new MultiplierConfig
                {
                    baseValue = -63f,
                    multiplierA = -0.28f,
                    multiplierB = 0.00f,
                    curveA = x => 1 - x,
                },
            };
            return parameter;
        }

        static PhysicsParameter InOutDamper(string side)
        {
            var parameter = new PhysicsParameter(
                TittyMagic.script,
                side,
                IN_OUT_DAMPER.NAME,
                IN_OUT_DAMPER.MIN,
                IN_OUT_DAMPER.MAX
            )
            {
                sync = value => SyncInOutDamper(Person.instance.pectoralJoints[side], Person.instance.pectoralRbs[side], value),
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 10f,
                    multiplierA = 0.50f,
                    multiplierB = -0.20f,
                },
                secondaryConfig = new MultiplierConfig
                {
                    baseValue = 10f,
                    multiplierA = 0.50f,
                    multiplierB = -0.20f,
                },
                quicknessConfig = new MultiplierConfig
                {
                    baseValue = -2.00f,
                    multiplierA = 1.33f,
                    multiplierB = 0.00f,
                },
                slownessConfig = new MultiplierConfig
                {
                    baseValue = 3.00f,
                    multiplierA = 1.33f,
                    multiplierB = 0.00f,
                },
            };
            return parameter;
        }

        static PhysicsParameter UpDownAngleTarget(string side)
        {
            var parameter = new PhysicsParameter(
                TittyMagic.script,
                side,
                UP_DOWN_ANGLE_TARGET.NAME,
                UP_DOWN_ANGLE_TARGET.MIN,
                UP_DOWN_ANGLE_TARGET.MAX
            )
            {
                primaryConfig = new MultiplierConfig(),
                secondaryConfig = new MultiplierConfig(),
            };
            return parameter;
        }

        static PhysicsParameter RightLeftAngleTarget(string side)
        {
            var parameter = new PhysicsParameter(
                TittyMagic.script,
                side,
                RIGHT_LEFT_ANGLE_TARGET.NAME,
                RIGHT_LEFT_ANGLE_TARGET.MIN,
                RIGHT_LEFT_ANGLE_TARGET.MAX
            )
            {
                primaryConfig = new MultiplierConfig(),
                secondaryConfig = new MultiplierConfig(),
            };
            return parameter;
        }

        static PhysicsParameter TwistAngleTarget(string side)
        {
            var parameter = new PhysicsParameter(
                TittyMagic.script,
                side,
                TWIST_ANGLE_TARGET.NAME,
                TWIST_ANGLE_TARGET.MIN,
                TWIST_ANGLE_TARGET.MAX
            )
            {
                primaryConfig = new MultiplierConfig(),
                secondaryConfig = new MultiplierConfig(),
            };
            TittyMagic.script.DeregisterFloat(parameter.offsetJsf);
            return parameter;
        }

#endregion

#region *** Sync Functions ***

        // Reimplements AdjustJoints.cs method SyncMass
        static void SyncBreastMass(Rigidbody rb, float value)
        {
            if(!TittyMagic.script.enabled)
            {
                return;
            }

            if(Mathf.Abs(rb.mass - value) <= 0.001f)
            {
                return;
            }

            rb.mass = value;
            rb.WakeUp();
        }

        // Reimplements AdjustJoints.cs method SyncCenterOfGravity
        static void SyncCenterOfGravity(Rigidbody rb, float value)
        {
            if(!TittyMagic.script.enabled)
            {
                return;
            }

            var newCenterOfMass = Vector3.Lerp(
                Person.instance.breastControl.lowCenterOfGravity,
                Person.instance.breastControl.highCenterOfGravity,
                value
            );
            if(Calc.VectorIsEqualWithin(1 / 100f, rb.centerOfMass, newCenterOfMass))
            {
                return;
            }

            rb.centerOfMass = newCenterOfMass;
            rb.WakeUp();
        }

        static void SyncSpring(ConfigurableJoint joint, Rigidbody rb, float value)
        {
            if(!TittyMagic.script.enabled)
            {
                return;
            }

            // see AdjustJoints.cs method ScaleChanged
            float scalePow = Mathf.Pow(1.7f, Person.instance.breastControl.scale - 1f);

            float scaledSpring = value * scalePow;
            if(Mathf.Abs(joint.slerpDrive.positionSpring - scaledSpring) <= 1)
            {
                return;
            }

            var slerpDrive = joint.slerpDrive;
            slerpDrive.positionSpring = scaledSpring;
            slerpDrive.maximumForce = DEFAULT_SLERP_MAX_FORCE * scalePow;
            joint.slerpDrive = slerpDrive;

            var angularXDrive = joint.angularXDrive;
            angularXDrive.positionSpring = scaledSpring;
            joint.angularXDrive = angularXDrive;

            var angularYZDrive = joint.angularYZDrive;
            angularYZDrive.positionSpring = scaledSpring;
            joint.angularYZDrive = angularYZDrive;

            var angularXLimitSpring = joint.angularXLimitSpring;
            angularXLimitSpring.spring = scaledSpring * Person.instance.breastControl.limitSpringMultiplier;
            joint.angularXLimitSpring = angularXLimitSpring;

            var angularYZLimitSpring = joint.angularYZLimitSpring;
            angularYZLimitSpring.spring = scaledSpring * Person.instance.breastControl.limitSpringMultiplier;
            joint.angularYZLimitSpring = angularYZLimitSpring;

            rb.WakeUp();
        }

        static void SyncDamper(ConfigurableJoint joint, Rigidbody rb, float value)
        {
            if(!TittyMagic.script.enabled)
            {
                return;
            }

            // see AdjustJoints.cs method ScaleChanged
            float scalePow = Mathf.Pow(1.7f, Person.instance.breastControl.scale - 1f);

            float scaledDamper = value * scalePow;
            if(Mathf.Abs(joint.slerpDrive.positionDamper - scaledDamper) <= 0.01f)
            {
                return;
            }

            var slerpDrive = joint.slerpDrive;
            slerpDrive.positionDamper = scaledDamper;
            slerpDrive.maximumForce = DEFAULT_SLERP_MAX_FORCE * scalePow;
            joint.slerpDrive = slerpDrive;

            var angularXDrive = joint.angularXDrive;
            angularXDrive.positionDamper = scaledDamper;
            joint.angularXDrive = angularXDrive;

            var angularYZDrive = joint.angularYZDrive;
            angularYZDrive.positionDamper = scaledDamper;
            joint.angularYZDrive = angularYZDrive;

            var angularXLimitSpring = joint.angularXLimitSpring;
            angularXLimitSpring.damper = scaledDamper * Person.instance.breastControl.limitDamperMultiplier;
            joint.angularXLimitSpring = angularXLimitSpring;

            var angularYZLimitSpring = joint.angularYZLimitSpring;
            angularYZLimitSpring.damper = scaledDamper * Person.instance.breastControl.limitDamperMultiplier;
            joint.angularYZLimitSpring = angularYZLimitSpring;

            rb.WakeUp();
        }

        // Reimplements AdjustJoints.cs method SyncJointPositionZDrive
        static void SyncInOutSpring(ConfigurableJoint joint, Rigidbody rb, float value)
        {
            if(!TittyMagic.script.enabled)
            {
                return;
            }

            var zDrive = joint.zDrive;
            if(Mathf.Abs(zDrive.positionSpring - value) <= 1)
            {
                return;
            }

            zDrive.positionSpring = value;
            joint.zDrive = zDrive;
            rb.WakeUp();
        }

        // Reimplements AdjustJoints.cs method SyncJointPositionZDrive
        static void SyncInOutDamper(ConfigurableJoint joint, Rigidbody rb, float value)
        {
            if(!TittyMagic.script.enabled)
            {
                return;
            }

            var zDrive = joint.zDrive;
            if(Mathf.Abs(zDrive.positionDamper - value) <= 1)
            {
                return;
            }

            zDrive.positionDamper = value;
            joint.zDrive = zDrive;
            rb.WakeUp();
        }

        static Dictionary<string, int> _rotationZZeroCount;

        static void SyncAngleTarget(string side, float rotationX, float rotationY, float rotationZ)
        {
            if(!TittyMagic.script.enabled)
            {
                return;
            }

            var rotation = Person.GetPectoralJointRotation(side);
            rotation.x = -rotationX;
            rotation.y = -rotationY;

            bool noZRotation = GravityPhysicsHandler.twistAngleTargetJsf.val == 0;
            if(_rotationZZeroCount[side] < 3)
            {
                rotation.z = rotationZ;
                if(noZRotation)
                {
                    _rotationZZeroCount[side]++;
                }
            }
            else if(!noZRotation)
            {
                _rotationZZeroCount[side] = 0;
            }

            Person.SetPectoralRotation(side, rotation);
        }

#endregion

        public static void SaveCalibrated()
        {
            breastMass.SaveCalibrated();
            foreach(var group in parameterGroups)
            {
                group.SaveCalibrated();
            }
        }

        public static void UpdatePhysics(bool updateOffsetMinMax = true)
        {
            foreach(var group in parameterGroups)
            {
                group.UpdateBaseValue(
                    normalizedMass,
                    TittyMagic.softness,
                    TittyMagic.quickness,
                    SoftPhysicsHandler.softPhysicsOn
                );
                if(updateOffsetMinMax)
                {
                    group.UpdateOffsetMinMax();
                }
            }
        }

        public static void SetGravitySimulationDamper()
        {
            float interpolant = Calc.InverseSmoothStep(normalizedMass, 1, 0.50f, 0.10f);
            damper.OverrideValue(Mathf.Lerp(2.5f, 10.00f, interpolant));
        }

        public static void UnsetGravitySimulationDamper()
        {
            damper.UpdateBaseValue(
                normalizedMass,
                TittyMagic.softness,
                TittyMagic.quickness,
                SoftPhysicsHandler.softPhysicsOn
            );
        }

        public static void RestoreOriginalPhysics()
        {
            if(!TittyMagic.IsInitialized())
            {
                return;
            }

            foreach(string name in new[]
            {
                "mass",
                "centerOfGravityPercent",
                "spring",
                "damper",
                "positionSpringZ",
                "positionDamperZ",
                "targetRotationX",
                "targetRotationY",
                "targetRotationZ",
            })
            {
                /* Set a value that is different from the original, then restore the original
                 * in order to trigger VaM's internal sync
                 */
                var paramJsf = Person.instance.breastControl.GetFloatJSONParam(name);
                float original = paramJsf.val;
                paramJsf.valNoCallback = Mathf.Abs(paramJsf.val - paramJsf.min) > 0.01f
                    ? paramJsf.min
                    : paramJsf.max;
                paramJsf.val = original;
            }
        }

        public static void Destroy()
        {
            parameterGroups = null;

            breastMass = null;
            centerOfGravity = null;
            spring = null;
            damper = null;
            inOutSpring = null;
            inOutDamper = null;
            rightLeftAngleTarget = null;
            upDownAngleTarget = null;
            twistAngleTarget = null;

            leftBreastIndices = null;
            rightBreastIndices = null;
        }
    }
}
