using System.Collections.Generic;
using System.Linq;
using Naturalis.Common;
using Naturalis.Common.Components;
using Naturalis.Common.Handlers.Configs;
using Naturalis.Common.Models;
using Naturalis.TM.Models;
using Naturalis.TM.UI;
using SimpleJSON;
using UnityEngine;
using static Side;
using static SoftVerticesGroupName;

namespace Naturalis.TM.Handlers
{
    static class SoftPhysicsHandler
    {
        public static ColliderVisualizer colliderVisualizer { get; private set; }
        public static SoftJointDiagnostics softJointDiagnostics { get; private set; }

        static Dictionary<string, SoftJoint[]> _allSoftJoints = new Dictionary<string, SoftJoint[]>();
        public static Dictionary<string, SoftJoint[]> outerSoftJoints { get; private set; }
        public static Dictionary<string, SoftJoint[]> mainSoftJoints { get; private set; }
        public static Dictionary<string, SoftJoint[]> areolaSoftJoints { get; private set; }
        public static Dictionary<string, SoftJoint> nippleSoftJoints { get; private set; }
        static Dictionary<string, SoftJoint[]> _nonGradientJoints = new Dictionary<string, SoftJoint[]>();

        static Dictionary<string, SoftJointGradientGroup[]> _gradientGroups = new Dictionary<string, SoftJointGradientGroup[]>();

        static Dictionary<string, Dictionary<string, DAZPhysicsMeshSoftVerticesGroup>> _softVerticesGroups =
            new Dictionary<string, Dictionary<string, DAZPhysicsMeshSoftVerticesGroup>>();

        public static SoftPhysicsParameterGroup[] softParameterGroups { get; private set; }
        public static PhysicsParameterGroup[] parameterGroups { get; private set; }

        public static BreastSoftPhysicsParameterGroup fatSpring { get; private set; }
        public static BreastSoftPhysicsParameterGroup fatDamper { get; private set; }
        public static BreastSoftPhysicsParameterGroup fatMass { get; private set; }
        public static BreastSoftPhysicsParameterGroup fatColliderRadius { get; private set; }
        public static BreastSoftPhysicsParameterGroup fatColliderDepth { get; private set; }
        public static BreastSoftPhysicsParameterGroup fatDistanceLimit { get; private set; }

        public static BreastPhysicsParameterGroup fatBackForce { get; private set; }
        public static BreastPhysicsParameterGroup fatBackForceMaxForce { get; private set; }
        public static BreastPhysicsParameterGroup fatBackForceThreshold { get; private set; }

        public static FrequencyRunner softJointsUpdater { get; private set; }

        public static JSONStorableBool breastSoftPhysicsOnJsb { get; private set; }
        public static JSONStorableBool breastSoftPhysicsSelfCollideJsb { get; private set; }
        public static StorableBool ldpEnabledJsb { get; private set; }

        public static JSONStorableBool showColliderPreviewsJsb { get; private set; }

        static bool _pluginSoftPhysicsToggled;
        public static bool softPhysicsSynced { get; set; }
        public static bool softPhysicsOn { get; private set; }
        static bool _ldpOn;

        public static bool LdpUpdateIsEnabled()
        {
            return _ldpOn && Settings.personMonitor.BreastSoftPhysicsIsEnabled();
        }

        public static void Init()
        {
            if(Person.IsFemale())
            {
                InitColliderVisualizer();
                SetupSoftGroupsAndJoints();
                EnableMultiplyFriction();
                softJointDiagnostics = new SoftJointDiagnostics(_allSoftJoints[LEFT]);
            }

            softJointsUpdater = new FrequencyRunner(Settings.instance.ldpUpdateRateJssc, () => UpdateSoftJointsDistortion());
            SetupStorables();
            LoadSettings();
            softPhysicsOn = breastSoftPhysicsOnJsb.val;
        }

        static void SetupStorables()
        {
            breastSoftPhysicsOnJsb = TittyMagic.script.NewStorableBool(
                Constant.BREAST_SOFT_PHYSICS_ENABLED,
                Person.IsFemale() && Person.instance.breastPhysicsMesh.on,
                Person.IsFemale()
            );
            ldpEnabledJsb = TittyMagic.script.NewStorableBool(
                "Local Distortion Physics Enabled",
                Person.IsFemale() && Person.instance.breastPhysicsMesh.on && Settings.instance.breastsLdpDefaultEnabledJsb.val,
                Person.IsFemale()
            );
            breastSoftPhysicsSelfCollideJsb = TittyMagic.script.NewStorableBool(
                "Breast Soft Physics Self Collide",
                Person.IsFemale(),
                Person.IsFemale()
            );

            if(Person.IsFemale())
            {
                breastSoftPhysicsOnJsb.setCallbackFunction = value =>
                {
                    ldpEnabledJsb.val = Settings.instance.breastsLdpDefaultEnabledJsb.val && value;

                    if(!_reverseSyncingSoftPhysics)
                    {
                        _pluginSoftPhysicsToggled = true;
                        if(TittyMagic.script.enabled)
                        {
                            Person.instance.breastPhysicsMesh.on = value;
                        }
                        else
                        {
                            TittyMagic.logger.LogMessage($"Enable the plugin to set {breastSoftPhysicsOnJsb.name}.");
                        }

                        if(TittyMagic.tabs.MainTabIsActive())
                        {
                            TittyMagic.tabs.mainWindow.Refresh();
                        }
                    }

                    _reverseSyncingSoftPhysics = false;
                };

                ldpEnabledJsb.setJSONCallbackFunction = jsb =>
                {
                    if(!TittyMagic.script.enabled)
                    {
                        TittyMagic.logger.LogMessage($"Enable the plugin to set {ldpEnabledJsb.name}.");
                        return;
                    }

                    var storable = (StorableBool) jsb;
                    if(storable.val)
                    {
                        if(!breastSoftPhysicsOnJsb.val)
                        {
                            TittyMagic.logger.LogMessage("Enable breast soft physics to use local distortion physics.");
                        }
                        else
                        {
                            TittyMagic.script.SetCalibrationPending(storable);
                        }
                    }
                    else
                    {
                        if(softPhysicsOn)
                        {
                            TittyMagic.script.SetCalibrationPending(storable);
                        }
                    }

                    softJointDiagnostics.visualizeLdpJsb.val = jsb.val;

                    if(TittyMagic.tabs.MainTabIsActive())
                    {
                        TittyMagic.tabs.mainWindow.Refresh();
                    }
                };

                if(ldpEnabledJsb.val)
                {
                    softJointDiagnostics.visualizeLdpJsb.val = true;
                }

                breastSoftPhysicsSelfCollideJsb.setCallbackFunction = SyncAllowSelfCollision;
            }
            else
            {
                breastSoftPhysicsOnJsb.valNoCallback = false;
                breastSoftPhysicsSelfCollideJsb.valNoCallback = false;
                ldpEnabledJsb.valNoCallback = false;
            }
        }

        static void InitColliderVisualizer()
        {
            colliderVisualizer = TittyMagic.script.gameObject.AddComponent<ColliderVisualizer>();
            var groups = new List<Group>
            {
                new Group("Off", @"$off"), //match nothing
                new Group("Both breasts", @"PhysicsMeshJoint(left|right)(areola|nipple|outer)?\d+$"),
                new Group("Left breast", @"PhysicsMeshJointleft(areola|nipple|outer)?\d+$"),
                new Group("Right breast", @"PhysicsMeshJointright(areola|nipple|outer)?\d+$"),
            };
            colliderVisualizer.Init(TittyMagic.script, groups);
            colliderVisualizer.PreviewOpacityJSON.val = 0.500f;
            colliderVisualizer.PreviewOpacityJSON.defaultVal = 0.500f;
            colliderVisualizer.SelectedPreviewOpacityJSON.val = 0.500f;
            colliderVisualizer.SelectedPreviewOpacityJSON.defaultVal = 0.500f;
            colliderVisualizer.GroupsJSON.val = "Left breast";
            colliderVisualizer.GroupsJSON.defaultVal = "Left breast";
            colliderVisualizer.HighlightMirrorJSON.val = true;
            colliderVisualizer.ShowPreviewsJSON.val = false;

            foreach(string option in new[] { "Select...", "Other", "All" })
            {
                colliderVisualizer.GroupsJSON.choices.Remove(option);
            }

            showColliderPreviewsJsb = new JSONStorableBool("Show Collider Previews", true, (bool _) =>
            {
                if(TittyMagic.tabs.PhysicsTabIsActive() && TittyMagic.tabs.GetActiveNestedWindow() is SoftParameterWindow)
                {
                    ((SoftParameterWindow) TittyMagic.tabs.GetActiveNestedWindow()).SetShowColliderPreviews();
                }
            });
        }

        static void SetupSoftGroupsAndJoints()
        {
            SetupSoftVerticesGroups(LEFT);
            SetupSoftVerticesGroups(RIGHT);
            outerSoftJoints = new Dictionary<string, SoftJoint[]>();
            mainSoftJoints = new Dictionary<string, SoftJoint[]>();
            areolaSoftJoints = new Dictionary<string, SoftJoint[]>();
            nippleSoftJoints = new Dictionary<string, SoftJoint>();
            SetupSoftJoints(LEFT);
            SetupSoftJoints(RIGHT);
        }

        static void SetupSoftVerticesGroups(string side)
        {
            var groups = Person.instance.breastPhysicsMesh.softVerticesGroups;
            _softVerticesGroups[side] = new Dictionary<string, DAZPhysicsMeshSoftVerticesGroup>
            {
                { OUTER, groups.Find(group => group.name == $"{side}{OUTER}") },
                { MAIN, groups.Find(group => group.name == side) },
                { AREOLA, groups.Find(group => group.name == $"{side}{AREOLA}") },
                { NIPPLE, groups.Find(group => group.name == $"{side}{NIPPLE}") },
            };
        }

        static void SetupSoftJoints(string side)
        {
            var allSJs = new List<SoftJoint>();
            var exceptNippleSJs = new List<SoftJoint>();
            var outerSJs = new List<SoftJoint>();
            var mainSJs = new List<SoftJoint>();
            var areolaSJs = new List<SoftJoint>();
            var nippleSJs = new List<SoftJoint>();

            foreach(var kvp in _softVerticesGroups[side])
            {
                string category = kvp.Key;
                var group = kvp.Value;
                foreach(var set in group.softVerticesSets)
                {
                    var softJoint = new SoftJoint(set)
                    {
                        id = $"{side}.{category}.{set.targetVertex.ToString()}",
                    };
                    switch(category)
                    {
                        case OUTER:
                            outerSJs.Add(softJoint);
                            break;
                        case MAIN:
                            mainSJs.Add(softJoint);
                            break;
                        case AREOLA:
                            areolaSJs.Add(softJoint);
                            break;
                        case NIPPLE:
                            nippleSJs.Add(softJoint);
                            break;
                    }

                    allSJs.Add(softJoint);
                    if(category != NIPPLE)
                    {
                        exceptNippleSJs.Add(softJoint);
                    }
                }
            }

            _allSoftJoints[side] = allSJs.ToArray();
            outerSoftJoints[side] = outerSJs.ToArray();
            mainSoftJoints[side] = mainSJs.ToArray();
            areolaSoftJoints[side] = areolaSJs.ToArray();
            nippleSoftJoints[side] = nippleSJs.First();

            var nearbyVerts = VertexIndices.NearbyVerts(side);
            var adjacentVerts = VertexIndices.AdjacentVerts(side);
            foreach(var softJoint in allSJs)
            {
                int target = softJoint.softVerticesSet.targetVertex;
                if(nearbyVerts.ContainsKey(target))
                {
                    int[] nearbyTarget = nearbyVerts[target];
                    softJoint.nearbyJoints = allSJs.Where(x => nearbyTarget.Contains(x.softVerticesSet.targetVertex)).ToArray();
                }
                else
                {
                    // nipple verts do not have nearby verts
                    softJoint.nearbyJoints = new SoftJoint[] { };
                }

                if(adjacentVerts.ContainsKey(target))
                {
                    int[] adjacentTarget = adjacentVerts[target];
                    softJoint.adjacentJoints = allSJs.Where(x => adjacentTarget.Contains(x.softVerticesSet.targetVertex)).ToArray();
                }
                else
                {
                    // areola and nipple verts do not have adjacent verts
                    softJoint.adjacentJoints = new SoftJoint[] { };
                }
            }

            var gradientGroupsList = new List<SoftJointGradientGroup>();
            foreach(int[] indexGroup in VertexIndices.CenterToOuterVerts(side))
            {
                var gradientGroup = new SoftJointGradientGroup
                {
                    areola = areolaSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[0])),
                    main1 = mainSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[1])),
                    main2 = mainSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[2])),
                    outer1 = outerSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[3])),
                };

                if(indexGroup.Length > 4)
                {
                    gradientGroup.outer2 = outerSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[4]));
                }

                if(indexGroup.Length > 5)
                {
                    gradientGroup.outer3 = outerSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[5]));
                }

                gradientGroupsList.Add(gradientGroup);
            }

            _gradientGroups[side] = gradientGroupsList.ToArray();

            // TODO hard code
            _nonGradientJoints[side] = exceptNippleSJs
                .Where(softJoint => !_gradientGroups[side].Any(x => x.HasJoint(softJoint)))
                .ToArray();
        }

        static bool MatchByVertex(SoftJoint softJoint, int vertex)
        {
            return softJoint.softVerticesSet.targetVertex == vertex;
        }

        static void LoadSettings()
        {
            fatSpring = new BreastSoftPhysicsParameterGroup
            {
                name = FAT_SPRING.NAME,
                left = FatSpring(LEFT),
                right = FatSpring(RIGHT),
                valueFormat = "F1",
                visualizeRelativeDistortion = true,
                visualizationLimit = 250f,
                distortionLimit = 500f,
                infoText = FAT_SPRING.NAME.Bold() +
                    " holds each soft joint in its target position." +
                    "\n\nLow fat spring makes breast fat soft and slow." +
                    " High fat spring makes it rigid and quick to return to its normal shape.",
            };

            fatDamper = new BreastSoftPhysicsParameterGroup
            {
                name = FAT_DAMPER.NAME,
                left = FatDamper(LEFT),
                right = FatDamper(RIGHT),
                valueFormat = "F2",
                visualizeRelativeDistortion = true,
                visualizationLimit = 2.50f,
                distortionLimit = 5.00f,
                infoText = FAT_DAMPER.NAME.Bold() +
                    " reduces oscillation of each soft joint around its target position." +
                    "\n\nLow fat damper makes breast fat jiggle more easily.",
            };

            fatMass = new BreastSoftPhysicsParameterGroup
            {
                name = FAT_MASS.NAME,
                left = FatMass(LEFT),
                right = FatMass(RIGHT),
                valueFormat = "F3",
                visualizeRelativeDistortion = true,
                visualizationLimit = 0.150f,
                distortionLimit = 0.300f,
                infoText = FAT_MASS.NAME.Bold() +
                    " determines the mass of each soft joint" +
                    " and the density of breast fat overall." +
                    " Increasing breast size without increasing fat mass reduces density.",
            };

            fatColliderRadius = new BreastSoftPhysicsParameterGroup
            {
                name = FAT_COLLIDER_RADIUS.NAME,
                left = FatColliderRadius(LEFT),
                right = FatColliderRadius(RIGHT),
                valueFormat = "F4",
                hasSoftColliderVisualization = true,
                visualizeRelativeDistortion = false,
                visualizationLimit = 0.0100f,
                distortionLimit = 0.0200f,
                infoText = FAT_COLLIDER_RADIUS.NAME.Bold() +
                    " is a scaling factor for the size of the" +
                    " colliders connected to each soft joint." +
                    "\n\nThe actual size varies per collider and automatically scales with breast size.",
            };

            fatColliderDepth = new BreastSoftPhysicsParameterGroup
            {
                name = FAT_COLLIDER_DEPTH.NAME,
                valueFormat = "F4",
                left = FatColliderDepth(LEFT),
                right = FatColliderDepth(RIGHT),
                hasSoftColliderVisualization = true,
                visualizeRelativeDistortion = false,
                visualizationLimit = 0.0020f,
                distortionLimit = 0.0020f,
                infoText = FAT_COLLIDER_DEPTH.NAME.Bold() +
                    " offsets collider positions relative to skin surface." +
                    "\n\nNegative values pull colliders out from the breast, positive values push them into the breast.",
            };

            fatDistanceLimit = new BreastSoftPhysicsParameterGroup
            {
                name = FAT_DISTANCE_LIMIT.NAME,
                valueFormat = "F3",
                left = FatDistanceLimit(LEFT),
                right = FatDistanceLimit(RIGHT),
                visualizeRelativeDistortion = true,
                visualizationLimit = 0.020f,
                distortionLimit = 0.020f,
                infoText = FAT_DISTANCE_LIMIT.NAME.Bold() + " is the maximum distance each soft joint can move away from its target position.",
            };

            softParameterGroups = new SoftPhysicsParameterGroup[]
            {
                fatSpring,
                fatDamper,
                fatMass,
                fatColliderRadius,
                fatColliderDepth,
                fatDistanceLimit,
            };

            fatBackForce = new BreastPhysicsParameterGroup
            {
                name = FAT_BACK_FORCE.NAME,
                valueFormat = "F2",
                left = FatBackForce(LEFT),
                right = FatBackForce(RIGHT),
                infoText = FAT_BACK_FORCE.NAME.Bold() +
                    " applies a force to the pectoral joint based on the movement of each soft joint." +
                    "\n\nLow back force (not 0) helps move the breast with collision, and adds a dampening effect." +
                    " High force can create a feedback loop that spirals out of control.",
            };

            fatBackForceMaxForce = new BreastPhysicsParameterGroup
            {
                name = FAT_BACK_FORCE_MAX_FORCE.NAME,
                valueFormat = "F2",
                left = FatBackForceMaxForce(LEFT),
                right = FatBackForceMaxForce(RIGHT),
                infoText = FAT_BACK_FORCE_MAX_FORCE.NAME.Bold() +
                    " sets an upper limit on the magnitude of back force." +
                    $"\n\nAlong with {FAT_BACK_FORCE_THRESHOLD.NAME}, this can be used to prevent an out of control feedback loop.",
            };

            fatBackForceThreshold = new BreastPhysicsParameterGroup
            {
                name = FAT_BACK_FORCE_THRESHOLD.NAME,
                valueFormat = "F3",
                left = FatBackForceThreshold(LEFT),
                right = FatBackForceThreshold(RIGHT),
                infoText = FAT_BACK_FORCE_THRESHOLD.NAME.Bold() +
                    " sets a minimum distance each soft joint needs to move" +
                    " for back force to be applied, ensuring that small movements don't cause the whole breast to move." +
                    $"\n\nAlong with {FAT_BACK_FORCE_MAX_FORCE.NAME}, this can be used to prevent an out of control feedback loop.",
            };

            parameterGroups = new PhysicsParameterGroup[]
            {
                fatBackForce,
                fatBackForceMaxForce,
                fatBackForceThreshold,
            };

            if(Person.IsFemale())
            {
                foreach(var group in softParameterGroups)
                {
                    group.SetCallbackFunctions();
                }

                foreach(var group in parameterGroups)
                {
                    group.SetCallbackFunctions();
                }

                var springJsf = fatSpring.left.baseValueJsf;
                var damperJsf = fatDamper.left.baseValueJsf;
                var massJsf = fatMass.left.baseValueJsf;
                var colliderRadiusJsf = fatColliderRadius.left.baseValueJsf;
                var colliderDepthJsf = fatColliderDepth.left.baseValueJsf;
                var distanceLimitJsf = fatDistanceLimit.left.baseValueJsf;

                SoftJoint.minSpring = springJsf.min;
                SoftJoint.maxSpring = springJsf.max * Constant.GRADIENT_MULTI_MAX;
                SoftJoint.minDamper = damperJsf.min;
                SoftJoint.maxDamper = damperJsf.max * Constant.GRADIENT_MULTI_MAX;
                SoftJoint.minMass = massJsf.min;
                SoftJoint.maxMass = massJsf.max * Constant.GRADIENT_MULTI_MAX;
                SoftJoint.minColliderRadius = colliderRadiusJsf.min;
                SoftJoint.maxColliderRadius = colliderRadiusJsf.max * Constant.GRADIENT_MULTI_MAX;
                SoftJoint.minColliderDepth = colliderDepthJsf.min * Constant.GRADIENT_MULTI_MAX;
                SoftJoint.maxColliderDepth = colliderDepthJsf.max * Constant.GRADIENT_MULTI_MAX;
                SoftJoint.minDistanceLimit = distanceLimitJsf.min;
                SoftJoint.maxDistanceLimit = distanceLimitJsf.max * Constant.GRADIENT_MULTI_MAX;
            }
        }

        public static Rigidbody[] GetPositionalRigidbodies(string side, string group)
        {
            var rigidbodies = new List<Rigidbody>();
            foreach(var set in _softVerticesGroups[side][group].softVerticesSets)
            {
                var indices = VertexIndices.PositionalSoftVerts(side, group);
                if(indices.Contains(set.targetVertex))
                {
                    rigidbodies.Add(set.jointRB);
                }
            }

            return rigidbodies.ToArray();
        }

#region *** Parameter Setup ***

        public static float FatSpringSoftnessCurve(float x)
        {
            return Calc.InverseSmoothStep(x, 1.00f, 0.35f, 0.11f);
        }

        static SoftPhysicsParameter NewSoftPhysicsParameter(string side, string name, float min, float max)
        {
            var parameter = new SoftPhysicsParameter(TittyMagic.script, Constant.TITTY_MAGIC, side, name, min, max);
            parameter.sync = (value, additionalOffset) => SyncGradientParameter(side, name, parameter.gradientConfig, value, additionalOffset);
            return parameter;
        }

        static SoftPhysicsParameter FatSpring(string side)
        {
            if(!Person.IsFemale())
            {
                return null;
            }

            var parameter = NewSoftPhysicsParameter(side, FAT_SPRING.NAME, FAT_SPRING.MIN, FAT_SPRING.MAX);
            parameter.primaryConfig = new MultiplierConfig
            {
                baseValue = 1000f,
                multiplierA = 0.00f,
                multiplierB = -0.962f,
                curveB = FatSpringSoftnessCurve,
            };
            parameter.secondaryConfig = new MultiplierConfig
            {
                baseValue = 1000f,
                multiplierA = 0.00f,
                multiplierB = -0.948f,
                curveB = FatSpringSoftnessCurve,
            };
            parameter.quicknessConfig = new MultiplierConfig
            {
                baseValue = 50f,
                multiplierA = 0.00f,
                multiplierB = -0.75f,
            };
            parameter.slownessConfig = new MultiplierConfig
            {
                baseValue = -125f,
                multiplierA = 0.00f,
                multiplierB = -0.960f,
            };

            var gradientConfig = parameter.gradientConfig;
            gradientConfig.middleShiftJsf.SetValueAndDefault(-0.40f);
            gradientConfig.outerMultiplierJsf.SetValueAndDefault(2.00f);
            gradientConfig.centerMultiplierJsf.SetValueAndDefault(1.50f);
            gradientConfig.nippleMultiplierJsf.SetValueAndDefault(2.25f);
            gradientConfig.outerToMiddleCurvature = 0.30f;
            gradientConfig.outerToMiddleMidPoint = 0.50f;
            gradientConfig.middleToCenterCurvature = 0.40f;
            gradientConfig.middleToCenterMidPoint = 0.70f;

            var ldpConfig = parameter.ldpConfig;
            ldpConfig.compressionEffectJsf.SetValueAndDefault(0.62f);
            ldpConfig.compressionSensitivityJsf.SetValueAndDefault(0.70f);
            ldpConfig.stretchingEffectJsf.SetValueAndDefault(0.62f);
            ldpConfig.stretchingSensitivityJsf.SetValueAndDefault(0.70f);

            return parameter;
        }

        static SoftPhysicsParameter FatDamper(string side)
        {
            if(!Person.IsFemale())
            {
                return null;
            }

            var parameter = NewSoftPhysicsParameter(side, FAT_DAMPER.NAME, FAT_DAMPER.MIN, FAT_DAMPER.MAX);
            parameter.primaryConfig = new MultiplierConfig
            {
                baseValue = 2.11f,
                multiplierA = 0.18f,
                multiplierB = -0.55f,
                curveA = x => Calc.Exponential2(x, 0.04f, 0.04f),
                curveB = x => Calc.Exponential1(x, 1.90f, 1.74f, 1.17f),
            };
            parameter.secondaryConfig = new MultiplierConfig
            {
                baseValue = 1.80f,
                multiplierA = 0.18f,
                multiplierB = -0.55f,
                curveA = x => Calc.Exponential2(x, 0.04f, 0.04f),
                curveB = x => Calc.Exponential1(x, 1.90f, 1.74f, 1.17f),
            };
            parameter.quicknessConfig = new MultiplierConfig
            {
                baseValue = -1.60f,
                multiplierA = 0.18f,
                multiplierB = -0.55f,
                curveA = x => Calc.Exponential2(x, 0.04f, 0.04f),
                curveB = x => Calc.Exponential1(x, 1.90f, 1.74f, 1.17f),
            };
            parameter.slownessConfig = new MultiplierConfig
            {
                baseValue = 0.80f,
                multiplierA = 0.18f,
                multiplierB = -0.55f,
                curveA = x => Calc.Exponential2(x, 0.04f, 0.04f),
                curveB = x => Calc.Exponential1(x, 1.90f, 1.74f, 1.17f),
            };

            var gradientConfig = parameter.gradientConfig;
            gradientConfig.middleShiftJsf.SetValueAndDefault(0.00f);
            gradientConfig.outerMultiplierJsf.SetValueAndDefault(1.80f);
            gradientConfig.centerMultiplierJsf.SetValueAndDefault(1.50f);
            gradientConfig.nippleMultiplierJsf.SetValueAndDefault(1.80f);
            gradientConfig.outerToMiddleCurvature = 0.30f;
            gradientConfig.outerToMiddleMidPoint = 0.50f;
            gradientConfig.middleToCenterCurvature = 0.40f;
            gradientConfig.middleToCenterMidPoint = 0.70f;

            var ldpConfig = parameter.ldpConfig;
            ldpConfig.compressionEffectJsf.SetValueAndDefault(-0.50f);
            ldpConfig.compressionSensitivityJsf.SetValueAndDefault(0.80f);
            ldpConfig.stretchingEffectJsf.SetValueAndDefault(-0.50f);
            ldpConfig.stretchingSensitivityJsf.SetValueAndDefault(0.80f);

            return parameter;
        }

        static SoftPhysicsParameter FatMass(string side)
        {
            if(!Person.IsFemale())
            {
                return null;
            }

            var parameter = NewSoftPhysicsParameter(side, FAT_MASS.NAME, FAT_MASS.MIN, FAT_MASS.MAX);
            parameter.primaryConfig = new MultiplierConfig
            {
                baseValue = 0.045f,
                multiplierA = 1.31f,
                multiplierB = 0.60f,
                curveA = x => Calc.Exponential1(x, 1.91f, 1.7f, 0.82f),
                curveB = x => Calc.Exponential1(x, 2.30f, 1.74f, 1.17f),
            };
            parameter.secondaryConfig = new MultiplierConfig
            {
                baseValue = 0.045f,
                multiplierA = 1.31f,
                multiplierB = 0.60f,
                curveA = x => Calc.Exponential1(x, 1.91f, 1.7f, 0.82f),
                curveB = x => Calc.Exponential1(x, 2.30f, 1.74f, 1.17f),
            };
            parameter.quicknessConfig = new MultiplierConfig
            {
                baseValue = -0.038f,
                multiplierA = 1.31f,
                multiplierB = 0.60f,
                curveB = x => Calc.Exponential1(x, 2.30f, 1.74f, 1.17f),
            };
            parameter.slownessConfig = new MultiplierConfig
            {
                baseValue = 0.038f,
                multiplierA = 1.31f,
                multiplierB = 0.60f,
                curveB = x => Calc.Exponential1(x, 2.30f, 1.74f, 1.17f),
            };

            var gradientConfig = parameter.gradientConfig;
            gradientConfig.middleShiftJsf.SetValueAndDefault(0.00f);
            gradientConfig.outerMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.centerMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.nippleMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.outerToMiddleCurvature = 0.30f;
            gradientConfig.outerToMiddleMidPoint = 0.50f;
            gradientConfig.middleToCenterCurvature = 0.40f;
            gradientConfig.middleToCenterMidPoint = 0.70f;

            var ldpConfig = parameter.ldpConfig;
            ldpConfig.compressionEffectJsf.SetValueAndDefault(0.70f);
            ldpConfig.compressionSensitivityJsf.SetValueAndDefault(0.80f);
            ldpConfig.stretchingEffectJsf.SetValueAndDefault(-0.35f);
            ldpConfig.stretchingSensitivityJsf.SetValueAndDefault(0.80f);

            return parameter;
        }

        static SoftPhysicsParameter FatColliderRadius(string side)
        {
            if(!Person.IsFemale())
            {
                return null;
            }

            var parameter = NewSoftPhysicsParameter(side, FAT_COLLIDER_RADIUS.NAME, FAT_COLLIDER_RADIUS.MIN, FAT_COLLIDER_RADIUS.MAX);
            parameter.primaryConfig = new MultiplierConfig
            {
                baseValue = 0.0142f,
                multiplierA = 1.80f,
                multiplierB = 0.20f,
                curveA = x => Calc.Exponential1(x, 1.42f, 4.25f, 1.17f),
            };
            parameter.secondaryConfig = new MultiplierConfig
            {
                baseValue = 0.0153f,
                multiplierA = 1.80f,
                multiplierB = 0.20f,
                curveA = x => Calc.Exponential1(x, 1.42f, 4.25f, 1.17f),
            };

            var gradientConfig = parameter.gradientConfig;
            gradientConfig.middleShiftJsf.SetValueAndDefault(0.00f);
            gradientConfig.outerMultiplierJsf.SetValueAndDefault(1.10f);
            gradientConfig.centerMultiplierJsf.SetValueAndDefault(1.25f);
            gradientConfig.nippleMultiplierJsf.SetValueAndDefault(0.33f);
            gradientConfig.outerToMiddleCurvature = 0.30f;
            gradientConfig.outerToMiddleMidPoint = 0.50f;
            gradientConfig.middleToCenterCurvature = 0.40f;
            gradientConfig.middleToCenterMidPoint = 0.70f;

            var ldpConfig = parameter.ldpConfig;
            ldpConfig.compressionEffectJsf.SetValueAndDefault(0.10f);
            ldpConfig.compressionSensitivityJsf.SetValueAndDefault(0.70f);
            ldpConfig.stretchingEffectJsf.SetValueAndDefault(0.40f);
            ldpConfig.stretchingSensitivityJsf.SetValueAndDefault(0.70f);

            parameter.usesRealMass = true;
            return parameter;
        }

        static SoftPhysicsParameter FatColliderDepth(string side)
        {
            if(!Person.IsFemale())
            {
                return null;
            }

            var parameter = NewSoftPhysicsParameter(side, FAT_COLLIDER_DEPTH.NAME, FAT_COLLIDER_DEPTH.MIN, FAT_COLLIDER_DEPTH.MAX);
            parameter.primaryConfig = new MultiplierConfig();
            parameter.secondaryConfig = new MultiplierConfig();

            var gradientConfig = parameter.gradientConfig;
            gradientConfig.middleShiftJsf.SetValueAndDefault(0.00f);
            gradientConfig.outerMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.centerMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.nippleMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.outerToMiddleCurvature = 0.30f;
            gradientConfig.outerToMiddleMidPoint = 0.50f;
            gradientConfig.middleToCenterCurvature = 0.30f;
            gradientConfig.middleToCenterMidPoint = 0.50f;

            var ldpConfig = parameter.ldpConfig;
            ldpConfig.compressionEffectJsf.SetValueAndDefault(0.00f);
            ldpConfig.compressionSensitivityJsf.SetValueAndDefault(0.80f);
            ldpConfig.stretchingEffectJsf.SetValueAndDefault(1.00f);
            ldpConfig.stretchingSensitivityJsf.SetValueAndDefault(0.90f);

            return parameter;
        }

        static SoftPhysicsParameter FatDistanceLimit(string side)
        {
            if(!Person.IsFemale())
            {
                return null;
            }

            var parameter = NewSoftPhysicsParameter(side, FAT_DISTANCE_LIMIT.NAME, FAT_DISTANCE_LIMIT.MIN, FAT_DISTANCE_LIMIT.MAX);
            parameter.primaryConfig = new MultiplierConfig
            {
                baseValue = 0.0155f,
                multiplierA = 3.68f,
                multiplierB = 0.52f,
            };
            parameter.secondaryConfig = new MultiplierConfig
            {
                baseValue = 0.020f,
                multiplierA = 3.68f,
                multiplierB = 0.52f,
            };

            var gradientConfig = parameter.gradientConfig;
            gradientConfig.middleShiftJsf.SetValueAndDefault(-0.80f);
            gradientConfig.outerMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.centerMultiplierJsf.SetValueAndDefault(0.80f);
            gradientConfig.nippleMultiplierJsf.SetValueAndDefault(0.86f);
            gradientConfig.outerToMiddleCurvature = 0.30f;
            gradientConfig.outerToMiddleMidPoint = 0.50f;
            gradientConfig.middleToCenterCurvature = 0.30f;
            gradientConfig.middleToCenterMidPoint = 0.50f;

            var ldpConfig = parameter.ldpConfig;
            ldpConfig.compressionEffectJsf.SetValueAndDefault(0.00f);
            ldpConfig.compressionSensitivityJsf.SetValueAndDefault(0.80f);
            ldpConfig.stretchingEffectJsf.SetValueAndDefault(1.00f);
            ldpConfig.stretchingSensitivityJsf.SetValueAndDefault(0.80f);

            parameter.usesRealMass = true;
            return parameter;
        }

        static PhysicsParameter FatBackForce(string side)
        {
            if(!Person.IsFemale())
            {
                return null;
            }

            var parameter = new PhysicsParameter(
                TittyMagic.script,
                side,
                FAT_BACK_FORCE.NAME,
                FAT_BACK_FORCE.MIN,
                FAT_BACK_FORCE.MAX
            )
            {
                sync = value => SyncBackForce(side, value),
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 0.72f,
                    multiplierA = 8.75f,
                    multiplierB = 35.00f,
                    curveA = x => Calc.InverseSmoothStep(x, 1.00f, 0.50f, 0.88f),
                    curveB = x => 1 - FatSpringSoftnessCurve(x),
                },
                secondaryConfig = new MultiplierConfig
                {
                    baseValue = 0.36f,
                    multiplierA = 17.50f,
                    multiplierB = 70.00f,
                    curveA = x => Calc.InverseSmoothStep(x, 1.00f, 0.50f, 0.88f),
                    curveB = x => 1 - FatSpringSoftnessCurve(x),
                },
                quicknessConfig = new MultiplierConfig
                {
                    baseValue = -0.36f,
                    multiplierA = 8.25f,
                    multiplierB = 0.00f,
                    curveA = x => Calc.InverseSmoothStep(x, 1.00f, 0.50f, 0.88f),
                },
                slownessConfig = new MultiplierConfig
                {
                    baseValue = 0.72f,
                    multiplierA = 8.25f,
                    multiplierB = 0.00f,
                },
            };

            return parameter;
        }

        static PhysicsParameter FatBackForceMaxForce(string side)
        {
            if(!Person.IsFemale())
            {
                return null;
            }

            var parameter = new PhysicsParameter(
                TittyMagic.script,
                side,
                FAT_BACK_FORCE_MAX_FORCE.NAME,
                FAT_BACK_FORCE_MAX_FORCE.MIN,
                FAT_BACK_FORCE_MAX_FORCE.MAX
            )
            {
                sync = value => SyncBackForceMaxForce(side, value),
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 50.00f,
                    multiplierA = 0.00f,
                    multiplierB = 0.00f,
                },
                secondaryConfig = new MultiplierConfig
                {
                    baseValue = 50.00f,
                    multiplierA = 0.00f,
                    multiplierB = 0.00f,
                },
            };

            return parameter;
        }

        static PhysicsParameter FatBackForceThreshold(string side)
        {
            if(!Person.IsFemale())
            {
                return null;
            }

            var parameter = new PhysicsParameter(
                TittyMagic.script,
                side,
                FAT_BACK_FORCE_THRESHOLD.NAME,
                FAT_BACK_FORCE_THRESHOLD.MIN,
                FAT_BACK_FORCE_THRESHOLD.MAX
            )
            {
                sync = value => SyncBackForceThresholdDistance(side, value),
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 0.001f,
                    multiplierA = 0.00f,
                    multiplierB = 0.00f,
                },
                secondaryConfig = new MultiplierConfig
                {
                    baseValue = 0.001f,
                    multiplierA = 0.00f,
                    multiplierB = 0.00f,
                },
            };

            return parameter;
        }

#endregion

#region *** Sync Functions ***

        static void SyncGradientParameter(string side, string paramName, GradientConfig gradientConfig, float value, float additionalOffset)
        {
            if(!TittyMagic.IsInitialized() || !TittyMagic.script.enabled || TittyMagic.restoringFromJson)
            {
                // Syncing parameter before first time calibration (e.g. due to offset storable callback) results in NaN errors
                return;
            }

            var gradientGroups = _gradientGroups[side];
            foreach(var group in gradientGroups)
            {
                group.SetGradientBaseValues(paramName, gradientConfig, value);
                group.SetAdditionalOffsets(paramName, additionalOffset);
            }

            /* Calculate remaining physics values for joints not included in gradients.
             * Values are based on neighboring (immediately adjacent) joints on the same "ring".
             */
            foreach(var joint in _nonGradientJoints[side])
            {
                joint.SetBaseValue(
                    paramName,
                    joint.adjacentJoints.Average(x => x.GetBaseValue(paramName))
                );
                joint.SetAdditionalOffset(paramName, additionalOffset);
            }

            nippleSoftJoints[side]
                .SetBaseValue(
                    paramName,
                    value * gradientConfig.nippleMultiplierJsf.val
                );
            nippleSoftJoints[side].SetAdditionalOffset(paramName, additionalOffset);
        }

        static void SyncBackForce(string side, float value)
        {
            if(!TittyMagic.script.enabled)
            {
                return;
            }

            foreach(var kvp in _softVerticesGroups[side])
            {
                kvp.Value.jointBackForce = value;
            }
        }

        static void SyncBackForceMaxForce(string side, float value)
        {
            if(!TittyMagic.script.enabled)
            {
                return;
            }

            foreach(var kvp in _softVerticesGroups[side])
            {
                kvp.Value.jointBackForceMaxForce = value;
            }
        }

        static void SyncBackForceThresholdDistance(string side, float value)
        {
            if(!TittyMagic.script.enabled)
            {
                return;
            }

            foreach(var kvp in _softVerticesGroups[side])
            {
                kvp.Value.jointBackForceThresholdDistance = value;
            }
        }

        static bool _reverseSyncingSoftPhysics;

        /* Update value if external value changed */
        public static void ReverseSyncSoftPhysicsOn()
        {
            if(!Person.IsFemale())
            {
                return;
            }

            if(!_pluginSoftPhysicsToggled)
            {
                _reverseSyncingSoftPhysics = true;
                breastSoftPhysicsOnJsb.valNoCallback = Person.instance.breastPhysicsMesh.on;
            }

            _pluginSoftPhysicsToggled = false;
        }

        /* Update value if external value changed */
        public static void ReverseSyncAllowSelfCollision()
        {
            if(!Person.IsFemale())
            {
                return;
            }

            breastSoftPhysicsSelfCollideJsb.valNoCallback = Person.instance.breastPhysicsMesh.allowSelfCollision;
        }

        public static void SyncSoftPhysicsOn()
        {
            softPhysicsOn = Settings.personMonitor.BreastSoftPhysicsIsEnabled();
            _ldpOn = softPhysicsOn && ldpEnabledJsb.val;
            if(!softPhysicsSynced && Person.IsFemale())
            {
                TittyMagic.script.leftBreast.SetSoftPhysicsBased(softPhysicsOn);
                TittyMagic.script.rightBreast.SetSoftPhysicsBased(softPhysicsOn);
            }

            softPhysicsSynced = true;
        }

        public static void SyncAllowSelfCollision()
        {
            SyncAllowSelfCollision(breastSoftPhysicsSelfCollideJsb.val);
        }

        static void SyncAllowSelfCollision(bool value)
        {
            if(!Person.IsFemale())
            {
                return;
            }

            if(!TittyMagic.script.enabled)
            {
                TittyMagic.logger.LogMessage($"Enable the plugin to set {breastSoftPhysicsSelfCollideJsb.name}.");
                return;
            }

            Person.instance.breastPhysicsMesh.allowSelfCollision = value;
        }

        public static void SyncFriction(string side, float friction)
        {
            // TODO convert to gradient
            float areolaFriction = friction + 0.33f * (1 - friction);
            float nippleFriction = friction + 0.5f * (1 - friction);

            foreach(var softJoint in mainSoftJoints[side])
            {
                SetSoftJointFriction(softJoint, friction);
            }

            foreach(var softJoint in outerSoftJoints[side])
            {
                SetSoftJointFriction(softJoint, friction);
            }

            foreach(var softJoint in areolaSoftJoints[side])
            {
                SetSoftJointFriction(softJoint, areolaFriction);
            }

            SetSoftJointFriction(nippleSoftJoints[side], nippleFriction);
        }

        static void SetSoftJointFriction(SoftJoint softJoint, float value)
        {
            var material = softJoint.softVerticesSet.jointCollider.material;
            material.dynamicFriction = value;
            material.staticFriction = value;
        }

#endregion

        public static void SetShowColliderPreviews(bool hasSoftColliderVisualization)
        {
            colliderVisualizer.ShowPreviewsJSON.val = showColliderPreviewsJsb.val && hasSoftColliderVisualization;
        }

        public static void SaveCalibrated()
        {
            ldpEnabledJsb.SaveCalibrated();
        }

        public static void UpdatePhysics(bool updateOffsetMinMax = true)
        {
            if(!softPhysicsOn)
            {
                return;
            }

            foreach(var kvp in _gradientGroups)
            {
                foreach(var group in kvp.Value)
                {
                    group.CalculateDistances();
                }
            }

            foreach(var group in softParameterGroups)
            {
                group.UpdateBaseValue(
                    group.left.usesRealMass
                        ? MainPhysicsHandler.normalizedRealMass
                        : MainPhysicsHandler.normalizedMass,
                    TittyMagic.softness,
                    TittyMagic.quickness,
                    LdpUpdateIsEnabled()
                );
                if(updateOffsetMinMax)
                {
                    group.UpdateOffsetMinMax();
                }
            }

            foreach(var group in parameterGroups)
            {
                group.UpdateBaseValue(
                    MainPhysicsHandler.normalizedMass,
                    TittyMagic.softness,
                    TittyMagic.quickness,
                    LdpUpdateIsEnabled()
                );
                if(updateOffsetMinMax)
                {
                    group.UpdateOffsetMinMax();
                }
            }
        }

        public static bool SoftJointsMoved(float tolerance)
        {
            return _allSoftJoints[LEFT].Any(softJoint => softJoint.Moved(tolerance)) ||
                _allSoftJoints[RIGHT].Any(softJoint => softJoint.Moved(tolerance));
        }

        public static void CalibrateSoftJointsDistortion(string direction)
        {
            foreach(var kvp in _allSoftJoints)
            {
                foreach(var softJoint in kvp.Value)
                {
                    softJoint.CalibrateDistortion(direction);
                }
            }
        }

        public static void UpdateSoftJointsDistortion(bool force = false)
        {
            UpdateSoftJointsDistortion(LEFT, force);
            UpdateSoftJointsDistortion(RIGHT, force);
        }

        public static void UpdateSoftJointsDistortion(string side, bool force = false)
        {
            if(!LdpUpdateIsEnabled())
            {
                return;
            }

            foreach(var softJoint in _allSoftJoints[side])
            {
                if(softJoint.UpdateDistortion(force))
                {
                    softJoint.SyncSpringAndDamperDistortion(side);
                    softJoint.SyncMassDistortion(side);
                    softJoint.SyncColliderRadiusAndDepthDistortion(side);
                    softJoint.SyncDistanceLimitDistortion(side);
                }
            }
        }

        public static void ResetSoftJointsDistortion()
        {
            if(!Person.IsFemale())
            {
                return;
            }

            foreach(var kvp in _allSoftJoints)
            {
                foreach(var softJoint in kvp.Value)
                {
                    softJoint.ResetDistortion();
                }
            }
        }

        static bool _originalAllowSelfCollision;

        public static void SaveOriginalBoolParamValues()
        {
            if(!Person.IsFemale())
            {
                return;
            }

            _originalAllowSelfCollision = Person.instance.breastPhysicsMesh.allowSelfCollision;
        }

        public static void EnableMultiplyFriction()
        {
            if(!Person.IsFemale())
            {
                return;
            }

            foreach(var kvp in _allSoftJoints)
            {
                foreach(var softJoint in kvp.Value)
                {
                    softJoint.EnableMultiplyFriction();
                }
            }
        }

        public static JSONClass SerializeCalibrationData()
        {
            var jc = new JSONClass();
            if(!LdpUpdateIsEnabled())
            {
                return jc;
            }

            foreach(var kvp in _allSoftJoints)
            {
                foreach(var softJoint in kvp.Value)
                {
                    jc[softJoint.id] = softJoint.SerializeCalibrationData();
                }
            }

            return jc;
        }

        public static void CalibrateFromJson(JSONClass jc)
        {
            if(!LdpUpdateIsEnabled())
            {
                return;
            }

            foreach(var kvp in _allSoftJoints)
            {
                foreach(var softJoint in kvp.Value)
                {
                    softJoint.CalibrateFromJson(jc[softJoint.id].AsObject);
                }
            }
        }

        public static void RestoreOriginalPhysics()
        {
            if(!TittyMagic.IsInitialized() || !Person.IsFemale())
            {
                return;
            }

            var breastPhysicsMesh = Person.instance.breastPhysicsMesh;
            breastPhysicsMesh.allowSelfCollision = _originalAllowSelfCollision;

            foreach(string name in breastPhysicsMesh.GetFloatParamNames())
            {
                /* Set a value that is different from the original, then restore the original
                 * in order to trigger VaM's internal sync
                 */
                var paramJsf = breastPhysicsMesh.GetFloatJSONParam(name);
                float original = paramJsf.val;
                paramJsf.valNoCallback = Mathf.Abs(paramJsf.val - paramJsf.min) > 0.01f
                    ? paramJsf.min
                    : paramJsf.max;
                paramJsf.val = original;
            }

            foreach(var kvp in _allSoftJoints)
            {
                foreach(var softJoint in kvp.Value)
                {
                    softJoint.RestoreOriginalFriction();
                }
            }
        }

        public static void Destroy()
        {
            Object.Destroy(colliderVisualizer);
            colliderVisualizer = null;
            softJointDiagnostics?.Destroy();
            softJointDiagnostics = null;
            showColliderPreviewsJsb = null;

            _allSoftJoints = null;
            outerSoftJoints = null;
            mainSoftJoints = null;
            areolaSoftJoints = null;
            nippleSoftJoints = null;

            _gradientGroups = null;
            _nonGradientJoints = null;
            _softVerticesGroups = null;

            parameterGroups = null;
            softParameterGroups = null;

            fatSpring = null;
            fatDamper = null;
            fatMass = null;
            fatColliderRadius = null;
            fatColliderDepth = null;
            fatDistanceLimit = null;

            fatBackForce = null;
            fatBackForceMaxForce = null;
            fatBackForceThreshold = null;

            softJointsUpdater = null;
            breastSoftPhysicsOnJsb = null;
            breastSoftPhysicsSelfCollideJsb = null;
            ldpEnabledJsb = null;
            showColliderPreviewsJsb = null;
        }
    }
}
