using System;
using System.Collections.Generic;
using System.Linq;
using System.Text.RegularExpressions;
using Naturalis.Common;

namespace Naturalis.TM
{
    sealed class Integration
    {
        public static Integration instance { get; private set; } = new Integration();
        readonly List<JSONStorable> _tittyMagicInstances = new List<JSONStorable>();
        readonly List<JSONStorable> _bootyMagicInstances = new List<JSONStorable>();

        public void Init()
        {
            TittyMagic.script.NewJSONStorableAction(Constant.INTERNAL_INTEGRATE, Integrate);
            Integrate();

            /* When the plugin is added to an existing atom and this method gets called during initialization,
             * other instances are told to update their knowledge on what other instances are in the network.
             */
            foreach(var pluginInstance in _tittyMagicInstances.Concat(_bootyMagicInstances).ToList())
            {
                pluginInstance.CallActionNullSafe(Constant.INTERNAL_INTEGRATE);
            }

            /* When the plugin is added as part of a new atom using e.g. scene or subscene merge load,
             * AddInstance adds the instance to this plugin's list of other instances.
             */
            SuperController.singleton.onAtomAddedHandlers += OnAtomAdded;
            SuperController.singleton.onAtomRemovedHandlers += OnAtomRemoved;
        }

        void Integrate()
        {
            _tittyMagicInstances.Prune();
            _bootyMagicInstances.Prune();
            if(Settings.bootyMagic)
            {
                _bootyMagicInstances.Add(Settings.bootyMagic);
            }

            var otherPersons = Utils.GetAtomsByType("Person").Where(atom => atom.uid != Person.GetId()).ToList();
            foreach(var atom in otherPersons)
            {
                FindAndAddOtherPluginStorables(atom);
            }
        }

        void FindAndAddOtherPluginStorables(Atom atom)
        {
            /* Find TittyMagic instance */
            try
            {
                var regex = new Regex($@"^plugin#\d+_{nameof(Naturalis)}.{Constant.TITTY_MAGIC}$", RegexOptions.Compiled);
                var otherPlugin = atom.FindStorablesByRegexMatch(regex).Find(storable => MinVersionStorableValue(storable, "1.0"));
                if(otherPlugin &&
                    !_tittyMagicInstances.Contains(otherPlugin) &&
                    !_tittyMagicInstances.Exists(pluginInstance => pluginInstance.containingAtom.uid == atom.uid))
                {
                    _tittyMagicInstances.Add(otherPlugin);
                }
            }
            catch(Exception e)
            {
                TittyMagic.logger.LogError($"Error finding {Constant.TITTY_MAGIC} instance on Atom {atom.uid}: {e}");
            }

            /* Find BootyMagic instance */
            try
            {
                var regex = new Regex($@"^plugin#\d+_{nameof(Naturalis)}.BootyMagic$", RegexOptions.Compiled);
                var otherPlugin = atom.FindStorablesByRegexMatch(regex).Find(storable => MinVersionStorableValue(storable, "1.0"));
                if(otherPlugin &&
                    !_bootyMagicInstances.Contains(otherPlugin) &&
                    !_bootyMagicInstances.Exists(pluginInstance => pluginInstance.containingAtom.uid == atom.uid))
                {
                    _bootyMagicInstances.Add(otherPlugin);
                }
            }
            catch(Exception e)
            {
                TittyMagic.logger.LogError($"Error finding BootyMagic instance on Atom {atom.uid}: {e}");
            }
        }

        static bool MinVersionStorableValue(JSONStorable storable, string ver)
        {
            if(storable.IsStringJSONParam(Constant.NATURALIS_VERSION))
            {
                string version = storable.GetStringParamValue(Constant.NATURALIS_VERSION);
                bool match = new SemVer(version, true).IsGte(ver);
                return Settings.envIsDevelopment ? match || version == "0.0.0" : match;
            }

            return false;
        }

        public bool OtherCalibrationInProgress()
        {
            try
            {
                if(
                    _tittyMagicInstances.Prune()
                    .Concat(_bootyMagicInstances.Prune())
                    .Any(instance => instance.GetBoolParamValueNullSafe(Constant.INTERNAL_IS_CALIBRATING))
                )
                {
                    return true;
                }
            }
            catch(Exception e)
            {
                TittyMagic.logger.LogError($"{Person.GetId()}: Error checking if other instances are calibrating: {e}");
            }

            return false;
        }

        void OnAtomAdded(Atom atom)
        {
            _tittyMagicInstances.Prune();
            _bootyMagicInstances.Prune();
            FindAndAddOtherPluginStorables(atom);
        }

        void OnAtomRemoved(Atom atom)
        {
            _tittyMagicInstances.Prune().RemoveAll(storable => storable.containingAtom.uid == atom.uid);
            _bootyMagicInstances.Prune().RemoveAll(storable => storable.containingAtom.uid == atom.uid);
        }

        void RemoveAtomHandlers()
        {
            SuperController.singleton.onAtomAddedHandlers -= OnAtomAdded;
            SuperController.singleton.onAtomRemovedHandlers -= OnAtomRemoved;
        }

        public static void Destroy()
        {
            instance.RemoveAtomHandlers();
            /* ensure GC */
            instance = null;
        }

        // ReSharper disable once UnusedMember.Global
        public new string ToString()
        {
            return $"TittyMagics:\n{string.Join("\n", _tittyMagicInstances.Select(x => $"  {InstanceString(x)}").ToArray())}" +
                $"\nBootyMagics:\n{string.Join("\n", _bootyMagicInstances.Select(x => $"  {InstanceString(x)}").ToArray())}";
        }

        static string InstanceString(JSONStorable instance)
        {
            return instance ? $"{instance.containingAtom.uid}: {instance.storeId}" : "null";
        }
    }
}
