using System;
using System.Linq;
using Naturalis.Common;
using Naturalis.TM.Handlers;
using SimpleJSON;
using UnityEngine;

namespace Naturalis.TM.Components
{
    sealed class Breast
    {
        readonly Rigidbody _chestRb;
        readonly Rigidbody _nippleRb;
        readonly Rigidbody _pectoralRb;

        readonly Rigidbody[] _softJointCenterRbs;
        readonly Rigidbody[] _softJointMainRbs;
        readonly Rigidbody[] _softJointOuterRbs;

        public Breast(string side)
        {
            _chestRb = Person.instance.chestRb;
            _pectoralRb = Person.instance.pectoralRbs[side];
            if(Person.IsFemale())
            {
                _nippleRb = Person.instance.nippleRbs[side];
                _softJointCenterRbs = SoftPhysicsHandler.GetPositionalRigidbodies(side, SoftVerticesGroupName.AREOLA);
                _softJointMainRbs = SoftPhysicsHandler.GetPositionalRigidbodies(side, SoftVerticesGroupName.MAIN);
                _softJointOuterRbs = SoftPhysicsHandler.GetPositionalRigidbodies(side, SoftVerticesGroupName.OUTER);
            }
            else
            {
                /* Position of nipple skin vertices relative to chest */
                _centerVertexIndices = VertexIndices.NippleSkinVerts(side);
                _calculateRelativePosition = SkinVerticesRelativePosition;
                _calculateRelativeDepth = () => Calc.RelativePosition(_chestRb, _pectoralRb.position).z;
            }
        }

        Func<Vector3> _calculateRelativePosition;
        Func<float> _calculateRelativeDepth;

        public void SetSoftPhysicsBased(bool softPhysicsOn)
        {
            if(softPhysicsOn)
            {
                var combinedRigidbodies = _softJointMainRbs.Concat(_softJointCenterRbs).ToArray();
                _calculateRelativePosition = () => Calc.RelativePosition(_chestRb, Calc.AverageRigidbodiesPosition(combinedRigidbodies));
                _calculateRelativeDepth = () =>
                {
                    var jointsRelativePosition = Calc.RelativePosition(_chestRb, _pectoralRb.position);
                    var centerToPectoral = Calc.RelativePosition(_pectoralRb, Calc.AverageRigidbodiesPosition(_softJointCenterRbs));
                    var outerToPectoral = Calc.RelativePosition(_pectoralRb, Calc.AverageRigidbodiesPosition(_softJointOuterRbs));
                    float softJointsDepth = (centerToPectoral - outerToPectoral).z;
                    return (jointsRelativePosition.z + softJointsDepth) / 2;
                };
            }
            else
            {
                _calculateRelativePosition = () => Calc.RelativePosition(_chestRb, _nippleRb.position);
                _calculateRelativeDepth = () => Calc.RelativePosition(_chestRb, _pectoralRb.position).z;
            }
        }

        readonly int[] _centerVertexIndices;

        Vector3 SkinVerticesRelativePosition()
        {
            var rawSkinnedVerts = Person.instance.skin.rawSkinnedVerts;
            var vertices = new Vector3[_centerVertexIndices.Length];
            for(int i = 0; i < _centerVertexIndices.Length; i++)
            {
                vertices[i] = rawSkinnedVerts[_centerVertexIndices[i]];
            }

            return Calc.RelativePosition(_chestRb, Calc.AveragePosition(vertices));
        }

        float _prevXAngle;
        float _prevYAngle;

        public bool Moved(float tolerance)
        {
            float xAngle = CalculateXAngle();
            float yAngle = CalculateYAngle();
            bool moved = Mathf.Abs(xAngle - _prevXAngle) > tolerance || Mathf.Abs(yAngle - _prevYAngle) > tolerance;
            _prevXAngle = xAngle;
            _prevYAngle = yAngle;
            return moved;
        }

        Vector3 _positionRelativeToOrigin;
        public float angleY { get; private set; }
        public float angleX { get; private set; }
        public float depthDiff { get; private set; }
        float _depthBase;
        Vector2 _angleXBaseVector;
        Vector2 _angleYBaseVector;

        public void Calibrate()
        {
            _positionRelativeToOrigin = _calculateRelativePosition();
            _angleXBaseVector = new Vector2(_positionRelativeToOrigin.z, _positionRelativeToOrigin.x);
            _angleYBaseVector = new Vector2(_positionRelativeToOrigin.z, _positionRelativeToOrigin.y);
            _depthBase = _calculateRelativeDepth();
        }

        public void Measure()
        {
            _positionRelativeToOrigin = _calculateRelativePosition();
            angleX = CalculateXAngle();
            angleY = CalculateYAngle();
            depthDiff = _depthBase - _calculateRelativeDepth();
        }

        float CalculateXAngle()
        {
            return Vector2.SignedAngle(_angleXBaseVector, new Vector2(_positionRelativeToOrigin.z, _positionRelativeToOrigin.x));
        }

        float CalculateYAngle()
        {
            return Vector2.SignedAngle(_angleYBaseVector, new Vector2(_positionRelativeToOrigin.z, _positionRelativeToOrigin.y));
        }

        public void ResetMeasurements()
        {
            angleY = 0;
            angleX = 0;
            depthDiff = 0;
        }

        public JSONClass SerializeCalibrationData()
        {
            var jc = new JSONClass
            {
                ["angle_x_base_vector"] = JSONUtils.SerializeVector2(_angleXBaseVector),
                ["angle_y_base_vector"] = JSONUtils.SerializeVector2(_angleYBaseVector),
                ["depth_base"] = { AsFloat = _depthBase },
            };
            return jc;
        }

        public void CalibrateFromJson(JSONClass jc)
        {
            _angleXBaseVector = JSONUtils.DeserializeVector2(jc["angle_x_base_vector"]);
            _angleYBaseVector = JSONUtils.DeserializeVector2(jc["angle_y_base_vector"]);
            _depthBase = jc["depth_base"].AsFloat;
        }
    }
}
