using Naturalis.Common.Components;
using UnityEngine;

namespace Naturalis.TM.Models
{
    sealed class HardColliderGroup
    {
        const float DEFAULT_MASS = 0.04f; // Seems to be a hard coded value in VaM.

        public string id { get; }

        public string GetJsonId()
        {
            return id.Replace(" ", "_").ToLower();
        }

        public string visualizerId { get; }

        readonly Scaler _baseRbMassScaler;
        readonly Scaler _radiusScaler;
        readonly Scaler _rightOffsetScaler;
        readonly Scaler _upOffsetScaler;
        readonly Scaler _lookOffsetScaler;

        readonly float _frictionMultiplier;

        public HardCollider left { get; }

        // ReSharper disable once MemberCanBePrivate.Global
        public HardCollider right { get; }

        public JSONStorableFloat massMultiplierJsf { get; set; }
        public JSONStorableFloat radiusJsf { get; set; }
        public JSONStorableFloat rightJsf { get; set; }
        public JSONStorableFloat upJsf { get; set; }
        public JSONStorableFloat lookJsf { get; set; }

        public HardColliderGroup(
            string id,
            string visualizerId,
            HardCollider left,
            HardCollider right,
            Scaler baseRbMassScaler,
            Scaler radiusScaler,
            Scaler rightOffsetScaler,
            Scaler upOffsetScaler,
            Scaler lookOffsetScaler,
            float frictionMultiplier
        )
        {
            this.id = id;
            this.visualizerId = visualizerId;
            this.left = left;
            this.right = right;
            _baseRbMassScaler = baseRbMassScaler;
            _radiusScaler = radiusScaler;
            _rightOffsetScaler = rightOffsetScaler;
            _upOffsetScaler = upOffsetScaler;
            _lookOffsetScaler = lookOffsetScaler;
            _frictionMultiplier = frictionMultiplier;
        }

        public void UpdateRigidbodyMass(float combinedMultiplier, float mass, float softness, bool softPhysicsOn)
        {
            float rbMass = softPhysicsOn
                ? combinedMultiplier * _baseRbMassScaler.Scale(DEFAULT_MASS, mass, softness)
                : combinedMultiplier * 0.125f * _baseRbMassScaler.Scale(DEFAULT_MASS, mass);
            left.UpdateRigidbodyMass(rbMass);
            right.UpdateRigidbodyMass(rbMass);
        }

        public void UpdateDimensions(float mass)
        {
            float radius = -_radiusScaler.Scale(radiusJsf.val, mass, TittyMagic.softness);
            left.UpdateDimensions(radius);
            right.UpdateDimensions(radius);
        }

        public void UpdatePosition(float mass)
        {
            float softness = TittyMagic.softness;
            float rightOffset = _rightOffsetScaler.Scale(rightJsf.val, mass, softness);
            float upOffset = -_upOffsetScaler.Scale(upJsf.val, mass, softness);
            float lookOffset = -_lookOffsetScaler.Scale(lookJsf.val, mass, softness);
            left.UpdatePosition(-rightOffset, upOffset, lookOffset);
            right.UpdatePosition(rightOffset, upOffset, lookOffset);
        }

        public void UpdateMaxFrictionalDistance(float sizeMultiplierLeft, float sizeMultiplierRight, float smallestRadius)
        {
            float halfSqrSmallestRadiusMillimeters = 0.5f * smallestRadius * smallestRadius * 1000;
            left.maxFrictionalDistance = sizeMultiplierLeft * halfSqrSmallestRadiusMillimeters;
            right.maxFrictionalDistance = sizeMultiplierRight * halfSqrSmallestRadiusMillimeters;
        }

        public void UpdateFriction(float max)
        {
            left.UpdateFriction(_frictionMultiplier, max);
            right.UpdateFriction(_frictionMultiplier, max);
        }

        public void AutoColliderSizeSet()
        {
            left.AutoColliderSizeSet();
            right.AutoColliderSizeSet();
        }

        public void Calibrate(Vector3 breastCenterLeft, Vector3 breastCenterRight, Rigidbody chestRb)
        {
            left.Calibrate(breastCenterLeft, chestRb);
            right.Calibrate(breastCenterRight, chestRb);
        }

        public void UpdateDistanceDiffs(Vector3 breastCenterLeft, Vector3 breastCenterRight, Rigidbody chestRb)
        {
            left.UpdateDistanceDiff(breastCenterLeft, chestRb);
            right.UpdateDistanceDiff(breastCenterRight, chestRb);
        }

        public void ResetDistanceDiffs()
        {
            left.ResetDistanceDiff();
            right.ResetDistanceDiff();
        }

        public void EnableMultiplyFriction()
        {
            left.EnableMultiplyFriction();
            right.EnableMultiplyFriction();
        }

        public void RestoreDefaults()
        {
            left.RestoreDefaults(DEFAULT_MASS);
            right.RestoreDefaults(DEFAULT_MASS);
        }

        public bool HasRigidbodies()
        {
            return left.collider.attachedRigidbody && right.collider.attachedRigidbody;
        }
    }
}
