using System;
using Naturalis.TM.Handlers;
using SimpleJSON;
using UnityEngine;

namespace Naturalis.TM.Models
{
    sealed class SoftJoint
    {
        public string id { get; set; }
        public DAZPhysicsMeshSoftVerticesSet softVerticesSet { get; }
        readonly ConfigurableJoint _joint;
        public Rigidbody jointRb { get; }
        readonly bool _enableSyncLinkParams;

        const float JOINT_SPRING_MAX_FORCE = 1E+07f;

        public static float minSpring { private get; set; }
        public static float maxSpring { private get; set; }
        public static float minDamper { private get; set; }
        public static float maxDamper { private get; set; }
        public static float minMass { private get; set; }
        public static float maxMass { private get; set; }
        public static float minColliderRadius { private get; set; }
        public static float maxColliderRadius { private get; set; }
        public static float minColliderDepth { private get; set; }
        public static float maxColliderDepth { private get; set; }
        public static float minDistanceLimit { private get; set; }
        public static float maxDistanceLimit { private get; set; }

        SoftJoint[] _nearbyJoints;
        float _nearbyJointsLength;

        public SoftJoint[] nearbyJoints
        {
            get { return _nearbyJoints; }
            set
            {
                _nearbyJoints = value;
                _nearbyJointsLength = _nearbyJoints.Length;
            }
        }

        public SoftJoint[] adjacentJoints { get; set; }

        public SoftJoint(DAZPhysicsMeshSoftVerticesSet softVerticesSet)
        {
            _joint = softVerticesSet.joint;
            _joint.xMotion = ConfigurableJointMotion.Limited;
            _joint.yMotion = ConfigurableJointMotion.Limited;
            _joint.zMotion = ConfigurableJointMotion.Limited;
            _joint.angularXMotion = ConfigurableJointMotion.Locked;
            _joint.angularYMotion = ConfigurableJointMotion.Locked;
            _joint.angularZMotion = ConfigurableJointMotion.Locked;

            this.softVerticesSet = softVerticesSet;
            _enableSyncLinkParams = softVerticesSet.linkJoints != null;

            jointRb = softVerticesSet.jointRB;
        }

        // values with user offset
        float _baseSpring;
        float _baseDamper;
        float _baseMass;
        float _baseColliderRadius;
        float _baseColliderDepth;
        float _baseDistanceLimit;

        public float GetBaseValue(string paramName)
        {
            switch(paramName)
            {
                case FAT_SPRING.NAME: return _baseSpring;
                case FAT_DAMPER.NAME: return _baseDamper;
                case FAT_MASS.NAME: return _baseMass;
                case FAT_COLLIDER_RADIUS.NAME: return _baseColliderRadius;
                case FAT_COLLIDER_DEPTH.NAME: return _baseColliderDepth;
                case FAT_DISTANCE_LIMIT.NAME: return _baseDistanceLimit;
                default: throw new Exception($"Invalid parameter name {paramName}");
            }
        }

        public void SetBaseValue(string paramName, float value)
        {
            if(float.IsNaN(value))
            {
                Debug.LogError($"TM {paramName} value is NaN");
                return;
            }

            switch(paramName)
            {
                case FAT_SPRING.NAME:
                    _baseSpring = value;
                    break;
                case FAT_DAMPER.NAME:
                    _baseDamper = value;
                    break;
                case FAT_MASS.NAME:
                    _baseMass = value;
                    break;
                case FAT_COLLIDER_RADIUS.NAME:
                    _baseColliderRadius = value;
                    break;
                case FAT_COLLIDER_DEPTH.NAME:
                    _baseColliderDepth = value;
                    break;
                case FAT_DISTANCE_LIMIT.NAME:
                    _baseDistanceLimit = value;
                    break;
                default: throw new Exception($"Invalid parameter name {paramName}");
            }
        }

        // offsets consisting of any additive gravity physics, force physics, friction based or other adjustments
        float _springOffset;
        float _damperOffset;
        float _massOffset;
        float _colliderRadiusOffset;
        float _colliderDepthOffset;
        float _distanceLimitOffset;

        // ReSharper disable once UnusedMember.Global
        public float GetOffsetValue(string paramName)
        {
            switch(paramName)
            {
                case FAT_SPRING.NAME: return _springOffset;
                case FAT_DAMPER.NAME: return _damperOffset;
                case FAT_MASS.NAME: return _massOffset;
                case FAT_COLLIDER_RADIUS.NAME: return _colliderRadiusOffset;
                case FAT_COLLIDER_DEPTH.NAME: return _colliderDepthOffset;
                case FAT_DISTANCE_LIMIT.NAME: return _distanceLimitOffset;
                default: throw new Exception($"Invalid parameter name {paramName}");
            }
        }

        public void SetAdditionalOffset(string paramName, float offset)
        {
            switch(paramName)
            {
                case FAT_SPRING.NAME:
                    _springOffset = offset;
                    SetRealValue(paramName);
                    break;
                case FAT_DAMPER.NAME:
                    _damperOffset = offset;
                    SetRealValue(paramName);
                    break;
                case FAT_MASS.NAME:
                    _massOffset = offset;
                    SetRealValue(paramName);
                    break;
                case FAT_COLLIDER_RADIUS.NAME:
                    _colliderRadiusOffset = offset;
                    SetRealValue(paramName);
                    break;
                case FAT_COLLIDER_DEPTH.NAME:
                    _colliderDepthOffset = offset;
                    SetRealValue(paramName);
                    break;
                case FAT_DISTANCE_LIMIT.NAME:
                    _distanceLimitOffset = offset;
                    SetRealValue(paramName);
                    break;
                default: throw new Exception($"Invalid parameter name {paramName}");
            }
        }

        // additional multipliers (used for nipple erection physics)
        float _additionalSpringMulti = 1f;
        float _additionalDamperMulti = 1f;

        public void SetAdditionalBaseMultiplier(string paramName, float value)
        {
            switch(paramName)
            {
                case FAT_SPRING.NAME:
                    _additionalSpringMulti = value;
                    SetRealValue(paramName);
                    break;
                case FAT_DAMPER.NAME:
                    _additionalDamperMulti = value;
                    SetRealValue(paramName);
                    break;
                default: throw new Exception($"Invalid parameter name {paramName}");
            }
        }

        // calculated values (not including real time distortion adjustments)
        float _spring;
        float _damper;
        float _mass;
        float _colliderRadius;
        float _colliderDepth;
        float _distanceLimit;

        // ReSharper disable once UnusedMember.Global
        public float GetCalculatedValue(string paramName)
        {
            switch(paramName)
            {
                case FAT_SPRING.NAME: return _spring;
                case FAT_DAMPER.NAME: return _damper;
                case FAT_MASS.NAME: return _mass;
                case FAT_COLLIDER_RADIUS.NAME: return _colliderRadius;
                case FAT_COLLIDER_DEPTH.NAME: return _colliderDepth;
                case FAT_DISTANCE_LIMIT.NAME: return _distanceLimit;
                default: throw new Exception($"Invalid parameter name {paramName}");
            }
        }

        // real-time values
        float _realSpring;
        float _realDamper;
        float _realMass;
        float _realColliderRadius;
        float _realColliderDepth;
        float _realDistanceLimit;

        public float GetRealValue(string paramName)
        {
            switch(paramName)
            {
                case FAT_SPRING.NAME:
                    return _realSpring;
                case FAT_DAMPER.NAME:
                    return _realDamper;
                case FAT_MASS.NAME:
                    return _realMass;
                case FAT_COLLIDER_RADIUS.NAME:
                    return _realColliderRadius;
                case FAT_COLLIDER_DEPTH.NAME:
                    return _realColliderDepth;
                case FAT_DISTANCE_LIMIT.NAME:
                    return _realDistanceLimit;
                default: throw new Exception($"Invalid parameter name {paramName}");
            }
        }

        void SetRealValue(string paramName)
        {
            switch(paramName)
            {
                case FAT_SPRING.NAME:
                    SetSpring();
                    break;
                case FAT_DAMPER.NAME:
                    SetDamper();
                    break;
                case FAT_MASS.NAME:
                    SetMass();
                    break;
                case FAT_COLLIDER_RADIUS.NAME:
                    SetColliderRadius();
                    break;
                case FAT_COLLIDER_DEPTH.NAME:
                    SetColliderDepth();
                    break;
                case FAT_DISTANCE_LIMIT.NAME:
                    SetJointLimits();
                    break;
                default: throw new Exception($"Invalid parameter name {paramName}");
            }
        }

        void SetSpring()
        {
            _spring = Mathf.Clamp(
                _additionalSpringMulti * _baseSpring + _springOffset,
                minSpring,
                maxSpring
            );
            _realSpring = _spring;
            float damper = _realDamper == 0 ? 1 : _realDamper;
            SetJointDrive(_realSpring, damper);

            if(_enableSyncLinkParams)
            {
                SyncLinkParams(_realSpring, damper);
            }
        }

        void SetDamper()
        {
            _damper = Mathf.Clamp(
                _additionalDamperMulti * _baseDamper + _damperOffset,
                minDamper,
                maxDamper
            );
            _realDamper = _damper;

            float spring = _realSpring == 0 ? 100 : _realSpring;
            SetJointDrive(spring, _realDamper);

            if(_enableSyncLinkParams)
            {
                SyncLinkParams(spring, _realDamper);
            }
        }

        void SetJointDrive(float spring, float damper)
        {
            _joint.xDrive = new JointDrive
            {
                positionSpring = spring,
                positionDamper = damper,
                maximumForce = JOINT_SPRING_MAX_FORCE,
            };
            _joint.yDrive = new JointDrive
            {
                positionSpring = spring,
                positionDamper = damper,
                maximumForce = JOINT_SPRING_MAX_FORCE,
            };
            _joint.zDrive = new JointDrive
            {
                positionSpring = spring,
                positionDamper = damper,
                maximumForce = JOINT_SPRING_MAX_FORCE,
            };
        }

        // Not sure what the link joints are, exactly
        void SyncLinkParams(float spring, float damper)
        {
            foreach(var linkJoint in softVerticesSet.linkJoints)
            {
                if(linkJoint)
                {
                    linkJoint.spring = spring;
                    linkJoint.damper = damper;
                }
            }
        }

        void SetMass()
        {
            _mass = Mathf.Clamp(_baseMass + _massOffset, minMass, maxMass);
            _realMass = _mass;
            jointRb.mass = _realMass;
        }

        void SetColliderRadius()
        {
            _colliderRadius = Mathf.Clamp(_baseColliderRadius + _colliderRadiusOffset, minColliderRadius, maxColliderRadius);
            _realColliderRadius = _colliderRadius;
            SetColliders();
        }

        void SetColliderDepth()
        {
            _colliderDepth = Mathf.Clamp(_baseColliderDepth + _colliderDepthOffset, minColliderDepth, maxColliderDepth);
            _realColliderDepth = _colliderDepth;
            SetColliders();
        }

        // Ignores second collider of nipple joint
        void SetColliders()
        {
            Vector3 colliderCenter;

            // lookAtOption != DAZPhysicsMeshSoftVerticesGroup.LookAtOption.Anchor
            colliderCenter.x = 0;
            colliderCenter.y = 0;
            colliderCenter.z = _realColliderRadius + _realColliderDepth;

            var jointCollider2 = (SphereCollider) softVerticesSet.jointCollider;
            jointCollider2.radius = _realColliderRadius * softVerticesSet.sizeMultiplier;
            jointCollider2.center = colliderCenter * softVerticesSet.sizeMultiplier;
        }

        void SetJointLimits()
        {
            _distanceLimit = Mathf.Clamp(_baseDistanceLimit + _distanceLimitOffset, minDistanceLimit, maxDistanceLimit);
            _realDistanceLimit = _distanceLimit;
            var linearLimit = _joint.linearLimit;
            linearLimit.limit = _realDistanceLimit;
            _joint.linearLimit = linearLimit;
        }

        Vector3 _prevPosition;

        public bool Moved(float tolerance)
        {
            var position = jointRb.position;
            bool moved = (position - _prevPosition).magnitude > tolerance;
            _prevPosition = position;
            return moved;
        }

        /* A combination of the average distance of this joint from its nearby joints
         * and the distance of this joint from the furthest of its nearby joints.
         */
        float _distortionBaseUp;
        float _distortionBaseDown;
        float _distortionBaseForward;
        float _distortionBaseBack;
        float _distortionBaseLeft;
        float _distortionBaseRight;

        public void CalibrateDistortion(string direction)
        {
            float result = CalculateDistances();
            switch(direction)
            {
                case Direction.UP:
                    _distortionBaseUp = result;
                    break;
                case Direction.DOWN:
                    _distortionBaseDown = result;
                    break;
                case Direction.FORWARD:
                    _distortionBaseForward = result;
                    break;
                case Direction.BACK:
                    _distortionBaseBack = result;
                    break;
                case Direction.LEFT:
                    _distortionBaseLeft = result;
                    break;
                case Direction.RIGHT:
                    _distortionBaseRight = result;
                    break;
                default: throw new ArgumentException($"Invalid direction {direction}");
            }
        }

        float _springDistortion;
        float _damperDistortion;
        float _massDistortion;
        float _colliderRadiusDistortion;
        float _colliderDepthDistortion;
        float _distanceLimitDistortion;

        public float GetDistortionValue(string paramName)
        {
            switch(paramName)
            {
                case FAT_SPRING.NAME: return _springDistortion;
                case FAT_DAMPER.NAME: return _damperDistortion;
                case FAT_MASS.NAME: return _massDistortion;
                case FAT_COLLIDER_RADIUS.NAME: return _colliderRadiusDistortion;
                case FAT_COLLIDER_DEPTH.NAME: return _colliderDepthDistortion;
                case FAT_DISTANCE_LIMIT.NAME: return _distanceLimitDistortion;
                default: throw new Exception($"Invalid parameter name {paramName}");
            }
        }

        float _distortionBase;
        public float distortion { get; private set; }
        float _prevAbsDistortion;

        public bool UpdateDistortion(bool force = false)
        {
            _distortionBase = CalculateDistortionBase();
            distortion = CalculateDistances() - _distortionBase;
            float absDistortion = Mathf.Abs(distortion);
            if(force || Mathf.Abs(absDistortion - _prevAbsDistortion) > 0.01f) // more than 0.1 mm
            {
                _prevAbsDistortion = absDistortion;
                return true;
            }

            return false;
        }

        public void SyncSpringAndDamperDistortion(string side)
        {
            _springDistortion = Mathf.Clamp(
                SoftPhysicsHandler.fatSpring.CalculateDistortionOffset(
                    side,
                    _distortionBase,
                    distortion
                ),
                minSpring - _spring,
                maxSpring - _spring
            );

            _damperDistortion = Mathf.Clamp(
                SoftPhysicsHandler.fatDamper.CalculateDistortionOffset(
                    side,
                    _distortionBase,
                    distortion
                ),
                minDamper - _damper,
                maxDamper - _damper
            );

            _realSpring = _spring + _springDistortion;
            _realDamper = _damper + _damperDistortion;
            SetJointDrive(_realSpring, _realDamper);
            if(_enableSyncLinkParams)
            {
                SyncLinkParams(_realSpring, _realDamper);
            }
        }

        public void SyncMassDistortion(string side)
        {
            _massDistortion = Mathf.Clamp(
                SoftPhysicsHandler.fatMass.CalculateDistortionOffset(
                    side,
                    _distortionBase,
                    distortion
                ),
                minMass - _mass,
                maxMass - _mass
            );

            _realMass = _mass + _massDistortion;
            jointRb.mass = _realMass;
        }

        public void SyncColliderRadiusAndDepthDistortion(string side)
        {
            _colliderRadiusDistortion = Mathf.Clamp(
                SoftPhysicsHandler.fatColliderRadius.CalculateDistortionOffset(
                    side,
                    _distortionBase,
                    distortion
                ),
                minColliderRadius - _colliderRadius,
                maxColliderRadius - _colliderRadius
            );

            _colliderDepthDistortion = Mathf.Clamp(
                SoftPhysicsHandler.fatColliderDepth.CalculateDistortionOffset(
                    side,
                    _distortionBase,
                    distortion
                ),
                minColliderDepth - _colliderDepth,
                maxColliderDepth - _colliderDepth
            );

            _realColliderRadius = _colliderRadius + _colliderRadiusDistortion;
            _realColliderDepth = _colliderDepth + _colliderDepthDistortion;
            SetColliders();
        }

        public void SyncDistanceLimitDistortion(string side)
        {
            _distanceLimitDistortion = Mathf.Clamp(
                SoftPhysicsHandler.fatDistanceLimit.CalculateDistortionOffset(
                    side,
                    _distortionBase,
                    distortion
                ),
                minDistanceLimit - _distanceLimit,
                maxDistanceLimit - _distanceLimit
            );

            _realDistanceLimit = _distanceLimit + _distanceLimitDistortion;
            var linearLimit = _joint.linearLimit;
            linearLimit.limit = _realDistanceLimit;
            _joint.linearLimit = linearLimit;
        }

        public void ResetDistortion()
        {
            distortion = 0;
        }

        float CalculateDistortionBase()
        {
            float roll = GravityEffectCalc.chestRoll;
            float pitch = GravityEffectCalc.chestPitch2X;
            float distortionBasePitch;
            if(pitch > 0) // leaning forward
            {
                distortionBasePitch = pitch < 1
                    ? Mathf.Lerp(_distortionBaseDown, _distortionBaseForward, pitch) // upright
                    : Mathf.Lerp(_distortionBaseForward, _distortionBaseUp, pitch - 1); // upside down
            }
            else // leaning back
            {
                distortionBasePitch = pitch > -1
                    ? Mathf.Lerp(_distortionBaseDown, _distortionBaseBack, -pitch) // upright
                    : Mathf.Lerp(_distortionBaseBack, _distortionBaseUp, -pitch - 1); // upside down
            }

            return roll > 0
                ? Mathf.Lerp(distortionBasePitch, _distortionBaseLeft, roll) // leaning left
                : Mathf.Lerp(distortionBasePitch, _distortionBaseRight, -roll); // leaning right
        }

        /* Average of the maximum distance to a nearby joint and the average of distances to every nearby joint. Result in centimeters. */
        float CalculateDistances()
        {
            float maxDistance = 0;
            float sumDistances = 0;
            for(int i = 0; i < _nearbyJointsLength; i++)
            {
                float current = 100f * Vector3.Distance(jointRb.position, nearbyJoints[i].jointRb.position);
                if(current - maxDistance > 0.01f)
                {
                    maxDistance = current;
                }

                sumDistances += current;
            }

            return (maxDistance + sumDistances / _nearbyJointsLength) / 2;
        }

        public void EnableMultiplyFriction()
        {
            var material = softVerticesSet.jointCollider.material;
            material.frictionCombine = PhysicMaterialCombine.Multiply;
        }

        public JSONClass SerializeCalibrationData()
        {
            var jc = new JSONClass
            {
                ["dist_up"] = { AsFloat = _distortionBaseUp },
                ["dist_down"] = { AsFloat = _distortionBaseDown },
                ["dist_forward"] = { AsFloat = _distortionBaseForward },
                ["dist_back"] = { AsFloat = _distortionBaseBack },
                ["dist_left"] = { AsFloat = _distortionBaseLeft },
                ["dist_right"] = { AsFloat = _distortionBaseRight },
            };
            return jc;
        }

        public void CalibrateFromJson(JSONClass jc)
        {
            _distortionBaseUp = jc["dist_up"].AsFloat;
            _distortionBaseDown = jc["dist_down"].AsFloat;
            _distortionBaseForward = jc["dist_forward"].AsFloat;
            _distortionBaseBack = jc["dist_back"].AsFloat;
            _distortionBaseLeft = jc["dist_left"].AsFloat;
            _distortionBaseRight = jc["dist_right"].AsFloat;
        }

        public void RestoreOriginalFriction()
        {
            var material = softVerticesSet.jointCollider.material;
            material.dynamicFriction = 0.6f;
            material.staticFriction = 0.6f;
            material.frictionCombine = PhysicMaterialCombine.Average;
        }
    }
}
