using Naturalis.Common;
using Naturalis.Common.UI;
using Naturalis.TM.Handlers;
using UnityEngine;

namespace Naturalis.TM.UI
{
    sealed class MainWindow : WindowBase
    {
        JSONStorableAction configureColliderFrictionAction { get; }
        JSONStorableAction openSettingsWindowAction { get; }
        public JSONStorableAction openDevWindowAction { get; }
        public JSONStorableAction openDevMorphWindowAction { get; }

        public MainWindow() : base(TittyMagic.script, nameof(MainWindow))
        {
            openSettingsWindowAction = new JSONStorableAction(
                "openSettings",
                () =>
                {
                    ClearSelf();
                    activeNestedWindow = nestedWindows.Find(window => window.GetId() == nameof(SettingsWindow));
                    activeNestedWindow.Rebuild();
                }
            );
            nestedWindows.Add(new SettingsWindow(TittyMagic.script, OnReturn));

            configureColliderFrictionAction = new JSONStorableAction(
                "configureColliderFriction",
                () =>
                {
                    if(!TittyMagic.script.enabled)
                    {
                        TittyMagic.logger.LogMessage("Enable the plugin to configure collider friction.");
                        return;
                    }

                    TittyMagic.bindings.actions[$"{Constant.OPEN_UI}_{Constant.FRICTION}"].actionCallback();
                }
            );

            if(Settings.envIsDevelopment)
            {
                openDevWindowAction = new JSONStorableAction(
                    "openDevWindow",
                    () =>
                    {
                        ClearSelf();
                        activeNestedWindow = nestedWindows.Find(window => window.GetId() == nameof(DevWindow));
                        activeNestedWindow.Rebuild();
                    }
                );

                openDevMorphWindowAction = new JSONStorableAction(
                    "openDevMorphWindow",
                    () =>
                    {
                        ClearSelf();
                        activeNestedWindow = nestedWindows.Find(window => window.GetId() == nameof(DevMorphWindow));
                        activeNestedWindow.Rebuild();
                    }
                );

                nestedWindows.Add(new DevWindow(OnReturn));
                nestedWindows.Add(new DevMorphWindow(OnReturn));
            }
        }

        protected override void OnBuild()
        {
            BuildLeftSide();
            BuildRightSide();
            Refresh();
        }

        void BuildLeftSide(bool rightSide = false)
        {
            AddSpacer(5, rightSide);
            AddElement(() =>
            {
                var slider = script.CreateSlider(TittyMagic.script.softnessJsf, rightSide);
                slider.valueFormat = "F0";
                return slider;
            });

            AddInfoTextField(
                "Softness, quickness and mass affect physics base values, the effect of gravity," +
                " the amount of morphing, and hard collider properties. Use them to determine" +
                " the overall behavior, then fine tune physics, morphs and colliders as needed.",
                rightSide,
                180
            );

            AddElement(SoftPhysicsHandler.breastSoftPhysicsOnJsb.name, () =>
                script.CreateToggle(SoftPhysicsHandler.breastSoftPhysicsOnJsb, rightSide)
            );

            string softPhysicsText = "Simulate soft-body dynamics for breasts.";
            softPhysicsText += Person.IsFemale()
                ? $" Soft physics is required to get the most out of {Constant.TITTY_MAGIC}."
                : " Soft physics is only supported on a female character.";
            AddInfoTextField(softPhysicsText, rightSide);

            AddElement(SoftPhysicsHandler.breastSoftPhysicsSelfCollideJsb.name, () =>
            {
                var toggle = script.CreateToggle(SoftPhysicsHandler.breastSoftPhysicsSelfCollideJsb, rightSide);
                toggle.SetActiveStyle(Person.IsFemale(), true, true);
                return toggle;
            });

            AddSpacer(405, rightSide);
            AddElement(() =>
            {
                var jsUrl = TittyMagic.script.presetHandler.savePresetJsu;
                var button = script.CreateButton(jsUrl.name, rightSide);
                jsUrl.fileBrowseButton = button.button;
                return button;
            });
            AddElement(() =>
            {
                var jsUrl = TittyMagic.script.presetHandler.loadPresetJsu;
                var button = script.CreateButton(jsUrl.name, rightSide);
                jsUrl.fileBrowseButton = button.button;
                return button;
            });
        }

        void BuildRightSide(bool rightSide = true)
        {
            AddSpacer(5, rightSide);
            AddElement(() =>
            {
                var slider = script.CreateSlider(TittyMagic.script.quicknessJsf, rightSide);
                slider.valueFormat = "F2";
                return slider;
            });

            AddElement(() =>
            {
                var slider = script.CreateSlider(MainPhysicsHandler.breastMass.left.offsetJsf, rightSide);
                slider.valueFormat = MainPhysicsHandler.breastMass.valueFormat;
                return slider;
            });

            AddElement(() =>
                CreateValueTextField(MainPhysicsHandler.breastMass.left.displayValueJss, 28, 45, rightSide)
            );

            AddElement(SoftPhysicsHandler.ldpEnabledJsb.name, () =>
                script.CreateToggle(SoftPhysicsHandler.ldpEnabledJsb, rightSide)
            );

            AddInfoTextField(
                "Adjust soft physics parameters of individual soft joints dynamically with stretching and compression.",
                rightSide
            );

            AddElement(configureColliderFrictionAction.name, () =>
            {
                var button = script.CreateButton(configureColliderFrictionAction.name, rightSide);
                configureColliderFrictionAction.RegisterButton(button);
                button.label = "Configure Collider Friction";
                button.SetActiveStyle(Person.IsFemale(), true);
                return button;
            });

            AddInfoTextField(
                "Collider friction controls how easily colliders stick to or slip away from hands or other colliding objects.",
                rightSide
            );

            AddSpacer(290, rightSide);

            AddElement(() =>
            {
                var button = script.CreateButton(openSettingsWindowAction.name, rightSide);
                openSettingsWindowAction.RegisterButton(button);
                button.label = "  Naturalis User Preferences";
                return button;
            });

            AddVersionTextField();
        }

        void AddVersionTextField()
        {
            var versionJss = new JSONStorableString("version", "");
            var versionTextField = CreateVersionTextField(versionJss);
            AddElement(versionTextField);
            TittyMagic.script.AddTextFieldToJss(versionTextField, versionJss);
        }

        public void Refresh()
        {
            var breastSoftPhysicsOnToggle = GetElement(SoftPhysicsHandler.breastSoftPhysicsOnJsb.name);
            var ldpEnabledtoggle = GetElement(SoftPhysicsHandler.ldpEnabledJsb.name);

            if(Person.IsFemale())
            {
                breastSoftPhysicsOnToggle.SetActiveStyle(
                    Settings.personMonitor.BreastSoftPhysicsIsEnabled(),
                    false,
                    SoftPhysicsHandler.breastSoftPhysicsOnJsb.val && script.enabled
                );
                ldpEnabledtoggle.SetActiveStyle(
                    SoftPhysicsHandler.softPhysicsOn && Settings.personMonitor.BreastSoftPhysicsIsEnabled(),
                    false,
                    TittyMagic.script.enabled
                );
            }
            else
            {
                breastSoftPhysicsOnToggle.SetActiveStyle(false, true);
                ldpEnabledtoggle.SetActiveStyle(false, true);
            }
        }

        protected override void OnClose()
        {
            if(activeNestedWindow is SettingsWindow)
            {
                activeNestedWindow.Clear();
                activeNestedWindow = null;
            }
        }
    }
}
