using Naturalis.Common;
using Naturalis.Common.Models;
using Naturalis.Common.UI;
using Naturalis.TM.Handlers;
using UnityEngine;
using UnityEngine.Events;

namespace Naturalis.TM.UI
{
    sealed class ParameterWindow : WindowBase
    {
        readonly PhysicsParameterGroup _parameterGroup;
        readonly PhysicsParameter _parameter;

        readonly JSONStorableString _gravityTargetsJss;

        public ParameterWindow(PhysicsParameterGroup parameterGroup, UnityAction onReturnToParent)
            : base(TittyMagic.script, parameterGroup.name, onReturnToParent)
        {
            _parameterGroup = parameterGroup;
            _parameter = _parameterGroup.left;
            _gravityTargetsJss = new JSONStorableString("gravityTargets", "");
        }

        protected override void OnBuild()
        {
            BuildLeftSide();
            BuildRightSide();
            Refresh();
        }

        void BuildLeftSide(bool rightSide = false)
        {
            AddInfoTextField(
                "\n".Size(8) + _parameterGroup.infoText,
                rightSide,
                300,
                28
            );

            if(_parameter.HasAdjustableGravityPhysics())
            {
                AddHeaderTextField("Gravity Physics", rightSide);

                AddInfoTextField(
                    "\n".Size(8) +
                    _parameterGroup.gravityInfoText +
                    "\n\nWhich slider takes effect depends on the person's orientation (or pose)." +
                    " For example, <i>Up</i> gravity is applied when the person is upside down" +
                    " and gravity pulls breasts <i>up</i> relative to the chest.",
                    rightSide,
                    350,
                    28
                );

                AddElement(() =>
                {
                    var textField = script.CreateTextField(_gravityTargetsJss, rightSide);
                    textField.UItext.fontSize = 28;
                    textField.height = 250;
                    textField.backgroundColor = Color.clear;
                    return textField;
                });
            }
            else if(GetId() == BREAST_JOINT_MASS.NAME)
            {
                AddHeaderTextField("Gravity Physics", rightSide);

                AddInfoTextField(
                    "\n".Size(8) +
                    "Compensate for the droop caused by gravity with a force countering the weight of the breasts." +
                    "\n\nAt 100 % counterforce, breasts are weightless, but still have the inertia granted by their mass.",
                    rightSide,
                    235,
                    28
                );

                AddElement(() =>
                    script.CreateToggle(GravityPhysicsHandler.autoAdjustGravityCounterforceJsb, rightSide)
                );

                AddElement(GravityPhysicsHandler.gravityCounterforceJsf.name, () =>
                {
                    var slider = script.CreateSlider(GravityPhysicsHandler.gravityCounterforceJsf, rightSide);
                    slider.valueFormat = "F0";
                    return slider;
                });
            }
        }

        void BuildRightSide(bool rightSide = true)
        {
            if(_parameterGroup.allowOffsetOnlyLeftParam)
            {
                AddElement(() =>
                {
                    var toggle = script.CreateToggle(_parameterGroup.applyChangesOnlyToLeftParamJsb, rightSide);
                    toggle.label = "Apply Changes Only To Left Breast";
                    return toggle;
                });
            }
            else
            {
                AddSpacer(50, rightSide);
            }

            AddElement(() =>
            {
                var storable = GetId() == TWIST_ANGLE_TARGET.NAME
                    ? GravityPhysicsHandler.twistAngleTargetJsf
                    : _parameter.offsetJsf;
                var slider = script.CreateSlider(storable, rightSide);
                slider.valueFormat = _parameterGroup.valueFormat;
                slider.label = "Offset";
                return slider;
            });

            AddElement(() =>
            {
                var slider = script.CreateSlider(_parameter.activeValueJsf, rightSide);
                slider.label = "Value";
                slider.SetActiveStyle(false, true);
                slider.valueFormat = _parameterGroup.valueFormat;
                return slider;
            });

            if(_parameter.HasAdjustableGravityPhysics())
            {
                AddSpacer(120, rightSide);
                foreach(var kvp in _parameter.gravityCalculators)
                {
                    AddElement(() =>
                    {
                        var slider = script.CreateSlider(kvp.Value.multiplierJsf, rightSide);
                        slider.label = $"{kvp.Key} Gravity";
                        return slider;
                    });
                }
            }
        }

        public void UpdateDiagnostics()
        {
            if(!_parameter.HasAdjustableGravityPhysics())
            {
                return;
            }

            if(Time.frameCount % 6 == 0)
            {
                string str = "\n".Size(8) + "Left breast gravity target values:\n";
                foreach(var kvp in _parameter.gravityCalculators)
                {
                    var config = kvp.Value;
                    float targetValue =
                        _parameter.offsetJsf.val +
                        _parameter.baseValueJsf.val +
                        config.CalculateDynamic(0.5f, MainPhysicsHandler.normalizedMass, TittyMagic.softness, SoftPhysicsHandler.softPhysicsOn);
                    str += $"\n{kvp.Key}: " + $"{Calc.RoundToDecimals(targetValue, _parameterGroup.roundFactor)}".Bold();
                }

                _gravityTargetsJss.val = str;
            }
        }

        public void Refresh()
        {
            var gravityCounterforceSlider = GetElement(GravityPhysicsHandler.gravityCounterforceJsf.name);
            if(gravityCounterforceSlider)
            {
                gravityCounterforceSlider.SetActiveStyle(!GravityPhysicsHandler.autoAdjustGravityCounterforceJsb.val, true);
            }
        }
    }
}
