using System.Collections.Generic;
using Naturalis.Common;
using Naturalis.Common.UI;
using UnityEngine;
using UnityEngine.Events;

namespace Naturalis.TM.UI
{
    sealed class Tabs
    {
        public IWindow activeTabWindow { get; private set; }

        public IWindow GetActiveNestedWindow()
        {
            return activeTabWindow?.GetActiveNestedWindow();
        }

        public MainWindow mainWindow { get; set; }
        public PhysicsWindow physicsWindow { get; set; }
        public MorphsWindow morphsWindow { get; set; }
        public HardCollidersWindow hardCollidersWindow { get; set; }

        // ReSharper disable UnusedMember.Global

        public bool MainTabIsActive()
        {
            return activeTabWindow == mainWindow;
        }

        public bool PhysicsTabIsActive()
        {
            return activeTabWindow == physicsWindow;
        }

        public bool MorphsTabIsActive()
        {
            return activeTabWindow == morphsWindow;
        }

        public bool HardCollidersTabIsActive()
        {
            return activeTabWindow == hardCollidersWindow;
        }

        public bool NestedWindowIsActive()
        {
            return GetActiveNestedWindow() != null;
        }

        // ReSharper restore UnusedMember.Global

        public bool IsActive(IWindow window)
        {
            if(window == activeTabWindow)
            {
                return !NestedWindowIsActive();
            }

            return GetActiveNestedWindow() == window;
        }

        readonly Dictionary<IWindow, NavigationButton> _tabButtons;
        readonly Transform _leftGroupTransform;
        readonly Transform _rightGroupTransform;

        public Tabs()
        {
            _leftGroupTransform = Utils.CreateHorizontalLayoutGroup(TittyMagic.script.GetLeftUIContent()).transform;
            _rightGroupTransform = Utils.CreateHorizontalLayoutGroup(TittyMagic.script.GetRightUIContent()).transform;
            _tabButtons = new Dictionary<IWindow, NavigationButton>();
        }

        public void CreateNavigationButtons()
        {
            CreateNavigationButton(mainWindow, Constant.TITTY_MAGIC, GoToMainWindow);
            CreateNavigationButton(physicsWindow, "Physics", GoToPhysicsWindow);
            CreateNavigationButton(morphsWindow, "Morphs", GoToMorphsWindow);
            CreateNavigationButton(hardCollidersWindow, "Hard Colliders", GoToHardCollidersWindow, !Person.IsFemale());
        }

        void CreateNavigationButton(IWindow window, string name, UnityAction callback, bool disabled = false)
        {
            var parent = _tabButtons.Count < 2 ? _leftGroupTransform : _rightGroupTransform;
            var button = new NavigationButton(
                TittyMagic.script.InstantiateButton(parent).GetComponent<UIDynamicButton>(),
                name,
                Colors.funkyCyan
            );
            if(disabled)
            {
                button.SetDisabled();
            }
            else
            {
                button.AddListener(callback);
                button.SetInactive();
            }

            _tabButtons[window] = button;
        }

        public void GoToMainWindow()
        {
            NavigateToTab(mainWindow);
        }

        public void GoToPhysicsWindow()
        {
            NavigateToTab(physicsWindow);
        }

        public void GoToMorphsWindow()
        {
            NavigateToTab(morphsWindow);
        }

        public void GoToHardCollidersWindow()
        {
            if(!Person.IsFemale())
            {
                TittyMagic.logger.LogMessage("Hard colliders are only supported on a female character.");
                return;
            }

            if(!TittyMagic.script.enabled)
            {
                TittyMagic.logger.LogMessage("Enable the plugin to configure hard colliders.");
                return;
            }

            NavigateToTab(hardCollidersWindow);
        }

        void NavigateToTab(IWindow window)
        {
            activeTabWindow?.Clear();
            activeTabWindow = window;
            window.Rebuild();

            var navigationButton = _tabButtons[window];
            navigationButton.SetActive();
            foreach(var kvp in _tabButtons)
            {
                if(kvp.Key != window)
                {
                    kvp.Value.SetInactive();
                }
            }
        }
    }
}
