using System;
using System.Collections;
using Naturalis.Common;
using Naturalis.Common.Components;
using Naturalis.TM.Handlers;
using UnityEngine;

namespace Naturalis.TM
{
    sealed class UIModManager : UIModManagerBase
    {
        static bool CheckPluginEnabled()
        {
            return TittyMagic.script.enabled;
        }

        public JSONStorableAction switchToTittyMagicAction { get; private set; }

        public void ModifyAtomUI()
        {
            parentUI = Person.instance.atomUIContent;
            TittyMagic.script.NewJSONStorableAction(Constant.INTERNAL_SHOW_FRICTION_UI, () => ShowFrictionUI(FrictionHandler.enabled));
            TittyMagic.script.NewJSONStorableAction(Constant.INTERNAL_HIDE_FRICTION_UI, () => HideFrictionUI(FrictionHandler.enabled));
            TittyMagic.script.NewJSONStorableAction(Constant.INTERNAL_SET_TOP_ELEMENTS_POSITION, () => StartCoroutine(SetTopElementsPositionCo()));
            skinMaterials2Modified = TittyMagic.script.NewStorableBool(Constant.INTERNAL_SKIN_MATERIALS_2_MODIFIED, false);
            switchToTittyMagicAction = new JSONStorableAction(
                $"Switch To {Constant.TITTY_MAGIC}",
                () =>
                {
                    Settings.bootyMagic.CallActionNullSafe(Constant.INTERNAL_HIDE_FRICTION_UI);
                    if(enabled)
                    {
                        ShowFrictionUI(FrictionHandler.enabled);
                    }
                }
            );

            NewUIMod("calibrationButton", null, CreateCalibrationButtonCo, CheckPluginEnabled, true);
            NewUIMod(Constant.M_PECTORAL_PHYSICS, "M Pectoral Physics", ReplaceWithPluginUIButtonCo, CheckPluginEnabled);
            NewUIMod(Constant.F_BREAST_PHYSICS_1, "F Breast Physics 1", ReplaceWithPluginUIButtonCo, CheckPluginEnabled);
            NewUIMod(Constant.F_BREAST_PHYSICS_2, "F Breast Physics 2", ReplaceWithPluginUIButtonCo, CheckPluginEnabled);
            NewUIMod(Constant.F_BREAST_PRESETS, "F Breast Presets", ReplaceWithPluginUIButtonCo, CheckPluginEnabled);
            NewUIMod(Constant.SKIN_MATERIALS_2_ADD_BTN, "Skin Materials 2", AddButtonToSkinMaterialsUICo, CheckPluginEnabled);
            NewUIMod(Constant.SKIN_MATERIALS_2_FRICTION, "Skin Materials 2", CreateFrictionUICo, CheckPluginEnabled);

            foreach(var uiMod in uiMods)
            {
                uiMod.Apply();
            }

            initialized = true;

            if(!TittyMagic.script.enabled)
            {
                enabled = false;
            }
        }

        RectTransform _calibrationRect;

        IEnumerator SetTopElementsPositionCo()
        {
            while(!_calibrationRect) // keeps waiting for person UI to be opened which triggers the elements to be created
            {
                yield return null;
            }

            yield return new WaitForSeconds(0.1f); // In case BootyMagic calls this action OnDestroy but isn't yet gone
            _calibrationRect.anchoredPosition = Settings.bootyMagic ? new Vector2(277, 155) : new Vector2(0, 155);
        }

        UIDynamicButton _calibrationButton;

        IEnumerator CreateCalibrationButtonCo(UIMod uiMod)
        {
            while(!TittyMagic.IsInitialized())
            {
                yield return null;
            }

            try
            {
                var customTransform = Utils.DestroyLayout(TittyMagic.script.InstantiateButton(parentUI));
                _calibrationRect = customTransform.GetComponent<RectTransform>();
                _calibrationRect.pivot = new Vector2(0, 0);
                _calibrationRect.anchoredPosition = Settings.bootyMagic ? new Vector2(277, 155) : new Vector2(0, 155);
                _calibrationRect.sizeDelta = new Vector2(-837, 80);
                uiMod.AddCustomObject(customTransform.gameObject);

                _calibrationButton = customTransform.GetComponent<UIDynamicButton>();
                _calibrationButton.button.onClick.AddListener(() => TittyMagic.script.StartCalibration());
                TittyMagic.script.calibrationHelper.calibrationButton = _calibrationButton;
            }
            catch(Exception e)
            {
                TittyMagic.logger.LogError($"Error modifying {uiMod.id} UI: {e}");
            }
        }

        static IEnumerator ReplaceWithPluginUIButtonCo(UIMod uiMod)
        {
            while(!TittyMagic.bindings && !TittyMagic.IsInitialized())
            {
                yield return null;
            }

            try
            {
                uiMod.InactivateChildren();
                uiMod.AddCustomObject(OpenPluginUIButton(uiMod.target).gameObject);
            }
            catch(Exception e)
            {
                TittyMagic.logger.LogError($"Error modifying {uiMod.id} UI: {e}");
            }
        }

        UIDynamicToggle enableAdaptiveFrictionToggle { get; set; }
        public UIDynamicSlider drySkinFrictionSlider { get; private set; }

        IEnumerator CreateFrictionUICo(UIMod uiMod)
        {
            while(!TittyMagic.IsInitialized())
            {
                yield return null;
            }

            if(Person.HasDummySkin())
            {
                while(!FrictionHandler.enabled)
                {
                    yield return null;
                }
            }
            else if(!FrictionHandler.enabled)
            {
                yield break;
            }

            try
            {
                var leftSide = uiMod.target.Find("LeftSide");
                var rightSide = uiMod.target.Find("RightSide");

                /* Collider friction title */
                {
                    var fieldTransform = Utils.DestroyLayout(TittyMagic.script.InstantiateTextField(leftSide));
                    var rectTransform = fieldTransform.GetComponent<RectTransform>();
                    rectTransform.pivot = new Vector2(0, 0);
                    rectTransform.anchoredPosition = new Vector2(20f, -930);
                    rectTransform.sizeDelta = new Vector2(rectTransform.sizeDelta.x + 10, 100);
                    uiMod.AddCustomObject(fieldTransform.gameObject);

                    var uiDynamic = fieldTransform.GetComponent<UIDynamicTextField>();
                    uiDynamic.UItext.alignment = TextAnchor.LowerCenter;
                    uiDynamic.text = $"{Constant.TITTY_MAGIC} Collider Friction".Size(32).Bold();
                    uiDynamic.backgroundColor = Color.clear;
                    uiDynamic.textColor = Color.white;
                }

                /* Collider friction info text area */
                {
                    var fieldTransform = Utils.DestroyLayout(TittyMagic.script.InstantiateTextField(leftSide));
                    var rectTransform = fieldTransform.GetComponent<RectTransform>();
                    rectTransform.pivot = new Vector2(0, 0);
                    rectTransform.anchoredPosition = new Vector2(20, -1290);
                    rectTransform.sizeDelta = new Vector2(rectTransform.sizeDelta.x + 10, 400);
                    uiMod.AddCustomObject(fieldTransform.gameObject);

                    var uiDynamic = fieldTransform.GetComponent<UIDynamicTextField>();
                    uiDynamic.text =
                        "Combined friction for hard colliders and soft colliders." +
                        "\n\n" +
                        "Adaptive friction reduces friction when <i>Gloss</i> is increased. The higher " +
                        "the gloss, the more <i>Specular Bumpiness</i> adds friction. Other skin material " +
                        "sliders are ignored." +
                        "\n\n" +
                        "Dry skin friction represents the friction value when gloss and specular bumpiness are low.";
                    uiDynamic.backgroundColor = Color.clear;
                    uiDynamic.textColor = Color.white;
                }

                /* Adaptive friction toggle */
                {
                    var customTransform = CustomToggle(rightSide, new Vector2(0, -880));
                    uiMod.AddCustomObject(customTransform.gameObject);

                    var jsb = FrictionHandler.instance.adaptiveFrictionJsb;
                    enableAdaptiveFrictionToggle = customTransform.GetComponent<UIDynamicToggle>();
                    enableAdaptiveFrictionToggle.label = jsb.name;
                    TittyMagic.script.AddToggleToJsb(enableAdaptiveFrictionToggle, jsb);
                }

                /* Dry skin friction slider */
                {
                    var customTransform = CustomSlider(rightSide, new Vector2(0, -1020));
                    uiMod.AddCustomObject(customTransform.gameObject);

                    drySkinFrictionSlider = customTransform.GetComponent<UIDynamicSlider>();
                    var storable = FrictionHandler.instance.drySkinFrictionJsf;
                    drySkinFrictionSlider.Configure(
                        storable.name,
                        storable.min,
                        storable.max,
                        storable.defaultVal,
                        storable.constrained,
                        "F3"
                    );
                    TittyMagic.script.AddSliderToJsf(drySkinFrictionSlider, storable);
                    drySkinFrictionSlider.SetActiveStyle(FrictionHandler.instance.adaptiveFrictionJsb.val, true);
                }

                /* Friction offset slider */
                {
                    var customTransform = CustomSlider(rightSide, new Vector2(0, -1160));
                    uiMod.AddCustomObject(customTransform.gameObject);

                    var uiDynamic = customTransform.GetComponent<UIDynamicSlider>();
                    var storable = FrictionHandler.instance.frictionOffsetJsf;
                    uiDynamic.Configure(
                        storable.name,
                        storable.min,
                        storable.max,
                        storable.defaultVal,
                        storable.constrained,
                        "F3"
                    );
                    TittyMagic.script.AddSliderToJsf(uiDynamic, storable);
                }

                /* Soft collider friction value slider */
                {
                    var customTransform = CustomSlider(rightSide, new Vector2(0, -1300));
                    uiMod.AddCustomObject(customTransform.gameObject);

                    var uiDynamic = customTransform.GetComponent<UIDynamicSlider>();
                    var storable = FrictionHandler.instance.frictionValueJsf;
                    uiDynamic.Configure(
                        storable.name,
                        storable.min,
                        storable.max,
                        storable.defaultVal,
                        storable.constrained,
                        "F3"
                    );
                    TittyMagic.script.AddSliderToJsf(uiDynamic, storable);
                    uiDynamic.SetActiveStyle(false, true);
                }

                if(FrictionHandler.enabled)
                {
                    HideBootyMagicFrictionUI();
                    skinMaterials2Modified.val = true;
                }
                else
                {
                    HideFrictionUI(false);
                }
            }
            catch(Exception e)
            {
                TittyMagic.logger.LogError($"Error modifying {uiMod.id} UI: {e}");
            }
        }

        IEnumerator AddButtonToSkinMaterialsUICo(UIMod uiMod)
        {
            while(!TittyMagic.IsInitialized())
            {
                yield return null;
            }

            if(Person.HasDummySkin())
            {
                while(!FrictionHandler.enabled)
                {
                    yield return null;
                }
            }
            else if(!FrictionHandler.enabled)
            {
                yield break;
            }

            try
            {
                var customTransform = CustomButton(uiMod.target.Find("LeftSide"), new Vector2(20f, -810));
                uiMod.AddCustomObject(customTransform.gameObject);
                var uiDynamicButton = customTransform.GetComponent<UIDynamicButton>();
                switchToTittyMagicAction.RegisterButton(uiDynamicButton);
                uiDynamicButton.label = switchToTittyMagicAction.name;
                uiMod.Disable();
            }
            catch(Exception e)
            {
                TittyMagic.logger.LogError($"Error modifying {uiMod.id}  UI: {e}");
            }
        }

        static Transform OpenPluginUIButton(Transform parent)
        {
            var button = Utils.DestroyLayout(TittyMagic.script.InstantiateButton(parent));
            var uiDynamic = button.GetComponent<UIDynamicButton>();
            uiDynamic.label = $"<b>Open {Constant.TITTY_MAGIC} UI</b>";
            uiDynamic.button.onClick.AddListener(() => TittyMagic.bindings.actions["OpenUI"].actionCallback());
            return button;
        }

        static Transform CustomToggle(Transform parent, Vector2 anchoredPosition)
        {
            var customTransform = Utils.DestroyLayout(TittyMagic.script.InstantiateToggle(parent));
            var rectTransform = customTransform.GetComponent<RectTransform>();
            rectTransform.pivot = new Vector2(0, 0);
            rectTransform.anchoredPosition = anchoredPosition;
            var sizeDelta = rectTransform.sizeDelta;
            rectTransform.sizeDelta = new Vector2(sizeDelta.x + 10, sizeDelta.y);
            return customTransform;
        }

        static Transform CustomSlider(Transform parent, Vector2 anchoredPosition)
        {
            var customTransform = Utils.DestroyLayout(TittyMagic.script.InstantiateSlider(parent));
            var rectTransform = customTransform.GetComponent<RectTransform>();
            rectTransform.pivot = new Vector2(0, 0);
            rectTransform.anchoredPosition = anchoredPosition;
            var sizeDelta = rectTransform.sizeDelta;
            rectTransform.sizeDelta = new Vector2(sizeDelta.x + 10, sizeDelta.y);
            return customTransform;
        }

        static Transform CustomButton(Transform parent, Vector2 anchoredPosition)
        {
            var customTransform = Utils.DestroyLayout(TittyMagic.script.InstantiateButton(parent));
            var rectTransform = customTransform.GetComponent<RectTransform>();
            rectTransform.pivot = new Vector2(0, 0);
            rectTransform.anchoredPosition = anchoredPosition;
            var sizeDelta = rectTransform.sizeDelta;
            rectTransform.sizeDelta = new Vector2(sizeDelta.x + 10, sizeDelta.y);
            return customTransform;
        }

        public void SetFrictionUIVisibility()
        {
            if(!enabled)
            {
                return;
            }

            if(FrictionHandler.enabled)
            {
                ShowFrictionUI(true);
                HideBootyMagicFrictionUI();
                Settings.instance.uiModManager.SetSkinMaterials2ModEnabled(true);
            }
            else
            {
                uiMods.Find(uiMod => uiMod.id == Constant.SKIN_MATERIALS_2_FRICTION).Disable();
                uiMods.Find(uiMod => uiMod.id == Constant.SKIN_MATERIALS_2_ADD_BTN).Disable();
                skinMaterials2Modified.val = false;
                if(!Settings.bootyMagic.GetBoolParamValueNullSafe(Constant.INTERNAL_SKIN_MATERIALS_2_MODIFIED))
                {
                    Settings.instance.uiModManager.SetSkinMaterials2ModEnabled(false);
                }
            }
        }

        static void HideBootyMagicFrictionUI()
        {
            if(!Settings.bootyMagic.IsEnabledNullSafe())
            {
                return;
            }

            if(Settings.bootyMagic.GetBoolParamValueNullSafe(Constant.INTERNAL_SKIN_MATERIALS_2_MODIFIED))
            {
                Settings.bootyMagic.CallActionNullSafe(Constant.INTERNAL_HIDE_FRICTION_UI);
            }
        }

        void OnEnable()
        {
            if(!initialized)
            {
                return;
            }

            foreach(var uiMod in uiMods)
            {
                if(uiMod.id == Constant.SKIN_MATERIALS_2_FRICTION)
                {
                    if(FrictionHandler.enabled)
                    {
                        uiMod.Enable();
                        skinMaterials2Modified.val = true;
                    }
                }
                else if(uiMod.id != Constant.SKIN_MATERIALS_2_ADD_BTN)
                {
                    uiMod.Enable();
                }
            }

            Settings.instance.uiModManager.SetEnabled(FrictionHandler.enabled);
            if(Settings.bootyMagic.IsEnabledNullSafe())
            {
                Settings.bootyMagic.CallActionNullSafe(Constant.INTERNAL_HIDE_FRICTION_UI);
            }
        }

        new void OnDisable()
        {
            base.OnDisable();
            skinMaterials2Modified.val = false;
            if(Settings.bootyMagic.IsEnabledNullSafe())
            {
                Settings.bootyMagic.CallActionNullSafe(Constant.INTERNAL_SHOW_FRICTION_UI);
            }
            else
            {
                Settings.instance.uiModManager.enabled = false;
            }
        }
    }
}
